/* -------------------------------------
   Includes, defines and local variables
   ------------------------------------- */

#include "Yap.h"
#if defined(YAPOR) || defined(TABLING)
#include "alloc.h"
#include <unistd.h>
#include <sys/mman.h>

#define K ((Int) 1024) /* as in 'alloc.c' */
#define MinHGap 256*K  /* as in 'alloc.c' */
#define KBYTES K
#define ALLOC_CHUNK_SIZE 1000 * KBYTES

#define MMAP_START_ADDR ((void *) 0x02000000) /* mcc address: 0xdf2c0000 */
#define ALIGN		3
#define ALIGNMASK	0xfffffffc
#define ADJUST_SIZE(X)	((X + ALIGN) & ALIGNMASK)
#define ADJUST_SIZE_TO_PAGE(size, pagesize)  (size - (size) % pagesize + pagesize)

#ifdef YAPOR
#define OR_FRAMES_ALLOC_NUMBER 50
#define OR_FRAME_SIZE ADJUST_SIZE(sizeof(struct or_frame))
#define ANSWER_FRAMES_ALLOC_NUMBER 50
#define ANSWER_FRAME_SIZE ADJUST_SIZE(sizeof(struct answer_frame))
#define SOLUTION_FRAMES_ALLOC_NUMBER 50
#define SOLUTION_FRAME_SIZE ADJUST_SIZE(sizeof(struct solution_frame))
#endif /* YAPOR */
#ifdef TABLING
#define TABLE_ENTRIES_ALLOC_NUMBER 10
#define TABLE_ENTRY_SIZE ADJUST_SIZE(sizeof(struct table_entry))
#define SUBGOAL_TRIE_NODES_ALLOC_NUMBER 50
#define SUBGOAL_TRIE_NODE_SIZE ADJUST_SIZE(sizeof(struct subgoal_trie_node))
#define SUBGOAL_FRAMES_ALLOC_NUMBER 50
#define SUBGOAL_FRAME_SIZE ADJUST_SIZE(sizeof(struct subgoal_frame))
#define ANSWER_TRIE_NODES_ALLOC_NUMBER 50
#define ANSWER_TRIE_NODE_SIZE ADJUST_SIZE(sizeof(struct answer_trie_node))
#define DEPENDENCY_FRAMES_ALLOC_NUMBER 50
#define DEPENDENCY_FRAME_SIZE ADJUST_SIZE(sizeof(struct dependency_frame))
#ifdef YAPOR
#define SUSPENSION_FRAMES_ALLOC_NUMBER 20
#define SUSPENSION_FRAME_SIZE ADJUST_SIZE(sizeof(struct suspension_frame))
#endif /* YAPOR */
#endif /* TABLING */


/* ----------------
   Global functions
   ---------------- */

int global_data_size(void) {
  return ADJUST_SIZE(sizeof(struct global_data));
}


int local_data_size(void) {
  return ADJUST_SIZE(sizeof(struct local_data));
}


/* HeapArea, GlobalLocalArea and TrailAuxArea in KBytes */
void map_memory(int HeapArea, int GlobalLocalArea, int TrailAuxArea, int fd, int n_workers) {
  int i;
  int pagesize;
  int WorkerArea;
  int OPTArea;
  int TotalArea;
  int offset;
  void *mmap_addr1, *mmap_addr2, *mmap_addr3;

  pagesize = sysconf(_SC_PAGESIZE);
  HeapArea = ADJUST_SIZE_TO_PAGE(HeapArea * KBYTES  + global_data_size(), pagesize);
  GlobalLocalArea = ADJUST_SIZE(GlobalLocalArea * KBYTES);
  TrailAuxArea = ADJUST_SIZE(TrailAuxArea * KBYTES);
#ifdef ACOW
  WorkerArea = ADJUST_SIZE_TO_PAGE(local_data_size(), pagesize);
  OPTArea = ADJUST_SIZE_TO_PAGE(ALLOC_CHUNK_SIZE, pagesize);
  TotalArea = HeapArea + WorkerArea * n_workers  + GlobalLocalArea + TrailAuxArea + OPTArea;
#else
  WorkerArea = ADJUST_SIZE_TO_PAGE(local_data_size() + GlobalLocalArea + TrailAuxArea, pagesize);
  OPTArea = ADJUST_SIZE_TO_PAGE(ALLOC_CHUNK_SIZE, pagesize);
  TotalArea = HeapArea + WorkerArea * n_workers + OPTArea;
#endif

  if (lseek(fd, TotalArea, SEEK_SET) < 0) 
    abort_optyap("lseek error in function map_memory");
  if (write(fd, "", 1) < 0) 
    abort_optyap("write error in function map_memory");
  /* mapping heap area */
  if ((mmap_addr1 = mmap(MMAP_START_ADDR, (size_t) HeapArea, PROT_READ|PROT_WRITE, 
                         MAP_SHARED|MAP_FIXED, fd, 0)) == (void *) -1)
    abort_optyap("mmap error in function map_memory");
#ifdef SBA /* sba_vsc */
  if (mmap(mmap_addr1+HeapArea, n_workers*WorkerArea, PROT_READ|PROT_WRITE, 
	   MAP_SHARED|MAP_FIXED, fd, HeapArea) == (void *) -1)
    abort_optyap("mmap error in function map_memory");
#endif /* SBA */
  /* mapping workers area */
  for (i = 0; i < n_workers; i++) {
    offset = HeapArea + i * WorkerArea;
    mmap_addr2 = mmap_addr1 + offset;
#ifndef SBA /* sba_vsc */
    if (mmap(mmap_addr2, WorkerArea, PROT_READ|PROT_WRITE, 
             MAP_SHARED|MAP_FIXED, fd, offset) == (void *) -1)
      abort_optyap("mmap error in function map_memory");
#endif /* SBA */
#ifdef YAPOR
    worker_area(i) = mmap_addr2;
    REMOTE[i] = (struct local_data *) mmap_addr2;
    worker_offset(i) = (int)(worker_area(i) - worker_area(0));
#endif /* YAPOR */
  }
  /* mapping OPT area */
  offset = HeapArea + n_workers * WorkerArea;
#ifdef ACOW
  mmap_addr2 = mmap_addr1 + offset;
  if (mmap(mmap_addr2 , GlobalLocalArea + TrailAuxArea, PROT_READ|PROT_WRITE, 
             MAP_PRIVATE|MAP_FIXED, fd, offset) == (void *) -1)
      abort_optyap("mmap error in function map_memory");
  offset += GlobalLocalArea + TrailAuxArea;
#endif /* ACOW */
  if((mmap_addr3 = mmap((void *)(mmap_addr1 + offset), (size_t) OPTArea, 
                        PROT_READ|PROT_WRITE, MAP_SHARED|MAP_FIXED, 
                        fd, offset)) == (void *) -1)
    abort_optyap("mmap error in function map_memory");
  /* initialize global vars */
  GLOBAL = (struct global_data *) (mmap_addr1 + HeapArea - global_data_size());
  BaseWorkArea = mmap_addr1;
  TopWorkArea = BaseWorkArea + TotalArea;
  FramesArea = mmap_addr3;
#if defined(YAPOR) && defined(TABLING)
  SuspensionArea = TopWorkArea;
#endif /* YAPOR && TABLING */

  HeapBase = BaseWorkArea;
  HeapMax = HeapUsed = 0;
  HeapTop = HeapBase + sizeof(SEG_SIZE);
  *((SEG_SIZE *) HeapBase) = InUseFlag;
  *((SEG_SIZE *) HeapTop) = InUseFlag;
  HeapPlus = HeapTop + MinHGap / CellSize;
  FreeBlocks = NIL;

  LOCAL = (struct local_data *) (mmap_addr1 + HeapArea);
#ifdef ACOW
  GlobalBase = mmap_addr2;
#else
  GlobalBase = mmap_addr1 + HeapArea + local_data_size();
#endif
  TrailBase = GlobalBase + GlobalLocalArea;
  LocalBase = TrailBase - CellSize;
  TrailTop = TrailBase + TrailAuxArea / 2;
  AuxTop = TrailBase + TrailAuxArea - CellSize;
  AuxSp = (CELL *) AuxTop;
#ifdef SBA /* sba_vsc */
  /* alloc space for the sparse binding array */
  if ((binding_array = (char *)malloc(WorkerArea*n_workers)) == NULL)
    abort_optyap("mmap error in function map_memory");
  sba_offset = binding_array-GlobalBase;
  sba_size = WorkerArea*n_workers;
  sba_end = binding_array+WorkerArea*n_workers;
#endif /* SBA */
#if USE_OFFSETS
  AtomBase = HeapBase;
#endif

  return;
}


#ifdef YAPOR
void remap_memory(int fd) {
#ifdef ACOW
  /* do nothing */
#else
#if SBA /* vsc_sba */
  /* setup workers so that they have different areas */
  int WorkerArea = worker_offset(1);

  GlobalBase +=  worker_id*WorkerArea;
  TrailBase += worker_id*WorkerArea;
  LocalBase += worker_id*WorkerArea;
  TrailTop += worker_id*WorkerArea;
  AuxTop += worker_id*WorkerArea;
  AuxSp = (CELL *) AuxTop; 
#else
  int i;
  void *start;
  int start_offset;
  int WorkerArea;
  int offset;

  WorkerArea = worker_offset(1);
  for (i = 0; i < number_workers; i++)
    if (munmap((void *)worker_area(i), (size_t)WorkerArea) == -1)
      abort_optyap("munmap error in function remap_memory");

  start = worker_area(0);
  start_offset = (int)(worker_area(0) - BaseWorkArea);
  for (i = 0; i < number_workers; i++) {
    worker_area(i) = start + WorkerArea * ((number_workers + i - worker_id) % number_workers);
    REMOTE[i] = (struct local_data *) worker_area(i);
    offset = start_offset + i * WorkerArea;
    if (mmap(worker_area(i), (size_t)WorkerArea, PROT_READ|PROT_WRITE, 
        MAP_SHARED|MAP_FIXED, fd, offset) == (void *) -1)
      abort_optyap("map error in function remap_memory");
  }
  for (i = 0; i < number_workers; i++) {
    worker_offset(i) = (int)(worker_area(i) - worker_area(worker_id));
  }
#endif /* SBA */
#endif /* ACOW */
}
#endif /* YAPOR */


void alloc_all_optyap_structs(void) {
#ifdef YAPOR
  alloc_or_frames();
  alloc_answer_frames();
  alloc_solution_frames();
#endif /* YAPOR */
#ifdef TABLING
  alloc_table_entries();
  alloc_subgoal_trie_nodes();
  alloc_subgoal_frames();
  alloc_answer_trie_nodes();
  alloc_dependency_frames();
#ifdef YAPOR
  alloc_suspension_frames();
#endif /* YAPOR */
#endif /* TABLING */
  return;
}


#ifdef YAPOR
void alloc_or_frames(void) {
  int i;
  or_fr_ptr aux;

  optyap_message_type1("allocating or frames");
  LOCK(GLOBAL_lock_free_area);
#ifdef TABLING
  if (OR_FRAME_SIZE * OR_FRAMES_ALLOC_NUMBER > SuspensionArea - FramesArea)
#else
  if (OR_FRAME_SIZE * OR_FRAMES_ALLOC_NUMBER > TopWorkArea - FramesArea)
#endif /* TABLING */
    abort_optyap("no more free alloc space (alloc_or_frames)");
  GLOBAL_free_or_frame = FramesArea;
  FramesArea += OR_FRAME_SIZE * OR_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_lock_free_area);

  aux = GLOBAL_free_or_frame;
  for (i = 1; i < OR_FRAMES_ALLOC_NUMBER; i++) {
    OrFr_next(aux) = (or_fr_ptr) ((void *)aux + OR_FRAME_SIZE);
    aux = OrFr_next(aux);
  }
  OrFr_next(aux) = NULL;
  return;
}


void alloc_answer_frames(void) {
  int i;
  ans_fr_ptr aux;

  optyap_message_type1("allocating answer frames");
  LOCK(GLOBAL_lock_free_area);
#ifdef TABLING
  if (ANSWER_FRAME_SIZE * ANSWER_FRAMES_ALLOC_NUMBER > SuspensionArea - FramesArea)
#else
  if (ANSWER_FRAME_SIZE * ANSWER_FRAMES_ALLOC_NUMBER > TopWorkArea - FramesArea)
#endif /* TABLING */
    abort_optyap("no more free alloc space (alloc_answer_frames)");
  GLOBAL_free_answer_frame = FramesArea;
  FramesArea += ANSWER_FRAME_SIZE * ANSWER_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_lock_free_area);

  aux = GLOBAL_free_answer_frame;
  for (i = 1; i < ANSWER_FRAMES_ALLOC_NUMBER; i++) {
    AnsFr_next(aux) = (ans_fr_ptr) ((void *)aux + ANSWER_FRAME_SIZE);
    aux = AnsFr_next(aux);
  }
  AnsFr_next(aux) = NULL;
  return;
}


void alloc_solution_frames(void) {
  int i;
  sol_fr_ptr aux;

  optyap_message_type1("allocating solution frames");
  LOCK(GLOBAL_lock_free_area);
#ifdef TABLING
  if (SOLUTION_FRAME_SIZE * SOLUTION_FRAMES_ALLOC_NUMBER > SuspensionArea - FramesArea)
#else
  if (SOLUTION_FRAME_SIZE * SOLUTION_FRAMES_ALLOC_NUMBER > TopWorkArea - FramesArea)
#endif /* TABLING */
    abort_optyap("no more free alloc space (alloc_solution_frames)");
  GLOBAL_free_solution_frame = FramesArea;
  FramesArea += SOLUTION_FRAME_SIZE * SOLUTION_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_lock_free_area);

  aux = GLOBAL_free_solution_frame;
  for (i = 1; i < SOLUTION_FRAMES_ALLOC_NUMBER; i++) {
    SolFr_next(aux) = (sol_fr_ptr) ((void *)aux + SOLUTION_FRAME_SIZE);
    aux = SolFr_next(aux);
  }
  SolFr_next(aux) = NULL;
  return;
}
#endif /* YAPOR */


#ifdef TABLING
void alloc_table_entries(void) {
  int i;
  tab_ent_ptr aux;

  optyap_message_type1("allocating table entries");
  LOCK(GLOBAL_lock_free_area);
#ifdef YAPOR
  if (TABLE_ENTRY_SIZE * TABLE_ENTRIES_ALLOC_NUMBER > SuspensionArea - FramesArea)
#else
  if (TABLE_ENTRY_SIZE * TABLE_ENTRIES_ALLOC_NUMBER > TopWorkArea - FramesArea)
#endif /* YAPOR */
    abort_optyap("no more free alloc space (alloc_table_entries)");
  GLOBAL_free_table_entry = FramesArea;
  FramesArea += TABLE_ENTRY_SIZE * TABLE_ENTRIES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_lock_free_area);

  aux = GLOBAL_free_table_entry;
  for (i = 1; i < TABLE_ENTRIES_ALLOC_NUMBER; i++) {
    TabEnt_next(aux) = (tab_ent_ptr) ((void *)aux + TABLE_ENTRY_SIZE);
    aux = TabEnt_next(aux);
  }
  TabEnt_next(aux) = NULL;
  return;
}


void alloc_subgoal_trie_nodes(void) {
  int i;
  sg_node_ptr aux;

  optyap_message_type1("allocating subgoal trie nodes");
  LOCK(GLOBAL_lock_free_area);
#ifdef YAPOR
  if (SUBGOAL_TRIE_NODE_SIZE * SUBGOAL_TRIE_NODES_ALLOC_NUMBER > SuspensionArea - FramesArea)
#else
  if (SUBGOAL_TRIE_NODE_SIZE * SUBGOAL_TRIE_NODES_ALLOC_NUMBER > TopWorkArea - FramesArea)
#endif /* YAPOR */
    abort_optyap("no more free alloc space (alloc_subgoal_trie_nodes)");
  GLOBAL_free_subgoal_trie_node = FramesArea;
  FramesArea += SUBGOAL_TRIE_NODE_SIZE * SUBGOAL_TRIE_NODES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_lock_free_area);

  aux = GLOBAL_free_subgoal_trie_node;
  for (i = 1; i < SUBGOAL_TRIE_NODES_ALLOC_NUMBER; i++) {
    TrNode_next(aux) = (sg_node_ptr) ((void *)aux + SUBGOAL_TRIE_NODE_SIZE);
    aux = TrNode_next(aux);
  }
  TrNode_next(aux) = NULL;
  return;
}


void alloc_subgoal_frames(void) {
  int i;
  sg_fr_ptr aux;

  optyap_message_type1("allocating subgoal frames");
  LOCK(GLOBAL_lock_free_area);
#ifdef YAPOR
  if (SUBGOAL_FRAME_SIZE * SUBGOAL_FRAMES_ALLOC_NUMBER > SuspensionArea - FramesArea)
#else
  if (SUBGOAL_FRAME_SIZE * SUBGOAL_FRAMES_ALLOC_NUMBER > TopWorkArea - FramesArea)
#endif /* YAPOR */
    abort_optyap("no more free alloc space (alloc_subgoal_frames)");
  GLOBAL_free_subgoal_frame = FramesArea;
  FramesArea += SUBGOAL_FRAME_SIZE * SUBGOAL_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_lock_free_area);

  aux = GLOBAL_free_subgoal_frame;
  for (i = 1; i < SUBGOAL_FRAMES_ALLOC_NUMBER; i++) {
    SgFr_next(aux) = (sg_fr_ptr) ((void *)aux + SUBGOAL_FRAME_SIZE);
    aux = SgFr_next(aux);
  }
  SgFr_next(aux) = NULL;
  return;
}


void alloc_answer_trie_nodes(void) {
  int i;
  ans_node_ptr aux;

  optyap_message_type1("allocating answer trie nodes");
  LOCK(GLOBAL_lock_free_area);
#ifdef YAPOR
  if (ANSWER_TRIE_NODE_SIZE * ANSWER_TRIE_NODES_ALLOC_NUMBER > SuspensionArea - FramesArea)
#else
  if (ANSWER_TRIE_NODE_SIZE * ANSWER_TRIE_NODES_ALLOC_NUMBER > TopWorkArea - FramesArea)
#endif /* YAPOR */
    abort_optyap("no more free alloc space (alloc_answer_trie_nodes)");
  GLOBAL_free_answer_trie_node = FramesArea;
  FramesArea += ANSWER_TRIE_NODE_SIZE * ANSWER_TRIE_NODES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_lock_free_area);

  aux = GLOBAL_free_answer_trie_node;
  for (i = 1; i < ANSWER_TRIE_NODES_ALLOC_NUMBER; i++) {
    TrNode_next(aux) = (ans_node_ptr) ((void *)aux + ANSWER_TRIE_NODE_SIZE);
    aux = TrNode_next(aux);
  }
  TrNode_next(aux) = NULL;
  return;
}


void alloc_dependency_frames(void) {
  int i;
  dep_fr_ptr aux;

  optyap_message_type1("allocating dependency frames");
  LOCK(GLOBAL_lock_free_area);
#ifdef YAPOR
  if (DEPENDENCY_FRAME_SIZE * DEPENDENCY_FRAMES_ALLOC_NUMBER > SuspensionArea - FramesArea)
#else
  if (DEPENDENCY_FRAME_SIZE * DEPENDENCY_FRAMES_ALLOC_NUMBER > TopWorkArea - FramesArea)
#endif /* YAPOR */
    abort_optyap("no more free alloc space (alloc_dependency_frames)");
  GLOBAL_free_dependency_frame = FramesArea;
  FramesArea += DEPENDENCY_FRAME_SIZE * DEPENDENCY_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_lock_free_area);

  aux = GLOBAL_free_dependency_frame;
  for (i = 1; i < DEPENDENCY_FRAMES_ALLOC_NUMBER; i++) {
    DepFr_next(aux) = (dep_fr_ptr) ((void *)aux + DEPENDENCY_FRAME_SIZE);
    aux = DepFr_next(aux);
  }
  DepFr_next(aux) = NULL;
  return;
}


#ifdef YAPOR
void alloc_suspension_frames(void) {
  int i;
  susp_fr_ptr aux;

  optyap_message_type1("allocating suspension frames");
  LOCK(GLOBAL_lock_free_area);
#ifdef YAPOR
  if (SUSPENSION_FRAME_SIZE * SUSPENSION_FRAMES_ALLOC_NUMBER > SuspensionArea - FramesArea)
#else
  if (SUSPENSION_FRAME_SIZE * SUSPENSION_FRAMES_ALLOC_NUMBER > TopWorkArea - FramesArea)
#endif /* YAPOR */
    abort_optyap("no more free alloc space (alloc_suspension_frames)");
  GLOBAL_free_suspension_frame = FramesArea;
  FramesArea += SUSPENSION_FRAME_SIZE * SUSPENSION_FRAMES_ALLOC_NUMBER;
  UNLOCK(GLOBAL_lock_free_area);

  aux = GLOBAL_free_suspension_frame;
  for (i = 1; i < SUSPENSION_FRAMES_ALLOC_NUMBER; i++) {
    SuspFr_next(aux) = (susp_fr_ptr) ((void *)aux + SUSPENSION_FRAME_SIZE);
    aux = SuspFr_next(aux);
  }
  SuspFr_next(aux) = NULL;
  return;
}


void *alloc_suspension_block(int size) {
  void *block;
  optyap_message_type1("allocating suspension block");
  LOCK(GLOBAL_lock_free_area);
  if (size > SuspensionArea - FramesArea)
    abort_optyap("no more free alloc space (alloc_suspension_block)");
  SuspensionArea -= size;
  block = SuspensionArea;
  UNLOCK(GLOBAL_lock_free_area);
  return block;
}


void reset_suspension_area(void) {
  SuspensionArea = TopWorkArea;
  return;
}
#endif /* YAPOR */
#endif /* TABLING */
#endif /* YAPOR || TABLING */
