/* --------
   Includes
   -------- */

#include "Yap.h"
#ifdef YAPOR
#include "or.macros.h"


/* ------------
   Local macros
   ------------ */

STD_PROTO(static inline void MOVE_UP_ONE_CP, (choiceptr));
STD_PROTO(static inline void MOVE_UP_TO_CP, (choiceptr));
STD_PROTO(static inline void PUT_NO_WORK_IN_UP_CPS, (void));
STD_PROTO(static inline void PUT_ME_IDLE, (int));
STD_PROTO(static inline void PUT_ME_BUSY, (int));
STD_PROTO(static inline void PUT_ME_IN_ROOT_CP, (int));
STD_PROTO(static inline void PUT_ME_OUT_ROOT_CP, (int));

/*
static inline 
void MOVE_UP_ONE_CP(choiceptr nearest_livenode) {
  LOCK_OR_FRAME(LOCAL_top_or_fr);
  BITMAP_delete(LOCAL_top_or_fr->members, worker_id);
  if (LOCAL_top_or_fr->pending_cut_cp != NULL &&
      GLOBAL_branch_ltt(worker_id, LOCAL_top_or_fr->depth) >= LOCAL_top_or_fr->pending_cut_ltt &&
      no_workers_left_pending_cut()) {
    CUT_continue_pending_cut();
  } else if (BITMAP_empty(LOCAL_top_or_fr->members)) {
    sol_fr_ptr solution;
    solution = LOCAL_top_or_fr->solutions;
    CUT_put_solutions_together(solution);
    FREE_OR_FRAME(LOCAL_top_or_fr);
    SCH_update_local_or_tops();
    if (solution != NULL) {
      choiceptr leftmost_cp;
      or_fr_ptr leftmost_or_fr;
      leftmost_cp = CUT_leftmost_until();
      leftmost_or_fr = leftmost_cp->cp_or_fr;
      LOCK_OR_FRAME(leftmost_or_fr);
      if (LOCAL_cut_cp >= LOCAL_top_cp) {
        UNLOCK_OR_FRAME(leftmost_or_fr);
        CUT_free_pending_solutions(solution);
        CUT_move_up_to_cut_cp();
      } else {
        CUT_put_inherted_solutions(solution, leftmost_or_fr);
        UNLOCK_OR_FRAME(leftmost_or_fr);
      }
    }
    CUT_clear_cut_cp();
  } else {
    if (nearest_livenode)
      LOCAL_top_or_fr->nearest_livenode = nearest_livenode;
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    SCH_update_local_or_tops();
    CUT_clear_cut_cp();
  }
  return;
}
*/


static inline 
void MOVE_UP_ONE_CP(choiceptr nearest_livenode) {
  LOCK_OR_FRAME(LOCAL_top_or_fr);
  BITMAP_delete(OrFr_members(LOCAL_top_or_fr), worker_id);
#ifdef TABLING
  if (LOCAL_top_cp >= B_FZ) {
    /* check for suspension frames with unconsumed answers */
/*
    if (BITMAP_empty(LOCAL_top_or_fr->members)) {
      collect_suspension_frames_with_unconsumed_answers();
    }
*/
    /* update top or frame */
    BITMAP_insert(OrFr_hidden_members(LOCAL_top_or_fr), worker_id);
    if (nearest_livenode)
      OrFr_nearest_livenode(LOCAL_top_or_fr) = nearest_livenode;
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    SCH_update_local_or_tops();
    CUT_clear_cut_cp();
  } else
#endif  /* TABLING */
  {
    if (OrFr_pend_cut_cp(LOCAL_top_or_fr) != NULL &&
        GLOBAL_branch_ltt(worker_id, OrFr_depth(LOCAL_top_or_fr)) >= OrFr_pend_cut_ltt(LOCAL_top_or_fr) &&
        no_workers_left_pending_cut()) {
      CUT_continue_pending_cut();
    } else if (BITMAP_empty(OrFr_members(LOCAL_top_or_fr))) {
      sol_fr_ptr solution;
#ifdef TABLING
/*
      susp_fr_ptr susp_fr;
      susp_fr = ?;
      while (susp_fr) {
        if (unconsumed_answers_in(susp_fr)) {
          resume(susp_fr);
        }
        susp_fr = susp_fr->next;
      }
      if (BITMAP_empty(LOCAL_top_or_fr->hidden_workers)) {
        complete_all_suspension_frames();
      }
*/
#endif /* TABLING */
      solution = OrFr_solutions(LOCAL_top_or_fr);
      CUT_put_solutions_together(solution);
      FREE_OR_FRAME(LOCAL_top_or_fr);
      SCH_update_local_or_tops();
      if (solution != NULL) {
        choiceptr leftmost_cp;
        or_fr_ptr leftmost_or_fr;
        leftmost_cp = CUT_leftmost_until();
        leftmost_or_fr = leftmost_cp->cp_or_fr;
        LOCK_OR_FRAME(leftmost_or_fr);
        if (LOCAL_cut_cp >= LOCAL_top_cp) {
          UNLOCK_OR_FRAME(leftmost_or_fr);
          CUT_free_pending_solutions(solution);
          CUT_move_up_to_cut_cp();
        } else {
          CUT_put_inherted_solutions(solution, leftmost_or_fr);
          UNLOCK_OR_FRAME(leftmost_or_fr);
        }
      }
      CUT_clear_cut_cp();
    } else {
      if (nearest_livenode)
        OrFr_nearest_livenode(LOCAL_top_or_fr) = nearest_livenode;
      UNLOCK_OR_FRAME(LOCAL_top_or_fr);
      SCH_update_local_or_tops();
      CUT_clear_cut_cp();
    }
  }
  return;
}


static inline
void MOVE_UP_TO_CP(choiceptr cp_to_move) {
  do {
    MOVE_UP_ONE_CP(cp_to_move);
  } while (LOCAL_top_cp < cp_to_move);
  return;
}


static inline
void PUT_NO_WORK_IN_UP_CPS(void) {
  or_fr_ptr aux2 = LOCAL_top_or_fr;
  choiceptr aux1 = OrFr_nearest_livenode(aux2);
  while (aux1 != NULL) {
    OrFr_nearest_livenode(aux2) = NULL;
    aux2 = aux1->cp_or_fr;
    aux1 = OrFr_nearest_livenode(aux2);
  }
  return;
}


static inline
void PUT_ME_IDLE(int worker_num) {
  LOCK(GLOBAL_lock_bm_idle_workers);
  BITMAP_insert(GLOBAL_bm_idle_workers, worker_num);
  UNLOCK(GLOBAL_lock_bm_idle_workers);
  return;
}


static inline
void PUT_ME_BUSY(int worker_num) {
  LOCK(GLOBAL_lock_bm_idle_workers);
  BITMAP_delete(GLOBAL_bm_idle_workers, worker_num);
  UNLOCK(GLOBAL_lock_bm_idle_workers);
  return;
}


static inline
void PUT_ME_IN_ROOT_CP(int worker_num) {
  LOCK(GLOBAL_lock_bm_root_cp_workers);
  BITMAP_insert(GLOBAL_bm_root_cp_workers, worker_num);
  UNLOCK(GLOBAL_lock_bm_root_cp_workers);
  return;
}


static inline
void PUT_ME_OUT_ROOT_CP(int worker_num) {
  LOCK(GLOBAL_lock_bm_root_cp_workers);
  BITMAP_delete(GLOBAL_bm_root_cp_workers, worker_num);
  UNLOCK(GLOBAL_lock_bm_root_cp_workers);
  return;
}


/* ---------------------------
   Local functions declaration
   --------------------------- */

static int get_work_below(void);
static int get_work_above(void);
static void find_a_better_position(void);
static int search_for_hiden_shared_work(bitmap stable_busy);


/* ----------------
   Global functions
   ---------------- */

int get_work(void){
  int counter;
  bitmap stable_busy;
  yamop *alt_with_work;
  choiceptr cp_with_work;
#ifdef TABLING
  choiceptr leader_cp;
#endif /* TABLING */

  while (LOCAL_reply_signal == install) continue;
  /* find nearest node with available work */
  cp_with_work = LOCAL_top_cp;
  /* versao sem predicados sequenciais 
    do {
      cp_with_work = cp_with_work->cp_or_fr->nearest_livenode;
    } while (cp_with_work != NULL && cp_with_work->cp_or_fr->alternative == NULL);
  */
  do {
    cp_with_work = OrFr_nearest_livenode(cp_with_work->cp_or_fr);
    if (cp_with_work == NULL) break;      
    alt_with_work = OrFr_alternative(cp_with_work->cp_or_fr);
  } while (alt_with_work == NULL || alt_with_work->YAMOP_SEQ);

#ifdef TABLING
  /* check if there is a leader node before the nearest cp with work */
  leader_cp = DepFr_leader_cp(LOCAL_top_dep_fr);
  if (leader_cp) {
    if (! cp_with_work || leader_cp < cp_with_work)
      cp_with_work = leader_cp;
      /* MOVE_UP_TO_CP(leader_cp (cp to move to), cp_with_work (nearest cp)); */
  }
#endif /* TABLING */
  if (cp_with_work != NULL) {
    MOVE_UP_TO_CP(cp_with_work);
    return 1;
  }
  PUT_NO_WORK_IN_UP_CPS();

  PUT_ME_OUT_ROOT_CP(worker_id);
  LOCK_WORKER(worker_id);
  PUT_ME_IDLE(worker_id);
  if (SCH_any_request()) {
    /* if any request to share work, refuse it */
    REMOTE_reply_signal(LOCAL_request) = refused;
    LOCAL_request = MAX_WORKERS;
  }
  UNLOCK_WORKER(worker_id);

  counter = 1;
  BITMAP_difference(stable_busy, OrFr_members(LOCAL_top_or_fr), GLOBAL_bm_idle_workers);
  while (1) {
    /* versao sem predicados sequenciais 
    if (BITMAP_subset(GLOBAL_bm_idle_workers, LOCAL_top_or_fr->members)) {
      while (BITMAP_subset(GLOBAL_bm_idle_workers, LOCAL_top_or_fr->members)
             && LOCAL_top_cp != GLOBAL_root_cp) {
        MOVE_UP_ONE_CP(NULL);
      }
    }
    */
    while (BITMAP_subset(GLOBAL_bm_idle_workers, OrFr_members(LOCAL_top_or_fr)) &&
           LOCAL_top_cp != GLOBAL_root_cp) {
      /* no busy workers here and below */
      if (OrFr_alternative(LOCAL_top_or_fr) != NULL) {
        /* This is a sequential choicepoint with unexploited alternatives */
        PUT_ME_BUSY(worker_id);
        return 1;   
      }
      MOVE_UP_ONE_CP(NULL);
    }
    /* versao sem predicados sequenciais 
    if (BITMAP_same(GLOBAL_bm_idle_workers, GLOBAL_bm_present_workers)) {
      return 0;
    }
    */
    if (LOCAL_top_cp == GLOBAL_root_cp) {
      PUT_ME_IN_ROOT_CP(worker_id);
      if (BITMAP_same(GLOBAL_bm_idle_workers, GLOBAL_bm_root_cp_workers) &&
          BITMAP_same(GLOBAL_bm_idle_workers, GLOBAL_bm_present_workers)) {
      /* All workers are idle in root choicepoint. Execution 
         must finish as there is no available computation. */
        return 0;
      }
    }
    if (get_work_below()) {
      PUT_ME_BUSY(worker_id);
      return 1;
    }
    if (get_work_above()) {
      PUT_ME_BUSY(worker_id);
      return 1;
    }
    find_a_better_position();
    if (counter++ == SCHEDULER_LOOP) {
      if (search_for_hiden_shared_work(stable_busy)) {
        PUT_ME_BUSY(worker_id);
        return 1;
      }
      BITMAP_difference(stable_busy, OrFr_members(LOCAL_top_or_fr), GLOBAL_bm_idle_workers);
      counter = 1;
    }
    BITMAP_minus(stable_busy, GLOBAL_bm_idle_workers);
  }
}


/* ---------------
   Local functions
   --------------- */

static int get_work_below(void){
  int i, p = -1, big_load = 0;
  bitmap busy_below, idle_below;
  BITMAP_difference(busy_below, OrFr_members(LOCAL_top_or_fr), GLOBAL_bm_idle_workers);
  BITMAP_difference(idle_below, OrFr_members(LOCAL_top_or_fr), busy_below);
  BITMAP_delete(idle_below, worker_id);
  for (i = 0; i < number_workers; i++)
    if (BITMAP_member(idle_below ,i) && REMOTE_top_cp(i) < LOCAL_top_cp)
      BITMAP_minus(busy_below, OrFr_members(REMOTE_top_or_fr(i)));
  if (BITMAP_empty(busy_below)) return 0;
  /* choose the worker with highest load */
  for (i = 0 ; i < number_workers; i++) {
    if (BITMAP_member(busy_below ,i) && REMOTE_load(i) > big_load) {
      p = i;
      big_load = REMOTE_load(i);
    }
  }
  if (p < 0) return 0;
  return (q_share_work(p));
}


static int get_work_above(void){
  int i, p = -1, big_load = 0;
  bitmap visible_busy_above, visible_idle_above;
  BITMAP_difference(visible_busy_above, GLOBAL_bm_present_workers, OrFr_members(LOCAL_top_or_fr));
  BITMAP_minus(visible_busy_above, GLOBAL_bm_invisible_workers);
  BITMAP_copy(visible_idle_above, visible_busy_above); 
  BITMAP_minus(visible_busy_above, GLOBAL_bm_idle_workers);
  BITMAP_and(visible_idle_above, GLOBAL_bm_idle_workers);
  BITMAP_insert(visible_busy_above, worker_id);
  for (i = 0 ; i < number_workers; i++)
    if (BITMAP_member(visible_idle_above, i))
      BITMAP_minus(visible_busy_above, OrFr_members(REMOTE_top_or_fr(i)));
  if (!BITMAP_member(visible_busy_above, worker_id) || 
      BITMAP_alone(visible_busy_above, worker_id)) return 0;
  BITMAP_delete(visible_busy_above, worker_id);
  /* choose the worker with higher load */
  for (i = 0; i < number_workers; i++)
    if (BITMAP_member(visible_busy_above ,i) && REMOTE_load(i) > big_load) {
      p = i;
      big_load = REMOTE_load(i);
    }
  if (p < 0) return 0;
  /* put workers invisibles */
  LOCK(GLOBAL_lock_bm_invisible_workers);
  if (BITMAP_member(GLOBAL_bm_invisible_workers, p)) {
    UNLOCK(GLOBAL_lock_bm_invisible_workers);
    return 0;
  }
  BITMAP_insert(GLOBAL_bm_invisible_workers, worker_id);
  BITMAP_insert(GLOBAL_bm_invisible_workers, p);
  UNLOCK(GLOBAL_lock_bm_invisible_workers);
  /* move up to cp with worker p */
  do {
    MOVE_UP_ONE_CP(NULL);
  } while (! BITMAP_member(OrFr_members(LOCAL_top_or_fr), p));
  /* put workers visibles */
  LOCK(GLOBAL_lock_bm_invisible_workers);
  BITMAP_delete(GLOBAL_bm_invisible_workers, worker_id);
  BITMAP_delete(GLOBAL_bm_invisible_workers, p);
  UNLOCK(GLOBAL_lock_bm_invisible_workers);
  return (q_share_work(p));
}


static int search_for_hiden_shared_work(bitmap stable_busy){
  int i;
  bitmap invisible_work, idle_below;
  BITMAP_intersection(invisible_work, stable_busy, GLOBAL_bm_requestable_workers);
  BITMAP_intersection(idle_below, OrFr_members(LOCAL_top_or_fr), GLOBAL_bm_idle_workers);
  BITMAP_delete(idle_below, worker_id);
  for (i = 0; i < number_workers; i++)
    if (BITMAP_member(idle_below ,i) && REMOTE_top_cp(i) < LOCAL_top_cp)
      BITMAP_minus(invisible_work, OrFr_members(REMOTE_top_or_fr(i)));
  if (BITMAP_empty(invisible_work)) return 0;
  /* choose the first available worker */
  for (i = 0; i < number_workers; i++ )
    if (BITMAP_member(invisible_work ,i)) break;
  return (q_share_work(i));
}


static void find_a_better_position(void){
  int i;
  bitmap busy_above, idle_above;
  BITMAP_difference(busy_above, GLOBAL_bm_present_workers, OrFr_members(LOCAL_top_or_fr));
  BITMAP_copy(idle_above, busy_above); 
  BITMAP_minus(busy_above, GLOBAL_bm_idle_workers);
  BITMAP_and(idle_above, GLOBAL_bm_idle_workers);
  for (i = 0; i < number_workers; i++) {
    if (BITMAP_member(idle_above, i)) {
      if (BITMAP_empty(busy_above)) break;
      if (BITMAP_member(OrFr_members(REMOTE_top_or_fr(i)), worker_id) )BITMAP_clear(busy_above);
      BITMAP_minus(busy_above, OrFr_members(REMOTE_top_or_fr(i)));
    }
  }
  if (BITMAP_empty(busy_above)) return;
  /* move up to cp with all workers of bitmap busy_above */
  do {
    MOVE_UP_ONE_CP(NULL);
  } while (! BITMAP_subset(OrFr_members(LOCAL_top_or_fr), busy_above));
  return;
}
#endif /* YAPOR */
