/* --------
   Includes
   -------- */

#include "Yap.h"
#ifdef TABLING
#include "Yatom.h"
#include "tab.macros.h"
#include <stdio.h>


/* -------------
   Local inlines
   ------------- */

STD_PROTO(static inline sg_node_ptr subgoal_trie_node_check_insert, (sg_node_ptr, Term));
STD_PROTO(static inline ans_node_ptr answer_trie_node_check_insert, (ans_node_ptr, Term, int));

static inline
sg_node_ptr subgoal_trie_node_check_insert(sg_node_ptr node, Term t) {
  sg_node_ptr aux_node;
  LOCK(TrNode_lock(node));
  aux_node = TrNode_child(node);
  if (aux_node == NULL) {
    new_subgoal_trie_node(aux_node, t, NULL, node, NULL, NULL);
    TrNode_child(node) = aux_node;
  } else {
    while (aux_node && TrNode_entry(aux_node) != t) {
      aux_node = TrNode_next(aux_node);
    }
    if (aux_node == NULL) {
      new_subgoal_trie_node(aux_node, t, NULL, node, NULL, TrNode_child(node));
      TrNode_previous(TrNode_child(node)) = aux_node;
      TrNode_child(node) = aux_node;
    }
  }
  UNLOCK(TrNode_lock(node));
  return aux_node;    
}


static inline 
ans_node_ptr answer_trie_node_check_insert(ans_node_ptr node, Term t, int instr) {
  ans_node_ptr aux_node;
  LOCK(TrNode_lock(node));
  aux_node = TrNode_child(node);
  if (aux_node == NULL) {
    new_answer_trie_node(aux_node, instr, 1, 0, 0, t, NULL, node, NULL);
    TrNode_child(node) = aux_node;
  } else {
    while (aux_node && TrNode_entry(aux_node) != t) {
      aux_node = TrNode_next(aux_node);
    }
    if (aux_node == NULL) {
      ans_node_ptr aux_child = TrNode_child(node);
      new_answer_trie_node(aux_node, instr + 2, TrNode_ltt(aux_child) + 1,
                           0, 0, t, NULL, node, aux_child);
      TrNode_instr(aux_child) = opcode(++TrNode_op_instr(aux_child));
      TrNode_child(node) = aux_node;
    }
  }
  UNLOCK(TrNode_lock(node));
  return aux_node;
}


/* ---------------------------
   Local functions declaration
   --------------------------- */

static void free_subgoal_trie_branch(sg_node_ptr child_node);
static void free_answer_trie_branch(ans_node_ptr node);
static void show_subgoal_trie(sg_node_ptr sg_node, int missing_nodes, int level);
static void show_answer_trie(ans_node_ptr ans_node, int level);


/* ----------------
   Global functions
   ---------------- */

sg_node_ptr subgoal_search(OPREG arity, sg_node_ptr current_sg_node, CELL **Yaddr) {
  int i, j, count_vars;
  CELL *stack_vars, *stack_terms_base, *stack_terms_top, *stack_terms;
  
  count_vars = 0;
  stack_vars = *Yaddr;
  stack_terms_top = (CELL *)TrailTop;
  stack_terms_base = stack_terms = AuxSp;

  for (i = 1; i <= arity; i++) {
    STACK_PUSH(XREGS[i], stack_terms, stack_terms_top);
    do {
      Term t = Deref(STACK_POP(stack_terms));
      int tag = t & TabTagBits;
      switch (tag) {
        case TabVarTagBits:
          if (IsTableVarTerm(t)) {
	    t = MakeTableVarTerm(VarIndexOfTerm(t));
            current_sg_node = subgoal_trie_node_check_insert(current_sg_node, t);
  	  } else {
            FREE_STACK_PUSH(t, stack_vars);
	    *((CELL *)t) = GLOBAL_table_var_enumerator(count_vars);
            t = MakeTableVarTerm(count_vars);
            count_vars++;
            current_sg_node = subgoal_trie_node_check_insert(current_sg_node, t);
	  }
          break;
        case TabAtomTagBits:
        case TabNumberTagBits:
          current_sg_node = subgoal_trie_node_check_insert(current_sg_node, t);
          break;
        case TabPairTagBits:
          current_sg_node = subgoal_trie_node_check_insert(current_sg_node, TabPairTagBits);
          STACK_PUSH(RepPair(t) + 1, stack_terms, stack_terms_top);
          STACK_PUSH(RepPair(t), stack_terms, stack_terms_top);
          break;
        case TabApplTagBits:
          current_sg_node = subgoal_trie_node_check_insert(current_sg_node, TAGGEDA(TabApplTagBits, FunctorOfTerm(t)));
          for (j = ArityOfFunctor(FunctorOfTerm(t)); j >= 1; j--)
            STACK_PUSH(RepAppl(t) + j, stack_terms, stack_terms_top);
          break;
        default :
          abort_optyap("unknown type tag in function subgoal_search");
      }
    } while (STACK_NOT_EMPTY(stack_terms, stack_terms_base));
  }

  FREE_STACK_PUSH(count_vars, stack_vars);
  *Yaddr = stack_vars++;
  /* restore variables */
  while (count_vars--)
    *((CELL *)*stack_vars) = STACK_POP(stack_vars);

  return current_sg_node;
}


ans_node_ptr answer_search(ans_node_ptr current_ans_node, CELL *subs_ptr) {
  int i, j, count_vars, subs_arity;
  CELL *stack_vars, *stack_terms_base, *stack_terms_top, *stack_terms;

  count_vars = 0;
  subs_arity = *subs_ptr;
  stack_vars = AuxSp;
  stack_terms_top = (CELL *)TrailTop;
  stack_terms_base = stack_terms = stack_vars - MAX_TABLE_VARS;

  for (i = subs_arity; i >= 1; i--) {
    STACK_PUSH(*(subs_ptr + i), stack_terms, stack_terms_top);
    do {
      Term t = Deref(STACK_POP(stack_terms));
      int tag = t & TabTagBits;
      switch (tag) {
        case TabVarTagBits:
          if (IsTableVarTerm(t)) {
	    t = MakeTableVarTerm(VarIndexOfTerm(t));
            current_ans_node = answer_trie_node_check_insert(current_ans_node, t, _trie_do_val);
  	  } else {
            FREE_STACK_PUSH(t, stack_vars);
	    *((CELL *)t) = GLOBAL_table_var_enumerator(count_vars);
            t = MakeTableVarTerm(count_vars);
            count_vars++;
            current_ans_node = answer_trie_node_check_insert(current_ans_node, t, _trie_do_var);
	  }
          break;
        case TabAtomTagBits:
        case TabNumberTagBits:
          current_ans_node = answer_trie_node_check_insert(current_ans_node, t, _trie_do_atom);
          break;
        case TabPairTagBits:
          current_ans_node = answer_trie_node_check_insert(current_ans_node, TabPairTagBits, _trie_do_list);
          STACK_PUSH(RepPair(t) + 1, stack_terms, stack_terms_top);
          STACK_PUSH(RepPair(t), stack_terms, stack_terms_top);
          break;
        case TabApplTagBits:
          current_ans_node = answer_trie_node_check_insert(current_ans_node,
                                    TAGGEDA(TabApplTagBits, FunctorOfTerm(t)),
                                    _trie_do_struct);
          for (j = ArityOfFunctor(FunctorOfTerm(t)); j >= 1; j--)
            STACK_PUSH(RepAppl(t) + j, stack_terms, stack_terms_top);
          break;
        default :
          abort_optyap("unknown type tag in function answer_search");
      }
    } while (STACK_NOT_EMPTY(stack_terms, stack_terms_base));
  }

  /* restore variables */
  while (count_vars--)
    *((CELL *)*stack_vars) = STACK_POP(stack_vars);

  return current_ans_node;
}


void show_trie(sg_node_ptr sg_node, int missing_nodes, int level) {
  fprintf(stdout, "[\n");
  show_subgoal_trie(sg_node, missing_nodes, level);
  fprintf(stdout, "]\n\n");
}


int p_check_incomplete_tables(void) {
  if (LOCAL_top_sg_fr != NULL)
    optyap_message_type0("unexpected incomplete subgoals");
  if (LOCAL_top_dep_fr != GLOBAL_root_dep_fr)
    optyap_message_type0("unexpected incomplete dependency frames");
  return(TRUE);
}


int p_abolish_incomplete_tables(void) {
  /* abolish incomplete subgoals */
  while (LOCAL_top_sg_fr != NULL) {
    sg_fr_ptr sg_fr;
    free_answer_trie_branch(SgFr_answer_trie(LOCAL_top_sg_fr));
    free_subgoal_trie_branch(SgFr_subgoal_trie(LOCAL_top_sg_fr));
    sg_fr = LOCAL_top_sg_fr;
    LOCAL_top_sg_fr = SgFr_next(LOCAL_top_sg_fr);
    optyap_message_type2("abolish incomplete subgoal (sg_fr %x)", (unsigned int)sg_fr);
printf("[ abolish incomplete subgoal ]\n");
    FREE_SUBGOAL_FRAME(sg_fr);
  }
  /* release dependency frames */
  while (LOCAL_top_dep_fr != GLOBAL_root_dep_fr) {
    dep_fr_ptr dep_fr = DepFr_next(LOCAL_top_dep_fr);
    FREE_DEPENDENCY_FRAME(LOCAL_top_dep_fr);
    LOCAL_top_dep_fr = dep_fr;
  }
  /* readjust registers */
  H_FZ = DepFr_h_fz(LOCAL_top_dep_fr);
  B_FZ = DepFr_b_fz(LOCAL_top_dep_fr);
  TR_FZ = DepFr_tr_fz(LOCAL_top_dep_fr);
  return(TRUE);
}


/* ---------------
   Local functions
   --------------- */

static void free_subgoal_trie_branch(sg_node_ptr child_node) {
  sg_node_ptr parent_node;
  TrNode_child(child_node) = NULL;
  parent_node = TrNode_parent(child_node);
  while (parent_node) {
    if (TrNode_child(child_node))
      return;
    if (TrNode_previous(child_node)) {
      TrNode_next(TrNode_previous(child_node)) = TrNode_next(child_node);
    }
    else {
      if (TrNode_next(child_node))
        TrNode_previous(TrNode_next(child_node)) = NULL;
      TrNode_child(parent_node) = TrNode_next(child_node);
    }
    FREE_SUBGOAL_TRIE_NODE(child_node);
    child_node = parent_node;
    parent_node = TrNode_parent(child_node);
  }
  return;
}


static void free_answer_trie_branch(ans_node_ptr node) {
  if (TrNode_child(node) && ! IS_ANSWER_LEAF_NODE(node))
    free_answer_trie_branch(TrNode_child(node));
  if (TrNode_next(node))
    free_answer_trie_branch(TrNode_next(node));
  FREE_ANSWER_TRIE_NODE(node);
  return;
}


static void show_subgoal_trie(sg_node_ptr sg_node, int missing_nodes, int level) {
  int i, tag, aux_nodes = -1;
  Term t;

  if (sg_node == NULL) return;
  if (missing_nodes == 0) {
    ans_node_ptr ans_node = SgFr_answer_trie((sg_fr_ptr)sg_node);
    if (! SgFr_completed((sg_fr_ptr)sg_node)) {
      for(i = 0; i < level; i++) fprintf(stdout, "  ");
      fprintf(stdout, "# ***** Potencial error: subgoal not completed !!! *****\n");
    }
    if (IS_ANSWER_LEAF_NODE(ans_node)) {
      for(i = 0; i < level; i++) fprintf(stdout, "  ");
      fprintf(stdout, "# yes\n");
    } else if (TrNode_child(ans_node) == NULL) {
      for(i = 0; i < level; i++) fprintf(stdout, "  ");
      fprintf(stdout, "# no\n");
    } else 
      show_answer_trie(TrNode_child(ans_node), level);
    return;
  }
  for(i = 0; i < level; i++) fprintf(stdout, "  ");

  t = TrNode_entry(sg_node);
  tag = t & TabTagBits;
  switch (tag) {
    case TabVarTagBits:
      fprintf(stdout, "var  --> %d\n", VarIndexOfTableTerm(t));
      break;
    case TabNumberTagBits:
      fprintf(stdout, "int  --> %d\n", IntOfTerm(t));
      break;
    case TabAtomTagBits:
      fprintf(stdout, "atom --> %s\n", AtomName(AtomOfTerm(t)));
      break;
    case TabPairTagBits:
      fprintf(stdout, "pair --> start\n");
      aux_nodes += 2;
      break;
    case TabApplTagBits:
      fprintf(stdout, "appl --> %s/%d\n",
              AtomName(NameOfFunctor((Functor)NonTagPart(t))),
              ArityOfFunctor((Functor)NonTagPart(t)));
      aux_nodes += ArityOfFunctor((Functor)NonTagPart(t));
      break;
    default:
      abort_optyap("unknown type tag in function show_subgoal_trie");
  }

  show_subgoal_trie(TrNode_child(sg_node), missing_nodes + aux_nodes, level + 1);
  show_subgoal_trie(TrNode_next(sg_node), missing_nodes, level);
  return;
}


static void show_answer_trie(ans_node_ptr ans_node, int level) {
  int i, tag;
  Term t;

  if (ans_node == NULL) return;
  for(i = 0; i < level; i++) fprintf(stdout, "  ");
  fprintf(stdout, "# ");

  t = TrNode_entry(ans_node);
  tag = t & TabTagBits;
  switch (tag) {
    case TabVarTagBits:
      fprintf(stdout, "var  --> %d\n", VarIndexOfTableTerm(t));
      break;
    case TabNumberTagBits:
      fprintf(stdout, "int  --> %d\n", IntOfTerm(t));
      break;
    case TabAtomTagBits:
      fprintf(stdout, "atom --> %s\n", AtomName(AtomOfTerm(t)));
      break;
    case TabPairTagBits:
      fprintf(stdout, "pair --> start\n");
      break;
    case TabApplTagBits:
      fprintf(stdout, "appl --> %s/%d\n",
              AtomName(NameOfFunctor((Functor)NonTagPart(t))),
              ArityOfFunctor((Functor)NonTagPart(t)));
      break;
    default:
      abort_optyap("unknown type tag in function show_answer_trie");
  }

  if (! IS_ANSWER_LEAF_NODE(ans_node))
    show_answer_trie(TrNode_child(ans_node), level + 1);
  show_answer_trie(TrNode_next(ans_node), level);
  return;
}
#endif /* TABLING */
