/*************************************************************************
*									 *
*	 YAP Prolog 	%W% %G% 					 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		Yap.h.m4						 *
* mods:									 *
* comments:	main header file for YAP				 *
* version:      $Id: Yap.h.m4,v 1.1 1997/06/02 16:32:00 vsc Exp $	 *
*************************************************************************/

#include "config.h"

/*

#define RATIONAL_TREES 1

#define DEPTH_LIMIT 1

#define YAP_ARRAYS 1

#define COROUTINING 1

#define YAPOR      1

#define TABLING      1

#define OPTYAP      1

#define ANALYST 1

*/

#define MULTI_ASSIGNMENT_VARIABLES 1

#ifdef OPTYAP
#define YAPOR
#define TABLING
#endif /* OPTYAP */

#if defined(YAPOR) || defined(TABLING)
#define FIXED_STACKS 1
#endif /* YAPOR || TABLING */

#if defined(TABLING) || defined(YAPOR)
#undef TRAILING_REQUIRES_BRANCH 
#undef MULTI_ASSIGNMENT_VARIABLES
#endif /* TABLING */

#if ANALYST
#ifdef USE_THREADED_CODE
#undef USE_THREADED_CODE
#endif
#endif

#ifdef  COROUTINING
#ifndef TERM_EXTENSIONS
#define TERM_EXTENSIONS 1
#endif
#endif

#if HAVE_GCC==0
#define inline
#endif

#ifndef ADTDEFS_C
#define EXTERN  static
#else
#define EXTERN
#endif

/*  truth-values							*/
#define	 TRUE	1
#define	 FALSE	0

/*  null pointer							*/
#define	 NIL	0

/* Basic types */

/* defines integer  types Int and UInt (unsigned) with the same size as a ptr
** and integer types Short and UShort with half the size of a ptr 
*/

#if SIZEOF_INT_P==4

#   if SIZEOF_INT==4
/*   */ typedef int Int;
/*   */ typedef unsigned int UInt;

#   elif SIZEOF_LONG_INT==4
/*   */ typedef long int Int;
/*   */ typedef unsigned long int UInt;

#   else
#	error Yap require integer types of the same size as a pointer
#   endif

#   if SIZEOF_SHORT_INT==2
/*   */ typedef short int Short;
/*   */ typedef unsigned short int UShort;

#   else
#	error Yap requires integer types half the size of a pointer
#   endif

#elif SIZEOF_INT_P==8

#   if SIZEOF_INT==8
/*   */ typedef int Int;
/*   */ typedef unsigned int UInt;

#   elif SIZEOF_LONG_INT==8
/*   */ typedef long int Int;
/*   */ typedef unsigned long int UInt;

#   elif SIZEOF_LONG_LONG_INT==8
/*   */ typedef long long int Int;
/*   */ typedef unsigned long long int UInt;

#   else
#	error Yap requires integer types of the same size as a pointer
#   endif

#   if SIZEOF_SHORT_INT==4
/*   */ typedef short int Short;
/*   */ typedef unsigned short int UShort;

#   elif SIZEOF_INT==4
/*   */ typedef int Short;
/*   */ typedef short int UShort;

#   else
#	error Yap requires integer types half the size of a pointer
#   endif

#else

#   error Yap requires pointers of size 4 or 8

#endif

/*   */ typedef double  Float;

#if SIZEOF_INT<SIZEOF_INT_P
#define SHORT_INTS 1
#else
#define SHORT_INTS 0
#endif

#if DEBUG
extern char     Option[20];
#endif

#define USE_OFFSETS	1
#ifndef SHORT_ADDRESSES
#	define LONG_ADDRESSES	1
#else
#	define LONG_ADDRESSES	0
#endif

#ifndef ALIGN_LONGS
#define ALIGN_LONGS 1
#endif

/*  basic data types  */

typedef UInt CELL;
typedef UShort BITS16;
typedef Short SBITS16;
typedef UInt BITS32;

#if ALIGN_LONGS
/*   */ typedef Int DISPREG;
/*   */ typedef CELL SMALLUNSGN;
/*   */ typedef Int  OPREG;
/*   */ typedef CELL UOPREG;
/*   */ typedef CELL OPCODE;

#else
/*   */ typedef Short DISPREG;
/*   */ typedef BITS16 SMALLUNSGN;
/*   */ typedef SBITS16 OPREG;
/*   */ typedef SBITS16 UOPREG;

#    if LOW_ABSMI
/*   */ typedef BITS16 OPCODE;

#    else
/*   */ typedef CELL OPCODE;

#    endif
#endif

typedef char   *ADDR;
typedef CELL OFFSET;
typedef unsigned char *CODEADDR;

#define WordSize     sizeof(BITS16)
#define CellSize     sizeof(CELL)
#define OpCodeSize   sizeof(OPCODE)
#define OpRegSize    sizeof(OPREG)
#define SmallSize    sizeof(SMALLUNSGN)

/* type	casting	macros		*/

#define	Addr(V)		((ADDR) (V))
#define	Unsigned(V)	((CELL) (V))
#define	Signed(V)	((Int) (V))

#define CodePtr(V)	((CODEADDR)(V))
#define	CellPtr(V) 	((CELL *)(V))
#define OpCodePtr(V)	((OPCODE *)(V))
#define OpRegPtr(V)	((OPREG *)(V))
#define SmallPtr(V)	((SMALLUNSGN *)(V))
#define	WordPtr(V)	((BITS16 *)(V))
#define DisplPtr(V)	((DISPREG *)(V))
#define TermPtr(V)	((Term *) (V))

/*	Abstract Type Definitions for YAPProlog			       */

typedef CELL Term;

/* ************************* Atoms  *************************************/

AbsType(Atom);

#if USE_OFFSETS
extern ADDR     AtomBase;

#endif

AbsType(Prop);

/*		 atom structure						*/
typedef struct {
  Atom            NextOfAE;	/* used	to build hash chains			*/
  Prop            PropOfAE;	/* property list for this atom			*/
  char            StrOfAE[1];	/* representation of atom as a string		*/
}

AtomEntry;

     /* constructors and destructors for Functor are defined after
	including Regs.h */

/* ************************* Functors  **********************************/

     /*		Functor	data type
      abstype Functor =	atom # int
      with MkFunctor(a,n) = ...
      and  NameOfFunctor(f) = ...
      and  ArityOfFunctor(f) = ...					*/

#define	MaxArity	    255

AbsType(Functor);


/* Support for co-routining */
#ifdef COROUTINING
#include "corout.h"
#endif

     /* constructors and destructors for Functor are defined in yatom.h */

/*********  abstract machine registers **********************************/

#include "Regs.h"

/*********  operations for atoms ****************************************/

/*    Atoms are assumed to be uniquely represented by an OFFSET and to have
    associated with them a struct of type AtomEntry
    The	two functions
	RepAtom	: Atom -> *AtomEntry
	AbsAtom	: *AtomEntry ->	Atom
    are used to encapsulate the implementation of atoms
*/

#if USE_OFFSETS
Constructor(Atom, Abs, AtomEntry *, p, Addr(p) - AtomBase)
Destructor(Atom, Rep, AtomEntry *, a, AtomBase + Unsigned(a))
#else
Constructor(Atom, Abs, AtomEntry *, p, p)
Destructor(Atom, Rep, AtomEntry *, a, a)
#endif

/* Props and Atoms are stored in chains, ending with a NIL */
#if USE_OFFSETS
#	define EndOfPAEntr(P)	( Addr(P) == AtomBase)
#else
#	define EndOfPAEntr(P)	( Addr(P) == NIL )
#endif

#define AtomName(at)	RepAtom(at)->StrOfAE

/************ variables	concerned with Error Handling *************/

#include <setjmp.h>

#if defined(SIMICS) || !HAVE_SIGSETJMP
#define sigjmp_buf jmp_buf
#define sigsetjmp(Env, Arg) setjmp(Env)
#define siglongjmp(Env, Arg) longjmp(Env, Arg)
#endif

extern sigjmp_buf    RestartEnv; /* used to restart after an abort */


/* Support for arrays */
#ifdef YAP_ARRAYS
#include "arrays.h"
#endif

/************ variables	concerned with Error Handling	*************/

/* Types of Errors */
typedef enum {
  FATAL_ERROR,
  INTERNAL_ERROR,
  PURE_ABORT,
  /* ISO_ERRORS */
  DOMAIN_ERROR_ARRAY_OVERFLOW,
  DOMAIN_ERROR_ARRAY_TYPE,
  DOMAIN_ERROR_IO_MODE,
  DOMAIN_ERROR_MUTABLE,
  DOMAIN_ERROR_NON_EMPTY_LIST,
  DOMAIN_ERROR_NOT_LESS_THAN_ZERO,
  DOMAIN_ERROR_NOT_NL,
  DOMAIN_ERROR_NOT_ZERO,
  DOMAIN_ERROR_OPERATOR_PRIORITY,
  DOMAIN_ERROR_OPERATOR_SPECIFIER,
  DOMAIN_ERROR_RADIX,
  DOMAIN_ERROR_SOURCE_SINK,
  DOMAIN_ERROR_STREAM,
  DOMAIN_ERROR_STREAM_OR_ALIAS,
  DOMAIN_ERROR_STREAM_POSITION,
  DOMAIN_ERROR_SYNTAX_ERROR_HANDLER,
  EVALUATION_ERROR_FLOAT_OVERFLOW,
  EVALUATION_ERROR_INT_OVERFLOW,
  EVALUATION_ERROR_UNDEFINED,
  EVALUATION_ERROR_UNDERFLOW,
  EVALUATION_ERROR_ZERO_DIVISOR,
  EXISTENCE_ERROR_ARRAY,
  EXISTENCE_ERROR_SOURCE_SINK,
  EXISTENCE_ERROR_STREAM,
  INSTANTIATION_ERROR,
  PERMISSION_ERROR_ACCESS_PRIVATE_PROCEDURE,
  PERMISSION_ERROR_CREATE_ARRAY,
  PERMISSION_ERROR_CREATE_OPERATOR,
  PERMISSION_ERROR_INPUT_BINARY_STREAM,
  PERMISSION_ERROR_INPUT_PAST_END_OF_STREAM,
  PERMISSION_ERROR_INPUT_STREAM,
  PERMISSION_ERROR_INPUT_TEXT_STREAM,
  PERMISSION_ERROR_MODIFY_STATIC_PROCEDURE,
  PERMISSION_ERROR_OPEN_SOURCE_SINK,
  PERMISSION_ERROR_OUTPUT_BINARY_STREAM,
  PERMISSION_ERROR_OUTPUT_STREAM,
  PERMISSION_ERROR_OUTPUT_TEXT_STREAM,
  PERMISSION_ERROR_REPOSITION_STREAM,
  REPRESENTATION_ERROR_CHARACTER_CODE,
  REPRESENTATION_ERROR_MAX_ARITY,
  SYSTEM_ERROR,
  TYPE_ERROR_ARRAY,
  TYPE_ERROR_ATOM,
  TYPE_ERROR_ATOMIC,
  TYPE_ERROR_CALLABLE,
  TYPE_ERROR_COMPOUND,
  TYPE_ERROR_DBREF,
  TYPE_ERROR_DBTERM,
  TYPE_ERROR_EVALUABLE,
  TYPE_ERROR_FLOAT,
  TYPE_ERROR_INTEGER,
  TYPE_ERROR_KEY,
  TYPE_ERROR_LIST,
  TYPE_ERROR_NUMBER,
  TYPE_ERROR_PREDICATE_INDICATOR,
  TYPE_ERROR_PTR,
  TYPE_ERROR_VARIABLE,
  UNKNOWN_ERROR
} yap_error_number;

extern char    *ErrorMessage;	/* used to pass error messages		*/
extern Term     Error_Term;	/* used to pass error terms */
extern yap_error_number  Error_TYPE;	/* used to pass the error */

#define YAP_INT_BOUNDED_FLAG	0
#define MAX_ARITY_FLAG		(YAP_INT_BOUNDED_FLAG+1)
#define INTEGER_ROUNDING_FLAG   (MAX_ARITY_FLAG+1)
#define YAP_MAX_INTEGER_FLAG	(INTEGER_ROUNDING_FLAG+1)
#define YAP_MIN_INTEGER_FLAG	(YAP_MAX_INTEGER_FLAG+1)
#define CHAR_CONVERSION_FLAG	(YAP_MIN_INTEGER_FLAG+1)
#define YAP_DOUBLE_QUOTES_FLAG	(CHAR_CONVERSION_FLAG+1)

#define STRING_AS_CHARS		0
#define STRING_AS_ATOM		2

#define NUMBER_OF_YAP_FLAGS     YAP_DOUBLE_QUOTES_FLAG+1

extern Int yap_flags[NUMBER_OF_YAP_FLAGS];

/************************  prototypes **********************************/

#include "Yapproto.h"

/***********************************************************************/

     /*
absrectype Term	= Int + Float +	Atom + Pair + Appl + Ref + Var

with AbsAppl(t)	: *CELL	-> Term
and  RepAppl(t)	: Term -> *CELL

and  AbsPair(t)	: *CELL	-> Term
and  RepPair(t)	: Term -> *CELL

and  IsIntTerm(t) = ...
and  IsAtomTerm(t) = ...
and  IsVarTerm(t) = ...
and  IsPairTerm(t) = ...
and  IsApplTerm(t) = ...
and  IsFloatTerm(t) = ...
and  IsRefTerm(t) = ...
and  IsNonVarTerm(t) = ! IsVar(t)
and  IsNumterm(t) = IsIntTerm(t) || IsFloatTerm(t)
and  IsAtomicTerm(t) = IsNumTerm(t) || IsAtomTerm(t)
and  IsPrimitiveTerm(t) = IsAtomicTerm(t) || IsRefTerm(t)

and  MkIntTerm(n) = ...
and  MkFloatTerm(f) = ...
and  MkAtomTerm(a) = ...
and  MkVarTerm(r) = ...
and  MkApplTerm(f,n,args) = ...
and  MkPairTerm(hd,tl) = ...
and  MkRefTerm(R) = ...

and  PtrOfTerm(t) : Term -> CELL * = ...
and  IntOfTerm(t) : Term -> int	= ...
and  FloatOfTerm(t) : Term -> flt = ...
and  AtomOfTerm(t) : Term -> Atom = ...
and  VarOfTerm(t) : Term -> *Term = ....
and  HeadOfTerm(t) : Term -> Term = ...
and  TailOfTerm(t) : Term -> Term = ...
and  FunctorOfTerm(t) :	Term ->	Functor	= ...
and  ArgOfTerm(i,t)  : Term -> Term= ...
and  RefOfTerm(t) : Term -> DBRef = ...

*/

/* 
   YAP can use several different tag schemes, according to the kind of
   machine we are experimenting with.
*/

#if LONG_ADDRESSES && defined(OLD_TAG_SCHEME)

#include "Tags_32bits.h"

#endif /* LONG_ADDRESSES && defined(OLD_TAG_SCHEME) */

/* AIX will by default place mmaped segments at 0x30000000. This is
	incompatible with the high tag scheme. Linux-ELF also does not like
	if you place things in the lower addresses (power to the libc people).
*/
#if (defined(_AIX) || defined(sparc)) && !defined(TABLING)
#define USE_LOW32_TAGS 1
#endif

#if LONG_ADDRESSES && SIZEOF_INT_P==4 && !defined(OLD_TAG_SCHEME) && !defined(USE_LOW32_TAGS)

#include "Tags_32Ops.h"

#endif /* LONG_ADDRESSES && !defined(OLD_TAG_SCHEME) && !defined(USE_LOW32_TAGS) */

#if LONG_ADDRESSES && SIZEOF_INT_P==4 && !defined(OLD_TAG_SCHEME) && defined(USE_LOW32_TAGS)

#include "Tags_32LowTag.h"

#endif /* LONG_ADDRESSES && !defined(OLD_TAG_SCHEME) */

#if LONG_ADDRESSES && SIZEOF_INT_P==8 && !defined(OLD_TAG_SCHEME)

#include "Tags_64bits.h"

#endif /* LONG_ADDRESSES && SIZEOF_INT_P==8 && !defined(OLD_TAG_SCHEME) */

#if !LONG_ADDRESSES

#include "Tags_24bits.h"

#endif /* !LONG_ADDRESSES */


#define	TermSize    sizeof(Term)

/* applies to unbound variables */
#if SBA
Inline0(MkVarTerm, Term, (*H = 0, H++))
#else
Inline0(MkVarTerm, Term, (*H = (CELL) H, H++))
#endif
Destructor(Term, VarOf, Term *, t, t)
Inline(IsUndefTerm, int, Term, t, *VarOfTerm(t) == t)
Inline(PtrOfTerm, CELL *, Term, t, *(CELL *)(t))

Inline(FunctorOfTerm, Functor, Term, t, *RepAppl(t))
Destructor(Term, AtomOf, Atom, t, NonTagPart(t))

Inline(MkAtomTerm, Term, Atom, a, TAGGEDA(AtomTag, a))
Inline(IsAtomTerm, int, Term, t, CHKTAG(t, AtomTag))

Inline(MkIntTerm, Term, Int, n, TAGGED(NumberTag, n))
Inline(IsIntTerm, int, Term, t, CHKTAG(t, NumberTag))

/* Needed to handle numbers:
   	these two macros are fundamental in the integer/float conversions */

#ifdef M_WILLIAMS
#define IntInBnd(X)	(TRUE)
#else
#ifdef TAGS_FAST_OPS
#define IntInBnd(X)	(Unsigned( ( (Int)(X) >> (32-7) ) + 1) <= 1)
#else
#define IntInBnd(X)	( (X) < (Int)MAX_ABS_INT && \
                          (X) > -(Int)MAX_ABS_INT-1 )
#endif
#endif
#ifdef C_PROLOG
#define FlIsInt(X)	( (X) == (Int)(X) && IntInBnd((X)) )
#else
#define FlIsInt(X)	( FALSE )
#endif



     /* There are two types of functors:

	o Special functors mark special terms
	on the heap that should be seen as constants.

	o Standard functors mark normal applications.
	*/



#include "TermExt.h"

#ifdef YAP_ARRAYS
#define IsAccessFunc(func)		((func) == FunctorAccess)

#else

#define IsAccessFunc(func)		(FALSE)

#endif /* YAP_ARRAYS */

Inline(MkIntegerTerm, Term, Int, n, IntInBnd(n) ? MkIntTerm(n) : MkLongIntTerm(n))
Inline(IsIntegerTerm, int, Term, t, IsIntTerm(t) || IsLongIntTerm(t))
Inline(IntegerOfTerm, Int, Term, t, IsIntTerm(t) ? IntOfTerm(t) : LongIntOfTerm(t))


/*************** unification routines ***********************************/

#if SBA
#include "sbaamiops.h"
#else
#include "amiops.h"
#endif

/*************** High level macros to access arguments ******************/

Inline2(ArgOfTerm, Term, int, i, Term, t, Derefa(RepAppl(t) + (i)))
Inline(HeadOfTerm, Term, Term, t, Derefa(RepPair(t)))
Inline(TailOfTerm, Term, Term, t, Derefa(RepPair(t) + 1))

Inline2(ArgOfTermCell, Term, int, i, Term, t, (CELL)(RepAppl(t) + (i)))
Inline(HeadOfTermCell, Term, Term, t, (CELL)(RepPair(t)))
Inline(TailOfTermCell, Term, Term, t, (CELL)(RepPair(t) + 1))

/*************** variables concerned with atoms	table *******************/
#define	MaxHash	    101

#ifndef YAPOR
extern Atom     HashChain[MaxHash];
#endif /* YAPOR */

/************* variables related to memory allocation *******************/
#ifndef YAPOR
extern Int      HeapUsed, HeapMax;
extern ADDR     HeapTop;
#endif /* YAPOR */
extern ADDR     HeapBase,
		     LocalBase,
		     GlobalBase,
		     TrailBase, TrailTop,
		     ForeignCodeBase, ForeignCodeTop, ForeignCodeMax;

/*************** Atoms And functors required by	the code ****************/
extern Atom     AtomNil, AtomComma, AtomAssert, AtomFail, AtomTrue,
                     AtomFalse, AtomCut, AtomRepeat, AtomLive, AtomPortray,
                     AtomDot, AtomBreak, AtomIndex, AtomFast, AtomAbol,
                     AtomCsult, AtomOtherwise, AtomPlus, AtomMinus,
		     AtomCatch, AtomThrow, AtomProfile, AtomSpyCreep,
		     AtomAlarm;
#ifdef YAP_ARRAYS
extern Atom	AtomArray, AtomArrayAccess;
extern Functor  FunctorArrayEntry;
#endif

extern Functor  FunctorOr, FunctorComma, FunctorAssert, FunctorQuery,
                     ListFunctor, FunctorNot, AltFunctorNot, FunctorCall,
                     FunctorVar, FunctorPortray, FunctorArrow,
                     FunctorCsult, FunctorBraces, FunctorGVar,
		     FunctorGAtom, FunctorGAtomic, FunctorGPrimitive,
		     FunctorGCompound, FunctorFFloat, FunctorGInteger,
		     FunctorGNumber, FunctorVBar, FunctorEq, FunctorModule;

extern Term FoundVarTerm, ReFoundVarTerm;

/************ variables	concerned with save and restore	*************/
extern int      splfild;

#ifdef EMACS

/******************** using Emacs mode ********************************/

extern int      emacs_mode;

#endif


/************ variable concerned with version number *****************/
extern char    *version_number;

/* consult stack management */

typedef union CONSULT_OBJ {
	Prop  p;
        union CONSULT_OBJ *c;
} consult_obj;

/********* common instructions codes*************************/

#define MAX_PROMPT  256

/* information that can be stored in Code Space */

typedef struct various_codes {
#ifdef YAPOR
     yamop getworkcode;
     yamop getworkcode_seq;
     yamop getworkfirsttimecode;
     int   seq_def;
#endif /* YAPOR */
#ifdef TABLING
     yamop tablecompletioncode;
     yamop tableanswerreturncode;
#endif /* TABLING */
     OPCODE failcode;
     OPCODE trustfailcode;
     struct {
	OPCODE op;
	COUNT s;
	CODEADDR l;
	CODEADDR l2;
     } env_for_yes_code; /* sla */
     OPCODE yescode;
#ifdef YAPOR
     yamop nocode;
     yamop rtrycode;
#else
     OPCODE nocode;
     /* yamop rtrycode; problem: we would have to publish yamop */
     CELL rtrycode[4];
#endif /* YAPOR */
     struct {
	OPREG arity;
	CODEADDR clause;
	Functor func;
     } clausecode;
     consult_obj *consultsp;
     consult_obj *consultbase;
     consult_obj *consultlow;
     Int   consultcapacity;
     int   profiling;
     Atom invisiblechain;
     OPCODE dummycode;
     Int maxdepth, maxlist;
     Atom atprompt;
     char prompt[MAX_PROMPT];
     yamop *retry_recorded_code,
	*retry_drecorded_code,
	*retry_c_recordedp_code;
     Int static_predicates_marked;
} all_heap_codes;

#ifdef YAPOR

#define  SEQUENTIAL_IS_DEFAULT      (heap_regs->seq_def                   )
#define  GETWORK            (&(heap_regs->getworkcode                    ))
#define  GETWORK_SEQ        (&(heap_regs->getworkcode_seq                ))
#define  GETWORK_FIRST_TIME (&(heap_regs->getworkfirsttimecode           ))
#endif /* YAPOR */
#ifdef TABLING
#define  COMPLETION         ((yamop *)&(heap_regs->tablecompletioncode   ))
#define  ANSWER_RETURN      ((yamop *)&(heap_regs->tableanswerreturncode ))
#endif /* TABLING */
#define  FAILCODE                 ((CODEADDR)&(heap_regs->failcode       ))
#define  TRUSTFAILCODE            ((CODEADDR)&(heap_regs->trustfailcode  ))
#define  TRUSTFAILCODE            ((CODEADDR)&(heap_regs->trustfailcode  ))
#define  YESCODE                  ((CODEADDR)&(heap_regs->yescode        ))
#define  NOCODE                   ((CODEADDR)&(heap_regs->nocode         ))
#define  RTRYCODE                 ((CODEADDR)&(heap_regs->rtrycode       ))
#define  DUMMYCODE                ((CODEADDR)&(heap_regs->dummycode      ))
#define  CLAUSECODE               (&(heap_regs->clausecode               ))
#define  INVISIBLECHAIN           (&(heap_regs->invisiblechain           ))
#define  max_depth                (&(heap_regs->maxdepth   	         ))
#define  max_list                 (&(heap_regs->maxlist                  ))
#define  AtPrompt                 (&(heap_regs->atprompt    	         ))
#define  Prompt                   (heap_regs->prompt    	          )
#define  PROFILING                (heap_regs->profiling 	          )
#define  RETRY_C_RECORDED_CODE    (heap_regs->retry_recorded_code         )
#define  RETRY_C_DRECORDED_CODE   (heap_regs->retry_drecorded_code        )
#define  RETRY_C_RECORDEDP_CODE   (heap_regs->retry_c_recordedp_code      )
#define  STATIC_PREDICATES_MARKED (heap_regs->static_predicates_marked    )

/* initially allow for files with up to 1024 predicates. This number
   is extended whenever needed */
#define  InitialConsultCapacity    1024
/* current consult stack */
#define  ConsultSp                (heap_regs->consultsp                   )
/* top of consult stack  */
#define  ConsultBase              (heap_regs->consultbase                 )
/* low-water mark for consult  */
#define  ConsultLow               (heap_regs->consultlow	          )
/* current maximum number of cells in consult stack */
#define  ConsultCapacity          (heap_regs->consultcapacity             )

#ifdef COROUTINING
#define  InitialSusRecCapacity    1024
#endif

extern struct various_codes *heap_regs; 

/******************* controlling the compiler ****************************/
extern int      optimizer_on;

/******************* the line for the current parse **********************/
extern int      StartLine;
extern int      StartCh;
extern int      CurFileNo;

/********************* how to write a Prolog term ***********************/

/********* Prolog may be in several modes *******************************/

#define BootMode  1		/* if booting or restoring */
#define UserMode  2		/* Normal mode */
#define CritMode  4		/* If we are meddling with the heap */
#define FullLMode 8		/* to access the hidden atoms chain */
#define AbortMode 16		/* expecting to abort */

extern int      PrologMode;

/************** Access to yap initial arguments ***************************/

extern char   **yap_args;
extern int      yap_argc;

#ifdef YAPOR
#define YAPEnterCriticalSection()                     \
	{                                             \
          if (worker_id != GLOBAL_who_locked_heap) {  \
	    LOCK(GLOBAL_lock_heap_access);            \
	    GLOBAL_who_locked_heap = worker_id;       \
	  }                                           \
          PrologMode |= CritMode;                     \
        }
#define YAPLeaveCriticalSection()                                        \
	{                                                                \
          if ((PrologMode ^= CritMode) & AbortMode) Abort((char *)NIL);  \
	  GLOBAL_who_locked_heap = MAX_WORKERS;                          \
          UNLOCK(GLOBAL_lock_heap_access);                               \
        }
#else
#define YAPEnterCriticalSection()     PrologMode |= CritMode;
#define YAPLeaveCriticalSection() \
	if((PrologMode ^= CritMode) & AbortMode) Abort((char *)NIL);
#endif /* YAPOR */

/********* sometimes we need to lookup the hidden atoms ******************/

#define NormalLookup 0
#define FullLookup   1

extern int      LookupMode;

/********* whether we should try to compile array references ******************/

#ifdef YAP_ARRAYS
extern int  compile_arrays;
#endif

/********* while debugging you may need some info ***********************/

#if DEBUG
extern int      output_msg;
#endif

#if EMACS
extern char     emacs_tmp[], emacs_tmp2[];
#endif

#if HAVE_SIGNAL
extern int      snoozing;
#endif

#if defined(YAPOR) || defined(TABLING)
#include "opt.structs.h"
#include "opt.macros.h"
#include "opt.proto.h"
#endif /* YAPOR || TABLING */

#if SBA
#include "sbaunify.h"
#endif
