/*************************************************************************
*									 *
*	 YAP Prolog 	%W% %G%						 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		c_interface.c						 *
* Last rev:	19/2/88							 *
* mods:									 *
* comments:	c_interface primitives definition 			 *
*									 *
*************************************************************************/

#define Bool int
#define flt double
#define C_INTERFACE

#include "Yap.h"
#include "Yatom.h"

Term    STD_PROTO(C_A,(int));
Term    STD_PROTO(C_MkVarTerm,(void));
Bool    STD_PROTO(C_IsVarTerm,(Term));
Bool    STD_PROTO(C_IsNonVarTerm,(Term));
Bool    STD_PROTO(C_IsIntTerm,(Term));
Bool    STD_PROTO(C_IsFloatTerm,(Term));
Bool    STD_PROTO(C_IsDbRefTerm,(Term));
Bool    STD_PROTO(C_IsAtomTerm,(Term));
Bool    STD_PROTO(C_IsPairTerm,(Term));
Bool    STD_PROTO(C_IsApplTerm,(Term));
Term    STD_PROTO(C_MkIntTerm,(Int));
Int     STD_PROTO(C_IntOfTerm,(Term));
Term    STD_PROTO(C_MkFloatTerm,(flt));
flt     STD_PROTO(C_FloatOfTerm,(Term));
Term    STD_PROTO(C_MkAtomTerm,(Atom));
Atom    STD_PROTO(C_AtomOfTerm,(Term));
Atom    STD_PROTO(LookupAtom,(char *));
char   *STD_PROTO(C_AtomName,(Atom));
Term    STD_PROTO(C_MkPairTerm,(Term,Term));
Term    STD_PROTO(C_HeadOfTerm,(Term));
Term    STD_PROTO(C_TailOfTerm,(Term));
Term    STD_PROTO(C_MkApplTerm,(Functor,unsigned int,Term *));
Functor STD_PROTO(C_FunctorOfTerm,(Term));
Term    STD_PROTO(C_ArgOfTerm,(Int,Term));
Functor STD_PROTO(C_MkFunctor,(Atom,Int));
Atom    STD_PROTO(C_NameOfFunctor,(Functor));
Int     STD_PROTO(C_ArityOfFunctor,(Functor));
void   *STD_PROTO(C_ExtraSpace,(void));
Int     STD_PROTO(C_cut_fail,(void));
Int     STD_PROTO(C_cut_succeed,(void));
Int     STD_PROTO(C_unify,(Term *,Term *));
Int     STD_PROTO(C_CallProlog,(Term));

Term
C_A(int i)
{
  return(Deref(XREGS[i]));
}

Bool 
C_IsIntTerm(Term t)
{
  return (IsIntTerm(t) || IsLongIntTerm(t));
}

Bool 
C_IsVarTerm(Term t)
{
  return (IsVarTerm(t));
}

Bool 
C_IsNonVarTerm(Term t)
{
  return (IsNonVarTerm(t));
}

Bool 
C_IsFloatTerm(Term t)
{
  return (IsFloatTerm(t));
}

Bool 
C_IsDbRefTerm(Term t)
{
  return (IsDBRefTerm(t));
}

Bool 
C_IsAtomTerm(Term t)
{
  return (IsAtomTerm(t));
}

Bool 
C_IsPairTerm(Term t)
{
  return (IsPairTerm(t));
}

Bool 
C_IsApplTerm(Term t)
{
  return (IsApplTerm(t) && !IsExtensionFunctor(FunctorOfTerm(t)));
}


Term 
C_MkIntTerm(Int n)
{
  Term I;

  restore_H();
  I = MkIntegerTerm(n);
  save_H();
  return(I);
}

Int 
C_IntOfTerm(Term t)
{
  if (!IsApplTerm(t))
    return (IntOfTerm(t));
  else
    return(LongIntOfTerm(t));
}

Term 
C_MkFloatTerm(double n)
{
  Term t;

  restore_H();
  t = MkFloatTerm(n);
  save_H();
  return(t);
}

flt 
C_FloatOfTerm(Term t)
{
  return (FloatOfTerm(t));
}

Term 
C_MkAtomTerm(Atom n)
{
  Term t;

  t = MkAtomTerm(n);
  return(t);
}

Atom 
C_AtomOfTerm(Term t)
{
  return (AtomOfTerm(t));
}


char           *
C_AtomName(Atom a)
{
  char *o;

  o = AtomName(a);
  return(o);
}

Term
C_MkVarTerm()
{
  CELL t; 
  restore_H();
  t = MkVarTerm();
  save_H();
  return(t);
}

Term
C_MkPairTerm(Term t1, Term t2)
{
  Term t; 

  restore_H();
  t = MkPairTerm(t1, t2);
  save_H();
  return(t);
}

Term 
C_HeadOfTerm(Term t)
{
  return (HeadOfTerm(t));
}

Term 
C_TailOfTerm(Term t)
{
  return (TailOfTerm(t));
}

Term
C_MkApplTerm(Functor f,unsigned int arity, Term args[])
{
  Term t; 

  restore_H();
  t = MkApplTerm(f, arity, args);
  save_H();
  return(t);
}

Functor 
C_FunctorOfTerm(Term t)
{
  return (FunctorOfTerm(t));
}


Term 
C_ArgOfTerm(Int n, Term t)
{
  return (ArgOfTerm(n, t));
}



Functor 
C_MkFunctor(Atom a, Int n)
{
  return (MkFunctor(a, n));
}

Atom 
C_NameOfFunctor(Functor f)
{
  return (NameOfFunctor(f));
}

Int 
C_ArityOfFunctor(Functor f)
{
  return (ArityOfFunctor(f));
}

void *
C_ExtraSpace()
{
  void *ptr;
  BACKUP_B();

  restore_B();
  /* find a pointer to extra space allocable */
  ptr = (void *)((CELL *)(B+1)+P->u.lds.s);
  save_B();
  RECOVER_B();
  return(ptr);
}

Int
C_cut_fail()
{
  BACKUP_B();
  restore_B();
  B = B->cp_b; /* cut_fail */
  save_B();
  RECOVER_B();
  return(FALSE);
}

Int
C_cut_succeed()
{
  BACKUP_B();
  restore_B();
  B = B->cp_b;
  save_B();    /* cut_succeed */
  RECOVER_B();
  return(TRUE);
}

Int
C_unify(Term *pt1, Term *pt2)
{
  Int out;
  BACKUP_MACHINE_REGS();

  restore_machine_regs();
  out = unify(*pt1, *pt2);
  save_machine_regs();

  RECOVER_MACHINE_REGS();
  return(out);
}

Int c_execute(CPredicate p)
{
  return((*p)());
}

Int C_CallProlog(Term t)
{
  Int out;
  BACKUP_MACHINE_REGS();

  restore_machine_regs();
  out = execute_goal(t,0);
  save_machine_regs();

  RECOVER_MACHINE_REGS();
  return(out);
}

