/*************************************************************************
*									 *
*	 YAP Prolog 	@(#)eval.h	1.2
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		eval.h							 *
* Last rev:								 *
* mods:									 *
* comments:	arithmetical functions info				 *
*									 *
*************************************************************************/

#include <stdlib.h>

/* C library used to implement floating point functions */
#include <math.h>
#if HAVE_IEEEFP_H
#include <ieeefp.h>
#endif

typedef union { Int IntMode; Float FloatMode; } fli;

#define FInt		0
#define FFloat		1

/* Needed to handle numbers:
   	these two macros are fundamental in the integer/float conversions */

#ifdef C_PROLOG
#define FlIsInt(X)	( (X) == (Int)(X) && IntInBnd((X)) )
#else
#define FlIsInt(X)	( FALSE )
#endif



#ifdef M_WILLIAMS
#define MkEvalFl(X)	MkFloatTerm(X)
#else
#define MkEvalFl(X)	( FlIsInt(X) ? MkIntTerm((Int)(X)) : MkFloatTerm(X) )
#endif


#if SIZEOF_DOUBLE == 2*SIZEOF_LONG_INT
#define EQUAL_FLOATS(pt0,pt1)    ((pt0)[1] == (pt1)[1] && (pt0)[2] == (pt1)[2])
#else
#define EQUAL_FLOATS(pt0,pt1)    ((pt0)[1] == (pt1)[1])
#endif

#define EQUAL_LONG_INTS(pt0,pt1)    ((pt0)[1] == (pt1)[1])

/* Macros used by some of the eval functions */
#define REvalInt(I)	{ reg.IntMode = (I); return(FInt); }
#define REvalFl(F)	{ reg.FloatMode = (F); return(FFloat); }
#define Iop1		reg.IntMode
#define Iop2		op2.IntMode
#define Fop1		reg.FloatMode
#define Fop2		op2.FloatMode


/* this macro, dependent on the particular implementation
	is used to interface the arguments into the C libraries */
#if	MPW
#define FL(X)		((extended)(X))
#else
#define FL(X)		((double)(X))
#endif

     /* arith.yap depends on these numbers */
#define e_plus		0
#define e_dif		1
#define e_times		2
#define	e_div		3
#define e_lshift	4
#define e_rshift	5
#define e_and		6
#define	e_or		7
#define	e_uminus	8
#define	e_not		9
#define e_exp		10
#define e_log		11
#define	e_log10		12
#define e_sqrt		13
#define e_sin		14
#define e_cos		15
#define e_tan		16
#define e_asin		17
#define e_acos		18
#define e_atan		19
#define e_mod		20
#define e_idiv		21
#define e_power		22
#define e_uplus		23
#define e_integer	24
#define e_float		25
#define e_abs		26
#define e_gcd		27
#define e_gcdmult	28
#define e_min		29
#define e_max		30
#define e_msb		31
#define e_floor		32
#define e_round		33
#define e_ceiling	34
#define e_truncate	35
#define e_sinh		36
#define e_cosh		37
#define e_tanh		38
#define e_asinh		39
#define e_acosh		40
#define e_atanh		41
#define e_atan2		42
#define e_fmodf		43
#define e_imodf		44

extern fli reg;

typedef struct INITEVAL {
	char           *OpName;
	short int       arity;
        unsigned short int eno;
}               InitEntry;

extern InitEntry InitTab[];

Int	STD_PROTO(cputime,(void));
double	STD_PROTO(yap_random,(void));
Int     STD_PROTO(gcd,(Int,Int));
Int     STD_PROTO(gcdmult,(Int,Int,Int *));
Int     STD_PROTO(msb,(Int));
Term	STD_PROTO(p_succ,(Term));
Term	STD_PROTO(p_predc,(Term));
Term	STD_PROTO(p_plus,(Term,Term));
Term	STD_PROTO(p_minus,(Term,Term));
Term	STD_PROTO(p_times,(Term,Term));
Term	STD_PROTO(p_div,(Term,Term));
Term	STD_PROTO(p_and,(Term,Term));
Term	STD_PROTO(p_or,(Term,Term));
Term	STD_PROTO(p_sll,(Term,Term));
Term	STD_PROTO(p_slr,(Term,Term));



