/**************************************************************************
*									 *
* File:		progol.h						 *
* Last rev:	27/4/96							 *
* mods:									 *
* comments:	interface to progol predicates				 *
*									 *
*************************************************************************/

#include  "Yap.h"
#include  "Yatom.h"
#include  "eval.h"

#include <sys/socket.h>
#include <sys/time.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <sys/types.h>
#include <limits.h>
#include <unistd.h>
#ifdef SOLARIS
#include <sys/file.h>
#include <strings.h>
#endif


#if SIZEOF_INT_P==4
#define ALIGN			3
#define ALIGNMASK		0xfffffff8  /* was 0xfffffffc */
#else
#define ALIGN			7
#define ALIGNMASK		0xfffffff8L
#endif	/* ALIGN_LONGS */

#define AdjustSize(X)	(((unsigned int)X) & ALIGNMASK)

#define SPLIT_ACC

/*************************
   #define
   **********************/

#define MAX_CELLS    1000000  /* was working with 1000000  */
#define MYHEAP_SIZE  2000000  /* was working with 2000000 */

#define MAX_POSITIVES   10000
#define MAX_NEGATIVES   10000
#define MAX_ARGS           20

#define 	CR	10
#define 	SP	32

#define NOISE    0
#define ACCURACY 1

#define         RAND            drand48()
#define         SRAND(x)        srand48((x))

#define ContType        float
#define Instance        AttrVal *

#define P_NIL           (Term)MkAtomTerm(LookupAtom("[]"))
#define P_EMPTY(l)      (IsAtomTerm((l)))
#define P_HEAD(l)       (Term)HeadOfTerm((l))
#define P_TAIL(l)       (Term)TailOfTerm((l))
#define P_ARG(arg)      (Term)Deref(arg)
#define P_LONG(t)       (LONG)IntOfTerm(t)
#define P_INT(t)        (int)IntegerOfTerm(t)
#define P_CHAR(t)       (CHAR)IntegerOfTerm(t)
#define P_DOUBLE(t)     (double)FloatOfTerm(t)
#define P_FLOAT(t)      (FLOAT)FloatOfTerm(t)
#define P_ATOM(str)     (ATOM)LookupAtom((str))
#define	P_NUM(t,n)	(n=(double)(P_ISTNUM(t)?P_NUMBER(ArgOfTerm(1,t)):P_NUMBER(t)))
#define	P_NUMBER(t)	(P_ISDOUBLE(t)?P_DOUBLE(t):P_INT(t))
#define P_MKINT(i)      (Term)MkIntegerTerm((i))
#define P_MKLONG(i)     (Term)MkIntegerTerm((i))
#define P_MKDOUBLE(i)   (Term)MkFloatTerm((i))
#define P_MKFLOAT(i)    (Term)MkFloatTerm((i))
#define P_MKATOM(a)     (Term)MkAtomTerm((a))
#define P_MKLIST(h,t)   (Term)MkPairTerm((h),(t))
#define	P_ISINT(t)	IsIntegerTerm((t))
#define	P_ISDOUBLE(t)	IsFloatTerm((t))
#define	P_ISTNUM(t)	(IsApplTerm((t)) && (ArityOfFunctor(FunctorOfTerm((t))) == 1) && (P_ISINT(ArgOfTerm(1,t)) || P_ISDOUBLE(ArgOfTerm(1,t))))
#define	P_ISNUM(t)	(P_ISINT(t) || P_ISDOUBLE(t) || P_ISTNUM(t))
#define P_BIND(arg,t)   unify((arg),(t))
#define P_ARRAY_TO_LIST(a,n)    array_to_pTerm(a,0,n)

/* macros */

#define equalf(n1, n2) (fabs(n1-n2) < 1e-8)

#define myfree (freeblock = &(mymemory[MAX_CELLS]))


/************************
Socket stuff
*************************/
/* protocol control identifiers (pci) */



#define CNR   1
#define CNI   2
#define CRS   3
#define CNC   4

#define ATR   5
#define ATI   6
#define ARS   7
#define ATC   8

#define RLR   9
#define RLI   10
#define RRS   11
#define RLC   12

#define UPR   13
#define UPI   14
#define URS   15
#define UPC   16

#define RER   17
/* #define REI and RES 18, 19 */
#define REC   20

#define ABR   21
#define ABI   22
#define ABC   23

/* pci for packets between HLCM-source or sink */
#define CON_R 26
#define CON_I 27
#define CON_RS 28
#define CON_C 29
#define UPD   30
#define UPD_I 31
#define UPD_RS 32
#define UPD_C 33
#define REL   34
#define REL_C 35
#define REL_I 36
#define REL_N 37
#define REL_RS 38
#define REA   39
#define REA_C 40
#define ATA   41
#define ATA_I 42
#define ATA_RP 43
#define ATA_RN 44
#define ACC   45
#define ABA   46

/* trigger constants */
#define UT_SOURCE      1   /* UPDATING : source initiated */
#define UT_SINK        2   /*            sink initiated */
#define UT_BOTH        3   /*            source & sink */
#define RT_UNCOND      1   /* READING  : provider initiated unconditional */
#define RT_CHANGE      2   /*            provider initiated upon change */
#define RT_SINK        3   /*            sink initiated */

/* result constants */
#define ACCEPT         0   /* accepted by ss-provider & ss-user */
#define REJECT         1   /* rejected by ss-provider or ss-user */
#define REJECT_UNSPEC  2   /* reason not specified */
#define REJECT_MEM     3   /* memory limitation */
#define REJECT_CONG    4   /* ss-provider congestion */
#define REJECT_SSAP    5   /* SSAP address unknown */

/* bitset type definition */
#define MAXBITS 64

#define INTBITS (8*sizeof(unsigned))

#define NUMINTS (MAXBITS/INTBITS)
#define SIZE ( (NUMINTS*INTBITS==MAXBITS) ? NUMINTS : NUMINTS+1 )
typedef unsigned bitset[SIZE] ;

/* limitations */
#define BUFSIZE     2048   /* length of TCP buffers */

/* Definition of 'user' list element, used by the CA to hold 
   useful data about connection parameters for every ss-user */
typedef struct _user {
   int sock ;
   struct sockaddr_in addr;
   char *path;
   int trig ;
   bitset dom ;
   struct _user *next ;
} USER ;

/*************************
   DATA TYPES 
   **********************/

enum AttrTypes {nominal,continuous};

typedef long int LONG;

union NormalVal {
  int               d;
  ContType          c;
};

typedef struct attr_val {
  union NormalVal   val;
} AttrVal;

typedef struct attr_info {
  enum AttrTypes    type_attr;
  int      nr_vals;
  void              *vals;
} AttrInfo;

typedef struct domain_info {
  char              *root_name;
  int               nr_attrs;
  AttrInfo          *attrs;
} DomainInfo;                                


typedef struct _ds{
  DomainInfo *domain;
  AttrVal    **instances;
  int        nr_exs;
}Dataset;

typedef struct ilist{
  int        value;
  struct ilist *next;
} IList;


/*******************
  Prototypes
  *****************/

STD_PROTO(static int l_lteq, (void));
STD_PROTO(static int l_gteq, (void));

STD_PROTO(static int l_interval, (void));
STD_PROTO(static int l_outsideInterval, (void));

STD_PROTO(static int c_lindiscrim2, (void));
STD_PROTO(static int c_lindiscrim3, (void));

STD_PROTO(static int l_regress1, (void));
STD_PROTO(static int l_regress2, (void));
STD_PROTO(static int l_regressn, (void));

STD_PROTO(static int negCoverage, (void));
STD_PROTO(static int posCoverage, (void));

STD_PROTO(static int create_socket, (void));


static unsigned char *mymalloc(int);
static unsigned char *mycalloc(int);
int discrim2alg(double *, double *, int *);
int discrim3alg(double *, double *, double *, int *);
DomainInfo *MyReadDomain2(void);
DomainInfo *MyReadDomain3(void);
AttrVal **ReadInstances2(double, int *, short int *);
AttrVal **ReadInstances3(double, int *, short int *);
AttrVal *mkInstance2(double, double, int, int);
AttrVal *mkInstance3(double, double, double, int, int);


/*******************
  Global Variables
  ******************/


static unsigned char mymemory[MAX_CELLS*sizeof(CELL)];
static CELL *my_buffer = NULL, *initial_buffer = NULL;;
static unsigned char *freeblock = &(mymemory[MAX_CELLS*sizeof(CELL)]);

static int noiseType = NOISE, noise = 0, bestCover = 1;
static float accuracy = 1.0;

/*** sockets **/
USER *ag;
int sock;
struct sockaddr_in  name;
char buf[BUFSIZE];
int SSAP;
struct sockaddr_un sname;
struct sockaddr_un hlname;


/*************** ARRAYs for pos and negs examples **************/

#ifdef DEBUG
static Term *positives, *negatives;
#else
static Term positives[MAX_POSITIVES], negatives[MAX_NEGATIVES];
#endif
int  numberPositives = 0, numberNegatives = 0;


/******************************
  Memory management routines
******************************/

static unsigned char *mymalloc(size)
{
  if(freeblock - size < mymemory) {
    printf("foreign BLOCKS exceeded\n");
    fflush(stdout);
    return(malloc(size));
  }

  freeblock -= size;
#ifdef SOLARIS
  freeblock = (unsigned char *) AdjustSize(freeblock);
#endif

  return(freeblock);
}

static unsigned char *mycalloc(size)
{
  if(freeblock - size < mymemory) {
    printf("foreign BLOCKS exceeded\n");
    fflush(stdout);
    return(calloc(size,1));
  }

  freeblock -= size;
#ifdef SOLARIS
  freeblock = (unsigned char *) AdjustSize(freeblock);
#endif
  bzero(freeblock, size);

  return(freeblock);
}


/***************************/

Term myCopyTerm(Term term)
{
  Term args[MAX_ARGS];
  int i;


  if(IsIntTerm(term)) return(term);
  if(IsFloatTerm(term)) return(MkFloatTerm(FloatOfTerm(term)));
  if(IsLongIntTerm(term)) return(MkLongIntTerm(LongIntOfTerm(term)));
  if(IsAtomTerm(term)) return(MkAtomTerm(LookupAtom(AtomName(AtomOfTerm(term)))));
  if(IsPairTerm(term))
     return(MkPairTerm(myCopyTerm(HeadOfTerm(term)),
		       myCopyTerm(TailOfTerm(term))));
  if(IsApplTerm(term)){
    /*printf("APPL term functor = %s arity = %d\n", AtomName(NameOfFunctor(FunctorOfTerm(term))), ArityOfFunctor(FunctorOfTerm(term)));
getchar();
*/
    for(i = 1; i <= ArityOfFunctor(FunctorOfTerm(term)); i++)
      args[i-1] = myCopyTerm(ArgOfTerm(i, term));
     return(MkApplTerm(MkFunctor(NameOfFunctor(FunctorOfTerm(term)),
				 ArityOfFunctor(FunctorOfTerm(term))),
		       ArityOfFunctor(FunctorOfTerm(term)),
		       args));
  }
  return((Term) NULL);
}

/*******************************
	set Positive Example
********************************/

static int setPositiveExample(void)
{
  int index;
  Term args[2];
  CELL *heapBackup;


  if(IsVarTerm(Deref(ARG2)) || ! IsIntegerTerm(Deref(ARG1))) return(0);
  
  index = IntOfTerm(Deref(ARG1));
  if(index < 1) return(0);

  /* backup real heap position */

  heapBackup = H;
  if(! my_buffer) {
    while ((my_buffer = (CELL *)AllocCodeSpace(MYHEAP_SIZE)) == NULL) {
      /* make sure we have enough space */
      if (!growheap()) {
	Abort("[ SYSTEM ERROR: YAP failed to reserve space in growheap ]\n");
	return(FALSE);
      }
      heapBackup = H;
    }
    initial_buffer = my_buffer;
  }
  H = my_buffer;

  args[1] = MkIntTerm(10);
  args[0] = myCopyTerm(Deref(ARG2));

  positives[index-1] = MkApplTerm(MkFunctor(LookupAtom("depth_bound_call"),2),2,args);

  if(H > initial_buffer + MYHEAP_SIZE){
    printf( "Need more space for the positive examples\n");
    exit(1);
  }
  my_buffer = H;
  H = heapBackup;

  return(1);
}

/***********************/

static int posExample(void)
{
  int index;
  Term term;
  Term t1 = Deref(ARG1);

  if(IsVarTerm(t1) || ! IsIntegerTerm(t1)) return(0);
  index = IntOfTerm(Deref(t1));
  if(index > 0) index--;
  else index = -index - 1;
  if(index > numberPositives || index < 0 ) return(0);
  
  term = ArgOfTerm(1, positives[index]);
  return(unify(term, ARG2));
}


/*******************************
	set Negative Example
********************************/

static int setNegativeExample(void)
{
  int index;
  Term args[2];
  CELL *heapBackup;


  if(IsVarTerm(Deref(ARG2)) || ! IsIntegerTerm(Deref(ARG1))) return(0);
  
  index = IntOfTerm(Deref(ARG1));
  if(index < 1) return(0);

  /* backup real heap position */

  heapBackup = H;
  if(! my_buffer) {
    while ((my_buffer = (CELL *)AllocCodeSpace(MYHEAP_SIZE)) == NULL) {
      /* make sure we have enough space */
      if (!growheap()) {
	Abort("[ SYSTEM ERROR: YAP failed to reserve space in growheap ]\n");
	return(FALSE);
      }
      heapBackup = H;
    }
    initial_buffer = my_buffer;
  }
  H = my_buffer;

  args[1] = MkIntTerm(10);
  args[0] = myCopyTerm(Deref(ARG2));

  negatives[index-1] = MkApplTerm(MkFunctor(LookupAtom("depth_bound_call"),2),2,args);

  if(H > initial_buffer + MYHEAP_SIZE){
    printf( "Need more space for the negative examples\n");
    exit(1);
  }

  my_buffer = H;
  H = heapBackup;

  return(1);
}

/***********************/

static int negExample(void)
{
  int index;
  Term term;

  if(! IsIntegerTerm(Deref(ARG1))) return(0);
  index = IntegerOfTerm(Deref(ARG1));
  if(index > 0) index--;
  else index = -index - 1;
  if(index > numberNegatives || index < 0 ) return(0);

  term = ArgOfTerm(1, negatives[index]);
  return(unify(term, ARG2));
}

/***********************/

static int setNumberOfPositives(void)
{
  if(IsVarTerm(Deref(ARG1))) return(0);

  numberPositives = IntOfTerm(Deref(ARG1));

  return(1);
}

/***********************/

static int numberOfPositives(void)
{
  Term t = MkIntegerTerm(numberPositives);

  return(unify(t, ARG1));
}

/***********************/

static int setNumberOfNegatives(void)
{
  if(IsVarTerm(Deref(ARG1))) return(0);

  numberNegatives = IntOfTerm(Deref(ARG1));

  return(1);
}

/***********************/

static int numberOfNegatives(void)
{
  Term t = MkIntegerTerm(numberNegatives);

  return(unify(t, ARG1));
}

/*******************************
	set Noise
********************************/

static int setNoise(void)
{
  if(! IsAtomTerm(Deref(ARG1))) return(0);
  
  if(strcmp("n", AtomName(AtomOfTerm(ARG1))) == 0){
    noiseType = NOISE;
    noise= IntOfTerm(Deref(ARG2));
    return(1);
  }
  else if(strcmp("acc", AtomName(AtomOfTerm(ARG1))) == 0){
    noiseType = ACCURACY;
    accuracy = FloatOfTerm(Deref(ARG2));
    return(1);
  }

  return(0);
}

/*******************************
	set Best Cover
********************************/

static int setBestCover(void)
{
  if(! IsIntTerm(Deref(ARG1))) return(0);
  
  bestCover = IntOfTerm(Deref(ARG1));

  return(1);
}


/*****************************************************************************/
/*                          COVERAGE PREDICATES                              */
/*****************************************************************************/


/*******************************
	posCoveage()
********************************/

static int posCoverage(void)
{
  Term inputList, outputList, dum;
  int  max, first;
  register IList *list = NULL, *aux = NULL, *outlist = NULL, *auxout = NULL;
  register int example, counter = 0, threshold;


  inputList = Deref(ARG1);
  outputList = MkAtomTerm(LookupAtom("[]"));
  if(P_EMPTY(inputList) || ! IsIntegerTerm(Deref(ARG2))){
    dum = MkIntegerTerm(counter);
    if(! unify(dum, ARG4)) return(FALSE);
    return(unify(outputList, ARG3));
  }

  threshold = IntOfTerm(Deref(ARG2));
  counter = 0;

  /* backup input list */

  while(! P_EMPTY(inputList)){
    if(!list){
      list = (IList *) mymalloc(sizeof(IList));
      list->value = IntegerOfTerm(P_HEAD(inputList));
      aux = list;
    }
    else{
      aux->next = (IList *) mymalloc(sizeof(IList));
      aux->next->value = IntegerOfTerm(P_HEAD(inputList));
      aux = aux->next;
    }
    inputList = P_TAIL(inputList);
  } /* END of while(! P_EMPTY(inputList) ..*/

  aux->next = NULL;
  aux = list;

  /* main cycle */

  while(aux && threshold > -1){

    example = (int) (aux->value / 65536);
    max = aux->value % 65536 + example;
    first = 0;

    for(;example <= max; example++){

      /* construct call */

      if(execute_goal(positives[example-1],4)){
	counter++;
	if(! first) first = example;
      }
      else if(first){
	threshold--;
	auxout = (IList *) mymalloc(sizeof(IList));
	auxout->value = first * 65536 + (example - 1) - first;
	auxout->next = outlist;
	outlist = auxout;
	first = 0;
      }
      else threshold--;
    } /* END of for( .... */

    if(threshold > -1 && first){ /* non empty interval */
      auxout = (IList *) mymalloc(sizeof(IList));
      auxout->value = first * 65536 + max - first;
      auxout->next = outlist;
      outlist = auxout;
    }

    aux = aux->next;
  }

  if(counter >= bestCover){
    /* reverse the list for the correct sequence */

    if(threshold > -1)
      while(outlist){
	outputList = MkPairTerm(MkIntegerTerm(outlist->value), outputList);
	outlist = outlist->next;
      }
  }
  myfree;
  dum = MkIntegerTerm(counter);

  if(! unify(dum, ARG4)) return(FALSE);

  if(counter >= bestCover) return(unify(outputList, ARG3));
  return(TRUE);
}

/*******************************
	negCoverage()
********************************/

static int negCoverage(void)
{
  Term inputList, outputList, dum;
  int  max, first;
  register IList *list = NULL, *aux = NULL, *outlist = NULL, *auxout = NULL;
  register int example, threshold, counter = 0;

  inputList = Deref(ARG1);
  outputList = MkAtomTerm(LookupAtom("[]"));
  if(P_EMPTY(inputList) || ! IsIntegerTerm(Deref(ARG2))){
    example = 0;
    dum = MkIntegerTerm(example);
    if(! unify(dum, ARG4)) return(FALSE);
    return(unify(outputList, ARG3));
  }
  threshold = IntOfTerm(Deref(ARG2));

  /* backup input list */

  while(! P_EMPTY(inputList)){
    if(!list){
      list = (IList *) mymalloc(sizeof(IList));
      list->value = IntegerOfTerm(P_HEAD(inputList));
      aux = list;
    }
    else{
      aux->next = (IList *) mymalloc(sizeof(IList));
      aux->next->value = IntegerOfTerm(P_HEAD(inputList));
      aux = aux->next;
    }
    inputList = P_TAIL(inputList);
  } /* END of while(! P_EMPTY(inputList) ..*/

  aux->next = NULL;
  aux = list;

  /* main cycle */

  while(aux && threshold > -1){

    example = (int) (aux->value / 65536);
    max = aux->value % 65536 + example;
    first = 0;

    for(;example <= max && threshold > -1; example++){

      /* construct call */

      if(execute_goal(negatives[example-1],4)){
	threshold--;
	counter++;
	if(! first) first = example;
      }
      else if(first){
	auxout = (IList *) mymalloc(sizeof(IList));
	auxout->value = first * 65536 + (example - 1) - first;
	auxout->next = outlist;
	outlist = auxout;
	first = 0;
      }
    } /* END of for( .... */

    if(first){ /* non empty interval */
	auxout = (IList *) mymalloc(sizeof(IList));
	auxout->value = first * 65536 + max - first;
	auxout->next = outlist;
	outlist = auxout;
    }

    aux = aux->next;
  } /* END of   while(aux && threshold > -1){  */

  /* in case of lazy evaluation, threshold < 0 */

  if(threshold < 0 && aux)
    while(aux){
      auxout = aux;
      aux = aux->next;
      auxout->next = outlist;
      outlist = auxout;
    }

  /* reverse the list for the correct sequence */

  while(outlist){
    outputList = MkPairTerm(MkIntegerTerm(outlist->value), outputList);
    outlist = outlist->next;
  }

  myfree;

  /*dum = MkIntegerTerm(examples);*/
  dum = MkIntegerTerm(counter);

  if(! unify(dum, ARG4)) return(FALSE);

  return(unify(outputList, ARG3));
}

/* ############################################################ */
/*		User defined C library functions		*/
/*		--------------------------------		*/
/* ############################################################ */


typedef struct vlist *VLIST;
typedef struct vlist {
	double val;
        unsigned int positives, negatives;
	struct vlist *next;
} vlist;




static VLIST
insertPositives(Term list, VLIST *values, int *number)
{
  register VLIST aux = NULL, aux1 = NULL;
  double val;
  Term term;

  *number = 0;
  while(! P_EMPTY(list)){

    term = P_HEAD(list);
    if(P_ISINT(term))         val = (double) P_INT(term);
    else if(P_ISDOUBLE(term)) val = P_DOUBLE(term);
    else return (*values);

    (*number)++;

    /* insert new positive term */

    if(! *values){
      *values = (vlist *) mymalloc(sizeof(vlist));
      (*values)->val = val;
      (*values)->positives = 1;
      (*values)->negatives = 0;
      (*values)->next = NULL;
    }
    else if(equalf(val, (*values)->val)) /* value exhists as first value */
      (*values)->positives++;
    else if(val < (*values)->val){ /* insert up front */
      aux = (vlist *) mymalloc(sizeof(vlist));
      aux->val = val;
      aux->positives = 1;
      aux->negatives = 0;
      aux->next = (*values);
      (*values) = aux;
    }
    else {
      aux = (*values);
      while(aux->next && ! equalf(val, aux->next->val) &&
	    val > aux->next->val) aux = aux->next;
      if(! aux->next){ /* insert at the end of the list */
	aux->next = (vlist *) mymalloc(sizeof(vlist));
	aux->next->val = val;
	aux->next->positives = 1;
	aux->next->negatives = 0;
	aux->next->next = NULL;
      }
      else if (equalf(val, aux->next->val)) /* = to exhisting value */
	aux->next->positives++;
      else{
	aux1 = (vlist *) mymalloc(sizeof(vlist));
	aux1->val = val;
	aux1->positives = 1;
	aux1->negatives = 0;
	aux1->next = aux->next;
	aux->next = aux1;
      }
    }
    
    list = P_TAIL(list); /* update term list */
  } /* END of while ! P_EMPTY(list) */

  return *values;
}


static VLIST
insertNegatives(Term list, VLIST *values, int *number)
{
  register VLIST aux = NULL, aux1 = NULL;
  double val;
  Term term;

  /* Assumes that the list "values" already has the positive samples and is,
     therefore, non-empty */

  *number = 0;
  while(! P_EMPTY(list)){

    term = P_HEAD(list);
    if(P_ISINT(term))         val = (double) P_INT(term);
    else if(P_ISDOUBLE(term)) val = P_DOUBLE(term);
    else return (*values);

    (*number)++; 
    /* insert new negative term */

    if(equalf(val, (*values)->val)) /* value exhists as first value */
      (*values)->negatives++;
    else if(val < (*values)->val && (*values)->positives > 0){ /* insert up front */
      aux = (vlist *) mymalloc(sizeof(vlist));
      aux->val = val;
      aux->positives = 0;
      aux->negatives = 1;
      aux->next = (*values);
      (*values) = aux;
    }
    else if(val < (*values)->val)
      (*values)->negatives++;
    else if(val > (*values)->val && (*values)->positives == 0 &&
	    val < (*values)->next->val){ /* insert up front */
      (*values)->val = val;
      (*values)->negatives++;
    }
    else {
      aux = (*values);
      while(TRUE){

	if(! aux->next){
	  aux->next = (vlist *) mymalloc(sizeof(vlist));
	  aux->next->val = val;
	  aux->next->positives = 0;
	  aux->next->negatives = 1;
	  aux->next->next = NULL;
	  break;
	}

	/* update exhisting block */
	if (equalf(val, aux->next->val) ||
	    (val < aux->next->val && ! aux->next->positives)){ /* = to exhisting value */
	  aux->next->negatives++;
	  break;
	}

	/* insert before next block with positives (create a neg only block)*/

	if(val < aux->next->val && aux->next->positives > 0){
	  aux1 = (vlist *) mymalloc(sizeof(vlist));
	  aux1->val = val;
	  aux1->positives = 0;
	  aux1->negatives = 1;
	  aux1->next = aux->next;
	  aux->next = aux1;
	  break;
	}

	/* last neg only block */
	if(! aux->next->next){
	  if(val > aux->next->val && aux->next->positives == 0){
	    aux->next->negatives++;
	    break;
	  }

	  aux->next->next = (vlist *) mymalloc(sizeof(vlist));
	  aux->next->next->val = val;
	  aux->next->next->positives = 0;
	  aux->next->next->negatives = 1;
	  aux->next->next->next = NULL;
	  break;
	}

	/* bigger than next but smaller thant next->next */

	if(val < aux->next->next->val && ! aux->next->positives){
	  aux->next->negatives++;
	  break;
	}

	aux = aux->next;
      } /* END of while(TRUE) */
    }
    
    list = P_TAIL(list); /* update term list */
  } /* END of while ! P_EMPTY(list) */

  return *values;
}

/* CONSERVATIVE constant determination */

#ifdef SPLIT_CONSERVATIVE


/*#define L_BETTER(np,nn,mp,mn)	((nn < mn) || (nn==mn && np > mp))*/

#define L_BETTER(np,nn,mp,mn)		(np - nn > mp - mn)

static int l_lteq(void)
{
  Term term, list;
  VLIST values = NULL;
  register VLIST pointer;
  int maxp = 0, minn = 0, nplteq, nnlteq, dumInt;
  double threshold = 1e10;

  if(IsVarTerm(P_ARG(ARG1))) return(FALSE);

  /* list of bindings from pos examples */

  list = P_HEAD(P_ARG(ARG1));

  if (P_EMPTY(list)) return (FALSE);

  values = insertPositives(list, &values, &dumInt);
  threshold = values->val;

  /* list of bindings from neg examples */
	
  list = P_HEAD(P_TAIL(P_ARG(ARG1)));

  values = insertNegatives(list, &values, &dumInt);

  nplteq = nnlteq = 0;
  minn = 1000; 
  maxp = 0;
  for (pointer=values; pointer; pointer = pointer->next) {
    nplteq += pointer->positives;
    nnlteq += pointer->negatives;

    if(nplteq >= bestCover &&
       pointer->positives && L_BETTER(nplteq, nnlteq, maxp, minn)) {
      minn = nnlteq;
      maxp = nplteq;
      threshold = pointer->val;
    }
  }

  myfree;

  if(nplteq < bestCover) return(FALSE);

  if(fabs(threshold - floor(threshold)) < 1e-8)
    term = P_MKINT((int)((float)threshold));
  else term = P_MKDOUBLE(threshold);
  
  if(!unify(ARG2,term)) return (FALSE);
  return (TRUE);
}

/************************/
	
/*#define	G_BETTER(np,nn,mp,mn) 	((nn < mn) || (nn==mn && np > mp))*/

#define	G_BETTER(np,nn,mp,mn) 	(np - nn > mp - mn)

static int
l_gteq(void)
{
  Term term, list;
  VLIST values = NULL, reverse = NULL;
  register VLIST pointer, aux;
  int maxp = 0, minn = 0, npgteq, nngteq, dumInt;
  double threshold;


  if(IsVarTerm(P_ARG(ARG1))) return(FALSE);

  /* list of bindings from pos examples */

  list = P_HEAD(P_ARG(ARG1));

  if (P_EMPTY(list)) return (FALSE);

  values = insertPositives(list, &values, &dumInt);
  threshold = values->val;

  /* list of bindings from neg examples */
	
  list = P_HEAD(P_TAIL(P_ARG(ARG1)));

  values = insertNegatives(list, &values, &dumInt);

  /* reverse list */

  reverse = NULL;
  while(values){
    aux=values;
    values = values->next;
    aux->next = reverse;
    reverse = aux;
  }

  npgteq = nngteq = 0;
  minn = 10000; 
  maxp = 0;
  for(pointer=reverse; pointer; pointer = pointer->next) {
    npgteq += pointer->positives;
    nngteq += pointer->negatives;

    if(npgteq >= bestCover && pointer->positives && 
       G_BETTER(npgteq, nngteq, maxp, minn)) {
      minn = nngteq;
      maxp = npgteq;
      threshold = pointer->val;
    }
  }

  myfree;

  if(fabs(threshold - floor(threshold)) < 1e-8)
    term = P_MKINT((int)((float)threshold));
  else term = P_MKDOUBLE(threshold);

  if(!P_BIND(ARG2,term)) return (FALSE);
  return (TRUE);
}

static int
l_interval(void)
{
  Term term, list;
  VLIST values = NULL;
  VLIST pointer = NULL, aux;
  int maxpos, maxnegs, pos, negs, dumInt;
  double lowBound, upBound;


  if(IsVarTerm(P_ARG(ARG1))) return(FALSE);

  /* list of bindings from pos examples */

  list = P_HEAD(P_ARG(ARG1));

  if (P_EMPTY(list)) return (FALSE);

  values = insertPositives(list, &values, &dumInt);
  lowBound = upBound = values->val;

  /* list of bindings from neg examples */
	
  list = P_HEAD(P_TAIL(P_ARG(ARG1)));

  values = insertNegatives(list, &values, &dumInt);

  maxpos = 0;
  maxnegs = 1000;

  for(pointer = values; pointer; pointer = pointer->next) {
    if(! pointer->positives) continue;
    
    pos = negs = 0;

    for(aux=pointer; aux; aux = aux->next) {

      pos += aux->positives;
      negs += aux->negatives;

      if(pos >= bestCover && aux->positives &&
	 G_BETTER(pos, negs, maxpos, maxnegs)) {
	maxpos = pos;
	maxnegs = negs;
	lowBound = pointer->val;
	upBound = aux->val;
      }
    } /* END of  for(aux=pointer; aux->next; aux = aux->next) { */
  }

  myfree;

  if(maxpos < bestCover) return(FALSE);

  if(fabs(lowBound - floor(lowBound)) < 1e-8 &&
     fabs(upBound - floor(upBound)) < 1e-8){
    term = P_MKINT((int)((float)lowBound));
    if(!P_BIND(ARG2,term)) return (FALSE);
    term = P_MKINT((int)((float)upBound));
    if(!P_BIND(ARG3,term)) return (FALSE);
  }
  else {
    term = P_MKDOUBLE(lowBound);
    if(!P_BIND(ARG2,term)) return (FALSE);
    term = P_MKDOUBLE(upBound);
    if(!P_BIND(ARG3,term)) return (FALSE);
  }

  return (TRUE);
}

static int l_outsideInterval(void)
{
  Term term, list;
  VLIST values = NULL;
  VLIST aux, pointer = NULL;
  int maxpos, maxnegs, pos, negs, allpos, allnegs;
  double lowBound, upBound;


  if(IsVarTerm(P_ARG(ARG1))) return(FALSE);

  /* list of bindings from pos examples */

  list = P_HEAD(P_ARG(ARG1));

  if (P_EMPTY(list)) return (FALSE);

  values = insertPositives(list, &values, &allpos);
  lowBound = upBound = values->val;

  /* list of bindings from neg examples */
	
  list = P_HEAD(P_TAIL(P_ARG(ARG1)));

  values = insertNegatives(list, &values, &allnegs);

  /* Main cycle */

  maxpos = 0;
  maxnegs = 10000;

  pos = negs = 0;

  for(pointer = values; pointer; pointer = pointer->next) {

    if(! pointer->positives) continue;

    pos = allpos;
    negs = allnegs;

    for(aux=pointer->next; aux; aux = aux->next) {

      if(pos >= bestCover && aux->positives &&
	 G_BETTER(pos, negs, maxpos, maxnegs)) {
	maxpos = pos;
	maxnegs = negs;
	lowBound = pointer->val;
	upBound = aux->val;
      }
      pos -= aux->positives;
      negs -= aux->negatives;
    } /* END of  for(aux=pointer; aux->next; aux = aux->next) { */
    
  } /* END of 	for(pointer=values; pointer.... */

  myfree;

  if(maxpos < bestCover) return(FALSE);

  if(fabs(lowBound - floor(lowBound)) < 1e-8 &&
     fabs(upBound - floor(upBound)) < 1e-8){
    term = P_MKINT((int)((float)lowBound));
    if(!P_BIND(ARG2,term)) return (FALSE);
    term = P_MKINT((int)((float)upBound));
    if(!P_BIND(ARG3,term)) return (FALSE);
  }
  else {
    term = P_MKDOUBLE(lowBound);
    if(!P_BIND(ARG2,term)) return (FALSE);
    term = P_MKDOUBLE(upBound);
    if(!P_BIND(ARG3,term)) return (FALSE);
  }

  return (TRUE);
}

#endif

#ifdef SPLIT_ACC

#ifdef SPLIT_ACC         /* ACCURACY constant determination  */
#define EVAL_FUNC(pppn, nnnp, pp, pn, nn, np)        (pp-pn)
#endif

#ifdef SPLIT_GAIN       /* C4.5 GAIN constant determination */
#define EVAL_FUNC(pppn, nnnp, pp, pn, nn, np)         (\
- (- (pp > 0?pp*log10(pp/(pppn)):0) - (pn > 0 ? pn*log10(pn/(pppn)):0)) -\
(- (nn > 0 ? nn*log10(nn/(nnnp)):0) - (np > 0 ? np*log10(np/(nnnp)):0))\
						)
#endif


static int l_lteq(void)
{
  Term term, list;
  VLIST values = NULL;
  register VLIST pointer;
  int bestPP = 0;
  int pp, nn, np, pn, pppn, nnnp, evaluation, best = -100000;
  float threshold;


  if(IsVarTerm(P_ARG(ARG1))) return(FALSE);

  /* list of bindings from pos examples */

  list = P_HEAD(P_ARG(ARG1));

  if (P_EMPTY(list)) return (FALSE);

  values = insertPositives(list, &values, &np);
  threshold = values->val;

  /* list of bindings from neg examples */
	
  list = P_HEAD(P_TAIL(P_ARG(ARG1)));

  values = insertNegatives(list, &values, &nn);

  pp = pn = 0;

  for (pointer=values; pointer; pointer = pointer->next) {

    pp += pointer->positives;
    pn += pointer->negatives;
    np -= pointer->positives;
    nn -= pointer->negatives;

    /*printf("val=%lf nplteq=%ld nnlteq=%ld\n",sl->val,nplteq,nnlteq); */

#ifdef SPLIT_ACC
    if(pp >= bestCover && pointer->positives > pointer->negatives){
#endif
      pppn = pp + pn;
      nnnp = nn + np;
      evaluation = EVAL_FUNC(pppn, nnnp, pp, pn, nn, np);
      if(best <= evaluation || (best == evaluation && pp > bestPP)) {
	best = evaluation;
	bestPP = pp;
	threshold = pointer->val;
      }
#ifdef SPLIT_ACC
    }
#endif
  }

  myfree;

  if(bestPP < bestCover)  return(FALSE);

  if(fabs(threshold - floor(threshold)) < 1e-8)
    term = P_MKINT((int)((float)threshold));
  else term = P_MKDOUBLE(threshold);

  if(!unify(ARG2,term)) return (FALSE);
  return (TRUE);
}

/************************/
	
static int
l_gteq(void)
{
  Term term, list;
  VLIST values = NULL;
  register VLIST pointer;
  int bestPP = 0;
  int pp, nn, np, pn, pppn, nnnp, evaluation, best;
  float threshold;


  if(IsVarTerm(P_ARG(ARG1))) return(FALSE);

  /* list of bindings from pos examples */

  list = P_HEAD(P_ARG(ARG1));

  if (P_EMPTY(list)) return (FALSE);

  values = insertPositives(list, &values, &pp);
  threshold = values->val;

  /* list of bindings from neg examples */
	
  list = P_HEAD(P_TAIL(P_ARG(ARG1)));

  values = insertNegatives(list, &values, &pn);

  np = nn = 0;
  pppn = pp + pn;
  nnnp = nn + np;
  best = EVAL_FUNC(pppn, nnnp, pp, pn, nn, np);

  for(pointer=values; pointer->next; pointer = pointer->next) {

    pp -= pointer->positives;
    pn -= pointer->negatives;
    np += pointer->positives;
    nn += pointer->negatives;

    /*printf("val=%lf nplteq=%ld nnlteq=%ld\n",sl->val,nplteq,nnlteq); */
    /*
#ifdef SPLIT_ACC
    if(pp >= bestCover &&
       pointer->positives >= pointer->negatives){
#endif
    */
      pppn = pp + pn;
      nnnp = nn + np;
      evaluation = EVAL_FUNC(pppn, nnnp, pp, pn, nn, np);
      if(pointer->positives && (best < evaluation ||
				(best == evaluation && pp > bestPP))){
	best = evaluation;
	bestPP = pp;
	threshold = pointer->next->val;
      }
      /*
#ifdef SPLIT_ACC
    }
#endif
      */
  }
  myfree;
  if(bestPP < bestCover) return(FALSE);

  if(fabs(threshold - floor(threshold)) < 1e-8)
    term = P_MKINT((int)((float)threshold));
  else term = P_MKDOUBLE(threshold);

  if(!unify(ARG2,term)) return (FALSE);
  return (TRUE);
}

/************************/

static int
l_interval(void)
{
  Term term, list;
  VLIST values = NULL;
  register VLIST pointer = NULL, aux;
  int pos, pp, nn, np, pn, pppn, nnnp, negs, bestPP = 0,
      evaluation, best =  -100000;
  float lowBound, upBound;


  if(IsVarTerm(P_ARG(ARG1))) return(FALSE);

  /* list of bindings from pos examples */

  list = P_HEAD(P_ARG(ARG1));

  if (P_EMPTY(list)) return (FALSE);

  values = insertPositives(list, &values, &pos);

  upBound = lowBound = values->val;

  /* list of bindings from neg examples */
	
  list = P_HEAD(P_TAIL(P_ARG(ARG1)));

  values = insertNegatives(list, &values, &negs);

  for(pointer=values; pointer; pointer = pointer->next) {

    if(! pointer->positives) continue;

    pp = pn = 0;
    nn = negs;
    np = pos;

    for(aux=pointer; aux; aux = aux->next) {

      nn -= aux->negatives;
      pn += aux->negatives;
      pp += aux->positives;
      np -= aux->positives;

#ifdef SPLIT_ACC
    if(pp >= bestCover && aux->positives > aux->negatives){
#endif

      pppn = pp + pn;
      nnnp = nn + np;
      evaluation = EVAL_FUNC(pppn, nnnp, pp, pn, nn, np);

      if(best < evaluation || (best == evaluation && pp > bestPP)) {
	best = evaluation;
	lowBound = pointer->val;
	upBound = aux->val;
	bestPP = pp;
      }
#ifdef SPLIT_ACC
    }
#endif

    } /* END of  for(aux=pointer; aux->next; aux = aux->next) { */
  }

  myfree;

  if(bestPP < bestCover)  return(FALSE);
  
  if(fabs(lowBound - floor(lowBound)) < 1e-8 &&
     fabs(upBound - floor(upBound)) < 1e-8){
    term = P_MKINT((int)((float)lowBound));
    if(!P_BIND(ARG2,term)) return (FALSE);
    term = P_MKINT((int)((float)upBound));
    if(!P_BIND(ARG3,term)) return (FALSE);
  }
  else {
    term = P_MKDOUBLE(lowBound);
    if(!P_BIND(ARG2,term)) return (FALSE);
    term = P_MKDOUBLE(upBound);
    if(!P_BIND(ARG3,term)) return (FALSE);
  }

  return (TRUE);
}

/************************/

static int l_outsideInterval(void)
{
  Term term, list;
  VLIST values = NULL;
  register VLIST pointer = NULL, aux;
  int pos, negs, bestPP = 0;
  int pp, nn, np, pn, pppn, nnnp, evaluation, best = -100000;
  float lowBound, upBound;


  if(IsVarTerm(P_ARG(ARG1))) return(FALSE);

  /* list of bindings from pos examples */

  list = P_HEAD(P_ARG(ARG1));

  if (P_EMPTY(list)) return (FALSE);

  values = insertPositives(list, &values, &pos);
  lowBound = upBound = values->val;

  /* list of bindings from neg examples */
	
  list = P_HEAD(P_TAIL(P_ARG(ARG1)));

  values = insertNegatives(list, &values, &negs);

  for(pointer=values; pointer; pointer = pointer->next) {

    if(! pointer->positives) continue;

    np = nn = 0;
    pn = negs;
    pp = pos;

    pppn = pp + pn;
    nnnp = nn + np;
    if(best < EVAL_FUNC(pppn, nnnp, pp, pn, nn, np)) {
      best = EVAL_FUNC(pppn, nnnp, pp, pn, nn, np);
      lowBound = upBound = pointer->val;
      bestPP = pp;
    }

    for(aux=pointer->next; aux; aux = aux->next) {

      /********????????? TESTING  !!!!!!!!! *
    if(((int) pp) < bestCover || pointer->positives > pointer->negatives) continue;
      */

      pppn = pp + pn;
      nnnp = nn + np;
      evaluation = EVAL_FUNC(pppn, nnnp, pp, pn, nn, np);
      if(aux->positives && (best < evaluation ||
			    (best == evaluation && pp > bestPP))) {
	best = evaluation;
	lowBound = pointer->val;
	upBound = aux->val;
	bestPP = pp;
      }

      nn += aux->negatives;
      pn -= aux->negatives;
      pp -= aux->positives;
      np += aux->positives;

    } /* END of  for(aux=pointer; aux->next; aux = aux->next) { */
  }

  myfree;

  if(bestPP < bestCover) return(FALSE);

  if(fabs(lowBound - floor(lowBound)) < 1e-8 &&
     fabs(upBound - floor(upBound)) < 1e-8){
    term = P_MKINT((int)((float)lowBound));
    if(!P_BIND(ARG2,term)) return (FALSE);
    term = P_MKINT((int)((float)upBound));
    if(!P_BIND(ARG3,term)) return (FALSE);
  }
  else {
    term = P_MKDOUBLE(lowBound);
    if(!P_BIND(ARG2,term)) return (FALSE);
    term = P_MKDOUBLE(upBound);
    if(!P_BIND(ARG3,term)) return (FALSE);
  }

  return (TRUE);
}

#endif

/*****************************************************************************/
/*                       Linear Discriminant                                 */
/*****************************************************************************/

static int
c_lindiscrim2(void)
{
  Term term;
  double c0, cx;
  int aboveLine = 0;


  if(!discrim2alg(&c0, &cx, &aboveLine)) {
    myfree;
    return(0);
  }

  /* bind the output args */

  term = MkFloatTerm((float)cx);
  unify( ARG4, term );

  term = MkFloatTerm((float)c0);
  unify( ARG5, term );

  if(aboveLine > 0)
    term = MkAtomTerm(LookupAtom("above"));
  else term = MkAtomTerm(LookupAtom("below"));
  unify( ARG6, term );

  myfree;

  return( 1 );
  
}

/**************************************************/

static int
c_lindiscrim3(void)
{
  Term term;
  double c0, cx, cy;
  int aboveLine = 0;


  if(! discrim3alg(&c0, &cx, &cy, &aboveLine)) {
    myfree;
    return(0);
  }

  /* bind the output args */

  term = MkFloatTerm((float)cy);
  unify( ARG5, term );

  term = MkFloatTerm((float)cx);
  unify( ARG6, term );

  term = MkFloatTerm((float)c0);
  unify( ARG7, term );

  if(aboveLine > 0)
    term = MkAtomTerm(LookupAtom("above"));
  else term = MkAtomTerm(LookupAtom("below"));
  unify( ARG8, term );

  myfree;
	
  return( 1 );
  
}

  /*++++++++++++++++++++++++++++++++++++++++++++++++*/
  /*++++++++++++++++++++++++++++++++++++++++++++++++*/
  /*++++++++++++++++++++++++++++++++++++++++++++++++*/
  /*++++++++++++++++++++++++++++++++++++++++++++++++*/
  /*++++++++++++++++++++++++++++++++++++++++++++++++*/



/*********************************************************************/
/*        "Discrim"  FOR PROPOSITIONAL LEARNING                      */
/*               Versao 2.0 (05/02/1997)                             */
/*               Developed by: Joao Gama                             */
/*                LIACC - Uni.do Porto                               */
/*                jgama@ncc.up.pt                                    */
/*********************************************************************/

int **imatrix(int nrl, int nrh, int ncl, int nch);
int *ivector(int nrl, int nrc);

double *dvector(int nrl, int nrh);
double **dmatrix(int nrl,int nrh, int ncl, int nch);
double ***dtensor(int nrm, int nrcol, int nrlin);

/* ======================================================================
   FUNCTIONS
   ---------------------------------------------------------------------- */

#define TypeAttr(D,i)         D->attrs[i].type_attr
#define ValsAttr(D,i)         ((char **)D->attrs[i].vals)
#define SValsAttr(D,i)        ((List *)D->attrs[i].vals)
#define NValsAttr(D,i)        D->attrs[i].nr_vals

#define DomainName(D)               D->root_name
#define NrAttrs(D)                  D->nr_attrs

#define DValAttEx(instance, att) instance[att].val.d
#define CValAttEx(instance, att) instance[att].val.c
#define Id(instance)             instance[0].val.c

/* U T I L S */

int ***Distribuitions2(Dataset *ds, int High);
int ***Distribuitions3(Dataset *ds, int High);
double ***Moments2(Dataset *, int, short int *);
double ***Moments3(Dataset *, int, short int *);

/* M A T R I C E S  */

int choldc(double **m, int n, double *p);
int cholsl(double **x, int n, double *d, double *y, double *r);
double **cholinv(double **x, int n, short int *error);


#define SQR(x)           ((x) * (x))
#define STD(x, x2,n)     sqrt((x2 / n - (x/n) * (x/n)))

#define HALF -0.5

/***************************************/
/*  Statistical Methods                */
/***************************************/

double ***covariance(Dataset *, int, double ***, int **, short int *);

/********************************/
/* Generation of the Classifier */
/********************************/

static double **coeficientes(DomainInfo *names, double ***mom, int ***dist, double **inversa, int nrexs, int nratt);
static double **matmul(DomainInfo *names, double **inversa, double ***momentos, int nratt);
static double *inner(DomainInfo *names, double ***momentos, double **prod, int nratt);
static void corrige_alpha(double *alfa, int ***dist, int nr_exs, int nratt);
static void corrige_beta(double **beta, int nratt);


/******************/

AttrVal **ReadInstances2(double scale, int *nrexs, short int *error)
{
  AttrVal **instances;
  Term term, listX, listY;
  register int counter;
  int dum;
  double x, y;

  /* counting the number of pos + neg examples */

  counter = 0;
  listX = P_HEAD(P_ARG(ARG1));  /* positives */
  while(! P_EMPTY(listX)){
    counter++;
    listX = P_TAIL(listX);
  }

  if(counter <= 1){
    *error = TRUE;
    return NULL;
  }
  dum = counter;

  listX = P_HEAD(P_TAIL(P_ARG(ARG1)));  /* negatives */
  while(! P_EMPTY(listX)){
    counter++;
    listX = P_TAIL(listX);
  }

  if(counter <= dum + 1){
    *error = TRUE;
    return NULL;
  }

  *nrexs = counter;
  instances = (AttrVal **) mymalloc(counter * sizeof(AttrVal *));

  counter = 0;

  /* processing positive examples */

  listX = P_HEAD(P_ARG(ARG1));
  listY = P_HEAD(P_ARG(ARG2));


  if(IsVarTerm(listX) || IsVarTerm(listY) || P_EMPTY(listX) || P_EMPTY(listY))
    return(NULL);

  while(! P_EMPTY(listX) && ! P_EMPTY(listY)){ 

    term = P_HEAD(listX);
    if(P_ISINT(term))
      x = ((double) P_INT(term)) * scale;
    else if(P_ISDOUBLE(term))
      x = P_DOUBLE(term) * scale;
    else return(FALSE);

    term = P_HEAD(listY);
    if(P_ISINT(term))
      y = ((double) P_INT(term)) * scale;
    else if(P_ISDOUBLE(term))
      y = P_DOUBLE(term) * scale;
    else return(FALSE);
    
    instances[counter] = mkInstance2(x, y, 1, counter); /* 1 - action; 2 - noaction */

    counter++;
    listX = P_TAIL(listX);
    listY = P_TAIL(listY);
      
  } /* END while */

  /* processing negative examples */

  listX = P_HEAD(P_TAIL(P_ARG(ARG1)));
  listY = P_HEAD(P_TAIL(P_ARG(ARG2)));

  if(IsVarTerm(listX) || IsVarTerm(listY) || P_EMPTY(listX) || P_EMPTY(listY))
    return (NULL);
    
  while(! P_EMPTY(listX) && ! P_EMPTY(listY)){ 

    term = P_HEAD(listX);
    if(P_ISINT(term))
      x = ((double) P_INT(term)) * scale;
    else if(P_ISDOUBLE(term))
      x = P_DOUBLE(term) * scale;
    else return(FALSE);

    term = P_HEAD(listY);
    if(P_ISINT(term))
      y = ((double) P_INT(term)) * scale;
    else if(P_ISDOUBLE(term))
      y = P_DOUBLE(term) * scale;
    else return(FALSE);

    instances[counter] = mkInstance2(x, y, 2, counter); /* 1 - action; 2 - noaction */

    counter++;
    listX = P_TAIL(listX);
    listY = P_TAIL( listY );
      
  } /* END while */

  return --instances;
}


 /* class = 1 - action; 2 - noaction */

AttrVal *mkInstance2(double x, double y, int class, int counter){

  AttrVal *attr_val = (AttrVal *) mycalloc(4 * sizeof(AttrVal));
  attr_val[0].val.c = (ContType) (counter+1);
  attr_val[1].val.c =  (float) x;
  attr_val[2].val.c =  (float) y;
  attr_val[3].val.d =  class;  /* 1 - action; 2 - noaction*/

  return attr_val;
}

/************************/

AttrVal **ReadInstances3(double scale, int *nrexs, short int *error)
{
  AttrVal **instances;
  Term term, listX, listY, listZ;
  register int counter;
  int dum;
  double x, y, z;

  /* counting the number of pos + neg examples */

  counter = 0;
  listX = P_HEAD(P_ARG(ARG1));  /* positives */
  while(! P_EMPTY(listX)){
    counter++;
    listX = P_TAIL(listX);
  }

  if(counter <= 1){
    *error = TRUE;
    return NULL;
  }
  dum = counter;

  listX = P_HEAD(P_TAIL(P_ARG(ARG1)));  /* negatives */
  while(! P_EMPTY(listX)){
    counter++;
    listX = P_TAIL(listX);
  }

  if(counter <= dum + 1){
    *error = TRUE;
    return NULL;
  }


  *nrexs = counter;
  instances = (AttrVal **) mymalloc(counter * sizeof(AttrVal *));

  counter = 0;

  /* processing positive examples */

  listX = P_HEAD(P_ARG(ARG1));
  listY = P_HEAD(P_ARG(ARG2));
  listZ = P_HEAD(P_ARG(ARG3));


  if(IsVarTerm(listX) || IsVarTerm(listY) || IsVarTerm(listZ) || P_EMPTY(listX) || P_EMPTY(listY) || P_EMPTY(listZ))
    return(NULL);

  while(! P_EMPTY(listX) && ! P_EMPTY(listY) && ! P_EMPTY(listZ)){ 

    term = P_HEAD(listX);
    if(P_ISINT(term))
      x = ((double) P_INT(term)) * scale;
    else if(P_ISDOUBLE(term))
      x = P_DOUBLE(term) * scale;
    else return(FALSE);

    term = P_HEAD(listY);
    if(P_ISINT(term))
      y = ((double) P_INT(term)) * scale;
    else if(P_ISDOUBLE(term))
      y = P_DOUBLE(term) * scale;
    else return(FALSE);

    term = P_HEAD(listZ);
    if(P_ISINT(term))
      z = ((double) P_INT(term)) * scale;
    else if(P_ISDOUBLE(term))
      z = P_DOUBLE(term) * scale;
    else return(FALSE);

    instances[counter] = mkInstance3(x, y, z, 1, counter); /* 1 - action; 2 - noaction */

    counter++;
    listX = P_TAIL(listX);
    listY = P_TAIL(listY);
    listZ = P_TAIL(listZ);
      
  } /* END while */

  /* processing negative examples */

  listX = P_HEAD(P_TAIL(P_ARG(ARG1)));
  listY = P_HEAD(P_TAIL(P_ARG(ARG2)));
  listZ = P_HEAD(P_TAIL(P_ARG(ARG3)));

  if(IsVarTerm(listX) || IsVarTerm(listY) || IsVarTerm(listZ) || P_EMPTY(listX) || P_EMPTY(listY) || P_EMPTY(listZ))
    return (NULL);
    
  while(! P_EMPTY(listX) && ! P_EMPTY(listY) && ! P_EMPTY(listZ)){ 

    term = P_HEAD(listX);
    if(P_ISINT(term))
      x = ((double) P_INT(term)) * scale;
    else if(P_ISDOUBLE(term))
      x = P_DOUBLE(term) * scale;
    else return(FALSE);

    term = P_HEAD(listY);
    if(P_ISINT(term))
      y = ((double) P_INT(term)) * scale;
    else if(P_ISDOUBLE(term))
      y = P_DOUBLE(term) * scale;
    else return(FALSE);

    term = P_HEAD(listZ);
    if(P_ISINT(term))
      z = ((double) P_INT(term)) * scale;
    else if(P_ISDOUBLE(term))
      z = P_DOUBLE(term) * scale;
    else return(FALSE);

    instances[counter] = mkInstance3(x, y, z, 2, counter); /* 2 -> noaction */

    counter++;
    listX = P_TAIL(listX);
    listY = P_TAIL(listY);
    listZ = P_TAIL(listZ);
      
  } /* END while */

  return --instances;
}

/*************************
 1 - action; 2 - noaction
 **************************/

AttrVal *mkInstance3(double x, double y, double z, int class, int counter){

  AttrVal *attr_val = (AttrVal *) mycalloc(5 * sizeof(AttrVal));
  attr_val[0].val.c = (ContType) (counter+1);
  attr_val[1].val.c =  (float) x;
  attr_val[2].val.c =  (float) y;
  attr_val[3].val.c =  (float) z;
  attr_val[4].val.d =  class;

  return attr_val;
}

/* **********************************************/
/*		MATRICES			*/
/*  Some of this functions comes from the book  */
/*  NUMERICAL RECIPES IN C (2. Edition)         */
/************************************************/

#define ROTATE(a,i,j,k,l) g=a[i][j]; h=a[k][l];a[i][j]=g-s*(h+g*tau);a[k][l]=h+s*(g-h*tau);
#define TRESH	1.0e-6;
#define RADIX	2

/****************************************************************/
/*		Cholesky  Matrices				*/
/*		Used for Invertion of COVARIANCE Matrices	*/
/****************************************************************/
/********************************************************/
/*	Cholesky decomposicao de Matrizes		*/
/*	A = L . Lt	A simetrica, positiva definite	*/
/* Input:  a[1..n][1..n]				*/
/* Output: a[1..n][1..n] como L no triangulo inferior	*/
/*	   p[1..n]	 diagonal de L			*/
/* Comments: 	The upper part of A is unchanged	*/
/********************************************************/
int choldc(double **a, int n, double *p)
{
  int i, j, k;
  double sum;
  for(i = 1; i <= n; i++) {
    for(j=i; j <= n; j++) {
      sum = a[i][j];
      for(k = i-1; k>=1; k--) sum -=a[i][k]*a[j][k];
      if( i == j) {
	if(sum <= 0.0 || equalf(sum, 0.0)) {
	  /*fprintf(stderr, "E)Choldc: Singular Matrix\n");*/
	  sum = 0.0;
	  return 1;
	}
	p[i]=sqrt(sum);
      }
      else
	if equalf(p[i], 0.0) return(1);
	else a[j][i] = sum / p[i];
    }
  }
  return 0;
}
/********************************************************/
/*	Resolucao de Equacoes	A . x = b		*/
/*	Input: 	a[1..n][1..n]	decomposta por choldc	*/
/*	 	p[1..n]		choldc			*/
/*		b[1..n]					*/
/*	Output:	x[1..n]					*/
/********************************************************/
int cholsl(double **a, int n, double *p, double *b, double *x)
{
  int i, k;
  double sum;
  for(i = 1; i <= n; i++) {
    for(sum=b[i], k = i - 1; k >= 1; k--)	sum -= a[i][k] * x[k];
    x[i] = sum / p[i];
  }
  for(i = n; i >= 1; i--) {
    for(sum = x[i], k = i+1; k <= n; k++)	sum -= a[k][i] * x[k];
    if equalf(p[i], 0.0) return(1);
    else x[i] = sum / p[i];
  }
  return(0);
}

double **cholinv(double **a, int n, short int *error)
{
  int i, j;
  double *p = dvector(1, n), *col;
  double **y = NULL;
  
  if (!choldc(a, n, p)) {
    y = dmatrix(1, n, 1, n);
    col = dvector(1, n);
    for(j = 1; j <= n; j++) {
      for(i = 1; i<= n; i++) col[i] = 0.0;
      col[j] = 1.0;
      if(cholsl(a,n,p,col, col)) {
	*error = TRUE;
	return NULL;
      }
      for(i = 1; i <= n; i++)	y[i][j] = col[i];
    }
  }
  else *error = TRUE;

  return(y);
}


/**************************************/
/*   Public Methods  Distributions    */
/**************************************/
int ***Distribuitions2(Dataset *ds, int High)
{
  register int classe, i;
  int ***stat;
  
  stat = (int ***) mycalloc(3 * sizeof(int **));  
  --stat;
  stat[1] = NULL;
  stat[2] = NULL;
  stat[3] = imatrix(0, 2, 1, 2);

  for(i = 1; i <= High; i++) {
    classe = DValAttEx(ds->instances[i], 3);
    ++stat[3][0][DValAttEx(ds->instances[i], 3)];
    ++stat[3][classe][DValAttEx(ds->instances[i], 3)];
  }
  return stat;
} 

int ***Distribuitions3(Dataset *ds, int High)
{
  register int classe, i;
  int ***stat;
  
  stat = (int ***) mycalloc(4 * sizeof(int **));  
  --stat;
  stat[1] = NULL;
  stat[2] = NULL;
  stat[3] = NULL;
  stat[4] = imatrix(0, 2, 1, 2);

  for(i = 1; i <= High; i++) {
    classe = DValAttEx(ds->instances[i], 4);
    ++stat[4][0][DValAttEx(ds->instances[i], 4)];
    ++stat[4][classe][DValAttEx(ds->instances[i], 4)];
  }
  return stat;
} 

/**************************************/
/*   Public Methods  Moments          */
/**************************************/
double ***Moments2(Dataset *ds, int High, short int *error)
{
  register int classe, valor, j, i;
  double cval, ***stat;
  int *class_freq;
  
  stat = (double ***) mymalloc(3 * sizeof(double **));  
  --stat;

  stat[3] = dmatrix(0, 3, 0, 3);
  stat[1] = dmatrix(0, 3, 1, 4);
      for(j = 0; j <= 3; j++){
        stat[1][j][1] = FLT_MAX;
        stat[1][j][2] = FLT_MIN;
      }
  stat[2] = dmatrix(0, 3, 1, 4);
      for(j = 0; j <= 3; j++){
        stat[2][j][1] = FLT_MAX;
        stat[2][j][2] = FLT_MIN;
      }

  class_freq = ivector(0, 2);
  for(i = 1; i <= High; i++) {
    classe = DValAttEx(ds->instances[i], 3);
    ++class_freq[classe]; 
    ++class_freq[0];

    valor = DValAttEx(ds->instances[i], 3);
    stat[3][0][1] += valor;
    stat[3][0][2] += SQR(valor);
    stat[3][classe][1] += valor;
    stat[3][classe][2] += SQR(valor);

    cval = CValAttEx(ds->instances[i], 1);
    if(stat[1][0][1] > cval) stat[1][0][1] = cval;
    if(stat[1][0][2] < cval) stat[1][0][2] = cval;
    stat[1][0][3] += cval;
    stat[1][0][4] += SQR(cval);
    if(stat[1][classe][1] > cval) stat[1][classe][1] = cval;
    if(stat[1][classe][2] < cval) stat[1][classe][2] = cval;
    stat[1][classe][3] += cval;
    stat[1][classe][4] += SQR(cval);

    cval = CValAttEx(ds->instances[i], 2);
    if(stat[2][0][1] > cval) stat[2][0][1] = cval;
    if(stat[2][0][2] < cval) stat[2][0][2] = cval;
    stat[2][0][3] += cval;
    stat[2][0][4] += SQR(cval);
    if(stat[2][classe][1] > cval) stat[2][classe][1] = cval;
    if(stat[2][classe][2] < cval) stat[2][classe][2] = cval;
    stat[2][classe][3] += cval;
    stat[2][classe][4] += SQR(cval);
  }

  for(i = 0; i <= 2; i++) {
    if(equalf(((double)class_freq[i]) - (stat[3][i][1]/((double)class_freq[i])) *  (stat[3][i][1]/((double)class_freq[i])), 0.0)) {
      *error = TRUE;
      return NULL;
    }
    else stat[3][i][2] = STD(stat[3][i][1],stat[3][i][2],(double) class_freq[i]);
    stat[3][i][1] /= (double) class_freq[i];
  }

  for(i = 0; i <= 2; i++) {
    if(equalf(((double)class_freq[i]) - (stat[1][i][3]/((double)class_freq[i])) *  (stat[1][i][3]/((double)class_freq[i])), 0.0)) {
      *error = TRUE;
      return NULL;
    }
    else stat[1][i][4] = STD(stat[1][i][3],stat[1][i][4],(double) class_freq[i]);
    stat[1][i][3] /= (double) class_freq[i];
  }

  for(i = 0; i <= 2; i++) {
    if(equalf(((double)class_freq[i]) - (stat[2][i][3]/((double)class_freq[i])) *  (stat[2][i][3]/((double)class_freq[i])), 0.0)) {
      *error = TRUE;
      return NULL;
    }
    else stat[2][i][4] = STD(stat[2][i][3],stat[2][i][4],(double) class_freq[i]);
    stat[2][i][3] /= (double) class_freq[i];
  }

  return stat;
} 


double ***Moments3(Dataset *ds, int High, short int *error)
{
  register int classe, valor, j, i;
  double cval, ***stat;
  int *class_freq;
  
  stat = (double ***) mymalloc(4 * sizeof(double **));  
  --stat;

  stat[4] = dmatrix(0, 3, 0, 3);

  stat[1] = dmatrix(0, 3, 1, 4);
  for(j = 0; j <= 3; j++){
    stat[1][j][1] = FLT_MAX;
    stat[1][j][2] = FLT_MIN;
  }

  stat[2] = dmatrix(0, 3, 1, 4);
  for(j = 0; j <= 3; j++){
    stat[2][j][1] = FLT_MAX;
    stat[2][j][2] = FLT_MIN;
  }

  stat[3] = dmatrix(0, 3, 1, 4);
  for(j = 0; j <= 3; j++){
    stat[3][j][1] = FLT_MAX;
    stat[3][j][2] = FLT_MIN;
  }

  class_freq = ivector(0, 2);

  for(i = 1; i <= High; i++) {
    classe = DValAttEx(ds->instances[i], 4);
    ++class_freq[classe]; 
    ++class_freq[0];

    valor = DValAttEx(ds->instances[i], 4);
    stat[4][0][1] += valor;
    stat[4][0][2] += SQR(valor);
    stat[4][classe][1] += valor;
    stat[3][classe][2] += SQR(valor);

    cval = CValAttEx(ds->instances[i], 1);
    if(stat[1][0][1] > cval) stat[1][0][1] = cval;
    if(stat[1][0][2] < cval) stat[1][0][2] = cval;
    stat[1][0][3] += cval;
    stat[1][0][4] += SQR(cval);
    if(stat[1][classe][1] > cval) stat[1][classe][1] = cval;
    if(stat[1][classe][2] < cval) stat[1][classe][2] = cval;
    stat[1][classe][3] += cval;
    stat[1][classe][4] += SQR(cval);


    cval = CValAttEx(ds->instances[i], 2);
    if(stat[2][0][1] > cval) stat[2][0][1] = cval;
    if(stat[2][0][2] < cval) stat[2][0][2] = cval;
    stat[2][0][3] += cval;
    stat[2][0][4] += SQR(cval);
    if(stat[2][classe][1] > cval) stat[2][classe][1] = cval;
    if(stat[2][classe][2] < cval) stat[2][classe][2] = cval;
    stat[2][classe][3] += cval;
    stat[2][classe][4] += SQR(cval);

    cval = CValAttEx(ds->instances[i], 3);
    if(stat[3][0][1] > cval) stat[3][0][1] = cval;
    if(stat[3][0][2] < cval) stat[3][0][2] = cval;
    stat[3][0][3] += cval;
    stat[3][0][4] += SQR(cval);
    if(stat[3][classe][1] > cval) stat[3][classe][1] = cval;
    if(stat[3][classe][2] < cval) stat[3][classe][2] = cval;
    stat[3][classe][3] += cval;
    stat[3][classe][4] += SQR(cval);
  }

  for(i = 0; i <= 2; i++) {
    if(equalf(((double)class_freq[i]) - (stat[4][i][1]/((double)class_freq[i])) *  (stat[4][i][1]/((double)class_freq[i])), 0.0)) {
      *error = TRUE;
      return NULL;
    }
    else stat[4][i][2] = STD(stat[4][i][1],stat[4][i][2],(double) class_freq[i]);
    stat[4][i][1] /= (double) class_freq[i];
  }

  for(i = 0; i <= 2; i++) {
    if(equalf(((double)class_freq[i]) - (stat[1][i][3]/((double)class_freq[i])) *  (stat[1][i][3]/((double)class_freq[i])), 0.0)) {
      *error = TRUE;
      return NULL;
    }
    else stat[1][i][4] = STD(stat[1][i][3],stat[1][i][4],(double) class_freq[i]);
    stat[1][i][3] /= (double) class_freq[i];
  }

  for(i = 0; i <= 2; i++) {
    if(equalf(((double)class_freq[i]) - (stat[2][i][3]/((double)class_freq[i])) *  (stat[2][i][3]/((double)class_freq[i])), 0.0)) {
      *error = TRUE;
      return NULL;
    }
    else stat[2][i][4] = STD(stat[2][i][3],stat[2][i][4],(double) class_freq[i]);
    stat[2][i][3] /= (double) class_freq[i];
  }

  for(i = 0; i <= 2; i++) {
    if(equalf(((double)class_freq[i]) - (stat[3][i][3]/((double)class_freq[i])) *  (stat[3][i][3]/((double)class_freq[i])), 0.0)) {
      *error = TRUE;
      return NULL;
    }
    else stat[3][i][4] = STD(stat[3][i][3],stat[3][i][4],(double) class_freq[i]);
    stat[3][i][3] /= (double) class_freq[i];
  }

  return stat;
} 

/* ******************************************* */
/*                Integers                     */
/* ******************************************* */
int **imatrix(int nrl,int nrh, int ncl, int nch)
/* allocate a int matrix m[nrl..nrh][ncl..nch]	*/
{
  int i, nrow = nrh - nrl+1, ncol = nch-ncl+1;
  int **m;
  
  m = (int **)  mymalloc((size_t) nrow * sizeof(int *));
  if (!m)	fprintf(stderr, "E)imatrix:Out of memory!\n");
  else {
    m -= nrl;
    m[nrl] = (int *) mycalloc( nrow*ncol * sizeof(int));
    if (!m[nrl])     fprintf(stderr, "E)imatrix: Out of memoy!\n");
    else {
      m[nrl] -= ncl;
      for(i = nrl+1; i<= nrh; i++) m[i]=m[i-1]+ncol;
    }
  }
  return m;
}

int *ivector(int nrl, int nrh)
{
  int *v;

  v = (int *) 	mycalloc( (nrh - nrl +1) * sizeof(int));
  if (!v)		fprintf(stderr, "E)fvector: Out of memory !");
  return(v-nrl);
}

/* ******************************************* */
/*                Doubles                      */
/* ******************************************* */
double *dvector(int nrl, int nrh)
{
  double *v;

  v = (double *) 	mycalloc((nrh - nrl +1) * sizeof(double));
  if (!v)		fprintf(stderr, "E)fvector: Out of memory !");
  return(v-nrl);
}

double **dmatrix(int nrl,int nrh, int ncl, int nch)
/* allocate a double matrix m[nrl..nrh][ncl..nch]	*/
{
  int i, nrow = nrh - nrl+1, ncol = nch-ncl+1;
  double **m;
  
  m = (double **) mycalloc(nrow * sizeof(double *));
  if (!m)	fprintf(stderr, "E)dmatrix:Out of memory!\n");
  else {
    m -= nrl;
    m[nrl] = (double *) mycalloc(nrow*ncol * sizeof(double));
    if (!m[nrl])	fprintf(stderr, "E)dmatrix: Out of memoy!\n");
    else {
      m[nrl] -= ncl;
      for(i = nrl+1; i<= nrh; i++) m[i]=m[i-1]+ncol;
    }
  }
  return m;
}


double ***dtensor(int nrm, int nrcol, int nrlin)
{
  double ***t;
  register int i;

  t = (double ***) mymalloc(nrm *  sizeof(double **));
  for(i = 0; i < nrm; i++)
    t[i] = dmatrix(1, nrcol, 1, nrlin);
  return t;
}


/***************************************/
/*  Statistical Methods                */
/***************************************/

double ***covariance(Dataset *ds, int goal, double ***momentos, int **classes,
		     short int *error)
{
  register int i, j, k, atj, atk, classe;
  double x, y, mx, my;
  double ***cov;
  int NR_CLASSES;

  NR_CLASSES = NValsAttr(ds->domain, goal);
  cov = dtensor(2+NR_CLASSES, NrAttrs(ds->domain)-1, NrAttrs(ds->domain)-1);
/* ************************************************
   Compute Class covariance matrices from Data
*************************************************** */
  for (i = 1; i <= ds->nr_exs; i++) {
    classe = DValAttEx(ds->instances[i], goal);
    for(j = 1; j <= NrAttrs(ds->domain); j++) {
      if (goal != j) {
	switch(TypeAttr(ds->domain, j)) {
	case continuous: 
	  x = CValAttEx(ds->instances[i], j); 
	  mx = momentos[j][classe][3];
	  break;
	case nominal:
	  x = (double) DValAttEx(ds->instances[i], j); 
	  mx = momentos[j][classe][1];
	  break;
	}
	for(k = j; k <= NrAttrs(ds->domain); k++) {
	  if (goal != k) {
	    switch(TypeAttr(ds->domain, k)) {
	    case continuous: 
	      y = CValAttEx(ds->instances[i], k); 
	      my = momentos[k][classe][3];
	      break;
	    case nominal:
	      y = (double) DValAttEx(ds->instances[i], k); 
	      my = momentos[k][classe][1];
	      break;
	    }
	    if (j > goal) atj = j - 1;     else          atj = j;
	    if (k > goal) atk = k - 1;     else          atk = k;
	    cov[classe][atj][atk] += (x - mx) * (y - my);
	    cov[NR_CLASSES+1][atj][atk] += (x - momentos[k][0][3]) * (y - momentos[k][0][3]);
	  }
	}
      }
    }
  }
/* ************************************************
   Normalize matrices and compute Pooled covariance Matrix
*************************************************** */
  for(k = 1; k <= NR_CLASSES; k++)
    for(i = 1; i < NrAttrs(ds->domain); i++)
      for(j = i; j < NrAttrs(ds->domain); j++) {
	if(equalf((classes[k][k] - 1.0), 0.0)) {
	  *error = TRUE;
	  return NULL;
	}
	else cov[k][i][j] /= (classes[k][k] - 1.0);
	cov[k][j][i] = cov[k][i][j];
	cov[0][i][j] += (cov[k][i][j] * (classes[k][k] - 1.0));
      }
  for(i = 1; i < NrAttrs(ds->domain); i++)
    for(j = i; j < NrAttrs(ds->domain); j++) {
      cov[0][i][j] /= (ds->nr_exs - NR_CLASSES);
      cov[0][j][i] = cov[0][i][j];
      cov[NR_CLASSES+1][i][j] /= (ds->nr_exs - 1.0);
      cov[NR_CLASSES+1][j][i] = cov[NR_CLASSES+1][i][j];
    }

  return cov;
}


/******************************/
/*    MAIN FUNCTION           */
/******************************/
int discrim2alg(double *c0, double *cx, int *aboveLine)
{
  Dataset ds;
  int ***dist;
  double ***mom, ***covar, **covinv, **coef, scale, cy;
  short int error = FALSE;
  register int i;


  scale = P_DOUBLE(P_ARG(ARG3));

  ds.domain = MyReadDomain2();
  ds.instances = ReadInstances2(scale, &(ds.nr_exs), &error);
  if(error) return(0);
  dist = Distribuitions2(&ds, ds.nr_exs);
  mom = Moments2(&ds, ds.nr_exs, &error);
  if(error) return(0);
  covar = covariance(&ds, 3, mom, dist[3], &error);
  if(error) return(0);
  covinv = cholinv(covar[0], 2, &error);
  if(error) return(0);
  coef = coeficientes(ds.domain, mom, dist, covinv, ds.nr_exs, 2);

  *c0 = coef[1][1];
  *cx = coef[1][2] * scale;
  cy = coef[1][3] * scale;

  /*printf("cx = %f cy = %f c0 = %f\n", (float) *cx, (float) cy, (float) *c0);*/

  if(!equalf(cy, 0.0)){ 
    *cx = - *cx / cy;
    *c0 = - *c0 / cy;
  }
  else /*{
    printf("Axis paralel boundary\n");*/
    return(FALSE);
  /*}*/

  *aboveLine = 0;
  for(i=1; i <= ds.nr_exs; i++)
    if(ds.instances[i][2].val.c > *cx * ds.instances[i][1].val.c + *c0)
      (*aboveLine)++;
    else (*aboveLine)--;

  return(1);
}

int discrim3alg(double *c0, double *cx, double *cy, int *aboveLine)
{
  Dataset ds;
  int ***dist;
  double ***mom, ***covar, **covinv, **coef, scale, cz;
  short int error = FALSE;
  register int i;


  scale = P_DOUBLE(P_ARG(ARG4));

  ds.domain = MyReadDomain3();
  ds.instances = ReadInstances3(scale, &ds.nr_exs, &error);
  if(error) return(0);
  dist = Distribuitions3(&ds, ds.nr_exs);
  mom = Moments3(&ds, ds.nr_exs, &error);
  if(error) return(0);
  covar = covariance(&ds, 4, mom, dist[4], &error);
  if(error) return(0);
  covinv = cholinv(covar[0], 3, &error);
  if(error) return(0);
  coef = coeficientes(ds.domain, mom, dist, covinv, ds.nr_exs, 3);

  *c0 = coef[1][1];
  *cx = coef[1][2] * scale;
  *cy = coef[1][3] * scale;
  cz = coef[1][4] * scale;

  /*printf("cx = %f cy = %f cz = %f c0 = %f\n", (float) *cx, (float) *cy, (float) cz, (float) *c0);*/

  if(!equalf(cz, 0.0)){ 
    *cx = - *cx / cz;
    *c0 = - *c0 / cz;
    *cy = - *cy / cz;
  }
  else{
    printf("Axis paralel boundary\n");
    return(FALSE);
  }

  *aboveLine = 0;
  for(i=1; i <= ds.nr_exs; i++)
    if(ds.instances[i][3].val.c > *cx * ds.instances[i][1].val.c + *cy * ds.instances[i][2].val.c + *c0) (*aboveLine)++;
    else (*aboveLine)--;

  return(1);
}
/******************************************/
/*          Reading Domain                */
/******************************************/

DomainInfo *MyReadDomain2(void)
{
  char **str;
  DomainInfo *Info = NULL;
  AttrInfo   *ats = NULL;

  Info = (DomainInfo *) mymalloc(sizeof(DomainInfo));
  Info->nr_attrs = 3;
  ats =(AttrInfo *) mymalloc(3 * sizeof(AttrInfo));
  Info->attrs = --ats;   

  Info->attrs[1].type_attr = continuous;
  Info->attrs[1].nr_vals = 0;
  Info->attrs[1].vals = NULL;

  Info->attrs[2].type_attr = continuous;
  Info->attrs[2].nr_vals = 0;
  Info->attrs[2].vals = NULL;

  Info->attrs[3].type_attr = nominal;

  Info->attrs[3].nr_vals = 2;

  str = (char **) mymalloc(2 * sizeof(char*));
  str[0] = (char *) mymalloc(7);
  strcpy(str[0], "action");
  str[1] = (char *) mymalloc(9);
  strcpy(str[1], "noaction");

  Info->attrs[3].vals = --str;

  return Info;
}

DomainInfo *MyReadDomain3(void)
{
  char **str;
  DomainInfo *Info = NULL;
  AttrInfo   *ats = NULL;

  Info = (DomainInfo *) mymalloc(sizeof(DomainInfo));
  Info->nr_attrs = 4;
  ats =(AttrInfo *) mymalloc(5 * sizeof(AttrInfo));
  Info->attrs = --ats;   

  Info->attrs[1].type_attr = continuous;
  Info->attrs[1].nr_vals = 0;
  Info->attrs[1].vals = NULL;

  Info->attrs[2].type_attr = continuous;
  Info->attrs[2].nr_vals = 0;
  Info->attrs[2].vals = NULL;


  Info->attrs[3].type_attr = continuous;
  Info->attrs[3].nr_vals = 0;
  Info->attrs[3].vals = NULL;

  Info->attrs[4].type_attr = nominal;
  Info->attrs[4].nr_vals = 2;

  str = (char **) mymalloc(2 * sizeof(char*));
  str[0] = (char *) mymalloc(7);
  strcpy(str[0], "action");
  str[1] = (char *) mymalloc(9);
  strcpy(str[1], "noaction");

  Info->attrs[4].vals = --str;

  return Info;
}


/******************************************/
/*   Generating the  Discriminant         */
/******************************************/
static double **coeficientes\
(DomainInfo *names, double ***mom, int ***dist, double **inversa, int nrexs, int nratt)
{
  register int i, j;
  double **beta, *alfa;
  double **coef = dmatrix(1, 1, 1, nratt+1);
  
  beta = matmul(names, inversa, mom, nratt);
  alfa = inner(names, mom, beta,  nratt);
  corrige_alpha(alfa, dist, nrexs, nratt);
  corrige_beta(beta, nratt);
  
  for(i = 1; i < 2; i++) {
    coef[i][1] = alfa[i];
    for(j = 1; j <= nratt; j++)
      coef[i][j+1] = beta[i][j];
  }

  return coef;
}

static double **matmul\
(DomainInfo *names, double **inversa, double ***momentos, int nratt)
{
  register int k1, k2, k3;
  double my, temp, **prod = dmatrix(1, 2, 1, nratt);
  
  for(k1 = 1; k1 <= nratt; k1++) 
    for(k3 = 1; k3 <= 2; k3++) {
      temp = 0.0;
      for(k2 = 1; k2 <= nratt; k2++) {
	switch(TypeAttr(names, k2)) {
	case continuous: 
	  my = momentos[k2][k3][3];
	  break;
	case nominal:
	  my = momentos[k2][k3][1];
	  break;
	}
	temp += inversa[k1][k2] *  my;
/*	printf("Matmul k1 %d k2 %d k3 %d Inv %f Med %f\n", k1, k2 , k3,inversa[k1][k2], my); */
      }
      prod[k3][k1] = temp;
    }
  return prod;
}

static double *inner\
(DomainInfo *names, double ***momentos, double **prod, int nratt)
{
  register int k1, k2;
  double my, temp, *alfa = dvector(1, 2);
  
  for(k2 = 1; k2 <= 2; k2++) {
    temp = 0.0;
    for(k1 = 1; k1 <= nratt; k1++) {
      switch(TypeAttr(names, k1)) {
      case continuous: 
	my = momentos[k1][k2][3];
	break;
      case nominal:
	my = momentos[k1][k2][1];
	break;
      }
      temp += prod[k2][k1] * my;
    }
    alfa[k2] = temp;
  }
  return alfa;
}

static void corrige_alpha\
(double *alfa, int ***dist, int nr_exs, int nratt)
{
  alfa[1] = HALF * alfa[1] + log((double) dist[nratt+1][1][1] / nr_exs);
  alfa[2] = HALF * alfa[2] + log((double) dist[nratt+1][2][2] / nr_exs);
 
  alfa[1] -= alfa[2];
  alfa[2] -= alfa[2];
}

static void corrige_beta(double **beta, int nratt)
{
  register int j;

    for(j = 1; j <= nratt; j++)
      beta[1][j] -= beta[2][j];
    for(j = 1; j <= nratt; j++)
      beta[2][j] -= beta[2][j];
}



  /*++++++++++++++++++++++++++++++++++++++++++++++++*/
  /*++++++++++++++++++++++++++++++++++++++++++++++++*/
  /*++++++++++++++++++++++++++++++++++++++++++++++++*/
  /*++++++++++++++++++++++++++++++++++++++++++++++++*/
  /*++++++++++++++++++++++++++++++++++++++++++++++++*/


/*****************************************************************************/
/*                       REGRESSION predicates                               */
/*****************************************************************************/


#define	ONE_TWO		0
#define	TWO_THREE	1
#define	ONE_THREE	2

static int
l_regress2(void)
{
	Term list[3], t;
	double r[3], s[3], b[4];
	double x1, x2, x3, x1s,x2s,x3s,x1sq,x2sq,x3sq;
	double x12, x13, x23;
	double num, denom;
	LONG n;

	list[0] = P_ARG(ARG1);
	list[1] = P_ARG(ARG2);
	list[2] = P_ARG(ARG3);

	if (P_EMPTY(list[0]) || P_EMPTY(list[1]) || P_EMPTY(list[2]))
		return (FALSE);

	x1s=x2s=x3s=x1sq=x2sq=x3sq=x12=x13=x23=(double)0.0;
	n = 0l;
	while (!P_EMPTY(list[0]) && !P_EMPTY(list[1]) && !P_EMPTY(list[2])) {

		if (P_ISNUM(P_HEAD(list[0])))
			P_NUM(P_HEAD(list[0]),x1);
      		else return(FALSE);

		if (P_ISNUM(P_HEAD(list[1])))
			P_NUM(P_HEAD(list[1]),x2);
      		else return(FALSE);

		if (P_ISNUM(P_HEAD(list[2])))
			P_NUM(P_HEAD(list[2]),x3);
      		else return(FALSE);

		x1s += x1; x2s += x2; x3s += x3;
		x1sq += x1*x1; x2sq += x2*x2; x3sq += x3*x3;
		x12 += x1*x2; x13 += x1*x3; x23 += x2*x3;
		n++;

		list[0] = P_TAIL(list[0]);
		list[1] = P_TAIL(list[1]);
		list[2] = P_TAIL(list[2]);
	}
	num = (double)n*x12 - x1s*x2s;
	denom = (((double)n*x1sq) - (x1s*x1s))*(((double)n*x2sq) - (x2s*x2s));
	if (fabs(denom) < 1e-6) return(0);
	r[ONE_TWO] =  (double)num/sqrt((double)denom);

	num = (double)n*x13 - x1s*x3s;
	denom = (((double)n*x1sq) - (x1s*x1s))*(((double)n*x3sq) - (x3s*x3s));
	if (fabs(denom) < 1e-6) return(0);
	r[ONE_THREE] =  (double)num/sqrt((double)denom);

	num = (double)n*x23 - x2s*x3s;
	denom = (((double)n*x2sq) - (x2s*x2s))*(((double)n*x3sq) - (x3s*x3s));
	if (fabs(denom) < 1e-6) return(0);
	r[TWO_THREE] =  (double)num/sqrt((double)denom);


	x1 = x1s/(double)n; x2 = x2s/(double)n; x3 = x3s/(double)n;

	s[0] = sqrt(x1sq/(double)n - x1*x1);
	s[1] = sqrt(x2sq/(double)n - x2*x2);
	s[2] = sqrt(x3sq/(double)n - x3*x3);
	/* printf("Std dev: s1 = %lf s2 = %lf s3 = %lf\n",s[0],s[1],s[2]);  */

	num = r[ONE_TWO] - r[ONE_THREE]*r[TWO_THREE];
	denom = 1-r[TWO_THREE]*r[TWO_THREE];
	if (fabs(denom) < 1e-6) return(0);
	if (fabs(s[1]) < 1e-6) return(0);
	if (fabs(s[2]) < 1e-6) return(0);

	b[0] = (num/denom)*(s[0]/s[1]);
	num = r[ONE_THREE] - r[ONE_TWO]*r[TWO_THREE];
	b[1] = (num/denom)*(s[0]/s[2]);
	b[2] = x1 - b[0]*x2 - b[1]*x3;

	num = 1-r[ONE_TWO]*r[ONE_TWO]-r[ONE_THREE]*r[ONE_THREE]-r[TWO_THREE]*r[TWO_THREE] +2*r[ONE_TWO]*r[ONE_THREE]*r[TWO_THREE];

	b[3] = s[0]*sqrt((double)num/denom);
	if (n > 3)
		b[3] = sqrt((double)n/(double)(n-3))*b[3];

	t = P_MKDOUBLE(b[0]);
	if (!P_BIND(ARG4,t))
		return (FALSE);
	t = P_MKDOUBLE(b[1]);
	if (!P_BIND(ARG5,t))
		return (FALSE);
	t = P_MKDOUBLE(b[2]);
	if (!P_BIND(ARG6,t))
		return (FALSE);

	t = P_MKDOUBLE(b[3]);
	if (!P_BIND(ARG7,t))
		return (FALSE);

	myfree;
	return (TRUE);
}


#define	DATABLOCK	100

static double **
enlarge_table(double **v, int incr, long ncols, long *szp)
{
	double **v1;
	long oldsz = *szp, newsz;
	register long int i;

	newsz = oldsz + incr;
	v1 = (double **) mymalloc(newsz*sizeof(double *));
	for (i=0; i<oldsz;i++) 
		v1[i] = v[i];
	for (i=oldsz; i<newsz;i++) 
		v1[i] = (double *) mymalloc(ncols*sizeof(double));
	*szp = newsz;

	return (v1);
}


#define	XCOL(col,dcol)	((dcol)-1>(col)?(col):(col)+1)

static double
sum_prod(double **data, long nrows, long dcol, long col, long row)
{
	double sum = 0.0;
	register long int i;
	long int xcol;
	double xki, xni;

/*	printf("Row = %ld Col = %ld\n",row,col); */
	for (i=0l;i < nrows; i++) {
		xcol = XCOL(col-1,dcol);
/*		printf("x%ld%ld ",i,xcol); */
		xni = data[i][xcol];
		xcol = XCOL(row-1,dcol);
		xki = (row?data[i][xcol]:(double)1.0);
/*		printf("x%ld%ld\n",i,xcol); */
		sum += (xki*xni);
	}

	return (sum);
}

static double *
gauss_solve(double **data, long nrows, long ncols, long dcol)
{

	double **x;
	double *coefs = NULL, *y;
	register long int row,col;
  	double big, dumflt, pivot;
  	int i, icol = 0, irow = 0, j, k, dumint, *coefspos;

/*	printf("in gauss solve\n"); */
	x = (double **)mymalloc(ncols*sizeof(double));
	coefs = (double *)mymalloc(ncols*sizeof(double));
	coefspos = (int *)mymalloc(ncols*sizeof(int));
	y = (double *)mymalloc(ncols*sizeof(double));
	for (row=0l;row < ncols;row++) 
		y[row] = coefs[row] = (double)0.0;
	for (row=0l;row < ncols;row++)
		x[row] = (double *)mymalloc(ncols*sizeof(double));

	x[0][0] = nrows;
	for (row=col=1l;col < ncols;row++,col++)
		x[row][0] = x[0][col] = sum_prod(data,nrows,dcol,col,0l); 
	for(row=1l;row < ncols;row++)
		for(col=1l;col < ncols;col++)
			x[row][col] = sum_prod(data,nrows,dcol,col,row);
/*	print_table(x,ncols,ncols); */

	for(row=0l;row<nrows;row++)
		y[0] += data[row][dcol-1];
	for(col=1l;col<ncols;col++) 
		for (row=0l;row<nrows;row++)
			y[col] += (data[row][dcol-1]*data[row][XCOL(col-1,dcol)]);

	for( i = 0; i < ncols; i++) coefspos[i] = i;

/*****************
	printf( "INITIAL MATRIX\n" );
	for( i = 0; i < ncols; i++) {
		for( j = 0; j < ncols; j++)
	printf( "%.2lf ", x[i][j] );
	printf( "|x[%d]|  %.2lf\n", coefspos[i], y[i] );
	}
	getchar();
*****************/

	for( i = 0; i < ncols - 1; i++) {
		big = 0.0;
		/* look for pivot */
	
		for( j = i; j < ncols; j++) {
			for( k = i; k < ncols; k++)
				if(fabs(x[j][k]) >= big) {
	  				big = fabs(x[j][k]);
	  				irow = j;
	  				icol = k;
				} /* end if */
		} /* END  of for( k = 0; k < ncols; k++) */
		if( irow != i && icol != i){ /* swap line and/or column */
			/* same line different column */
			if(irow == i){ /* update coeficient positions */
		
				dumint = coefspos[i];
				coefspos[i] = icol;
				coefspos[icol] = dumint;
				for( j = 0; j < ncols; j++) {
					dumflt = x[j][i];
					x[j][i] = x[j][icol];
					x[j][icol] = dumflt;
				}
			}
			else if (icol == i) { /* same column */
				for( j = i; j < ncols; j++) {
					dumflt = x[i][j];
					x[i][j] = x[irow][j];
					x[irow][j] = dumflt;
				}
				dumflt = y[i];
				y[i] = y[irow];
				y[irow] = dumflt;
			}
			else { /* different column and row */
      				dumint = coefspos[i];
      				coefspos[i] = icol;
      				coefspos[icol] = dumint;
      				for( j = 0; j < ncols; j++) {
					dumflt = x[j][i];
					x[j][i] = x[j][icol];
					x[j][icol] = dumflt;
      				}
      				for( j = i; j < ncols; j++) {
					dumflt = x[i][j];
					x[i][j] = x[irow][j];
					x[irow][j] = dumflt;
      				}
      				dumflt = y[i];
      				y[i] = y[irow];
      				y[irow] = dumflt;
    			}
		} 
	
		for( j = i + 1; j < ncols; j++){
			pivot = x[j][i] / x[i][i];
			for( k = i; k < ncols; k++)
			x[j][k] -= x[i][k] * pivot;
			y[j] -= y[i] * pivot;
		}

/*****************
		printf( "MATRIX(%d) \n", i );
			for( k = 0; k < ncols; k++) {
				for( j = 0; j < ncols; j++)
				printf( "%.2lf ", x[k][j] );
				printf( "|x[%d]|  %.2lf\n", coefspos[k], y[k] );
			}
		getchar();
*****************/

	} 


/*****************
	printf( "FINAL MATRIX\n" );
	for( i = 0; i < ncols; i++) {
	for( j = 0; j < ncols; j++)
		printf( "%.2lf ", x[i][j] );
		printf( "|x[%d]|  %.2lf\n", coefspos[i], y[i]);
	}
	getchar();
*****************/

	for( j = ncols - 1; j >= 0; j--) {
    		dumflt = y[j];
    		for( k = j+1; k < ncols; k++)
      		dumflt -= coefs[coefspos[k]]*x[j][k];
    		coefs[coefspos[j]] = dumflt / x[j][j];
  	}

/***************** 

	for( i = 0; i < ncols; i++)
		printf( "Coef[%d] = %.2f\n", i, coefs[i] );
	getchar();
*****************/

	return( coefs );
}


static int
l_regressn(void)
{
  	Term *lists, *dp, arg2, arg3, t;
	long int ncols; /*, dcol;*/
	register long col = 0l, nrows = 0l, i;
	double **data = (double **)NULL, *coefs = NULL, sd, err, dum;
	double sumy = 0.0, sumysq = 0.0, ssr, sst, sse, f, rsq;
	long int datasize = 0l;

	if (P_ISINT(P_ARG(ARG1)))
	  ncols = (long int) P_INT(P_ARG(ARG1));
	else return (FALSE);
	if (ncols <= 1) return (FALSE);
	lists = (Term *)mymalloc(ncols*sizeof(Term));
	dp = (Term *)mymalloc(ncols*sizeof(Term));
	arg2 = P_ARG( ARG2 );
	for (col=0l; col< ncols;col++) {
  		lists[col] = P_HEAD(arg2);
		dp[col] = lists[col];
		arg2 = P_TAIL( arg2 );
	}
	while (!P_EMPTY(dp[0])) {
		nrows++;
		if (nrows > datasize)
			data = enlarge_table(data,DATABLOCK,ncols,&datasize);
		for(col=0l;col<ncols;col++) {
			if (!P_ISNUM(P_HEAD(dp[col]))) return (FALSE); 
			P_NUM(P_HEAD(dp[col]),data[nrows-1][col]);
			dp[col] = P_TAIL(dp[col]);
		      }
	}
	coefs = gauss_solve(data,nrows,ncols,1l);
	arg3 = P_NIL;
	for (col=ncols-1l;col >= 0;col--)
	  arg3 = P_MKLIST( P_MKDOUBLE(coefs[col]), arg3);

	/* Computing the error */

	sumy = sumysq = sd = 0.0;

	for( i = 0; i < nrows; i++) {

	  P_NUM(P_HEAD(lists[0]),err);
	  lists[0] = P_TAIL(lists[0]);
	  sumy += err; sumysq += err*err;

	  for (col=1l; col< ncols;col++) {
	    P_NUM(P_HEAD(lists[col]),dum);
	    err -= dum * coefs[col];
	    lists[col] = P_TAIL(lists[col]);
	  }
	  err -= coefs[0];  /* final constant */

	  sd += err * err;
	} 
	/* sd at this point has SSE */
	sse = sd;
	sst = sumysq - (sumy*sumy)/(double)nrows;
	ssr = sst - sse; /* SSR = SST - SSE */


	if (nrows > ncols + 2)
	  sd = sqrt( sd / (double) (nrows-ncols-1-1)); /* unbiased estimate of sigma */
	else sd = sqrt( sd / (double) nrows); /* biased estimate of sigma */

	f = ssr/(sd*sd);		/* Fval */
	/* rsq = (double)1.0 - (sse/(double)(nrows-ncols-1))/(sst/(double)(nrows-1));*/
	rsq = (double)1.0 - sse/sst;


	t = P_MKDOUBLE( sd );
	P_BIND( ARG4, t);
	t = P_MKDOUBLE( f );
	P_BIND( ARG5, t);
	t = P_MKDOUBLE( rsq );
	P_BIND( ARG6, t);

	myfree;
	return (P_BIND( ARG3, arg3));
}
	

static int
l_regress1(void)
{
  Term term, listX, listY;
  double x, y, xx,xy, m, b, dumX,dumY, sd =0.0;
  double sumy, sumysq, ssr, f;
  int counter, i;


  listY = P_ARG(ARG1);
  listX = P_ARG(ARG2);

  if(IsVarTerm(listX) || IsVarTerm(listY) ||
     P_EMPTY(listX) || P_EMPTY(listY))
    return (FALSE);
  else {
    x = 0.0; y = 0.0; counter = 0; xx = 0.0; xy= 0.0;
    dumX = dumY = 0.0;
    
    while(! P_EMPTY(listX) && ! P_EMPTY(listY)){ 

      term = P_HEAD(listX);
      if(P_ISINT(term))
	dumX = (double) P_INT(term);
      else if(P_ISDOUBLE(term))
	dumX = P_DOUBLE(term);
      else return(FALSE);

      term = P_HEAD(listY);
      if(P_ISINT(term))
	dumY = (double) P_INT(term);
      else if(P_ISDOUBLE(term))
	dumY = P_DOUBLE(term);
      else return(FALSE);


      x += dumX;
      y += dumY;
      xy += dumX * dumY;
      xx += dumX * dumX;

      listX = P_TAIL(listX);
      listY = P_TAIL( listY );
      counter++;
      
    } /* END while */
  } /* END else */

  if ( fabs(xx * (double) counter - x * x) < 1e-6) return( 0 );
  m = ( xy * (double) counter - x * y) / ( xx * (double) counter - x * x);
  b = (y - m * x) / (double) counter;

  /* Computing the error */

  listY = P_ARG(ARG1);
  listX = P_ARG(ARG2);

  sumy = sumysq = 0.0;

  for( i = 0; i < counter; i++) {

      term = P_HEAD(listX);
      if(P_ISINT(term))
	dumX = (double) P_INT(term);
      else if(P_ISDOUBLE(term))
	dumX = P_DOUBLE(term);

      term = P_HEAD(listY);
      if(P_ISINT(term))
	dumY = (double) P_INT(term);
      else if(P_ISDOUBLE(term))
	dumY = P_DOUBLE(term);


      listX = TailOfTerm( listX );
      listY = TailOfTerm( listY );

    sd += (dumY - m * dumX - b) * (dumY - m * dumX - b);
    sumy += dumY; sumysq += (dumY*dumY);

  } 

  /* sd at this point has SSE */
  ssr = (sumysq - (sumy*sumy)/(double)(counter)) - sd;
  if (counter > 2)
    sd = sqrt( sd / (double) (counter - 2)); /* unbiased estimate of sigma */
  else sd = sqrt( sd / (double) counter); /* biased estimate of sigma */

  term = MkFloatTerm( m );
  unify( ARG3, term );

  term = MkFloatTerm( b );
  unify( ARG4, term );

  f = ssr/(sd*sd);
  term = MkFloatTerm( f );
  unify( ARG5, term );

  myfree;
  return( 1 );
  
}

/************************************************************
*************************************************************
*************************************************************

static VLIST
pterm_to_ulist(list,n)
Term list;
LONG *n;
{
	register VLIST unsorted = NULL, ul;
	double v;
	LONG i;
	Term pt;

	if (P_EMPTY(list)) {
		*n = 0l;
		return unsorted;
	}
	pt = P_HEAD(list);
	if (P_ISINT(pt)) {
		i = P_LONG(pt);
		v = (double)i;
	}
	else if (P_ISDOUBLE(pt))
		v = P_DOUBLE(pt);
	else return (unsorted);
	unsorted = ul = (vlist *)mymalloc(sizeof(vlist));
	unsorted->val = v;
	unsorted->next = NULL;
	*n = 1l;
	list = P_TAIL(list);
	while (!P_EMPTY(list)) {
		pt = P_HEAD(list);
		if (P_ISINT(pt)) {
			i = P_LONG(pt);
			v = (double)i;
		}
		else if (P_ISDOUBLE(pt))
			v = P_DOUBLE(pt);
		else return (unsorted);
		ul->next = (vlist *)mymalloc(sizeof(vlist));
		ul->next->val = v;
		ul->next->next = NULL;
		ul = ul->next;
		*n = *n + 1l;
		list = P_TAIL(list);
	}
	return (unsorted);
}

#define DIMENSION   3


static int
gaussj( x, dim, y, coefs)
double x[DIMENSION][DIMENSION], y[DIMENSION], coefs[DIMENSION];
int    dim;
{
  double big, dumFlt, pivot;
  int i, icol, irow, j, k, dumInt, coefsPos[DIMENSION];


  for( i = 0; i < dim; i++) coefsPos[i] = i;

*****************
    printf( "INITIAL MATRIX\n" );
    for( i = 0; i < dim; i++) {
      for( j = 0; j < dim; j++)
	printf( "%.2lf ", x[i][j] );
      printf( "|x[%d]|  %.2lf\n", coefsPos[i], y[i] );
    }
getchar();
*****************

  for( i = 0; i < dim - 1; i++) {

    big = 0.0;

    * look for pivot *

    for( j = i; j < dim; j++) {
      for( k = i; k < dim; k++)
	if(fabs(x[j][k]) >= big) {
	  big = fabs(x[j][k]);
	  irow = j;
	  icol = k;
	} * end if *
    } * END  of for( k = 0; k < dim; k++) *

  if( irow != i && icol != i){ * swap line and/or column *
    
    * same line different column *

    if(irow == i){

      * update coeficient positions *

      dumInt = coefsPos[i];
      coefsPos[i] = icol;
      coefsPos[icol] = dumInt;

      for( j = 0; j < dim; j++) {
	dumFlt = x[j][i];
	x[j][i] = x[j][icol];
	x[j][icol] = dumFlt;
      }
    }
    else if (icol == i) { * same column *

      for( j = i; j < dim; j++) {
	dumFlt = x[i][j];
	x[i][j] = x[irow][j];
	x[irow][j] = dumFlt;
      }
      dumFlt = y[i];
      y[i] = y[irow];
      y[irow] = dumFlt;
    }
    else{ * different column and row *

      dumInt = coefsPos[i];
      coefsPos[i] = icol;
      coefsPos[icol] = dumInt;

      for( j = 0; j < dim; j++) {
	dumFlt = x[j][i];
	x[j][i] = x[j][icol];
	x[j][icol] = dumFlt;
      }
      for( j = i; j < dim; j++) {
	dumFlt = x[i][j];
	x[i][j] = x[irow][j];
	x[irow][j] = dumFlt;
      }
      dumFlt = y[i];
      y[i] = y[irow];
      y[irow] = dumFlt;
    }
  } * END   if( irow != i && icol != i){ *

    for( j = i + 1; j < dim; j++){
      pivot = x[j][i] / x[i][i];
      for( k = i; k < dim; k++)
	x[j][k] -= x[i][k] * pivot;
      y[j] -= y[i] * pivot;
    }

*****************
    printf( "MATRIX(%d) \n", i );
    for( k = 0; k < dim; k++) {
      for( j = 0; j < dim; j++)
	printf( "%.2lf ", x[k][j] );
      printf( "|x[%d]|  %.2lf\n", coefsPos[k], y[k] );
    }
    getchar();
*****************

  } * END of   for( i = 0; i < dim - 1; i++) { *


*****************
    printf( "FINAL MATRIX\n" );
    for( i = 0; i < dim; i++) {
      for( j = 0; j < dim; j++)
	printf( "%.2lf ", x[i][j] );
      printf( "|x[%d]|  %.2lf\n", coefsPos[i], y[i]);
    }
    getchar();
*****************

  for( j = dim - 1; j >= 0; j--) {
    dumFlt = y[j];
    for( k = j+1; k < dim; k++)
      dumFlt -= coefs[coefsPos[k]]*x[j][k];
    coefs[coefsPos[j]] = dumFlt / x[j][j];
  }

*****************
  for( i = 0; i < dim; i++)
    printf( "Coef[%d] = %.2f\n", i, coefs[i] );
  getchar();
*****************

  return( 1 );
}


* A very specific polynomial regression: needs to be changed *
static Int
l_polyregress2()
{
   double x[DIMENSION][DIMENSION], y[DIMENSION], coefs[DIMENSION], dumX, dumY, sd;
   Term t, listX, listY;
   double sse, ssr, sst, sumy, sumysq, f, rsq;
   register int counter, i, j;

   listY = P_HEAD(P_ARG(ARG1));
   listX = P_HEAD(P_ARG(ARG2));

   if (P_EMPTY(listX) || P_EMPTY(listY))
     return (FALSE);
   else {
     for( i = 0;i < DIMENSION; i++){
       for( j = 0; j < DIMENSION; j++)
	 x[i][j] = 0.0; 
       y[i] = 0.0;
       coefs[i] = 0.0;
     }
     counter = 0;

     while(! P_EMPTY( listX ) && ! P_EMPTY( listY )){ 
       if (P_ISDOUBLE(P_HEAD(listX)))
	 dumX = (double) P_DOUBLE( HeadOfTerm( listX ));
       else if (P_ISINT(P_HEAD(listX)))
	 dumX = (double)P_INT(P_HEAD(listX));
       else return(FALSE);
       
       listX = P_TAIL(listX);
	
       if (P_ISDOUBLE(P_HEAD(listY)))
	 dumY = (double) P_DOUBLE( HeadOfTerm( listY ));
       else if (P_ISINT(P_HEAD( listY )))
	 dumY = (double) P_INT(P_HEAD(listY));
       else return(FALSE);
	
       listY = P_TAIL( listY );

       x[0][1] += dumX;
       x[0][2] += dumX * dumX;
       y[0]    += dumY;

       x[1][0] += dumX;
       x[1][1] += dumX * dumX;
       x[1][2] += dumX * dumX * dumX;
       y[1]    += dumX * dumY;

       x[2][0] += dumX * dumX;
       x[2][1] += dumX * dumX * dumX;
       x[2][2] += dumX * dumX * dumX * dumX;
       y[2]    += dumX * dumX * dumY;

       counter++;
	
     } * END while *

     x[0][0] = (double) counter;
   } * END else *


   gaussj( x, DIMENSION, y, coefs);

  * Computing the error *

  listY = P_ARG(ARG1);
  listY = P_HEAD(listY);
  listX = P_ARG(ARG2);
  listX = P_HEAD(listX);
  sumy = sumysq = sd = (double)0.0;
  for( i = 0; i < counter; i++) {

      if (IsFloatTerm( HeadOfTerm( listX )))
	dumX = (double) FloatOfTerm( HeadOfTerm( listX ));
      else if (IsIntTerm( HeadOfTerm( listX )))
	dumX = (double) IntOfTerm( HeadOfTerm( listX ));

      listX = TailOfTerm( listX );

      if (IsFloatTerm( HeadOfTerm( listY )))
	dumY = (double) FloatOfTerm( HeadOfTerm( listY ));
      else if (IsIntTerm( HeadOfTerm( listY )))
	dumY = (double) IntOfTerm( HeadOfTerm( listY ));

	sumy += dumY ; sumysq += dumY*dumY;

      listY = TailOfTerm( listY );

    sd += (dumY - coefs[0] - coefs[1] * dumX - coefs[2] * dumX * dumX) *
          (dumY - coefs[0] - coefs[1] * dumX - coefs[2] * dumX * dumX);

  } * END for( i = .... *
  * sd at this point has SSE *
  sse = sd;
  sst = sumysq - (sumy*sumy)/(double)counter;
  ssr = sst - sse; * SSR = SST - SSE *


   sd = sqrt( sd / (double) (counter - 2)); * unbiased estimate of sigma *
   f = ssr/(sd*sd);		* Fval *
   * rsq = (double)1.0 - (sse/(double)(nrows-ncols-1))/(sst/(double)(nrows-1));*
   rsq = (double)1.0 - sse/sst;

  t = MkFloatTerm( coefs[0] );
  unify( ARG3, t );

  t = MkFloatTerm( coefs[1] );
  unify( ARG4, t );

  t = MkFloatTerm( coefs[2] );
  unify( ARG5, t );

  t = MkFloatTerm( sd );
  unify( ARG6, t );

  t = MkFloatTerm( f );
  unify( ARG7, t );

  t = MkFloatTerm( rsq );
  unify( ARG8, t );

  return( 1 );
  
}

#define MAX_SAMPLES	25
#define MAX_CALLS	100


static int
l_circular()
{
	Term pt, list[2];
	register VLIST sl, sl1, xl = NULL, yl = NULL;
	LONG nxl=0l, nyl=0l, t, sel[3];
	register LONG i,j, maxsamples, nsamples, samples;
	LONG point;
	double x[3], y[3];
	double alpha, beta, gamma, delta, c2, c1, a, b, r;
	double sr, ssr, sa, sb, dr;
	static LONG callno=0l;

	list[0] = P_HEAD(P_ARG(ARG1));
	list[1] = P_HEAD(P_ARG(ARG2));

	if (P_EMPTY(list[0]) || P_EMPTY(list[1])) return (FALSE);

	xl = pterm_to_ulist(list[0],&nxl);
	yl = pterm_to_ulist(list[1],&nyl);


	if (nxl != nyl) return (FALSE);
	if (nxl < 3l) return (FALSE);
	maxsamples = nxl*(nxl-1)*(nxl-2)/6;	* n-chose-3 *

	maxsamples = (maxsamples > MAX_SAMPLES)?MAX_SAMPLES:maxsamples;

	if (!callno) {
		time(&t);
		SRAND(t);
		callno++;
	}
	sr = ssr = sa = sb = (double)0.0;
	for (samples=nsamples=0l;(nsamples<maxsamples)&&(samples<1000l);samples++) {
		sel[0] = sel[1] = sel[2] = -1l;
		for (i=0l; i < 3l;i++) {
			point = (LONG)floor((double)RAND*nxl);
			for (j=0l; j<i;)
				if (sel[j] == point) {
					j = 0l;
					point = (LONG)floor((double)RAND*nxl);
				}
				else j++;
			sel[i] = point;
			for (j = 0l,sl=xl,sl1=yl;j<point;j++) {
				sl = sl->next;
				sl1 = sl1->next;
			}
			x[i] = sl->val;
			y[i] = sl1->val;
		}
		* printf("selected %ld %ld %ld [%lf,%lf],[%lf,%lf],[%lf,%lf]\n",sel[0],sel[1],sel[2],x[0],y[0],x[1],y[1],x[2],y[2]);  *
		alpha = 2*(x[1]-x[0]); beta = 2*(y[1]-y[0]);
		gamma = 2*(x[2]-x[0]); delta = 2*(y[2]-y[0]);
		c1 = (x[1]*x[1]-x[0]*x[0]) + (y[1]*y[1]-y[0]*y[0]);
		c2 = (x[2]*x[2]-x[0]*x[0]) + (y[2]*y[2]-y[0]*y[0]);
	
		if ((alpha*delta == gamma*beta) || !(alpha))
			continue;
		nsamples++;
		b = (c2*alpha - c1*gamma)/(alpha*delta - gamma*beta);
		a = (c1 - beta*b)/alpha;
		r = (x[0]-a)*(x[0]-a) + (y[0]-b)*(y[0]-b);
		ssr += r;
		r = sqrt((double)r);
		* printf("estimated %lf %lf %lf\n",a,b,r);  *
		sa += a; sb += b; sr += r;
	}
	*
	for (sl=xl;sl;sl=sl1) {
		sl1= sl->next;
		free((char *)sl);
	}
	for (sl=yl;sl;sl=sl1) {
		sl1= sl->next;
		free((char *)sl);
	}
	*

	if (!nsamples) return (FALSE);

	* printf("Calculating means based on %ld samples\n",nsamples); *

	a = sa/(double)nsamples;
	b = sb/(double)nsamples;
	r = sr/(double)nsamples;
	* printf("sums %lf %lf %lf estimates %lf %lf %lf\n",sa,sb,sr,a,b,r); *
	* dr = ssr/(double)nsamples - (r*r);
	dr = sqrt((double)dr); *

	pt = P_MKDOUBLE(a);
	if (!P_BIND(ARG3,pt))
		return (FALSE);

	pt = P_MKDOUBLE(b);
	if (!P_BIND(ARG4,pt))
		return (FALSE);

	pt = P_MKDOUBLE(r);
	if (!P_BIND(ARG5,pt))
		return (FALSE);

*
	pt = P_MKDOUBLE(dr);
	if (!P_BIND(ARG6,pt))
		return (FALSE);
*

	return (TRUE);
}
	
static int
l_parabolic1()
{
	Term pt, list[2];
	register VLIST sl, sl1, sl2, xl = NULL, yl = NULL;
	LONG nxl=0l, nyl=0l, t, sel[3];
	register LONG i,j, maxsamples, nsamples, samples;
	LONG point;
	double x[3], y[3];
	double alpha, beta, gamma, delta, c2, c1, a, b, r;
	double sr, ssr, sa, sb, dr, dy;
	static LONG callno=0l;

	list[0] = P_HEAD(P_ARG(ARG1));
	list[1] = P_HEAD(P_ARG(ARG2));

	if (P_EMPTY(list[0]) || P_EMPTY(list[1])) return (FALSE);

	xl = pterm_to_ulist(list[0],&nxl);
	yl = pterm_to_ulist(list[1],&nyl);


	if (nxl != nyl) return (FALSE);
	if (nxl < 3l) return (FALSE);
	maxsamples = nxl*(nxl-1)*(nxl-2)/6;	* n-chose-3 *

	maxsamples = (maxsamples > MAX_SAMPLES)?MAX_SAMPLES:maxsamples;

	if (!callno) {
		time(&t);
		SRAND(t);
		callno++;
	}
	sr = ssr = sa = sb = (double)0.0;
	for (samples=nsamples=0l;(nsamples<maxsamples)&&(samples<1000l);samples++) {
		sel[0] = sel[1] = sel[2] = -1l;
		for (i=0l; i < 3l;i++) {
			point = (LONG)floor((double)RAND*nxl);
			for (j=0l; j<i;)
				if (sel[j] == point) {
					j = 0l;
					point = (LONG)floor((double)RAND*nxl);
				}
				else j++;
			sel[i] = point;
			for (j = 0l,sl=xl,sl1=yl;j<point;j++) {
				sl = sl->next;
				sl1 = sl1->next;
			}
			x[i] = sl->val;
			y[i] = sl1->val;
		}
		if (!(x[0]-x[1]) || !(x[0]-x[2]))
			continue;
		* printf("selected %ld %ld %ld [%lf,%lf],[%lf,%lf],[%lf,%lf]\n",sel[0],sel[1],sel[2],x[0],y[0],x[1],y[1],x[2],y[2]); *
		alpha = (y[0]-y[1])/(x[0]-x[1]); beta = (y[0]-y[2])/(x[0]-x[2]);
		gamma = x[0]+x[1]; delta = x[0]+x[2];

		if (!(alpha-beta)) continue;
		a = (alpha*delta-gamma*beta)/2*(alpha-beta);
		r = alpha/(gamma-2*a);
		b = y[0]-r*(x[0]-a)*(x[0]-a);
		ssr += r*r;
		* printf("estimated %lf %lf %lf\n",a,b,r); *
		sa += a; sb += b; sr += r;
		nsamples++;
	}
	if (!nsamples) return (FALSE);
	* printf("Calculating means based on %ld samples\n",nsamples); *

	a = sa/(double)nsamples;
	b = sb/(double)nsamples;
	r = sr/(double)nsamples;
	* printf("sums %lf %lf %lf estimates %lf %lf %lf\n",sa,sb,sr,a,b,r); *
	* dr = ssr/(double)nsamples - (r*r);
	dr = sqrt((double)dr); *

	dy = 0.0;
	for (sl2=yl,sl=xl;sl;sl=sl1) {
		alpha = r*(sl->val-a)*(sl->val-a) +b;
		beta  = sl2->val;
		dy += (alpha-beta)*(alpha-beta);
		sl1= sl->next;
		sl2= sl2->next;
		* free((char *)sl); *
	}
	*
	for (sl=yl;sl;sl=sl1) {
		sl1= sl->next;
		free((char *)sl);
	}
	*
	dy = dy/(double)nyl;
	dy = sqrt((double)dy);

	pt = P_MKDOUBLE(a);
	if (!P_BIND(ARG4,pt))
		return (FALSE);

	pt = P_MKDOUBLE(b);
	if (!P_BIND(ARG5,pt))
		return (FALSE);

	pt = P_MKDOUBLE(r);
	if (!P_BIND(ARG3,pt))
		return (FALSE);

	pt = P_MKDOUBLE(dy);
	if (!P_BIND(ARG6,pt))
		return (FALSE);

	return (TRUE);
}

***************************************
static int
show_list(l)
VLIST l;
{
	register VLIST vl;

	for (vl = l; vl ; vl= vl->next)
		printf("%lf \n",vl->val);
	printf("\n");
}
***************************************




static int
print_table(m,nrows,ncols)
double **m;
long nrows, ncols;
{
	register int i,j;

	for (i=0l; i< nrows; i++) {
		for (j=0l;j<ncols;j++) 
			printf("%0.3lf ",m[i][j]);
		printf("\n");
	}
}

***************************************
static int
print_vector(v,nrows)
double *v;
long nrows;
{
	register int i;

	for (i=0l; i< nrows; i++) {
			printf("%0.3lf  ",v[i]);
		printf("\n");
	}
}


static double *
enlarge_vector(v,incr,szp)
double *v;
int incr;
long *szp;
{
	double *v1;
	long oldsz = *szp;
	register long int i;

	v1 = (double *)mymalloc((incr+oldsz)*sizeof(double));
	for (i=0; i < oldsz; i++) v1[i] = v[i];
	*szp = oldsz + incr;
	*free((char *)v); *

	return (v1);
}

************************************************************
*************************************************************
*************************************************************/



/****************************
SOCKETS
*****************************/



static int close_socket()
{
  Term nm_term = Deref(ARG1);
  char *nm;

  nm=AtomName(AtomOfTerm(nm_term));
  close(sock);
  close(SSAP);

  return(1);
}

static int create_socket()
{
	Term nm_term = Deref(ARG1);
	Term host_term = Deref(ARG2);

	char *nm;
	char  *host ;
	struct hostent *hp ;


	nm=AtomName(AtomOfTerm(nm_term));
	host=AtomName(AtomOfTerm(host_term));

	sock = socket(PF_INET, SOCK_DGRAM, 0);
		if (sock < 0 )
			   {

			   perror("Socket");
			   printf("error is %d \n",errno);	
			   exit(1);
			    }

	fflush(stdout);
 	if ((hp = gethostbyname (host)) == NULL) perror("host");
	fflush(stdout); 

   	name.sin_family = AF_INET;
	fflush(stdout);

	bcopy(hp->h_addr, (void *) &name.sin_addr, hp->h_length); 
	fflush(stdout);

     	name.sin_port = htons(IntOfTerm( Deref(ARG3) ) );
	fflush(stdout);


      if ( bind(sock, &name, sizeof(ag->addr) ) )
	{ perror("Bind");
		exit(1);
	}
      if ( fcntl (sock,F_SETFL,FNDELAY) ) {
      		perror ("Fcntl") ; 
		exit (1) ;
				      }
return(1);

}


int sendcl()
{
	Term nm_term = Deref(ARG1);
	Term msg1_term = Deref(ARG2);
	Term msg2_term = Deref(ARG3);
	Term host_term = Deref(ARG4);

	char *nm;
	char *msg1;
	char *msg2;
	char *host ;
	char buf[BUFSIZE], *s;
	struct sockaddr_in hname;
	struct hostent *hp ;
	int len=0;

	nm=AtomName(AtomOfTerm(nm_term));
	msg1=AtomName(AtomOfTerm(msg1_term));
	msg2=AtomName(AtomOfTerm(msg2_term));
	host=AtomName(AtomOfTerm(host_term));


 hp =  gethostbyname (host);

 hname.sin_family = AF_INET;
 bcopy(hp->h_addr, (void *) &hname.sin_addr, hp->h_length);
 hname.sin_port =  htons( IntOfTerm( Deref(ARG5) ) );

 
 strcpy(buf, msg1);
 s = buf+strlen(msg1);
 len += strlen(msg1);
 *s++ = '\0';
 ++len;

 strcpy(s, msg2);
 s = s+strlen(msg2);
 len+= strlen(msg2);
 *s++ = '\0';
 len++;


 if (sendto(sock, buf, len,0, &hname, sizeof(hname )) <0 )
 perror("send buf to agent");
return(1);
}


static int recl(){

	Term msg1_term;
	Term msg2_term;
	char msg1[2048];
	char msg2[2048];
	char buf[BUFSIZE];
	char *s;

        if (read (sock,buf, sizeof(buf)) >0 )
	   { 
	    strcpy(msg1, buf);
	    s = msg1 + strlen(buf);
	    s = buf + strlen(msg1)+1;
	    strcpy(msg2, s);
	    msg1_term=MkAtomTerm(LookupAtom(msg1));
	    msg2_term=MkAtomTerm(LookupAtom(msg2));
	    unify(ARG1,msg1_term);
	    unify(ARG2,msg2_term);
	    return(1);
	   }
	else
	   { 
	    strcpy(msg1, "noagent");
	    strcpy(msg2, "nomessage");
	    msg1_term=MkAtomTerm(LookupAtom(msg1));
	    msg2_term=MkAtomTerm(LookupAtom(msg2));
	    unify(ARG1,msg1_term);
	    unify(ARG2,msg2_term);
	    return(1); /* era -1 */
	   }
}


int create_SSAP()
{
	Term ssap_term = Deref(ARG1);
	char *ssap;
	ssap=AtomName(AtomOfTerm(ssap_term));
        SSAP = socket(PF_UNIX, SOCK_DGRAM, 0);
	if (SSAP < 0)
           {
                perror("Socket");
		 exit(1);
           }
        sname.sun_family = AF_UNIX;
        strcpy( sname.sun_path, ssap);
        if ( bind(SSAP, &sname, sizeof(sname) ) )
           {    
		perror("Bind");
                exit(1);
          }
	  if ( fcntl (SSAP,F_SETFL,FNDELAY) ) {
			  perror ("Fcntl") ;
			  exit (1) ;

						   }

return(1);
}


/* ############################################################ */

void InitForeignPreds(void)
  {
    /*******
    InitCPred("loadprot",1,l_loadprot,SafePredFlag);
    InitCPred("unloadprot",1,l_unloadprot,SafePredFlag);
    InitCPred("prob",4,l_prob,SafePredFlag);
    InitCPred("loadbayes",1,l_loadbayes,SafePredFlag);
    InitCPred("getbayes",5,l_bayes,SafePredFlag);
    InitCPred("l_polyregress2",8,l_polyregress2,SafePredFlag);
    InitCPred("l_circular",5,l_circular,SafePredFlag);
    InitCPred("l_parabolic1",6,l_parabolic1,SafePredFlag);
    ********/

    /* reading and accessing the examples */

#ifdef DEBUG
    positives = (Term *)AllocCodeSpace(MAX_POSITIVES*sizeof(Term));
    negatives = (Term *)AllocCodeSpace(MAX_NEGATIVES*sizeof(Term));
#endif
    InitCPred("negExample", 2, negExample, SafePredFlag|SyncPredFlag);
    InitCPred("setPositiveExample", 2, setPositiveExample, SafePredFlag|SyncPredFlag);
    InitCPred("setNegativeExample", 2, setNegativeExample, SafePredFlag|SyncPredFlag);
    InitCPred("posExample", 2, posExample, SafePredFlag|SyncPredFlag);
    InitCPred("numberOfNegatives", 1, numberOfNegatives, SafePredFlag|SyncPredFlag);
    InitCPred("numberOfPositives", 1, numberOfPositives, SafePredFlag|SyncPredFlag);
    InitCPred("setNumberOfPositives", 1, setNumberOfPositives, SafePredFlag|SyncPredFlag);
    InitCPred("setNumberOfNegatives", 1, setNumberOfNegatives, SafePredFlag|SyncPredFlag);

    /* coverage predicates */

    InitCPred("posCoverage", 4, posCoverage, SyncPredFlag);
    InitCPred("negCoverage", 4, negCoverage, SyncPredFlag);

    /* IndLog settings */

    InitCPred("setNoise", 2, setNoise, SafePredFlag);
    InitCPred("setBestCover", 1, setBestCover, SafePredFlag);

    InitCPred("l_lteq",2,l_lteq, SafePredFlag);
    InitCPred("l_gteq", 2, l_gteq,SafePredFlag);
    InitCPred("l_interval", 3, l_interval,SafePredFlag);
    InitCPred("l_outsideInterval", 3, l_outsideInterval, SafePredFlag);

    InitCPred("c_lindiscrim2",6,c_lindiscrim2,SafePredFlag);
    InitCPred("c_lindiscrim3",8,c_lindiscrim3,SafePredFlag);

    InitCPred("l_regress1",5,l_regress1,SafePredFlag|SyncPredFlag);
    InitCPred("l_regress2",7,l_regress2,SafePredFlag|SyncPredFlag);
    InitCPred("l_regressn",6,l_regressn,SafePredFlag|SyncPredFlag);

    /* sockets */

    InitCPred("create_socket",3, create_socket, SafePredFlag);
    InitCPred("create_SSAP", 1, create_SSAP, SafePredFlag);
    InitCPred("sendcl", 5, sendcl, SafePredFlag);
    InitCPred("recl", 2, recl, SafePredFlag);
    InitCPred("close_socket", 1, close_socket, SafePredFlag);
}

