/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		grow.c							 *
* Last rev:	Thu Feb 23 1989		vv				 *
* mods:									 *
* comments:	Shifting the stacks down				 *
*									 *
*************************************************************************/

#include "Yap.h"
#include "Yatom.h"
#include "io.h"

static int heap_overflows = 0;
static Int total_heap_overflow_time = 0;

int stack_overflows = 0;
static Int total_stack_overflow_time = 0;

static int trail_overflows = 0;
static Int total_trail_overflow_time = 0;

STATIC_PROTO(Int p_growheap, (void));
STATIC_PROTO(Int p_growstack, (void));
STATIC_PROTO(Int p_inform_trail_overflows, (void));
STATIC_PROTO(Int p_inform_heap_overflows, (void));
STATIC_PROTO(Int p_inform_stack_overflows, (void));

#ifdef FIXED_STACKS

/* Used by do_goal() when we're short of stack space */
int
growtrail(long size)
{
  abort_optyap("notrailleft in function absmi");
  return(FALSE);
}

/* Used by do_goal() when we're short of stack space */
int
growstack(long size)
{
  abort_optyap("nostackleft in function absmi");
  return(FALSE);
}

/* Used by do_goal() when we're short of stack space */
int
growheap()
{
  abort_optyap("noheapleft in function absmi");
  return(FALSE);
}

#else

/* #define undf7  */
/* #define undf5 */

STATIC_PROTO(void MoveGlobal, (void));
STATIC_PROTO(void MoveLocalAndTrail, (void));
STATIC_PROTO(void SetHeapRegs, (void));
STATIC_PROTO(void AdjustGrowHeap, (void));
STATIC_PROTO(void SetStackRegs, (void));
STATIC_PROTO(void AdjustGrowStack, (void));
STATIC_PROTO(int  local_growheap, (long));
STATIC_PROTO(void cpcellsd, (CELL *, CELL *, CELL));
STATIC_PROTO(CELL AdjustAppl, (CELL));
STATIC_PROTO(CELL AdjustPair, (CELL));
STATIC_PROTO(void AdjustRegs, (int));

#define IsHeapP(t) ( CharP(t)>=HeapBase && CharP(t)<=HeapTop )

/* Adjusting cells and pointers to cells */
#define PtoGloAdjust(ptr)	((CELL *)(Unsigned(ptr)+GDiff))
#define PtoTRAdjust(ptr)	((CELL *)(Unsigned(ptr)+Diff))
#define PtoLocAdjust(ptr)	((CELL *)(Unsigned(ptr)+Diff))

#define GlobalAdjust(ptr) 	((ptr)+GDiff)
#define LocalAdjust(ptr)	((ptr)+Diff)
#define TrailAdjust(ptr)	((ptr)+Diff)

/* Testing where a pointer points to */
#define IsOldLocal(ptr)	( CharP(ptr) <= CharP(OldLCL0) && CharP(ptr) > CharP(OldASP))
/* require because the trail might contain dangling pointers */
#define IsOldLocalInTR(ptr)	( CharP(ptr) <= CharP(OldLCL0) && CharP(ptr) > CharP(OldH))
#define IsOldGlobal(ptr)  ( CharP(ptr) >= CharP(OldH0) && CharP(ptr) <= CharP(OldH) )
#define IsOldTrail(ptr)	( CharP(ptr) >= CharP(OldTrailBase) && CharP(ptr) <= CharP(OldTR) )
#define IsOldH(ptr)	( CharP(ptr) == CharP(OldH) )
#define IsGlobal(ptr)	( CharP(ptr) >= CharP(H0) && CharP(ptr) <= CharP(H) )
#define IsHeap(ptr)	( CharP(ptr)<CharP(HeapTop)&&CharP(ptr)>=CharP(HeapBase))

#define CharP(ptr)	((char *) (ptr))

static void
cpcellsd(register CELL *Dest, register CELL *Org, CELL NOf)
{
#if HAVE_MEMMOVE
  memmove((void *)Dest, (void *)Org, NOf*sizeof(CELL));
#else
  register Int    n_of = NOf;
  for (; n_of >= 0; n_of--)
    *--Dest = *--Org;
#endif
}


/* The old stack pointers */
static CELL    *OldASP, *OldLCL0;
static tr_fr_ptr OldTR;
static CELL    *OldH0, *OldH;
static ADDR     OldTrailBase, OldTrailTop;
static ADDR     OldAuxTop;

/* The difference between the old stack pointers and the new ones */
static CELL     Diff, GDiff;

static void
SetHeapRegs(void)
{
#ifdef undf7
	YP_fprintf(YP_stderr,"HeapBase = %x\tHeapTop=%x\nGlobalBase=%x\tGlobalTop=%x\nLocalBase=%x\tLocatTop=%x\n", HeapBase, HeapTop, GlobalBase, H, LCL0, ASP);
#endif
	/* The old stack pointers */
	OldLCL0 = LCL0;
	OldASP = ASP;
	OldH0 = H0;
	OldH = H;
	OldTrailBase = TrailBase;
 	OldTrailTop = TrailTop;
	OldTR = TR;
	OldAuxTop = AuxTop;
	/* Adjust stack addresses */
	TrailBase = TrailAdjust(TrailBase);
	TrailTop = TrailAdjust(TrailTop);
	GlobalBase += Diff;
	LocalBase += Diff;
	AuxSp = PtoGloAdjust(AuxSp);
	AuxTop = (ADDR)PtoGloAdjust(((CELL *)AuxTop));
	/* The registers pointing to one of the stacks */
	ENV = PtoLocAdjust(ENV);
	ASP = PtoLocAdjust(ASP);
	H0 = PtoGloAdjust(H0);
	LCL0 = PtoLocAdjust(LCL0);
	H = PtoGloAdjust(H);
	HB = PtoGloAdjust(HB);
	B = (choiceptr)PtoLocAdjust(CellPtr(B));
	TR = (tr_fr_ptr)PtoTRAdjust(TR);
	YENV = PtoLocAdjust(YENV);
	if (IsOldGlobal(S))
		S = PtoGloAdjust(S);
	if (MyTR)
		MyTR = (tr_fr_ptr)PtoTRAdjust(MyTR);
#ifdef COROUTINING
#ifdef MULTI_ASSIGNMENT_VARIABLES
	MutableList = AbsAppl(PtoGloAdjust(RepAppl(MutableList)));
	WokenGoals = AbsAppl(PtoGloAdjust(RepAppl(WokenGoals)));
#endif
#endif

}

static void
SetStackRegs(void)
{
	/* The old local stack pointers */
	OldLCL0 = LCL0;
	OldASP = ASP;
	OldTrailTop = TrailTop;
	OldTrailBase = TrailBase;
	OldTR = TR;
	OldH = H;
	/* The local and aux stack addresses */
	LocalBase += Diff;
	TrailTop = (ADDR)PtoLocAdjust(TrailTop);
	TrailBase = (ADDR)PtoLocAdjust(TrailBase);
	TR = (tr_fr_ptr)PtoLocAdjust(TR);
	/* The registers pointing to the local stack */
	ENV = PtoLocAdjust(ENV);
	ASP = PtoLocAdjust(ASP);
	LCL0 = PtoLocAdjust(LCL0);
	B = (choiceptr)PtoLocAdjust(CellPtr(B));
	YENV = PtoLocAdjust(YENV);
}

static void
MoveLocalAndTrail(void)
{
	/* cpcellsd(To,From,NOfCells) - copy the cells downwards  */
#if HAVE_MEMMOVE
  cpcellsd(ASP, OldASP, (CELL *)OldTR - OldASP);
#else
  cpcellsd((CELL *)TR, (CELL *)OldTR, (CELL *)OldTR - OldASP);
#endif
}

static void
MoveGlobal(void)
{
	/*
	 * cpcellsd(To,From,NOfCells) - copy the cells downwards - in
	 * absmi.asm 
	 */
#if HAVE_MEMMOVE
  cpcellsd(H0, OldH0, OldH - OldH0);  
#else
  cpcellsd(H, OldH, OldH - OldH0);
#endif
}

static inline CELL
AdjustAppl(register CELL t0)
{
  register CELL   t = (CELL) RepAppl(t0);

  if (IsOldGlobal(t))
    return (AbsAppl((CELL *)(GlobalAdjust(t))));
  else if (IsHeapP(t))
    return (AbsAppl((CELL *)t));
#ifdef undf7
  else {
    YP_fprintf(YP_stderr,"[ Fatal error: garbage %lx found in stacks by stack shifter ]\n", t0);
    return (t0);
  }
#else
  return(t0);
#endif
}

static inline CELL
AdjustPair(register CELL t0)
{
  register CELL   t = (CELL) RepPair(t0);

  if (IsOldGlobal(t))
    return (AbsPair((CELL *)(GlobalAdjust(t))));
  else if (IsHeapP(t))
    return (AbsPair((CELL *)t));
#ifdef undf7
  else {
    YP_fprintf(YP_stderr,"[ Fatal error: garbage %lx found in stacks by stack shifter ]\n", t0);
    return (t0);
  }
#else
  return(t0);
#endif
}

/*
 * When growing the stack we need to adjust: the local stack cells pointing
 * to the local; the local stack cells and the X terms pointing to the global
 * (just once) the trail cells pointing both to the global and to the local
 */
static void
AdjustGrowHeap(void)
{
  register CELL  *pt;
  register tr_fr_ptr ptt;

  /* Phase 1 - Adjust the local stack */

  ptt = TR;
  /* moving the trail is simple */
  while (ptt != (tr_fr_ptr)TrailBase) {
    register CELL reg = TrailTerm(ptt-1);

    ptt--;
    if (IsVarTerm(reg)) {
      if (IsOldLocalInTR(reg))
	TrailTerm(ptt) = LocalAdjust(reg);
      else if (IsOldGlobal(reg))
	TrailTerm(ptt) = GlobalAdjust(reg);
    }
#ifdef MULTI_ASSIGNMENT_VARIABLES /* does not work with new structures */
    /* check it whether we are protecting a
       multi-assignment */
    else if (IsPairTerm(reg)) {
      reg = (CELL)RepPair(reg);
      if (IsOldGlobal(reg) || IsOldLocal(reg)) {
	/* multi-assignment variable */
	/* we don't actually need to do anything here,
	   unless we are pointing to a mutable variable,
	   which itself is pointing to the trail. In this
	   case, we need to fix the mutable variable */
	if (IsOldGlobal(reg))
	  TrailTerm(ptt) = AbsPair((CELL *)GlobalAdjust(reg));
	else
	  TrailTerm(ptt) = AbsPair((CELL *)LocalAdjust(reg));
	/* now, go to the value field */
	reg = TrailTerm(ptt-1);
	ptt--;
	if (IsVarTerm(reg)) {
	  if (IsOldLocal(reg))
	    TrailTerm(ptt) = LocalAdjust(reg);
	  else if (IsOldGlobal(reg))
	    TrailTerm(ptt) = GlobalAdjust(reg);
	  else if (IsOldTrail(reg))
	    TrailTerm(ptt) = TrailAdjust(reg);
	} else if (IsApplTerm(reg))
	  TrailTerm(ptt) = AdjustAppl(reg);
	else if (IsPairTerm(reg))
	  TrailTerm(ptt) = AdjustPair(reg);
      }
    }
#endif
  }

  pt = LCL0;

  do {
    register CELL reg = *--pt;
    if (IsVarTerm(reg)) {
      if (IsOldLocal(reg))
	*pt = LocalAdjust(reg);
      else if (IsOldGlobal(reg))
	*pt = GlobalAdjust(reg);
      else if (IsOldTrail(reg))
	*pt = TrailAdjust(reg);
    } else if (IsApplTerm(reg)) {
      *pt = AdjustAppl(reg);
    } else if (IsPairTerm(reg)) {
      *pt = AdjustPair(reg);
    }
  } while (pt > ASP);


  /*
   * to clean the heap now that functors are just variables pointing to
   * the code 
   */
  pt = CellPtr(H0);
  while (pt < H) {
    register CELL reg;
    
    reg = *pt;
    if (reg == (CELL)FunctorDouble) {
      pt += 1 + sizeof(Float)/sizeof(CELL);
      if (pt[0] != (CELL)FunctorEndDouble) {
	pt -= 1 + sizeof(Float)/sizeof(CELL);
	goto normal_cell;
      }
    } else if (reg == (CELL)FunctorLongInt) {
      pt += 2;
      if (pt[0] != (CELL)FunctorEndLongInt) {
	pt -= 2;
	goto normal_cell;
      }
    } else {
    normal_cell:
      if (IsVarTerm(reg)) {
	if (IsOldGlobal(reg))
	  *pt = GlobalAdjust(reg);
#ifdef MULTI_ASSIGNMENT_VARIABLES
	else if (IsOldTrail(reg))
	  *pt = TrailAdjust(reg);
#endif
      } else if (IsApplTerm(reg))
	*pt = AdjustAppl(reg);
      else if (IsPairTerm(reg))
	*pt = AdjustPair(reg);
    }
    pt++;
  }
}

/*
 * When growing the stack we need to adjust: the local cells pointing to the
 * local; the trail cells pointing to the local
 */
static void
AdjustGrowStack(void)
{
  register CELL   reg, *pt;
  register tr_fr_ptr ptt;

  /* Adjusting the trail */
  ptt = TR;
  while (ptt != (tr_fr_ptr)TrailBase) {
    register CELL reg = TrailTerm(ptt-1);
    ptt --;
    if (IsVarTerm(reg)) {
      if (IsOldLocalInTR(reg)) {
	TrailTerm(ptt) = LocalAdjust(reg);
      } else if (IsOldTrail(reg)) {
	TrailTerm(ptt)  = TrailAdjust(reg);
      }
    }
#if MULTI_ASSIGNMENT_VARIABLES
    /* check it whether we are protecting a
       multi-assignment */
    else if (IsPairTerm(reg)) {
      reg = (CELL)RepPair(reg);
      if (IsOldGlobal(reg) || IsOldLocal(reg)) {
	/* multi-assignment variable */
	/* we don't actually need to do anything here,
	   unless we are pointing to a mutable variable,
	   which itself is pointing to the trail. In this
	   case, we need to fix the mutable variable */
	if (IsOldGlobal(reg))
	  TrailTerm(ptt) = AbsPair((CELL *)GlobalAdjust(reg));
	else
	  TrailTerm(ptt) = AbsPair((CELL *)LocalAdjust(reg));
	/* we are not going to adjust the global, but we
	   still need to adjust pointers from the global
	   to the trail (the updatable variables)
	*/
	if (IsOldTrail(*(CELL *)reg)) {
	  *(CELL *)reg = TrailAdjust(*(CELL *)reg);
	}
	/* now, go to the value field */
	reg = TrailTerm(ptt-1);
	ptt--;
	if (IsVarTerm(reg)) {
	  if (IsOldLocal(reg))
	    TrailTerm(ptt) = LocalAdjust(reg);
	  else if (IsOldGlobal(reg))
	    TrailTerm(ptt) = GlobalAdjust(reg);
	  else if (IsOldTrail(reg))
	    TrailTerm(ptt) = TrailAdjust(reg);
	} else if (IsApplTerm(reg))
	  TrailTerm(ptt) = AdjustAppl(reg);
	else if (IsPairTerm(reg))
	  TrailTerm(ptt) = AdjustPair(reg);
      }
    }
#endif
  }

  /* Adjusting the local */
  pt = LCL0;
  while (pt > ASP) {
    reg = *--pt;
    if (IsVarTerm(reg)) {
      if (IsOldLocal(reg))
	*pt = LocalAdjust(reg);
      else if (IsOldTrail(reg))
	*pt = TrailAdjust(reg);
    }
  }

}

static void
AdjustRegs(int n)
{
	int            i;
	CELL	       reg;
	for (i = 1; i < n; ++i) {
	  reg = (CELL) XREGS[i];
	  if (IsVarTerm(reg)) {
	    if (IsOldLocal(reg))
	      reg = LocalAdjust(reg);
	    else if (IsOldGlobal(reg))
	      reg = GlobalAdjust(reg);
	    else if (IsOldTrail(reg))
	      reg = TrailAdjust(reg);
	  } else if (IsApplTerm(reg))
	    reg = AdjustAppl(reg);
	  else if (IsPairTerm(reg))
	    reg = AdjustPair(reg);
	  XREGS[i] = (Term) reg;
	}
}

/* Used by do_goal() when we're short of heap space */
static int
local_growheap(long size)
{
  Int start_growth_time, growth_time;
  int gc_verbose;

  if (!ExtendWorkSpace(size)) {
    return(FALSE);
  }
  start_growth_time = cputime();
  gc_verbose = is_gc_verbose();
  heap_overflows++;
  if (gc_verbose) {
    YP_fprintf(YP_stderr, "[HO] Heap overflow %d\n", heap_overflows);
    YP_fprintf(YP_stderr, "[HO]   growing the heap %ld bytes\n", size);
  }
  ASP -= 256;
  Diff = size;
  GDiff = Diff;
  YAPEnterCriticalSection();
  SetHeapRegs();
  MoveLocalAndTrail();
  MoveGlobal();
  AdjustGrowHeap();
  AdjustRegs(MaxTemps);
  YAPLeaveCriticalSection();
  ASP += 256;
  growth_time = cputime()-start_growth_time;
  total_heap_overflow_time += growth_time;
  if (gc_verbose) {
    YP_fprintf(YP_stderr, "[HO]   took %g sec\n", (double)growth_time/1000);
    YP_fprintf(YP_stderr, "[HO] Total of %g sec expanding heap \n", (double)total_heap_overflow_time/1000);
  }
  return(TRUE);
}

int
growheap()
{
  long size = sizeof(CELL) * 16 * 1024L;
  int shift_factor = (heap_overflows > 8 ? 8 : heap_overflows);
  int sz =  size << shift_factor;
  while(sz >= sizeof(CELL) * 16 * 1024L && !local_growheap(sz)) {
    size = size/2;
    sz =  size << shift_factor;
  }
  return(sz >= sizeof(CELL) * 16 * 1024L);
}

/* Used by do_goal() when we're short of stack space */
int
growstack(long size)
{
  Int start_growth_time, growth_time;
  int gc_verbose;

  if (!ExtendWorkSpace(size)) {
    return(FALSE);
  }
  start_growth_time = cputime();
  gc_verbose = is_gc_verbose();
  stack_overflows++;
  if (gc_verbose) {
    YP_fprintf(YP_stderr, "[SO] Stack overflow %d\n", stack_overflows);
#if SHORT_INTS
    YP_fprintf(YP_stderr, "[SO] Heap: %8ld cells (%p-%p)\n", H-H0,H0,H);
    YP_fprintf(YP_stderr, "[SO] Local:%8ld cells (%p-%p)\n", LCL0-ASP,LCL0,ASP);
    YP_fprintf(YP_stderr, "[SO] Trail:%8ld cells (%p-%p)\n",
	       TR-(CELL *)TrailBase,TrailBase,TR);
#else
    YP_fprintf(YP_stderr, "[SO] Heap: %8d cells (%p-%p)\n", H-H0,H0,H);
    YP_fprintf(YP_stderr, "[SO] Local:%8d cells (%p-%p)\n", LCL0-ASP,LCL0,ASP);
    YP_fprintf(YP_stderr, "[SO] Trail:%8d cells (%p-%p)\n",
	       TR-(tr_fr_ptr)TrailBase,TrailBase,TR);
#endif
    YP_fprintf(YP_stderr, "[SO]    growing the stacks %ld bytes\n", size);
  }
  Diff = size;
  GDiff = 0;
  ASP -= 256;
  YAPEnterCriticalSection();
  SetStackRegs();
  MoveLocalAndTrail();
  AdjustGrowStack();
  AdjustRegs(MaxTemps);
  YAPLeaveCriticalSection();
  CreepFlag = MinStackGap*(stack_overflows+1);
  ASP += 256;
  growth_time = cputime()-start_growth_time;
  total_stack_overflow_time += growth_time;
  if (gc_verbose) {
    YP_fprintf(YP_stderr, "[SO]    took %g sec\n", (double)growth_time/1000);
    YP_fprintf(YP_stderr, "[SO] Total of %g sec expanding stacks \n", (double)total_stack_overflow_time/1000);
  }
  return(TRUE);
}

/* Used by do_goal() when we're short of stack space */
int
growtrail(long size)
{
  Int start_growth_time = cputime(), growth_time;
  int gc_verbose = is_gc_verbose();

  trail_overflows++;
  if (gc_verbose) {
    YP_fprintf(YP_stderr, "[TO] Trail overflow %d\n", trail_overflows);
    YP_fprintf(YP_stderr, "[TO]    growing the trail %ld bytes\n", size);
  }
  if (!ExtendWorkSpace(size)) {
    return(FALSE);
  }
  YAPEnterCriticalSection();
  TrailTop += size;
  YAPLeaveCriticalSection();
  growth_time = cputime()-start_growth_time;
  total_trail_overflow_time += growth_time;
  if (gc_verbose) {
    YP_fprintf(YP_stderr, "[TO]   took %g sec\n", (double)growth_time/1000);
    YP_fprintf(YP_stderr, "[TO] Total of %g sec expanding stacks \n", (double)total_stack_overflow_time/1000);
  }
  return(TRUE);
}


#endif /* FIXED_STACKS */

static Int
p_inform_trail_overflows()
{
  Term tn = MkIntTerm(trail_overflows);
  Term tt = MkIntegerTerm(total_trail_overflow_time);
 
  return(unify(tn, ARG1) && unify(tt, ARG2));
}

/* :- grow_heap(Size) */
static Int
p_growheap(void)
{
#ifndef FIXED_STACKS
  Int             diff;
  Term t1 = Deref(ARG1);

  if (IsVarTerm(t1)) {
    Error(INSTANTIATION_ERROR, t1, "grow_heap/1");
    return(FALSE);
  } else if (!IsIntTerm(t1)) {
    Error(TYPE_ERROR_INTEGER, t1, "grow_heap/1");
    return(FALSE);
  }
  diff = IntOfTerm(t1);
  if (diff < 0) {
    Error(DOMAIN_ERROR_NOT_LESS_THAN_ZERO, t1, "grow_heap/1");
  }
  return(local_growheap(diff));
#endif /* FIXED_STACKS */
  return (TRUE);
}

static Int
p_inform_heap_overflows()
{
  Term tn = MkIntTerm(heap_overflows);
  Term tt = MkIntegerTerm(total_heap_overflow_time);
 
  return(unify(tn, ARG1) && unify(tt, ARG2));
}

/* :- grow_stack(Size) */
static Int
p_growstack(void)
{
#ifndef FIXED_STACKS
  Int             diff;
  Term t1 = Deref(ARG1);

  if (IsVarTerm(t1)) {
    Error(INSTANTIATION_ERROR, t1, "grow_stack/1");
    return(FALSE);
  } else if (!IsIntTerm(t1)) {
    Error(TYPE_ERROR_INTEGER, t1, "grow_stack/1");
    return(FALSE);
  }
  diff = IntOfTerm(t1);
  if (diff < 0) {
    Error(DOMAIN_ERROR_NOT_LESS_THAN_ZERO, t1, "grow_stack/1");
  }
  return(growstack(diff));
#endif /* FIXED_STACKS */
  return (TRUE);
}

static Int
p_inform_stack_overflows()
{
  Term tn = MkIntTerm(stack_overflows);
  Term tt = MkIntegerTerm(total_stack_overflow_time);
 
  return(unify(tn, ARG1) && unify(tt, ARG2));

}

Int total_stack_shift_time(void)
{
  return(total_heap_overflow_time+
	 total_stack_overflow_time+
	 total_trail_overflow_time);
}

void
InitGrowPreds(void)
{
	InitCPred("$grow_heap", 1, p_growheap, SafePredFlag);
	InitCPred("$grow_stack", 1, p_growstack, SafePredFlag);
	InitCPred("$inform_trail_overflows", 2, p_inform_trail_overflows, SafePredFlag);
	InitCPred("$inform_heap_overflows", 2, p_inform_heap_overflows, SafePredFlag);
	InitCPred("$inform_stack_overflows", 2, p_inform_stack_overflows, SafePredFlag);
	init_gc();
}
