/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		init.c							 *
* Last rev:								 *
* mods:									 *
* comments:	initializing a prolog session				 *
*									 *
*************************************************************************/
#ifdef SCCS
static char     SccsId[] = "%W% %G%";
#endif

/*
 * The code from this file is used to initialize the environment for prolog 
 *
 */

#include "Yap.h"
#include "io.h"
#include "clause.h"
#ifdef LOW_LEVEL_TRACER
#include "tracer.h"
#endif

#ifdef DEBUG
#ifdef MACC
STATIC_PROTO(void  InTTYLine, (char *));
#endif
#endif
STATIC_PROTO(void  SetOp, (int, int, char *));
STATIC_PROTO(void  InitOps, (void));
STATIC_PROTO(void  InitDebug, (void));
STATIC_PROTO(void  CleanBack, (PredEntry *, CPredicate, CPredicate));
STATIC_PROTO(void  InitFlags, (void));
STATIC_PROTO(void  InitCodes, (void));
STATIC_PROTO(void  InitFuncs, (void));
STATIC_PROTO(void  InitVersion, (void));

STD_PROTO(void  InitModules, (void));
STD_PROTO(void  exit, (int));

#ifdef DEBUG
char            Option[20];
#endif

/**************	YAP PROLOG GLOBAL VARIABLES *************************/

/*************** variables supporting transparent waking up of goals
  from library predicates ***************/
#ifdef COROUTINING
CELL *SusRecTop, *SusRecBase;       
#endif

/*************** variables concerned with atoms	table ***************/
#ifndef YAPOR
Atom            HashChain[MaxHash];
#endif /* YAPOR */


/************* variables related to memory allocation ***************/
#ifndef YAPOR
Int             HeapUsed, HeapMax;
ADDR		HeapTop;
#endif /* YAPOR */
ADDR            HeapBase, LocalBase, GlobalBase, TrailBase, TrailTop, ForeignCodeBase, ForeignCodeTop, ForeignCodeMax;

#if      USE_OFFSETS
ADDR            AtomBase;
#endif

/*************** atoms and functors required by	the code ************/
Atom            AtomNil,
		AtomComma,
		AtomAssert,
		AtomFail,
		AtomTrue,
		AtomFalse,
		AtomCut,
		AtomRepeat,
		AtomLive,
		AtomPortray,
		AtomBreak,
		AtomDot,
		AtomIndex,
		AtomFast,
		AtomAbol,
		AtomCsult,
		AtomOtherwise,
		AtomVersionNumber,
		AtomPlus,
		AtomMinus,
		AtomCatch,
		AtomThrow,
                AtomProfile,
                AtomSpyCreep,
                AtomAlarm;

#ifdef YAP_ARRAYS
Atom		AtomArray, AtomArrayAccess;
Functor		FunctorArrayEntry;
#endif
Functor         FunctorOr,
		FunctorComma,
		FunctorAssert,
		FunctorQuery,
		ListFunctor,
		FunctorNot,
		AltFunctorNot,
		FunctorCall,
		FunctorVar,
		FunctorPortray,
		FunctorArrow,
		FunctorCsult,
		FunctorBraces,
		FunctorGVar,
		FunctorGAtom,
		FunctorGInteger,
		FunctorGNumber,
		FunctorGAtomic,
		FunctorGPrimitive,
		FunctorGCompound,
		FunctorGFloat,
		FunctorVBar,
		FunctorEq,
                FunctorModule,
                FunctorNot;

#ifdef MULTI_ASSIGNMENT_VARIABLES
Functor		FunctorMutable;
#endif

/* Functor		FunctorDouble, FunctorLongInt, FunctorGeneralRef, FunctorDBRef; */

/************ variables	concerned with Error Handling *************/
char           *ErrorMessage;	/* used to pass error messages */
Term           Error_Term;	/* used to pass error terms */
yap_error_number  Error_TYPE;	/* used to pass the error */

/********* flags for comunication with prolog system predicates	*****/
int             Parm[10];

#ifdef MPWSHELL
/********** informing if we are in the MPW shell ********************/

int             mpwshell = FALSE;

#endif

#ifdef EMACS

int             emacs_mode = FALSE;
char            emacs_tmp[256], emacs_tmp2[256];

#endif

#if HAVE_SIGNAL

int             snoozing = FALSE;

#endif

/********* common code for heap registers *********************/
struct various_codes *heap_regs;

/************ variable concerned with version number *****************/
char           *version_number;

/******************** compiler information ***************************/

int             optimizer_on = TRUE;

/******************* the line for the current parse **********************/

int             StartLine = 1;
int             StartCh = 0;
int             CurFileNo = 0;

int             compile_mode = 0;

/********* Prolog State ********************************************/

int             PrologMode = BootMode;

int             LookupMode = NormalLookup;

#if PUSH_REGS
REGSTORE *regp;

#else

REGSTORE REGS;

#endif

/* module data */

Term            ModuleName[MaxModules];
int             NoOfModules = 1;
SMALLUNSGN      CurrentModule = 0;
int             UserModule = 1;


/************** Access to yap initial arguments ***************************/

char          **yap_args;
int             yap_argc;

/************** Extensions to Terms ***************************************/

/* array with the ops for your favourite extensions */
ext_op attas[last_ext];

/************** YAP Flags ***************************************/

Int yap_flags[NUMBER_OF_YAP_FLAGS];

/**************	declarations local to init.c ************************/
static char    *optypes[] =
{"", "xfx", "xfy", "yfx", "xf", "yf", "fx", "fy"};

#define	xfx	1
#define	xfy	2
#define	yfx	3
#define	xf	4
#define	yf	5
#define	fx	6
#define	fy	7

#define Nill NULL

#ifdef DEBUG

#define	LOGFILE	"logfile"

 /* static */ FILE *logfile;

int 
DebugPutc(ch)
	int             ch;
{
	if (Option['l' - 96])
		(void) YP_putc(ch, logfile);
	return (YP_putchar(ch));
}

static char     my_line[200] = {0};
static char    *lp = my_line;

static YP_File     curfile;

#ifdef MACC

static void 
InTTYLine(char *line)
{
	char           *p = line;
	char            ch;
	while ((ch = InKey()) != '\n' && ch != '\r')
		if (ch == 8) {
			if (line < p)
				BackupTTY(*--p);
		} else
			TTYChar(*p++ = ch);
	TTYChar('\n');
	*p = 0;
}

#endif

void 
DebugSetIFile(char *fname)
{
	if (curfile)
		YP_fclose(curfile);
	curfile = YP_fopen(fname, "r");
	if (curfile == Nill) {
		curfile = YP_stdin;
		YP_fprintf(YP_stderr,"[ Warning: can not open %s for input]\n", fname);
	}
}

void 
DebugEndline()
{
	*lp = 0;

}

static       int   eolflg = 1;

int 
DebugGetc()
{
	int             ch;
	if (eolflg) {
		if (curfile != Nill) {
			if (YP_fgets(my_line, 200, curfile) == 0)
				curfile = Nill;
		}
		if (curfile == Nill)
			YP_fgets(my_line, 200, YP_stdin);
		eolflg = 0;
		lp = my_line;
	}
	if ((ch = *lp++) == 0)
		ch = '\n', eolflg = 1;
	if (Option['l' - 96])
		YP_putc(ch, logfile);
	return (ch);
}

#endif


void 
UserCPredicate(char *name, CPredicate def, unsigned int arity)
{
	InitCPred(name, arity, def, UserCPredFlag);
}

void 
UserBackCPredicate(char *name, CPredicate init, CPredicate cont,
		   unsigned int arity, int extra)
{
	InitCPredBack(name, arity, extra, init, cont, UserCPredFlag);
}

int IsOpType(char *type)
{
  int i;

  for (i = 1; i <= 7; ++i)
    if (strcmp(type, optypes[i]) == 0)
      break;
  return (i <= 7);
}

int 
OpDec(int p, char *type, Atom a)
{
  int             i;
  OpEntry        *info = RepOpProp(GetAProp(a, OpProperty));

  for (i = 1; i <= 7; ++i)
    if (strcmp(type, optypes[i]) == 0)
      break;
  if (i > 7) {
    Error(DOMAIN_ERROR_OPERATOR_SPECIFIER,MkAtomTerm(LookupAtom(type)),"op/3");
    return(FALSE);
  }
  if (EndOfPAEntr(info)) {
    info = (OpEntry *) AllocAtomSpace(sizeof(OpEntry));
    info->KindOfPE = Ord(OpProperty);
    info->NextOfPE = RepAtom(a)->PropOfAE;
    RepAtom(a)->PropOfAE = AbsOpProp(info);
    info->Prefix = info->Infix = info->Posfix = 0;
  }
  if (p) {
    if (i == 1 || i == 2 || i == 4)
      p |= DcrlpFlag;
    if (i == 1 || i == 3 || i == 6)
      p |= DcrrpFlag;
  }
  if (i <= 3) {
    if (info->Posfix != 0) /* there is a posfix operator */ {
      /* ISO dictates */
      Error(PERMISSION_ERROR_CREATE_OPERATOR,MkAtomTerm(a),"op/3");
      return(FALSE);
    }
    info->Infix = p;
  } else if (i <= 5) {
    if (info->Infix != 0) /* there is an infix operator */ {
      /* ISO dictates */
      Error(PERMISSION_ERROR_CREATE_OPERATOR,MkAtomTerm(a),"op/3");
      return(FALSE);
    }
    info->Posfix = p;
  } else {
    info->Prefix = p;
  }
  return (TRUE);
}

static void 
SetOp(int p, int type, char *at)
{
#ifdef DEBUG
  if (Option[5])
    YP_fprintf(YP_stderr,"[setop %d %s %s]\n", p, optypes[type], at);
#endif
  OpDec(p, optypes[type], LookupAtom(at));
}

/* Gets the info about an operator in a prop */
Atom 
GetOp(OpEntry *pp, int *prio, int fix)
{
  int             n;
  SMALLUNSGN      p;

  if (fix == 0) {
    p = pp->Prefix;
    if (p & DcrrpFlag)
      n = 6, *prio = (p ^ DcrrpFlag);
    else
      n = 7, *prio = p;
  } else if (fix == 1) {
    p = pp->Posfix;
    if (p & DcrlpFlag)
      n = 4, *prio = (p ^ DcrlpFlag);
    else
      n = 5, *prio = p;
  } else {
    p = pp->Infix;
    if ((p & DcrrpFlag) && (p & DcrlpFlag))
      n = 1, *prio = (p ^ (DcrrpFlag | DcrlpFlag));
    else if (p & DcrrpFlag)
      n = 3, *prio = (p ^ DcrrpFlag);
    else if (p & DcrlpFlag)
      n = 2, *prio = (p ^ DcrlpFlag);
    else
      n = 4, *prio = p;
  }
  return (LookupAtom(optypes[n]));
}

typedef struct OPSTRUCT {
	char           *opName;
	short int       opType, opPrio;
}               Opdef;

static Opdef    Ops[] = {
  {":-", xfx, 1200},
  {"-->", xfx, 1200},
  {"?-", fx, 1200},
  {":-", fx, 1200},
  {"dynamic", fx, 1150},
  {"mode", fx, 1150},
  {"public", fx, 1150},
  {"multifile", fx, 1150},
  {"meta_predicate", fx, 1150},
#ifdef YAPOR
  {"sequential", fx, 1150},
#endif /* YAPOR */
#ifdef TABLING
  {"table", fx, 1150},
#endif /* TABLING */
  {";", xfy, 1100},
  {"|", xfy, 1100},
  /*  {";", yf, 1100}, not allowed in ISO */
  {"->", xfy, 1050},
  {",", xfy, 1000},
  {".", xfy, 999},
  {"\\+", fy, 900},
  {"not", fy, 900},
  {"=", xfx, 700},
  {"\\=", xfx, 700},
  {"is", xfx, 700},
  {"=..", xfx, 700},
  {"==", xfx, 700},
  {"\\==", xfx, 700},
  {"@<", xfx, 700},
  {"@>", xfx, 700},
  {"@=<", xfx, 700},
  {"@>=", xfx, 700},
  {"=:=", xfx, 700},
  {"=\\=", xfx, 700},
  {"<", xfx, 700},
  {">", xfx, 700},
  {"=<", xfx, 700},
  {">=", xfx, 700},
  {":", xfy, 600},
  {"+", yfx, 500},
  {"-", yfx, 500},
  {"/\\", yfx, 500},
  {"\\/", yfx, 500},
  {"+", fx, 500},
  {"-", fx, 500},
  {"\\", fx, 500},
  {"*", yfx, 400},
  {"/", yfx, 400},
  {"<<", yfx, 400},
  {">>", yfx, 400},
  {"mod", xfx, 300},
  {"//", yfx, 400},
  {"**", xfx, 200},
  {"^", xfy, 200}
};

static void 
InitOps()
{
	unsigned int             i;
	for (i = 0; i < sizeof(Ops) / sizeof(*Ops); ++i)
		SetOp(Ops[i].opPrio, Ops[i].opType, Ops[i].opName);
}

static void 
InitDebug()
{
	/* Set at half leash */
	Atom            At = LookupAtom("$leash");
	PutValue(At, MkIntTerm(10));
}

void 
InitCPred(char *Name, int Arity, CPredicate code, int flags)
{
	Atom            atom = LookupAtom(Name);
	PredEntry      *pe = RepPredProp(PredProp(atom, Arity));
	pe->PredFlags = flags | StandardPredFlag | CPredFlag;
	pe->TrueCodeOfPred = pe->CodeOfPred = (CODEADDR) code;
}

void 
InitCmpPred(char *Name, int Arity, CmpPredicate cmp_code, CPredicate code, int flags)
{
	Atom            atom = LookupAtom(Name);
	PredEntry      *pe = RepPredProp(PredProp(atom, Arity));
	pe->PredFlags = flags | StandardPredFlag | CPredFlag;
	pe->CodeOfPred = (CODEADDR) code;
	pe->TrueCodeOfPred = (CODEADDR) cmp_code;
}


static void 
CleanBack(PredEntry *pe, CPredicate Start, CPredicate Cont)
{
  yamop   *code;
  if (pe->FirstClause != pe->LastClause || pe->TrueCodeOfPred !=
      pe->FirstClause || pe->CodeOfPred != pe->FirstClause) {
    Error(SYSTEM_ERROR,TermNil,
	  "initiating a C Pred with backtracking");
    return;
  }
  code = (yamop *)(pe->FirstClause);
  if (pe->PredFlags & UserCPredFlag)
    code->opc = opcode(_try_userc);
  else
    code->opc = opcode(_try_c);
#ifdef YAPOR
  code->u.lds.ltt = 2;
  code->u.lds.cut = 0;
  code->u.lds.seq = TRUE;
#endif /* YAPOR */
  code->u.lds.d = (CODEADDR) Start;
  code = NEXTOP(code,lds);
  if (pe->PredFlags & UserCPredFlag)
    code->opc = opcode(_retry_userc);
  else
    code->opc = opcode(_retry_c);
#ifdef YAPOR
  code->u.lds.ltt = 1;
  code->u.lds.cut = 0;
  code->u.lds.seq = TRUE;
#endif /* YAPOR */
  code->u.lds.d = (CODEADDR) Cont;
}


void 
InitCPredBack(char *Name, int Arity, int Extra, CPredicate Start, CPredicate Cont, int flags)
{
  PredEntry      *pe;
  Atom            atom = LookupAtom(Name);

  pe = RepPredProp(PredProp(atom, Arity));
  if (pe->FirstClause != NIL)
    CleanBack(pe, Start, Cont);
  else {
    Clause *cl;
    yamop      *code = ((Clause *)NIL)->ClCode;
    pe->PredFlags = CompiledPredFlag | StandardPredFlag;
#ifdef YAPOR
    pe->PredFlags |= SequentialPredFlag;
#endif /* YAPOR */
    cl = (Clause
	  *)AllocCodeSpace((CELL)NEXTOP(NEXTOP(code,lds),lds));
    if (cl == NIL) {
      Error(SYSTEM_ERROR,TermNil,"No Heap Space in InitCPredBack");
      return;
    }
    cl->u.ClValue = 0;
    cl->ClFlags = 0;
    code = cl->ClCode;
    pe->TrueCodeOfPred = pe->CodeOfPred =
      pe->FirstClause = pe->LastClause = (CODEADDR)code;
    if (flags & UserCPredFlag)
      pe->OpcodeOfPred = code->opc = opcode(_try_userc);
    else
      pe->OpcodeOfPred = code->opc = opcode(_try_c);
    code->u.lds.d = (CODEADDR) Start;
    code->u.lds.p = (CODEADDR) pe;
    code->u.lds.s = Arity;
    code->u.lds.extra = Extra;
#ifdef YAPOR
    code->u.lds.ltt = 2;
    code->u.lds.cut = 0;
    code->u.lds.seq = TRUE;
#endif /* YAPOR */
    code = NEXTOP(code,lds);
    if (flags & UserCPredFlag)
      code->opc = opcode(_retry_userc);
    else
      code->opc = opcode(_retry_c);
    code->u.lds.d = (CODEADDR) Cont;
    code->u.lds.p = (CODEADDR) pe;
    code->u.lds.s = Arity;
    code->u.lds.extra = Extra;
#ifdef YAPOR
    code->u.lds.ltt = 1;
    code->u.lds.cut = 0;
    code->u.lds.seq = TRUE;
#endif /* YAPOR */
  }
}


void 
InitStdPreds()
{
  InitCPreds();
}

static void
InitFlags()
{
  yap_flags[YAP_INT_BOUNDED_FLAG] = 1;
  yap_flags[MAX_ARITY_FLAG] = -1;
  yap_flags[INTEGER_ROUNDING_FLAG] = 0;
  yap_flags[YAP_MAX_INTEGER_FLAG] = (Int)(~((CELL)1 << (sizeof(Int)*8-1)));
  yap_flags[YAP_MIN_INTEGER_FLAG] = (Int)(((CELL)1 << (sizeof(Int)*8-1)));
  yap_flags[CHAR_CONVERSION_FLAG] = 1;
  yap_flags[YAP_DOUBLE_QUOTES_FLAG] = 1;
}

static void 
InitCodes()
{
  heap_regs = (struct various_codes *)AllocCodeSpace(sizeof(struct various_codes));
  if (heap_regs == NIL) {
    Error(SYSTEM_ERROR,TermNil,"No Heap Space in InitCodes");
    return;
  }
#ifdef YAPOR
  heap_regs->seq_def = TRUE;
  heap_regs->getworkfirsttimecode.opc = opcode(_getwork_first_time);
  heap_regs->getworkcode.opc = opcode(_getwork);
  heap_regs->getworkcode.u.ld.p = (CODEADDR)RepPredProp(PredProp(LookupAtom("$getwork"), 0));
  heap_regs->getworkcode.YAMOP_LTT = 0;
  heap_regs->getworkcode.YAMOP_CUT = 0;
  heap_regs->getworkcode.YAMOP_SEQ = FALSE;
  heap_regs->getworkcode_seq.opc = opcode(_getwork_seq);
  heap_regs->getworkcode_seq.YAMOP_LTT = 0;
  heap_regs->getworkcode_seq.YAMOP_CUT = 0;
  heap_regs->getworkcode_seq.YAMOP_SEQ = FALSE;
  heap_regs->getworkcode_seq.u.ld.p = (CODEADDR)RepPredProp(PredProp(LookupAtom("$getwork_seq"), 0));
#endif /* YAPOR */
#ifdef TABLING
  heap_regs->tablecompletioncode.opc = opcode(_table_completion);
  heap_regs->tableanswerreturncode.opc = opcode(_table_answer_return);
#ifdef YAPOR
  heap_regs->tablecompletioncode.YAMOP_LTT = 0;
  heap_regs->tablecompletioncode.YAMOP_CUT = 0;
  heap_regs->tablecompletioncode.YAMOP_SEQ = FALSE;
  heap_regs->tableanswerreturncode.YAMOP_LTT = 0;
  heap_regs->tableanswerreturncode.YAMOP_CUT = 0;
  heap_regs->tableanswerreturncode.YAMOP_SEQ = FALSE;
#endif /* YAPOR */
#endif /* TABLING */
  heap_regs->failcode = opcode(_op_fail);
  
  heap_regs->trustfailcode = opcode(_trust_fail);

  heap_regs->env_for_yes_code.op = opcode(_call);
  heap_regs->env_for_yes_code.s = -Signed(RealEnvSize);
  heap_regs->env_for_yes_code.l = NULL;
  heap_regs->env_for_yes_code.l2 = NULL;
  heap_regs->yescode = opcode(_Ystop);

#ifdef YAPOR
  heap_regs->nocode.opc = opcode(_Nstop);
  heap_regs->nocode.YAMOP_LTT = 1;
  heap_regs->nocode.YAMOP_CUT = 0;
  heap_regs->nocode.YAMOP_SEQ = FALSE;
#else
  heap_regs->nocode = opcode(_Nstop);
#endif /* YAPOR */

  ((yamop *)(&heap_regs->rtrycode))->opc = opcode(_retry_and_mark);
  ((yamop *)(&heap_regs->rtrycode))->u.ld.s = 0;
  ((yamop *)(&heap_regs->rtrycode))->u.ld.d = NIL;
#ifdef YAPOR
  heap_regs->rtrycode.YAMOP_LTT = 1;
  heap_regs->rtrycode.YAMOP_CUT = 0;
  heap_regs->rtrycode.YAMOP_SEQ = FALSE;
#endif /* YAPOR */

  heap_regs->clausecode.arity = 0;
  heap_regs->clausecode.clause = NIL;
  heap_regs->clausecode.func = NIL;

  heap_regs->invisiblechain = 0;
  
  heap_regs->consultlow = (consult_obj *)AllocCodeSpace(sizeof(consult_obj)*InitialConsultCapacity);
  if (heap_regs->consultlow == NIL) {
    Error(SYSTEM_ERROR,TermNil,"No Heap Space in InitCodes");
   return;
  }
  heap_regs->consultcapacity = InitialConsultCapacity;
  heap_regs->profiling = FALSE;
  heap_regs->consultbase = heap_regs->consultsp =
    heap_regs->consultlow + heap_regs->consultcapacity;
#ifdef COROUTINING
  SusRecBase = (CELL *)
    AllocCodeSpace(sizeof(CELL)*InitialSusRecCapacity);
  if (SusRecBase == NIL) {
    Error(SYSTEM_ERROR,TermNil,"No Heap Space in InitCodes");
    return;
  }
  SusRecBase += InitialSusRecCapacity;
  SusRecTop   = SusRecBase;
#endif

  heap_regs->maxdepth = 0;
  heap_regs->maxlist  = 0;

  heap_regs->atprompt = 0;

  /* system boots in compile mode */
  heap_regs->static_predicates_marked = TRUE;
}

static void 
InitFuncs()
{
  Atom            AtomSemic, AtomQuery, AtomNot, AltAtomNot, AtomCall, AtomVar, AtomArrow, AtomBraces, AtomGVar, AtomEq;

  AtomNil = LookupAtom("[]");
  AtomComma = LookupAtom(",");
  AtomAssert = LookupAtom(":-");
  AtomTrue = LookupAtom("true");
  AtomOtherwise = LookupAtom("otherwise");
  AtomFail = LookupAtom("fail");
  AtomFalse = LookupAtom("false");
  AtomCut = LookupAtom("!");
  AtomSemic = LookupAtom(";");
  AtomQuery = LookupAtom("?-");
  AtomRepeat = LookupAtom("repeat");
  AtomDot = LookupAtom(".");
  AtomNot = LookupAtom("\\+");
  AltAtomNot = LookupAtom("not");
  AtomCall = LookupAtom("call");
  AtomVar = LookupAtom("$VAR");
  AtomGVar = LookupAtom("var");
  AtomPortray = LookupAtom("$portray");
  AtomLive = LookupAtom("$live");
  AtomBreak = LookupAtom("$break");
  AtomIndex = LookupAtom("$doindex");
  AtomFast = LookupAtom("$fast");
  AtomAbol = LookupAtom("$abol");
  AtomCsult = LookupAtom("$csult");
  AtomArrow = LookupAtom("->");
  AtomPlus = LookupAtom("+");
  AtomMinus = LookupAtom("-");
  AtomBraces = LookupAtom("{}");
  AtomEq = LookupAtom("=");
  AtomCatch = LookupAtom("$catch");
  AtomThrow = LookupAtom("$throw");
  AtomProfile = LookupAtom("$profile");
  AtomSpyCreep = LookupAtom("spy_creep");
  AtomAlarm = LookupAtom("$alarm");
#ifdef YAP_ARRAYS
  AtomArray = LookupAtom("$array");
  AtomArrayAccess = LookupAtom("$array_arg");
#endif
  FunctorOr = MkFunctor(AtomSemic, 2);
  FunctorComma = MkFunctor(AtomComma, 2);
  FunctorAssert = MkFunctor(AtomAssert, 2);
  FunctorQuery = MkFunctor(AtomQuery, 1);
  ListFunctor = MkFunctor(AtomDot, 2);
  FunctorNot = MkFunctor(AtomNot, 1);
  AltFunctorNot = MkFunctor(AltAtomNot, 1);
  FunctorCall = MkFunctor(AtomCall, 1);
  FunctorVar = MkFunctor(AtomVar, 1);
  FunctorGVar = MkFunctor(AtomGVar, 1);
  FunctorGAtom = MkFunctor(LookupAtom("atom"), 1);
  FunctorGInteger = MkFunctor(LookupAtom("integer"), 1);
  FunctorGNumber = MkFunctor(LookupAtom("number"), 1);
  FunctorGAtomic = MkFunctor(LookupAtom("atomic"), 1);
  FunctorGPrimitive = MkFunctor(LookupAtom("primitive"), 1);
  FunctorGCompound = MkFunctor(LookupAtom("compound"), 1);
  FunctorGFloat = MkFunctor(LookupAtom("compound"), 1);
  FunctorPortray = MkFunctor(AtomPortray, 1);
  FunctorArrow = MkFunctor(AtomArrow, 2);
  FunctorCsult = MkFunctor(AtomCsult, 1);
  FunctorBraces = MkFunctor(AtomBraces, 1);
  FunctorVBar = MkFunctor(LookupAtom("|"), 2);
  FunctorEq = MkFunctor(AtomEq, 2);
  FunctorModule = MkFunctor(LookupAtom(":"), 2);
#ifdef YAP_ARRAYS
  FunctorArrayEntry  = MkFunctor(AtomArrayAccess, 3);
#endif
#ifdef MPW
  InitAbsmi(REGS, ListFunctor);
#endif
}

static void 
InitVersion()
{
  PutValue(LookupAtom("$version_name"),
	   MkAtomTerm(LookupAtom(version_number)));
}


void 
InitStaff()
{

  InitModules();
  InitFuncs();
  InitStdPreds();
#ifdef LOW_LEVEL_TRACER
  InitLowLevelTrace();
#endif
  InitVersion();
  TermDot = MkAtomTerm(NameOfFunctor(ListFunctor));
  TermNil = MkAtomTerm(AtomNil);
  PutValue(LookupAtom("$catch_counter"),
	   MkIntTerm(0));
  InitSysPath();
}

void 
Init()
{
  int             i;

  for (i = 0; i < MaxHash; ++i)
    HashChain[i] = NIL;
  /* make sure these functors are initialised before anyone
     else uses the heap */
  /* FunctorDBRef = MkFunctor(LookupAtom("$dbref"), 0); */
  /* FunctorGeneralRef = MkFunctor(LookupAtom("$ref"), 1); */
  /* FunctorLongInt = MkFunctor(LookupAtom("$long_int$"), 2); */
  /* FunctorDouble = MkFunctor(LookupAtom("$float$"), 3); */
  CreepFlag = MinStackGap;
#ifdef MPW
  InitAbsmi(REGS, ListFunctor);
#else
  InitAbsmi();
#endif
#ifdef DEBUG
  for (i = 1; i < 20; ++i)
    Option[i] = 0;
  if (output_msg) {
    char            ch;
    opcode(_Ystop);
#if !SHORT_INTS
    YP_fprintf(YP_stderr,"absmi address:%x\n", Unsigned(FunAdr(absmi)));
#else
    YP_fprintf(YP_stderr,"absmi address:%lx\n", Unsigned(FunAdr(absmi)));
#endif
    YP_fprintf(YP_stderr,"Set	Trace Options:\n");
    YP_fprintf(YP_stderr,"a getch\t\tb token\t\tc Lookup\td LookupVar\ti Index\n");
    YP_fprintf(YP_stderr,"e SetOp\t\tf compile\tg icode\t\th boot\t\tl log\n");
    YP_fprintf(YP_stderr,"m Machine\n");
    while ((ch = YP_putchar(YP_getchar())) != '\n')
      if (ch >= 'a' && ch <= 'z')
	Option[ch - 'a' + 1] = 1;
    if (Option['l' - 96]) {
      logfile = YP_fopen(LOGFILE, "w");
      if (logfile == Nill) {
	YP_fprintf(YP_stderr,"can not open %s\n", LOGFILE);
	YP_getchar();
	exit(0);
      }
      YP_fprintf(YP_stderr,"logging session to file 'logfile'\n");
#ifdef MAC
      SetTextFile(LOGFILE);
      lp = my_line;
      curfile = Nill;
#endif
    }
  }
#endif
  InitFlags();
  InitCodes();
  InitStaff();
  InitBackCPreds();
  InitOps();
  InitDebug();
#if CAMACHO
  { extern void InitForeignPreds(void);
  
  InitForeignPreds();
  }
#endif
}
