/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		iopreds.c						 *
* Last rev:	5/2/88							 *
* mods:									 *
* comments:	Input/Output C implemented predicates			 *
*									 *
*************************************************************************/
#ifdef SCCS
static char SccsId[] = "%W% %G%";
#endif

/*
 * This file includes the definition of a miscellania of standard predicates
 * for yap refering to: Files and Streams, Simple Input/Output, 
 *
 */

#include "Yap.h"
#include "Yatom.h"
#include "io.h"

#include <stdlib.h>
#include <unistd.h>

typedef struct
  {
    union {
      struct {
	Atom name;
	Term user_name;
	YP_File file;
      } file;
      struct {
	int domain;
	int flags;
	int fd;
      } socket;
    } u;
    Int charcount, linecount, linepos;
    Int status;
  }
StreamDesc;

typedef struct AliasDescS
  {
    Atom name;
    int my_stream;
    struct AliasDescS *next;
  }
* AliasDesc;

STATIC_PROTO (Int p_always_prompt_user, (void));
STATIC_PROTO (void InitStdStream, (int, SMALLUNSGN, YP_File, Atom));
STATIC_PROTO (void InitPlIO, (void));
STATIC_PROTO (Int PlIOError, (yap_error_number, Term, char *));
STATIC_PROTO (int PlPutc, (int, int));
STATIC_PROTO (int PlPutchar, (int));
STATIC_PROTO (Int p_setprompt, (void));
STATIC_PROTO (Int p_prompt, (void));
STATIC_PROTO (int UserGetc, (void));
STATIC_PROTO (Int UserUnGetc, (Int));
STATIC_PROTO (int PlGetc, (int));
STATIC_PROTO (void PlUnGetc, (int,Int));
STATIC_PROTO (Term MkStream, (int));
STATIC_PROTO (Int p_stream_flags, (void));
STATIC_PROTO (int find_csult_file, (char *, char *, StreamDesc *, char *));
STATIC_PROTO (Int p_open, (void));
STATIC_PROTO (void AddAlias, (Atom, int));
STATIC_PROTO (void PurgeAlias, (int));
STATIC_PROTO (int CheckAlias, (Atom));
STATIC_PROTO (Atom FetchAlias, (int));
STATIC_PROTO (int FindAliasForStream, (int, Atom));
STATIC_PROTO (int CheckStream, (Term, int, char *));
STATIC_PROTO (Int p_check_stream, (void));
STATIC_PROTO (Int p_check_if_stream, (void));
STATIC_PROTO (Int init_cur_s, (void));
STATIC_PROTO (Int cont_cur_s, (void));
STATIC_PROTO (Int p_close, (void));
STATIC_PROTO (Int p_set_input, (void));
STATIC_PROTO (Int p_set_output, (void));
STATIC_PROTO (Int p_current_input, (void));
STATIC_PROTO (Int p_current_output, (void));
STATIC_PROTO (Int p_write, (void));
STATIC_PROTO (Int p_write2, (void));
STATIC_PROTO (void checkcol, (int));
STATIC_PROTO (Int p_inform_of_clause, (void));
STATIC_PROTO (Int p_set_read_error_handler, (void));
STATIC_PROTO (Int p_get_read_error_handler, (void));
STATIC_PROTO (Int p_read, (void));
STATIC_PROTO (Int p_cur_line_no, (void));
STATIC_PROTO (Int p_get, (void));
STATIC_PROTO (Int p_get0, (void));
STATIC_PROTO (Int p_get_byte, (void));
STATIC_PROTO (Int p_ungetc, (void));
STATIC_PROTO (Int p_past_eof, (void));
STATIC_PROTO (Int p_put, (void));
STATIC_PROTO (Int p_put_byte, (void));
STATIC_PROTO (Int p_skip, (void));
STATIC_PROTO (Int p_flush, (void));
STATIC_PROTO (Int p_write_depth, (void));
STATIC_PROTO (Int p_open_null_stream, (void));
STATIC_PROTO (Int p_user_file_name, (void));
STATIC_PROTO (Int p_line_position, (void));
STATIC_PROTO (Int p_character_count, (void));
STATIC_PROTO (Int p_show_stream_flags, (void));
STATIC_PROTO (Int p_show_stream_position, (void));
STATIC_PROTO (Int p_set_stream_position, (void));
STATIC_PROTO (Int p_add_alias_to_stream, (void));
STATIC_PROTO (Int p_check_if_valid_new_alias, (void));
STATIC_PROTO (Int p_fetch_stream_alias, (void));
STATIC_PROTO (Int GetArgSizeFromThirdArg, (char **, Term *));
STATIC_PROTO (Int GetArgSizeFromChars, (char **pptr, Int *, Term *));
STATIC_PROTO (void format_print_str, (Int, Int, Term));
STATIC_PROTO (Int p_format, (void));
STATIC_PROTO (Int p_startline, (void));
STATIC_PROTO (Int p_change_type_of_char, (void));
STATIC_PROTO (Int p_type_of_char, (void));
STATIC_PROTO (Int GetArgSizeFromChar, (Term *));


#define YAP_ERROR NIL


#define MaxStreams 32

StreamDesc Stream[MaxStreams];

#define	Free_Stream_f		0x000001
#define Output_Stream_f		0x000002
#define Input_Stream_f		0x000004
#define Append_Stream_f		0x000008
#define Eof_Stream_f		0x000010
#define Null_Stream_f		0x000020
#define Tty_Stream_f		0x000040
#define Socket_Stream_f		0x000080
#define Binary_Stream_f		0x000100
#define Eof_Error_Stream_f	0x000200
#define Reset_Eof_Stream_f	0x000400
#define Past_Eof_Stream_f	0x000800
#define Push_Eof_Stream_f	0x001000
#define Seekable_Stream_f	0x002000
#define Promptable_Stream_f	0x004000
#define Client_Socket_Stream_f	0x008000
#define Server_Socket_Stream_f	0x010000

int c_input_stream, c_output_stream, c_csult_stream;

/* used to communicate with PlPutchar by bypassing
   c_output_stream, towards making write/2 and friends a bit faster
 */

static int PlPutchar_output_stream;

#if EMACS
static int first_char;
#endif

#define FAIL_ON_PARSER_ERROR      0
#define QUIET_ON_PARSER_ERROR     1
#define CONTINUE_ON_PARSER_ERROR  2
#define EXCEPTION_ON_PARSER_ERROR 3

static int parser_error_style = FAIL_ON_PARSER_ERROR;

#define StdInStream	0
#define StdOutStream	1
#define	StdErrStream	2


static Atom ReadAtom, WriteAtom, AppendAtom, UserAtom, UsrInAtom, UsrOutAtom,
  UsrErrAtom, EofAtom;
static Functor StreamFunctor, StreamPosFunctor, StreamEOSFunctor;

extern int YP_sockets_io;

static AliasDesc AliasList = NULL;

static void
unix_upd_stream_info (StreamDesc * s)
{
  int filedes;

#ifdef HAVE_SOCKET
  if (YP_sockets_io &&
      (YP_fileno (s->u.file.file) == 0))
    {
      s->status |= Socket_Stream_f;
      s->u.socket.domain = af_inet;
      s->u.socket.flags = client_socket;
      s->u.socket.fd = 0;
      return;
    }
#endif /* HAVE_SOCKET */
  filedes = YP_fileno (s->u.file.file);
#if HAVE_ISATTY
  if (isatty (filedes))
    {
      s->u.file.name = LookupAtom (ttyname (filedes));
      s->status |= Tty_Stream_f|Reset_Eof_Stream_f|Promptable_Stream_f;
      return;
    }
#endif /* HAVE_ISATTY */
  s->status |= Seekable_Stream_f;
}

static Int
p_always_prompt_user(void)
{
  Stream[StdInStream].status |= Promptable_Stream_f;
  return(TRUE);
}


static void
InitStdStream (int sno, SMALLUNSGN flags, YP_File file, Atom name)
{
  StreamDesc *s = &Stream[sno];
  s->u.file.file = file;
  s->status = flags;
  s->linepos = 0;
  s->linecount = 1;
  s->charcount = 0;
  unix_upd_stream_info (s);
  if (s->status & Socket_Stream_f) {
    
  } else {
    s->u.file.name = name;
    s->u.file.user_name = MkAtomTerm (name);
  }
#if LIGHT
  s->status |= Tty_Stream_f|Promptable_Stream_f;
#endif
#if HAVE_SETBUF
  if ((s->status & Tty_Stream_f) && file == stdin)
    /* make sure input is unbuffered if it comes from stdin, this
       makes life simpler for interrupt handling */
    YP_setbuf (YP_stdin, NULL); 
#endif /* HAVE_SETBUF */

}


static void
InitPlIO (void)
{
  Int i;
  for (i = 0; i < MaxStreams; ++i)
    Stream[i].status = Free_Stream_f;
  InitStdStream (StdInStream, Input_Stream_f, YP_stdin, UsrInAtom);
  InitStdStream (StdOutStream, Output_Stream_f, YP_stdout, UsrOutAtom);
  InitStdStream (StdErrStream, Output_Stream_f, YP_stderr, UsrErrAtom);
  c_input_stream = StdInStream;
  c_output_stream = StdOutStream;
  AliasList = NULL;
}

static Int
PlIOError (yap_error_number type, Term culprit, char *who)
{
  if (GetValue(LookupAtom("fileerrors")) == MkIntTerm(1)) {
    Error(type, culprit, who);
    /* and fail */
    return(FALSE);
  } else {
    return(FALSE);
  }
}

/*
 * Used by the prompts to check if they are after a newline, and then a
 * prompt should be output, or if we are in the middle of a line 
 */
static int newline = TRUE;

/* static */
static int
PlPutc (int sno, int ch)
{
  StreamDesc *s = &Stream[sno];
#if MAC
  if (ch == 10)
    {
      ch = '\n';
    }
#endif
  if (!(s->status & (Null_Stream_f|Socket_Stream_f)))
    YP_putc (ch, s->u.file.file);
  else {
    if (s->status & Socket_Stream_f) {
      char c = ch;
      write(s->u.socket.fd,  &c, sizeof(c));
    }
    ++s->charcount;
  }
  if (ch == '\n')
    {
#if MPWSHELL
      if (mpwshell && (sno == StdOutStream || sno ==
		       StdErrStream) &&
	  !(s->status & Null_Stream_f))
	{
	  YP_putc (MPWSEP, s->u.file.file);
	  if (!(Stream[c_output_stream].status & Null_Stream_f))
	    YP_fflush (YP_stdout);
	}
#endif
      ++s->linecount;
      s->linepos = 0;
      /* Inform that we have written a newline */
      if (sno == StdOutStream || sno == StdErrStream)
	newline = TRUE;
      /* Inform we are not at the start of a newline */
    }
  else {
    if (sno == StdOutStream || sno == StdErrStream)
    {
      newline = FALSE;
    }
#if MAC
    if ((sno == StdOutStream || sno == StdErrStream)
	&& s->linepos > 200)
      PlPutc (sno, '\n');
#endif
    ++s->linepos;
  }
  if (s->status & Tty_Stream_f)
    ++s->charcount;
  return ((int) ch);
}

static int
PlPutchar (int ch)
{
  return (PlPutc (PlPutchar_output_stream, ch));
}


static Int
p_setprompt (void)
{				/* 'prompt(Atom)                 */
  Term t = Deref(ARG1);
  if (IsVarTerm (t) || !IsAtomTerm (t))
    return (FALSE);
  *AtPrompt = AtomOfTerm (t);
  return (TRUE);
}

static Int
p_prompt (void)
{				/* prompt(Old,New)       */
  Term t = Deref (ARG2);
  Atom a;
  if (!unify_constant (ARG1, MkAtomTerm (*AtPrompt)))
    return (FALSE);
  if (IsVarTerm (t) || !IsAtomTerm (t))
    return (FALSE);
  a = AtomOfTerm (t);
  if (strlen (RepAtom (a)->StrOfAE) > MAX_PROMPT) {
    Error(SYSTEM_ERROR,t,"prompt %s is too long", RepAtom (a)->StrOfAE);
    return(FALSE);
  }
  strcpy(Prompt, RepAtom (a)->StrOfAE);
  *AtPrompt = a;
  return (TRUE);
}

#if !HAVE_LIBREADLINE
static char ttybuf[1024];
#endif
static char *ttyptr = 0;

#if HAVE_LIBREADLINE
#include <readline/readline.h>
extern void add_history (const char *);

static char *_line = (char *) NULL;

int in_readline = FALSE;

#endif

static int
UserGetc (void)
{
  if (ttyptr == NIL) {
#if HAVE_LIBREADLINE
    /* Do it the gnu way */
    YP_fflush (YP_stdout);
    /* Only sends a newline if we are at the start of a line */
    if (_line != (char *) NULL && _line != (char *) EOF)
      free (_line);
    rl_instream = YP_stdin;
    rl_outstream = YP_stderr;
    /* window of vulnerability opened */
    in_readline = TRUE;
    if (newline)
      _line = readline (Prompt);
    else
      _line = readline ("");
    newline=FALSE;
    in_readline = FALSE;
    strcpy (Prompt, RepAtom (*AtPrompt)->StrOfAE);
    /* window of vulnerability closed */
    if (PrologMode & AbortMode) {
      Abort ((char *) NULL);
    }
    if (_line == NULL || _line == (char *) EOF)
      return (EOF);
    if (_line[0] != '\0' && _line[1] != '\0')
      add_history (_line);
    ttyptr = _line;
#else /* HAVE_LIBREADLINE */
    if (pfgets (ttybuf, sizeof (ttybuf), YP_stdin) == NULL)
      return (EOF);
    ttyptr = ttybuf;
#endif /* HAVE_LIBREADLINE */
  }
#if HAVE_LIBREADLINE
  if (*ttyptr == '\0')
#else
  if (*ttyptr == '\n')
#endif
    {
      ttyptr = NIL;
      return ('\n');
    }
  return (*ttyptr++);
}

static char
ch_backup[] = "\n";

static Int
UserUnGetc (Int ch)
{
  if (ttyptr == NIL) /* newline */
    {
      /* go back one line */
      ttyptr = ch_backup;
      return (ch);
    } else {
      /* back one character */
      ttyptr--;
      return(*ttyptr);
    }
}


static int
PlGetc (int sno)
{
  register StreamDesc *s = &Stream[sno];
  register int ch;
 restart_plgetc:
  if (s->status & Eof_Stream_f)
    {
      if (s->status & Push_Eof_Stream_f) {
	/* ok, we have pushed an EOF, send it away */
	s->status &= ~Push_Eof_Stream_f;
	return(EOF);
      }	  
      if (s->status & Eof_Error_Stream_f) {
	Error(PERMISSION_ERROR_INPUT_PAST_END_OF_STREAM,MkAtomTerm(s->u.file.name),
	    "GetC");
      } else if (s->status & Reset_Eof_Stream_f) {
	/* reset the eof indicator on file */
	if (YP_feof (s->u.file.file))
	  YP_clearerr (s->u.file.file);
	/* next, reset our own error indicator */
	s->status &= ~Eof_Stream_f;
	goto restart_plgetc;
      } else
	s->status |= Past_Eof_Stream_f;
      return (EOF);
    }
  else if (s->status & Socket_Stream_f) 
    {
      char c;
      int count;
	/* should be able to use a buffer */
      count = read(s->u.socket.fd, &c, sizeof(char));
      if (count == 0) {
	ch = EOF;
	s->status |= Eof_Stream_f;
      } else if (count > 0) {
	++s->charcount;
	ch = c;
      } else {
	Error(SYSTEM_ERROR, TermNil, "read");
	return(EOF);
      }
    }
  else if (sno == StdInStream) {
#if !HAVE_LIBREADLINE
    if (newline && s->status & Promptable_Stream_f) {
      YP_fputs (Prompt, YP_stderr);
      strcpy (Prompt, RepAtom (*AtPrompt)->StrOfAE);
      newline = FALSE;
    }
#endif
    if (s->status & (Tty_Stream_f))
    {
      ch = UserGetc ();
    } else {
      ch = YP_getc (s->u.file.file);
      if (ch != EOF)
	++s->charcount;
    }
    if (ch != '\n')
      newline = FALSE;
  } else
    ch = YP_getc (s->u.file.file);
  if (ch == '\n')
    {
      ++s->linecount;
      s->linepos = 0;
      /* don't convert if the stream is binary */
      if (!(Stream[sno].status & Binary_Stream_f))
	ch = 10;
      if (sno == StdOutStream || StdErrStream)
	newline = TRUE;
    }
  else if (ch == EOF)
    {
      s->status |= Eof_Stream_f;
      return (EOFCHAR);
    }
  else
    ++s->linepos;
  return (ch);
}

static void
PlUnGetc (int sno, Int ch)
{
  register StreamDesc *s = &Stream[sno];

  if (s->status & Socket_Stream_f)
    {
      /* past the point of no return */
      if (s->status & Past_Eof_Stream_f)
	return;
      s->status &= ~Eof_Stream_f;
      ch = EOF;
      return;
    }
  if (s->status & Eof_Stream_f)
    {
      /* past the point of no return */
      if (s->status & Past_Eof_Stream_f)
	return;
      s->status &= ~Eof_Stream_f;
      ch = YP_ungetc (ch, s->u.file.file);
      return;
    }
  else if (s->status & Tty_Stream_f)
    {
      if (sno == StdInStream) {
	ch = UserUnGetc (ch);
      } else
	ch = YP_ungetc (ch, s->u.file.file);
      --s->charcount;
    }
  else
    ch = YP_ungetc (ch, s->u.file.file);
  if (ch == 10)
    {
      --s->linecount;
      /* we don't know where we are in the line */
      s->linepos = -1;
    }
}


#if EMACS
int
GetCurInpPos ()
{
  if (Stream[c_input_stream].status & (Tty_Stream_f|Socket_Stream_f))
    return (Stream[c_input_stream].charcount);
  else
    return (YP_ftell (Stream[c_input_stream].u.file.file));
}
#endif /* EMACS */

int
PlGetchar (Int * ptr)
{
  return (PlGetc (c_input_stream));
}

static StreamDesc *s_inp_s_ptr;
static YP_File s_inp_s_fl;

int
PlFGetchar (void)
{
  register int ch;

 restart_plfgetc:
  if (s_inp_s_ptr->status & Eof_Stream_f)
    {
      if (s_inp_s_ptr->status & Eof_Error_Stream_f) {
	Error(PERMISSION_ERROR_INPUT_PAST_END_OF_STREAM,MkAtomTerm(s_inp_s_ptr->u.file.name),
	    "GetC");
      } else if (s_inp_s_ptr->status & Reset_Eof_Stream_f) {
	/* reset the eof indicator on file */
	if (YP_feof (s_inp_s_ptr->u.file.file))
	  YP_clearerr (s_inp_s_ptr->u.file.file);
	/* next, reset our own error indicator */
	s_inp_s_ptr->status &= ~Eof_Stream_f;
	goto restart_plfgetc;
      } else {
	s_inp_s_ptr->status |= Past_Eof_Stream_f;
      }
      return (EOFCHAR);
    }
  else
    ch = YP_getc (s_inp_s_fl);
  ++s_inp_s_ptr->charcount;
  if (ch == '\n')
    {
      ++s_inp_s_ptr->linecount;
      s_inp_s_ptr->linepos = 0;
      ch = 10;
    }
  else if (ch == EOF)
    {
      s_inp_s_ptr->status |= Eof_Stream_f;
      return (EOFCHAR);
    }
  else
    ++s_inp_s_ptr->linepos;
  return (ch);
}

/*
 * get a character that is not a newline, we know that it does not start from
 * a EOF 
 */
/*
   int
   PlFGetcharnonl(void)
   {
   ++s_inp_s_ptr->linepos;
   return (YP_getc(s_inp_s_fl));
   }
 */

static Term
MkStream (int n)
{
  /*  switch (n)
    {
    case StdInStream:
      return (MkAtomTerm (UsrInAtom));
    case StdOutStream:
      return (MkAtomTerm (UsrOutAtom));
    case StdErrStream:
      return (MkAtomTerm (UsrErrAtom));
    default:
      {
  */
	Term t[1];
	t[0] = MkIntTerm (n);
	return (MkApplTerm (StreamFunctor, 1, t));
   /*
      }
    }
   */
}

static Int
p_stream_flags (void)
{				/* '$stream_flags'(+N,-Flags)            */
  Term trm;
  trm = Deref (ARG1);
  if (IsVarTerm (trm) || !IsIntTerm (trm))
    return (FALSE);
  return (unify_constant (ARG2, MkIntTerm (Stream[IntOfTerm (trm)].status)));
}

static int
find_csult_file (char *source, char *buf, StreamDesc * st, char *io_mode)
{

  char *cp = source, ch;
  while (*cp++);
  while ((ch = *--cp) != '.' && !dir_separator((int)ch) && cp != source);
  if (ch == '.')
    return (FALSE);
  strcpy (buf, source);
  strcat (buf, ".yap");
  if ((st->u.file.file = YP_fopen (buf, io_mode)) != YAP_ERROR)
    return (TRUE);
  strcpy (buf, source);
  strcat (buf, ".pl");
  if ((st->u.file.file = YP_fopen (buf, io_mode)) != YAP_ERROR)
    return (TRUE);
  return (FALSE);
}

/* given a stream index, get the corresponding fd */
int
GetStreamFd(int sno)
{
  if (Stream[sno].status & Socket_Stream_f) {
    return(Stream[sno].u.socket.fd);
  } else
    return(YP_fileno(Stream[sno].u.file.file));
}

/* given a socket file descriptor, get the corresponding stream descripor */
int
CheckIOStream(Term stream, char * error)
{
  return(CheckStream(stream, Input_Stream_f|Output_Stream_f|Socket_Stream_f, error));
}

#if HAVE_SOCKET

Term
InitSocketStream(int fd, socket_info flags, socket_domain domain) {
  StreamDesc *st;
  int sno;

  for (sno = 0; sno < MaxStreams; ++sno)
    if (Stream[sno].status & Free_Stream_f)
      break;
  if (sno == MaxStreams)
    return (PlIOError (SYSTEM_ERROR,TermNil, "new stream not available for socket/4"));
  st = &Stream[sno];
  st->u.socket.domain = domain;
  st->u.socket.flags = flags;
  if (flags & (client_socket|server_session_socket)) {
    /* I can read and write from these sockets */
    st->status = (Socket_Stream_f|Input_Stream_f|Output_Stream_f);
  } else {
    /* oops, I cannot */
    st->status = Socket_Stream_f;
  }
  st->u.socket.fd = fd;
  st->charcount = 0;
  st->linecount = 1;
  st->linepos = 0;
  return(MkStream(sno));
}

/* given a socket file descriptor, get the corresponding stream descripor */
int
CheckSocketStream(Term stream, char * error)
{
  return(CheckStream(stream, Socket_Stream_f, error));
}

/* given a stream index, get the corresponding domain */
socket_domain
GetSocketDomain(int sno)
{
  return(Stream[sno].u.socket.domain);
}

/* given a stream index, get the corresponding status */
socket_info
GetSocketStatus(int sno)
{
  return(Stream[sno].u.socket.flags);
}

/* update info on a socket, eg, new->server or new->client */
void
UpdateSocketStream(int sno, socket_info flags, socket_domain domain) {
  StreamDesc *st;

  st = &Stream[sno];
  st->u.socket.domain = domain;
  st->u.socket.flags = flags;
  if (flags & (client_socket|server_session_socket)) {
    /* I can read and write from these sockets */
    st->status = (Socket_Stream_f|Input_Stream_f|Output_Stream_f);
  } else {
    /* oops, I cannot */
    st->status = Socket_Stream_f;
  }
}

#endif

static Int
p_open (void)
{				/* '$open'(+File,+Mode,?Stream,-ReturnCode)      */
  Term file_name, t, t2, topts;
  Atom mode;
  int sno;
  SMALLUNSGN s;
  char *io_mode;
  StreamDesc *st;
  char true_name[256];
  Int opts;

  file_name = Deref(ARG1);
  /* we know file_name is bound */
  if (!IsAtomTerm (file_name)) {
    Error(DOMAIN_ERROR_SOURCE_SINK,file_name, "open/3");
    return(FALSE);
  }
  t2 = Deref (ARG2);
  if (!IsAtomTerm (t2)) {
    Error(TYPE_ERROR_ATOM,t2, "open/3");
    return(FALSE);
  }
  mode = AtomOfTerm (t2);
  if (mode == ReadAtom || mode == AtomCsult)
    io_mode = "r", s = Input_Stream_f;
  else if (mode == WriteAtom)
    {
      io_mode = "w", s = Output_Stream_f;
    }
  else if (mode == AppendAtom)
    io_mode = "a", s = Append_Stream_f | Output_Stream_f;
  else {
    Error(DOMAIN_ERROR_IO_MODE, t2, "open/3");
    return(FALSE);   
  }
  if (!TrueFileName (RepAtom (AtomOfTerm (file_name))->StrOfAE, true_name))
    return (PlIOError (EXISTENCE_ERROR_SOURCE_SINK,file_name,"open/3"));
  for (sno = 0; sno < MaxStreams; ++sno)
    if (Stream[sno].status & Free_Stream_f)
      break;
  if (sno == MaxStreams)
    return (PlIOError (SYSTEM_ERROR,TermNil, "new stream not available for open/3"));
  st = &Stream[sno];
  if ((st->u.file.file = YP_fopen (true_name, io_mode)) == YAP_ERROR)
    {
      if (mode == AtomCsult)
	{
	  char buf[256];
	  if (!find_csult_file (true_name, buf, st, io_mode))
	    return (PlIOError (EXISTENCE_ERROR_SOURCE_SINK,file_name, "open/3"));
	  strcpy (true_name, buf);
	}
      else {
	if (errno == ENOENT)
	  return (PlIOError(EXISTENCE_ERROR_SOURCE_SINK,file_name,"open/3"));
	else
	  return (PlIOError(PERMISSION_ERROR_OPEN_SOURCE_SINK,file_name,"open/3"));
      }
    }
#if MAC
  if (mode == WriteAtom)
    {
      SetTextFile (RepAtom (AtomOfTerm (file_name))->StrOfAE);
    }
#endif
  st->status = s;
  st->charcount = 0;
  st->linecount = 1;
  st->u.file.name = LookupAtom (true_name);
  st->u.file.user_name = file_name;
  st->linepos = 0;
  unix_upd_stream_info (st);
  topts = Deref(ARG4);
  /* can never happen */
  if (IsVarTerm(topts) || !IsIntTerm(topts))
    return(FALSE);
  opts = IntOfTerm(topts);
  if (opts != 0) {
    if (opts & 2)
      st->status |= Binary_Stream_f;
    if (opts & 4) {
      if (st->status & (Tty_Stream_f|Socket_Stream_f)) {
	Term ta[1], t;
	
	ta[1] = MkAtomTerm(AtomTrue);
	t = MkApplTerm(MkFunctor(LookupAtom("reposition"),1),1,ta);
	Error(PERMISSION_ERROR_OPEN_SOURCE_SINK,t,"open/4");
	return(FALSE);
      }
      /* useless crap */
      st->status |= Seekable_Stream_f;
    }
    if (opts & 8) {
      /* There may be one reason why one wouldn't want to seek in a
	 file, maybe .... */
      st->status &= ~Seekable_Stream_f;
    }
    if (opts & 16) {
      st->status &= ~(Reset_Eof_Stream_f);
      st->status |= Eof_Error_Stream_f;
    }
    if (opts & 32) {
      st->status &= ~(Eof_Error_Stream_f|Reset_Eof_Stream_f);
    }
    if (opts & 64) {
      st->status &= ~(Eof_Error_Stream_f);
      st->status |= Reset_Eof_Stream_f;
    }
  }
  t = MkStream (sno);
  return (unify (ARG3, t));
}


static Int p_add_alias_to_stream (void)
{
  Term tname = Deref(ARG1);
  Term tstream = Deref(ARG2);
  Atom at;
  Int sno;

  if (IsVarTerm(tname)) {
    Error(INSTANTIATION_ERROR, tname, "$add_alias_to_stream");
    return (FALSE);
  } else if (!IsAtomTerm (tname)) {
    Error(TYPE_ERROR_ATOM, tname, "$add_alias_to_stream");
    return (FALSE);
  }
  if (IsVarTerm(tstream)) {
    Error(INSTANTIATION_ERROR, tstream, "$add_alias_to_stream");
    return (FALSE);
  } else if (!IsApplTerm (tstream) || FunctorOfTerm (tstream) != StreamFunctor ||
	     !IsIntTerm(ArgOfTerm(1,tstream))) {
    Error(DOMAIN_ERROR_STREAM_OR_ALIAS, tstream, "$add_alias_to_stream");
    return (FALSE);
  }
  at = AtomOfTerm(tname);
  sno = (int)IntOfTerm(ArgOfTerm(1,tstream));
  AddAlias(at, sno);
  return(TRUE);
}

static Int p_check_if_valid_new_alias (void)
{
  Term tname = Deref(ARG1);
  Atom at;

  if (IsVarTerm(tname)) {
    Error(INSTANTIATION_ERROR, tname, "$add_alias_to_stream");
    return (FALSE);
  } else if (!IsAtomTerm (tname)) {
    Error(TYPE_ERROR_ATOM, tname, "$add_alias_to_stream");
    return (FALSE);
  }
  at = AtomOfTerm(tname);
  return(CheckAlias(at) == -1);
}


static Int
p_fetch_stream_alias (void)
{				/* '$check_stream'(Stream)                  */
  int sno;
  Term t2 = Deref(ARG2);

  if ((sno = CheckStream (ARG1, Input_Stream_f | Output_Stream_f,
			  "check_stream/1"))  == -1)
    return(FALSE);
  switch (sno) {
  case StdInStream:
    return(unify_constant(t2, MkAtomTerm(UsrInAtom)));
  case StdOutStream:
    return(unify_constant(t2, MkAtomTerm(UsrOutAtom)));
  case StdErrStream:
    return(unify_constant(t2, MkAtomTerm(UsrErrAtom)));
  default:
    if (IsVarTerm(t2)) {
      Atom at = FetchAlias(sno);
      if (at == UsrInAtom)
	return(FALSE);
      else 
	return(unify_constant(t2, MkAtomTerm(at)));
    } else if (IsAtomTerm(t2)) {
      Atom at = AtomOfTerm(t2);
      return((Int)FindAliasForStream(sno,at));
    } else {
      Error(TYPE_ERROR_ATOM, t2, "predicate_property/2");
      return(FALSE);
    }
  }
}

static Int
p_open_null_stream (void)
{
  Term t;
  StreamDesc *st;
  int sno;
  for (sno = 0; sno < MaxStreams; ++sno)
    if (Stream[sno].status & Free_Stream_f)
      break;
  if (sno == MaxStreams)
    return (PlIOError (SYSTEM_ERROR,TermNil, "new stream not available for open_null_stream/1"));
  st = &Stream[sno];
  st->status = Append_Stream_f | Output_Stream_f | Null_Stream_f;
  st->linepos = 0;
  st->charcount = 0;
  st->linecount = 1;
  st->u.file.user_name = MkAtomTerm (st->u.file.name = LookupAtom ("/dev/null"));
  t = MkStream (sno);
  return (unify (ARG1, t));
}

/* create a new alias arg for stream sno */
static void
AddAlias (Atom arg, int sno)
{
  
  AliasDesc aliasp;

  if (CheckAlias(arg) != -1) {
    /* find out which error to give here */
    Error(INSTANTIATION_ERROR, MkAtomTerm(arg), "double alias");
  }
  aliasp = (AliasDesc)AllocAtomSpace(sizeof(struct AliasDescS));
  aliasp->name = arg;
  aliasp->my_stream = sno;
  aliasp->next = AliasList;
  AliasList = aliasp;
}

/* purge all aliases for stream sno */
static void
PurgeAlias (int sno)
{
  
  AliasDesc aliasp = AliasList,
    prev = NULL;
  
  while (aliasp != NULL) {
    if (aliasp->my_stream == sno) {
      if (prev == NULL)
	AliasList = aliasp->next;
      else
	prev->next = aliasp->next;
      FreeAtomSpace((char *)aliasp);
    } else
      prev = aliasp;
    aliasp = aliasp->next;
  }
}

/* check if arg is an alias */
static int
CheckAlias (Atom arg)
{
  AliasDesc aliasp = AliasList;
  while (aliasp != NULL) {
    if (aliasp->name == arg)
      return(aliasp->my_stream);
    else aliasp = aliasp->next;
  }
  return(-1);
}

/* check if arg is an alias */
static Atom
FetchAlias (int sno)
{
  AliasDesc aliasp = AliasList;
  while (aliasp != NULL) {
    if (aliasp->my_stream == sno)
      return(aliasp->name);
    else aliasp = aliasp->next;
  }
  return(UsrInAtom);
}

/* check if arg is an alias */
static int
FindAliasForStream (int sno, Atom al)
{
  AliasDesc aliasp = AliasList;
  while (aliasp != NULL) {
    if (aliasp->my_stream == sno && aliasp->name == al)
      return(TRUE);
    else aliasp = aliasp->next;
  }
  return(FALSE);
}

static int
CheckStream (Term arg, int kind, char *msg)
{
  int sno = -1;
  arg = Deref (arg);
  if (IsVarTerm (arg)) {
    Error(INSTANTIATION_ERROR, arg, msg);
    return (-1);
  }
  else if (IsAtomTerm (arg))
    {
      Atom sname = AtomOfTerm (arg);
      if (sname == UsrInAtom)
	sno = StdInStream;
      else if (sname == UsrOutAtom)
	sno = StdOutStream;
      else if (sname == UsrErrAtom)
	sno = StdErrStream;
      else if (sname == UserAtom)
	{
	  if (kind & Input_Stream_f)
	    sno = StdInStream;
	  if (kind & Output_Stream_f) {
	    if (sno >= 0)
	      {
		Error(PERMISSION_ERROR_INPUT_STREAM, arg,
		      "ambiguous use of 'user' as a stream");
		return (-1);
	      }
	    else {
	      sno = StdOutStream;
	    }
	  }
	}
      else {
	if ((sno = CheckAlias(sname)) == -1) {
	  Error(EXISTENCE_ERROR_STREAM, arg, msg);
	  return(-1);
	}
      }
    }
  else if (IsApplTerm (arg) && FunctorOfTerm (arg) == StreamFunctor)
    {
      arg = ArgOfTerm (1, arg);
      if (!IsVarTerm (arg) && IsIntTerm (arg))
	sno = IntOfTerm (arg);
    }
  if (sno < 0)
    {
      Error(DOMAIN_ERROR_STREAM_OR_ALIAS, arg, msg);
      return (-1);
    }
  if (Stream[sno].status & Free_Stream_f)
    {
      Error(EXISTENCE_ERROR_STREAM, arg, msg);
      return (-1);
    }
  if ((Stream[sno].status & kind) == 0)
    {
      if (kind & Input_Stream_f)
	Error(PERMISSION_ERROR_INPUT_STREAM, arg, msg);
      else
	Error(PERMISSION_ERROR_OUTPUT_STREAM, arg, msg);
      return (-1);
    }
  return (sno);
}

static Int
p_check_stream (void)
{				/* '$check_stream'(Stream,Mode)                  */
  Term mode = Deref (ARG2);
  return (CheckStream (ARG1,
   AtomOfTerm (mode) == ReadAtom ? Input_Stream_f : Output_Stream_f,
   "check_stream/2") != -1);
}

static Int
p_check_if_stream (void)
{				/* '$check_stream'(Stream)                  */
  return (CheckStream (ARG1, Input_Stream_f | Output_Stream_f | Socket_Stream_f,  "check_stream/1")
	  != -1);
}

static Int
init_cur_s (void)
{				/* Init current_stream */
  EXTRA_CBACK_ARG (3, 1) = MkIntTerm (0);
  return (cont_cur_s ());
}

static Int
cont_cur_s (void)
{				/* current_stream */
  Term t1, t2, t3;
  int i = IntOfTerm (EXTRA_CBACK_ARG (3, 1));
  while (i < MaxStreams)
    {
      if (Stream[i].status & Free_Stream_f)
	{
	  ++i;
	  continue;
	}
      if (Stream[i].status & Socket_Stream_f)
	t1 = MkAtomTerm(LookupAtom("socket"));
      else
	t1 = Stream[i].u.file.user_name;
      t2 = (Stream[i].status & Input_Stream_f ?
	    MkAtomTerm (ReadAtom) :
	    MkAtomTerm (WriteAtom));
      t3 = MkStream (i++);
      EXTRA_CBACK_ARG (3, 1) = Unsigned (MkIntTerm (i));
      if (unify (ARG3, t3) && unify_constant (ARG1, t1) && unify_constant (ARG2, t2))
	{
	  return (TRUE);
	}
      else
	{
	  return(FALSE);
	}
    }
  cut_fail();
}


/*
 * Called when you want to close all open streams, except for stdin, stdout
 * and stderr 
 */
void
CloseStreams (int loud)
{
  int sno;
  for (sno = 3; sno < MaxStreams; ++sno)
    {
      if (Stream[sno].status & Free_Stream_f)
	continue;
      if (!(Stream[sno].status & (Null_Stream_f|Socket_Stream_f)))
	YP_fclose (Stream[sno].u.file.file);
      else if (Stream[sno].status & (Socket_Stream_f)) {
	close(Stream[sno].u.socket.fd);
      }
      Stream[sno].status = Free_Stream_f;
      if (loud)
	YP_fprintf (YP_stderr, "[ Error: while closing stream: %s ]\n", RepAtom (Stream[sno].u.file.name)->StrOfAE);
      if (c_input_stream == sno)
	{
	  c_input_stream = StdInStream;
	}
      else if (c_output_stream == sno)
	{
	  c_output_stream = StdOutStream;
	}
    }
}


void
CloseStream(int sno)
{
  if (!(Stream[sno].status & (Null_Stream_f|Socket_Stream_f)))
    YP_fclose (Stream[sno].u.file.file);
  else if (Stream[sno].status & (Socket_Stream_f)) {
    CloseSocket(Stream[sno].u.socket.fd,
		Stream[sno].u.socket.flags,
		Stream[sno].u.socket.domain);
  }
  Stream[sno].status = Free_Stream_f;
  PurgeAlias(sno);
  if (c_input_stream == sno)
    {
      c_input_stream = StdInStream;
    }
  else if (c_output_stream == sno)
    {
      c_output_stream = StdOutStream;
    }
  /*  if (st->status == Socket_Stream_f|Input_Stream_f|Output_Stream_f) {
    CloseSocket();
  }
  */
}

static Int
p_close (void)
{				/* '$close'(+Stream) */
  Int sno = CheckStream (ARG1, (Input_Stream_f | Output_Stream_f | Socket_Stream_f), "close/2");
  if (sno < 0)
    return (FALSE);
  if (sno <= StdErrStream)
    return (TRUE);
  CloseStream(sno);
  return (TRUE);
}

static Int
p_past_eof (void)
{				/* at_end_of_stream */
  /* the next character is a EOF */ 
  int sno = CheckStream (ARG1, Input_Stream_f, "past_eof/1");

  if (sno < 0)
    return (FALSE);
  return(Stream[sno].status & Eof_Stream_f);
}

static Int
p_ungetc (void)
{				/* at_end_of_stream */
  /* the next character is a EOF */ 
  Term t = Deref(ARG2);
  int sno = CheckStream (ARG1, Input_Stream_f, "ungetc/1");

  if (sno < 0)
    return(FALSE);
  if (IsVarTerm(t))
    Error(INSTANTIATION_ERROR, t, "ungetc/1");
  if (!IsIntTerm(t))
    Error(TYPE_ERROR_INTEGER, t, "ungetc/1");
  PlUnGetc(sno, IntOfTerm(t));
  return(TRUE);
}

static Int
p_set_input (void)
{				/* '$set_input'(+Stream,-ErrorMessage)   */
  Int sno = CheckStream (ARG1, Input_Stream_f, "set_input/1");
  if (sno < 0)
    return (FALSE);
  c_input_stream = sno;
  return (TRUE);
}

static Int
p_set_output (void)
{				/* '$set_output'(+Stream,-ErrorMessage)  */
  Int sno = CheckStream (ARG1, Output_Stream_f, "set_output/1");
  if (sno < 0)
    return (FALSE);
  c_output_stream = sno;
  return (TRUE);
}

static Int
p_current_input (void)
{				/* current_input(?Stream)                */
  Term t1 = Deref(ARG1);
  if (IsVarTerm(t1)) {
    Term t = MkStream (c_input_stream);
    Bind((CELL *)t1, t);
    return(TRUE);
  } else if (!IsApplTerm(t1) ||
	     FunctorOfTerm(t1) != StreamFunctor ||
	     !IsIntTerm((t1=ArgOfTerm(1,t1)))) {
    Error(DOMAIN_ERROR_STREAM,t1,"current_input/1");
    return(FALSE);
  } else {
    return(c_input_stream == IntOfTerm(t1));
  }
}

static Int
p_current_output (void)
{				/* current_output(?Stream)               */
  Term t1 = Deref(ARG1);
  if (IsVarTerm(t1)) {
    Term t = MkStream (c_output_stream);
    Bind((CELL *)t1, t);
    return(TRUE);
  } else if (!IsApplTerm(t1) ||
	     FunctorOfTerm(t1) != StreamFunctor ||
	     !IsIntTerm((t1=ArgOfTerm(1,t1)))) {
    Error(DOMAIN_ERROR_STREAM,t1,"current_output/1");
    return(FALSE);
  } else {
    return(c_output_stream == IntOfTerm(t1));
  }
}

static Int
p_write (void)
{				/* '$write'(+Flags,?Term) */
  PlPutchar_output_stream = c_output_stream;
  plwrite (ARG2, PlPutchar, (int) IntOfTerm (Deref (ARG1)));
  return (TRUE);
}

static Int
p_write2 (void)
{				/* '$write'(+Flags,?Term) */
  PlPutchar_output_stream = CheckStream (ARG1, Output_Stream_f, "write/2");
  if (PlPutchar_output_stream == -1) {
    return(FALSE);
  }
  plwrite (ARG3, PlPutchar, (int) IntOfTerm (Deref (ARG2)));
  return (TRUE);
}

static int error_col;

static void
checkcol (int l)
{
  if (error_col + l > 70)
    YP_putc ('\n', YP_stderr), error_col = 0;
  error_col += l;
}

void
syntax_error (TokEntry * tokptr)
{
  Term info;
  char *s;
  while (1)
    {
      if (tokptr == toktide)
	{
	  YP_fprintf (YP_stderr, "\n<==== HERE ====>\n");
	  error_col = 0;
	}
      info = tokptr->TokInfo;
      switch (tokptr->Tok)
	{
	case Name_tok:
	  s = RepAtom (((Atom) info))->StrOfAE;
	  checkcol (strlen (s) + 1);
	  YP_fprintf (YP_stderr, " %s", s);
	  break;
	case Number_tok:
	  checkcol (6);
#if SHORT_INTS
	  YP_fprintf (YP_stderr, " %ld", IntOfTerm (info));
#else
	  YP_fprintf (YP_stderr, " %d", IntOfTerm (info));
#endif
	  break;
	case Var_tok:
	  s = ((VarEntry *) info)->VarRep;
	  checkcol (strlen (s) + 1);
	  YP_fprintf (YP_stderr, " %s", s);
	  break;
	case String_tok:
	  s = (char *) info;
	  checkcol (strlen (s) + 2);
	  YP_fprintf (YP_stderr, "\"%s\"", s);
	  break;
	case Ponctuation_tok:
	  if (Ord (info) == 'l')
	    {
	      checkcol (1);
	      YP_putc ('(', YP_stderr);
	    }
	  else
	    {
	      checkcol (2);
	      YP_fprintf (YP_stderr, " %c", (int) info);
	    }
	}
      if (tokptr->Tok == Ord (eot_tok))
	break;
      tokptr = tokptr->TokNext;
    }
  YP_putc ('.', YP_stderr);
  YP_putc ('\n', YP_stderr);
}

void
FirstLineInParse (void)
{
  StartLine = Stream[c_input_stream].linecount;
  StartCh = Stream[c_input_stream].charcount;
}

static Int
p_startline (void)
{
  return (unify_constant (ARG1, MkIntTerm (StartLine)));
}

#if EMACS
Atom
YapConsultingFile ()
{
  Term t1 = Deref (ARG1);
  Term t2 = Deref (ARG2);
  if (IsVarTerm (t1) || !IsAtomTerm (t1))
    return (FALSE);
  if (IsVarTerm (t2) || !IsAtomTerm (t2))
    return (FALSE);
  if (emacs_mode)
    YP_fprintf (YP_stdout, "\001(yap-consulting-file \"%s\" \"%s\" )\002\n",
		RepAtom (AtomOfTerm (t2))->StrOfAE,
		RepAtom (AtomOfTerm (t1))->StrOfAE);
}

#endif /* EMACS */

Atom
YapConsultingFile (void)
{
  return (Stream[c_csult_stream].u.file.name);
}

static Int
p_inform_of_clause (void)
{				/* '$inform_of_clause'(Func,Mode)        */
#if EMACS
  unsigned int arity;
  int clause_no;
  Atom at;
  Prop pred_prop;
  if (emacs_mode)
    {
      Term t1 = Deref (ARG1);
      Term t2 = Deref (ARG2);
      if (IsVarTerm (t1))
	return (FALSE);
      else if (IsAtomTerm (t1))
	{
	  arity = 0;
	  at = AtomOfTerm (t1);
	}
      else if (IsApplTerm (t1))
	{
	  Functor func = FunctorOfTerm (t1);
	  arity = ArityOfFunctor (func);
	  at = NameOfFunctor (func);
	}
      else
	return (FALSE);
      if (IsVarTerm (t2) || !IsIntTerm (t2))
	return (FALSE);
      YP_fprintf (YP_stdout, "\001(yap-consult-clause \"%s\" %d %d %d)\002\n",
		  RepAtom (at)->StrOfAE, arity,
      where_new_clause (PredProp (at, arity), (int) (IntOfTerm (t2) % 4)),
		  first_char);
    }
#endif
  return (TRUE);
}

/* control the parser error handler */
static Int
p_set_read_error_handler(void)
{
  Term t = Deref(ARG1);
  char *s;
  if (IsVarTerm(t)) {
    Error(INSTANTIATION_ERROR,t,"set_read_error_handler");
    return(FALSE);
  }
  if (!IsAtomTerm(t)) {
    Error(TYPE_ERROR_ATOM,t,"bad syntax_error handler");
    return(FALSE);
  }
  s = RepAtom(AtomOfTerm(t))->StrOfAE;
  if (!strcmp(s, "fail")) {
    parser_error_style = FAIL_ON_PARSER_ERROR;
  } else if (!strcmp(s, "error")) {
    parser_error_style = EXCEPTION_ON_PARSER_ERROR;
  } else if (!strcmp(s, "quiet")) {
    parser_error_style = QUIET_ON_PARSER_ERROR;
  } else if (!strcmp(s, "dec10")) {
    parser_error_style = CONTINUE_ON_PARSER_ERROR;
  } else {
    Error(DOMAIN_ERROR_SYNTAX_ERROR_HANDLER,t,"bad syntax_error handler");
    return(FALSE);
  }
  return(TRUE);
}

/* return the status for the parser error handler */
static Int
p_get_read_error_handler(void)
{
  Term t;

  switch (parser_error_style) {
  case FAIL_ON_PARSER_ERROR:
    t = MkAtomTerm(LookupAtom("fail"));
    break;
  case EXCEPTION_ON_PARSER_ERROR:
    t = MkAtomTerm(LookupAtom("error"));
    break;
  case QUIET_ON_PARSER_ERROR:
    t = MkAtomTerm(LookupAtom("quiet"));
    break;
  case CONTINUE_ON_PARSER_ERROR:
    t = MkAtomTerm(LookupAtom("dec10"));
    break;
  default:
    Error(SYSTEM_ERROR,TermNil,"corrupted syntax_error handler");
    return(FALSE);
  }
  return (unify_constant (ARG1, t));
}

CELL *ParserAuxSp;

static Int
p_read (void)
{				/* '$read'(+Flag,?Term,?Vars)    */
  Term t, v;
  TokEntry *tokstart, *fast_tokenizer (void);
#if EMACS
  int emacs_cares = FALSE;
#endif
  if (Stream[c_input_stream].status & Binary_Stream_f) {
    Error(PERMISSION_ERROR_INPUT_BINARY_STREAM, MkAtomTerm(Stream[c_input_stream].u.file.name), "read_term/2");
    return(FALSE);
  } else if (Stream[c_input_stream].status & Eof_Error_Stream_f) {
    Error(PERMISSION_ERROR_INPUT_PAST_END_OF_STREAM, MkAtomTerm(Stream[c_input_stream].u.file.name), "read_term/2");
    return(FALSE);
  }
  while (TRUE)
    {
      c_csult_stream = c_input_stream;
      /* Scans the term using stack space */
      ParserAuxSp = ASP - 64;
      eot_before_eof = FALSE;
      if (Stream[c_input_stream].status & (Promptable_Stream_f|Socket_Stream_f))
	tokstart = tokptr = toktide = tokenizer (PlGetchar);
      else
	{
	  s_inp_s_ptr = &Stream[c_input_stream];
	  s_inp_s_fl = s_inp_s_ptr->u.file.file;
	  tokstart = tokptr = toktide = fast_tokenizer ();
	}
      if ((Stream[c_input_stream].status & Eof_Stream_f)
	  && !eot_before_eof)
	{
	  if (tokstart != NIL && tokstart->Tok != Ord (eot_tok)) {
	    /* we need to force the next reading to also give end of file.*/
	    Stream[c_input_stream].status |= Push_Eof_Stream_f;
	    ErrorMessage = "[ Error: end of file found before end of term ]";
	  } else {
	    return (unify_constant (ARG2, MkAtomTerm (EofAtom)));
	  }
	}
      if (ErrorMessage || (t = Parse ()) == 0)
	{
	  if (parser_error_style == QUIET_ON_PARSER_ERROR) {
	    return(FALSE);
	  }
	  if (c_input_stream != StdInStream)
	    {
#if MPW
	      if (mpwshell)
		{
		  YP_fprintf (YP_stderr, "File \"%s\" ;",
			    RepAtom (Stream[c_input_stream].u.file.name)->StrOfAE);
		  YP_fprintf (YP_stderr, " line %d # syntax error \n",
			      StartLine);
		}
	      else
		{
		  YP_fprintf (YP_stderr, "<==== syntax error ");
		  YP_fprintf (YP_stderr, " line %d ====>\n", StartLine);
		}
#else
#if EMACS
	      if (emacs_mode)
		{
		  YP_fprintf (YP_stdout, "\001(yap-error \"%s\" %d \"yap syntax error\")\002\n",
			      RepAtom (Stream[c_input_stream].u.file.name)->StrOfAE,
			      toktide->TokPos);
		  emacs_cares = TRUE;
		}
	      else
		{
		  YP_fprintf (YP_stderr, "[ Syntax error: ");
		  YP_fprintf (YP_stderr, "line %d ]\n", StartLine);
		}
#else
	      YP_fprintf (YP_stderr, "[ Syntax error: ");
	      YP_fprintf (YP_stderr, "line %d ]\n", StartLine);
#endif /* EMACS */
#endif /* MPW */
	    }
	  else
	    YP_fprintf (YP_stderr, "[ Syntax error ]\n");
#if EMACS
	  if (emacs_cares)
	    {
	      char tmp[256], *s = tmp;
	      long i = 0, fd_char;
	      YP_File fd;
	      snoozing = TRUE;
	      YP_fclose (Stream[c_input_stream].u.file.file);
	      sleep (3600);	/* snooze until a nasty
				 * interrupt wakes Yap for
				 * breakfast */
	      if ((fd = YP_fopen (emacs_tmp, "r")) == NULL)
		YP_fprintf (YP_stderr, "Unable to communicate with Emacs: could not open %s\n",
			    emacs_tmp);
	      while ((fd_char = YP_getc (fd)) != ' '
		     && fd_char != EOF)
		i = i * 10 + fd_char - '0';
	      Stream[c_input_stream].u.file.file =
		YP_fopen (RepAtom (Stream[c_input_stream].u.file.name)->StrOfAE, "r");
	      if (YP_fseek (Stream[c_input_stream].u.file.file, i, 0) < 0)
		YP_fprintf (YP_stderr, "Problems while working with Emacs\n");
	      else
		{

		  /*
		   * YP_fprintf(YP_stderr,"gone to %d with
		   * %d\n",YP_ftell(Stream[c_input_stream].
		   * file),i); 
		   */
		  Stream[c_input_stream].linepos = 0;
		  Stream[c_input_stream].linecount = 1;
		  Stream[c_input_stream].charcount = 0;
		}
	      YP_fclose (fd);
	      unlink (emacs_tmp);
	    }
	  else
	    {
	      if (ErrorMessage)
		YP_fprintf (YP_stderr, "%s\n", ErrorMessage);
	      else
		syntax_error (tokstart);
	    }
#else
	  if (ErrorMessage)
	    YP_fprintf (YP_stderr, "%s\n", ErrorMessage);
	  else
	    syntax_error (tokstart);
#endif /* EMACS */
	  if (parser_error_style == FAIL_ON_PARSER_ERROR) {
	    return (FALSE);
	  } else if (parser_error_style == EXCEPTION_ON_PARSER_ERROR) {
	    Error(SYSTEM_ERROR,TermNil,NULL);
	    return(FALSE);
	  }
	}
      else
	break;
    }
#if EMACS
  first_char = tokstart->TokPos;
#endif /* EMACS */
  if (AtomOfTerm (Deref (ARG1)) == AtomTrue)
    {
      v = VarNames ();
      unify (v, ARG3);
    }
  return (unify (ARG2, t));
}

static Int
p_user_file_name (void)
{
  Term tout;
  int sno = CheckStream (ARG1, Input_Stream_f | Output_Stream_f | Append_Stream_f,"user_file_name/2");
  if (sno < 0)
    return (FALSE);
  if (Stream[sno].status & Socket_Stream_f)
    tout = MkAtomTerm(LookupAtom("socket"));
  else
    tout = Stream[sno].u.file.user_name;
  return (unify_constant (ARG2, tout));
}

static Int
p_cur_line_no (void)
{				/* '$current_line_number'(+Stream,-N) */
  Term tout;
  int sno =
  CheckStream (ARG1, Input_Stream_f | Output_Stream_f | Append_Stream_f,"current_line_number/2");
  if (sno < 0)
    return (FALSE);
  /* one has to be somewhat more careful because of terminals */
  if (Stream[sno].status & Tty_Stream_f)
    {
      Int no = 1;
      int i;
      Atom my_stream;
      if (Stream[sno].status & Socket_Stream_f)
	my_stream = LookupAtom("socket");
      else
	my_stream = Stream[sno].u.file.name;
      for (i = 0; i < MaxStreams; i++)
	{
	  if (!(Stream[i].status & (Free_Stream_f|Socket_Stream_f)) &&
	      Stream[i].u.file.name == my_stream)
	    no += Stream[i].linecount - 1;
	}
      tout = MkIntTerm (no);
    }
  else
    tout = MkIntTerm (Stream[sno].linecount);
  return (unify_constant (ARG2, tout));
}

static Int
p_line_position (void)
{				/* '$line_position'(+Stream,-N) */
  Term tout;
  int sno =
  CheckStream (ARG1, Input_Stream_f | Output_Stream_f | Append_Stream_f, "line_position/2");
  if (sno < 0)
    return (FALSE);
  if (Stream[sno].status & Tty_Stream_f)
    {
      Int no = 0;
      int i;
      Atom my_stream = Stream[sno].u.file.name;
      for (i = 0; i < MaxStreams; i++)
	{
	  if (!(Stream[i].status & Free_Stream_f) &&
	      Stream[i].u.file.name == my_stream)
	    no += Stream[i].linepos;
	}
      tout = MkIntTerm (no);
    }
  else
    tout = MkIntTerm (Stream[sno].linepos);
  return (unify_constant (ARG2, tout));
}

static Int
p_character_count (void)
{				/* '$character_count'(+Stream,-N) */
  Term tout;
  int sno =
  CheckStream (ARG1, Input_Stream_f | Output_Stream_f | Append_Stream_f, "character_count/2");
  if (sno < 0)
    return (FALSE);
  if (Stream[sno].status & Tty_Stream_f)
    {
      Int no = 0;
      int i;
      Atom my_stream = Stream[sno].u.file.name;
      for (i = 0; i < MaxStreams; i++)
	{
	  if (!(Stream[i].status & Free_Stream_f) &&
	      Stream[i].u.file.name == my_stream)
	    no += Stream[i].charcount;
	}
      tout = MkIntTerm (no);
    }
  else if (Stream[sno].status & Null_Stream_f)
    tout = MkIntTerm (Stream[sno].charcount);
  else
    tout = MkIntTerm (YP_ftell (Stream[sno].u.file.file));
  return (unify_constant (ARG2, tout));
}

static Int
p_show_stream_flags(void)
{				/* '$show_stream_flags'(+Stream,Pos) */
  Term tout;
  int sno =
    CheckStream (ARG1, Input_Stream_f | Output_Stream_f | Append_Stream_f, "stream_property/2");
  if (sno < 0)
    return (FALSE);
  tout = MkIntTerm(Stream[sno].status);
  return (unify (ARG2, tout));
}

static Int
p_show_stream_position (void)
{				/* '$show_stream_position'(+Stream,Pos) */
  Term sargs[3], tout;
  int sno =
  CheckStream (ARG1, Input_Stream_f | Output_Stream_f | Append_Stream_f, "stream_position/2");
  if (sno < 0)
    return (FALSE);
  if (Stream[sno].status & (Tty_Stream_f|Socket_Stream_f))
    sargs[0] = MkIntTerm (Stream[sno].charcount);
  else if (Stream[sno].status & Null_Stream_f)
    sargs[0] = MkIntTerm (Stream[sno].charcount);
  else
    sargs[0] = MkIntTerm (YP_ftell (Stream[sno].u.file.file));
  sargs[1] = MkIntTerm (Stream[sno].linecount);
  sargs[2] = MkIntTerm (Stream[sno].linepos);
  tout = MkApplTerm (StreamPosFunctor, 3, sargs);
  return (unify (ARG2, tout));
}

static Int
p_set_stream_position (void)
{				/* '$set_stream_position'(+Stream,Pos) */
  Term tin, tp;
  Int char_pos;
  int sno = CheckStream (ARG1, Input_Stream_f | Output_Stream_f | Append_Stream_f, "set_stream_position/2");
  if (sno < 0) {
    return (FALSE);
  }
  tin = Deref (ARG2);
  if (IsVarTerm (tin)) {
    Error(INSTANTIATION_ERROR, tin, "set_stream_position/2");
    return (FALSE);
  } else if (!(IsApplTerm (tin))) {
    Error(DOMAIN_ERROR_STREAM_POSITION, tin, "set_stream_position/2");
    return (FALSE);
  }
  if (FunctorOfTerm (tin) == StreamPosFunctor) {
    if (IsVarTerm (tp = ArgOfTerm (1, tin))) {
      Error(INSTANTIATION_ERROR, tp, "set_stream_position/2");
      return (FALSE);    
    } else if (!IsIntTerm (tp)) {
      Error(DOMAIN_ERROR_STREAM_POSITION, tin, "set_stream_position/2");
      return (FALSE);
    }
    if (!(Stream[sno].status & Seekable_Stream_f) ) {
      Error(PERMISSION_ERROR_REPOSITION_STREAM, ARG1,"set_stream_position/2");
      return(FALSE);
    }
    char_pos = IntOfTerm (tp);
    if (IsVarTerm (tp = ArgOfTerm (2, tin))) {
      Error(INSTANTIATION_ERROR, tp, "set_stream_position/2");
      return (FALSE);    
    } else if (!IsIntTerm (tp)) {
      Error(DOMAIN_ERROR_STREAM_POSITION, tin, "set_stream_position/2");
      return (FALSE);
    }
    Stream[sno].charcount = char_pos;
    Stream[sno].linecount = IntOfTerm (tp);
    if (IsVarTerm (tp = ArgOfTerm (2, tin))) {
      Error(INSTANTIATION_ERROR, tp, "set_stream_position/2");
      return (FALSE);    
    } else if (!IsIntTerm (tp)) {
      Error(DOMAIN_ERROR_STREAM_POSITION, tin, "set_stream_position/2");
      return (FALSE);
    }
    Stream[sno].linepos = IntOfTerm (tp);
    if (YP_fseek (Stream[sno].u.file.file, (long) (char_pos), 0) == -1) {
      Error(SYSTEM_ERROR, tp, 
	    "fseek failed for set_stream_position/2");
      return(FALSE);
    }
  } else if (FunctorOfTerm (tin) == StreamEOSFunctor) {
    if (IsVarTerm (tp = ArgOfTerm (1, tin))) {
      Error(INSTANTIATION_ERROR, tp, "set_stream_position/2");
      return (FALSE);    
    } else if (tp != MkAtomTerm(LookupAtom("at"))) {
      Error(DOMAIN_ERROR_STREAM_POSITION, tin, "set_stream_position/2");
      return (FALSE);
    }
    if (!(Stream[sno].status & Seekable_Stream_f) ) {
      Error(PERMISSION_ERROR_REPOSITION_STREAM, ARG1,"set_stream_position/2");
      return(FALSE);
    }
    if (YP_fseek (Stream[sno].u.file.file, 0L, SEEK_END) == -1) {
      Error(SYSTEM_ERROR, tp, 
	    "fseek failed for set_stream_position/2");
      return(FALSE);
    }
    /* reset the counters */
    Stream[sno].linepos = 0;
    Stream[sno].linecount = 0;
    Stream[sno].charcount = 0;
  }
  return (TRUE);
}

static Int
p_get (void)
{				/* '$get'(Stream,-N)                     */
  int sno = CheckStream (ARG1, Input_Stream_f, "get/2");
  Int ch;
  Int status;

  if (sno < 0)
    return(FALSE);
  status = Stream[sno].status;
  if (status & (Binary_Stream_f|Eof_Stream_f)) {
    if (status & Binary_Stream_f) {
      Error(PERMISSION_ERROR_INPUT_BINARY_STREAM, ARG1, "get/2");
      return(FALSE);
    } else if (status & Eof_Error_Stream_f) {
      Error(PERMISSION_ERROR_INPUT_PAST_END_OF_STREAM, ARG1, "get/2");
      return(FALSE);
    }
  }
  while ((ch = PlGetc (sno)) <= 32 && ch >= 0);
  return (unify_constant (ARG2, MkIntTerm (ch)));
}

static Int
p_get0 (void)
{				/* '$get0'(Stream,-N)                    */
  int sno = CheckStream (ARG1, Input_Stream_f, "get0/2");
  Int status;

  if (sno < 0)
    return(FALSE);
  status = Stream[sno].status;
  if (status & (Binary_Stream_f|Eof_Stream_f)) {
    if (status & Binary_Stream_f) {
      Error(PERMISSION_ERROR_INPUT_BINARY_STREAM, ARG1, "get0/2");
      return(FALSE);
    } else if (status & (Eof_Error_Stream_f)) {
      Error(PERMISSION_ERROR_INPUT_PAST_END_OF_STREAM, ARG1, "get0/2");
      return(FALSE);
    }
  }
  return (unify_constant (ARG2, MkIntTerm (PlGetc (sno))));
}

static Int
p_get_byte (void)
{				/* '$get_byte'(Stream,-N)                    */
  int sno = CheckStream (ARG1, Input_Stream_f, "get_byte/2");
  Int status;

  if (sno < 0)
    return(FALSE);
  status = Stream[sno].status;
  if (!(status & Binary_Stream_f)) {
    Error(PERMISSION_ERROR_INPUT_TEXT_STREAM, ARG1, "get_byte/2");
    return(FALSE);
  }
  if ((status & (Eof_Stream_f|Eof_Error_Stream_f)) == (Eof_Stream_f|Eof_Error_Stream_f)) {
    Error(PERMISSION_ERROR_INPUT_PAST_END_OF_STREAM, ARG1, "get_byte/2");
    return(FALSE);
  }
  return (unify_constant (ARG2, MkIntTerm (PlGetc (sno))));
}

static Int
p_put (void)
{				/* '$put'(Stream,N)                      */
  int sno = CheckStream (ARG1, Output_Stream_f, "put/2");
  if (sno < 0)
    return (FALSE);
  if (Stream[sno].status & Binary_Stream_f) {
    Error(PERMISSION_ERROR_OUTPUT_BINARY_STREAM, ARG1, "get0/2");
    return(FALSE);
  }
  PlPutc (sno, (int) IntOfTerm (Deref (ARG2)));
  /*
   * if (!(Stream[sno].status & Null_Stream_f))
   * YP_fflush(Stream[sno].u.file.file); 
   */
  return (TRUE);
}

static Int
p_put_byte (void)
{				/* '$put_byte'(Stream,N)                 */
  int sno = CheckStream (ARG1, Output_Stream_f, "put/2");
  if (sno < 0)
    return (FALSE);
  if (!(Stream[sno].status & Binary_Stream_f)) {
    Error(PERMISSION_ERROR_OUTPUT_TEXT_STREAM, ARG1, "get0/2");
    return(FALSE);
  }
  PlPutc (sno, (int) IntOfTerm (Deref (ARG2)));
  /*
   * if (!(Stream[sno].status & Null_Stream_f))
   * YP_fflush(Stream[sno].u.file.file); 
   */
  return (TRUE);
}


static int format_error = FALSE;

static Int
GetArgSizeFromThirdArg (char **pptr, Term * termptr)
{
  Term args = *termptr, arghd;
  char *ptr = *pptr;
  Int res;
  if (IsVarTerm(args)) {
    format_error = TRUE;
    Error(INSTANTIATION_ERROR, args, "format/2");
    return(0);
  } else if (!IsPairTerm (args)) {
    format_error = TRUE;
    Error(TYPE_ERROR_LIST, args, "format/2");
    return(0);
  }
  arghd = HeadOfTerm (args);
  args = TailOfTerm (args);
  if (IsVarTerm(arghd)) {
    format_error = TRUE;
    Error(INSTANTIATION_ERROR, arghd, "format/2");
    return(0);
  } else if (!IsIntTerm (arghd)) {
    format_error = TRUE;
    Error(TYPE_ERROR_LIST, arghd, "format/2");
    return(0);
  }
  res = IntOfTerm (arghd);
  if (res < 0) {
    format_error = TRUE;
    Error(DOMAIN_ERROR_NOT_LESS_THAN_ZERO, arghd, "format/2");
    return (0);
  }
#if SHORT_INTS
  sprintf (ptr, "%ld", res);
#else
  sprintf (ptr, "%d", res);
#endif
  while (*ptr)
    ptr++;
  *pptr = ptr;
  *termptr = args;
  return (res);
}

static Int
GetArgSizeFromChar (Term * args)
{
  Term arghd, argtl = *args;
  Int val;
  if (IsVarTerm(argtl)) {
    format_error = TRUE;
    Error(INSTANTIATION_ERROR, argtl, "format/2");
    return(0);
  } else if (!IsPairTerm (argtl)) {
    format_error = TRUE;
    Error(TYPE_ERROR_LIST, argtl, "format/2");
    return(0);
  }
  arghd = HeadOfTerm (argtl);
  argtl = TailOfTerm (argtl);
  if (IsVarTerm(arghd)) {
    format_error = TRUE;
    Error(INSTANTIATION_ERROR, arghd, "format/2");
    return(0);
  } else if (!IsIntTerm (arghd)) {
    format_error = TRUE;
    Error(TYPE_ERROR_LIST, arghd, "format/2");
    return(0);
  }
  val = IntOfTerm (arghd);
  if (IsVarTerm(argtl)) {
    format_error = TRUE;
    Error(INSTANTIATION_ERROR, argtl, "format/2");
    return(0);
  } else if (!IsPairTerm (argtl)) {
    format_error = TRUE;
    Error(TYPE_ERROR_LIST, argtl, "format/2");
    return(0);
  }
  arghd = HeadOfTerm (argtl);
  argtl = TailOfTerm (argtl);
  if (IsVarTerm(arghd)) {
    format_error = TRUE;
    Error(INSTANTIATION_ERROR, arghd, "format/2");
    return(0);
  } else if (!IsIntTerm (arghd)) {
    format_error = TRUE;
    Error(TYPE_ERROR_LIST, arghd, "format/2");
    return(0);
  }
  if (IntOfTerm (arghd) != 't') {
    format_error = TRUE;
    Error(TYPE_ERROR_LIST, arghd, "format/2");
    return (0);
  }
  *args = argtl;
  return (val);
}

static Int
GetArgSizeFromChars (char **pptr, Int * intptr, Term * termptr)
{
  Term args = *termptr, arghd;
  char *ptr = *pptr;
  Int up_to_now = *intptr;
  int ch;
  if (IsVarTerm(args)) {
    format_error = TRUE;
    Error(INSTANTIATION_ERROR, args, "format/2");
    return(0);
  } else if (!IsPairTerm (args)) {
    format_error = TRUE;
    Error(TYPE_ERROR_LIST, args, "format/2");
    return(0);
  }
  arghd = HeadOfTerm (args);
  args = TailOfTerm (args);
  if (IsVarTerm(arghd)) {
    format_error = TRUE;
    Error(INSTANTIATION_ERROR, arghd, "format/2");
    return(0);
  } else if (!IsIntTerm (arghd)) {
    format_error = TRUE;
    Error(TYPE_ERROR_LIST, arghd, "format/2");
    return(0);
  }
  ch = IntOfTerm (arghd);
  while (ch >= '0' && ch <= '9')
    {
      *ptr++ = ch;
      up_to_now = up_to_now * 10 + ch - '0';
      if (IsVarTerm(args)) {
	format_error = TRUE;
	Error(INSTANTIATION_ERROR, args, "format/2");
	return(0);
      } else if (!IsPairTerm (args)) {
	format_error = TRUE;
	Error(TYPE_ERROR_LIST, args, "format/2");
	return(0);
      }
      arghd = HeadOfTerm (args);
      args = TailOfTerm (args);
      if (IsVarTerm(arghd)) {
	format_error = TRUE;
	Error(INSTANTIATION_ERROR, arghd, "format/2");
	return(0);
      } else if (!IsIntTerm (arghd)) {
	format_error = TRUE;
	Error(TYPE_ERROR_LIST, arghd, "format/2");
	return(0);
      }
      ch = (int) IntOfTerm (arghd);
    }
  *intptr = up_to_now;
  *termptr = args;
  *pptr = ptr;
  return (ch);
}

static void
format_print_str (Int sno, Int size, Term args)
{
  Term arghd;
  Int always_flag = !size;
  while (always_flag || size > 0)
    {
      if (IsVarTerm(args)) {
	format_error = TRUE;
	Error(INSTANTIATION_ERROR, args, "format/2");
	return;
      } else if (IsAtomTerm (args) && args == TermNil)
	break;
      else if (!IsPairTerm (args)) {
	format_error = TRUE;
	Error(TYPE_ERROR_LIST, args, "format/2");
	return;
      }
      arghd = HeadOfTerm (args);
      args = TailOfTerm (args);
      if (IsVarTerm(arghd)) {
	format_error = TRUE;
	Error(INSTANTIATION_ERROR, arghd, "format/2");
	return;
      } else if (!IsIntTerm (arghd)) {
	format_error = TRUE;
	Error(TYPE_ERROR_LIST, arghd, "format/2");
	return;
      }
      PlPutc (sno, (int) IntOfTerm (arghd));
      size--;
    }
}

static Int
format(Term tail, Term args, int sno)
{
  char tmp1[256], tmp2[256], *ptr;
  Term head, arghd;
  int ch;
  Int int2, i;
  int column_boundary = 0, column_fill = ' ';
  Int size_args, arg_size;
  Float float_tmp;

  format_error = FALSE;
  if (IsVarTerm(tail)) {
    Error(INSTANTIATION_ERROR,tail,"format/2");
    return(FALSE);
  } else if (!IsPairTerm (tail)) {
    Error(TYPE_ERROR_LIST,tail,"format/2");
    return(FALSE);
  }
  head = HeadOfTerm (tail);
  tail = TailOfTerm (tail);
  if (IsVarTerm (args) || !IsPairTerm (args))
    args = MkPairTerm (args, TermNil);
  while (!IsVarTerm (head) && IsIntTerm (head))
    {
      ch = IntOfTerm (head);
      if (ch == '~')
	{
	  size_args = FALSE;
	  arg_size = 0;
	  ptr = tmp1;
	  *ptr++ = '%';
	  if (IsVarTerm (tail = Deref (tail)) ) {
	    Error(INSTANTIATION_ERROR,tail,"format/2");
	    return(FALSE);
	  } else if (!IsPairTerm (tail)) {
	    Error(TYPE_ERROR_LIST,tail,"format/2");
	    return(FALSE);
	  }
	  head = HeadOfTerm (tail);
	  tail = TailOfTerm (tail);
	  if (IsVarTerm (head)) {
	    Error(INSTANTIATION_ERROR,tail,"format/2");
	    return(FALSE);
	  } else  if ( !IsIntTerm (head)) {
	    Error(TYPE_ERROR_INTEGER,tail,"format/2");
	    return(FALSE);
	  } else
	    ch = IntOfTerm (head);
	  if (ch == '*')
	    {
	      size_args = TRUE;
	      arg_size = GetArgSizeFromThirdArg (&ptr, &args);
	      if (format_error) {
		return(FALSE);
	      }
	      if (IsVarTerm (tail = Deref (tail)) ) {
		Error(INSTANTIATION_ERROR,tail,"format/2");
		return(FALSE);
	      } else if (!IsPairTerm (tail)) {
		Error(TYPE_ERROR_LIST,tail,"format/2");
		return(FALSE);
	      }
	      head = HeadOfTerm (tail);
	      tail = TailOfTerm (tail);
	      if (IsVarTerm (head)) {
		Error(INSTANTIATION_ERROR,tail,"format/2");
		return(FALSE);
	      } else  if ( !IsIntTerm (head)) {
		Error(TYPE_ERROR_INTEGER,tail,"format/2");
		return(FALSE);
	      } else
		ch = IntOfTerm (head);
	    }
	  else if (ch >= '0' && ch <= '9')
	    {
	      arg_size = ch - '0';
	      size_args = TRUE;
	      ch = GetArgSizeFromChars (&ptr, &arg_size, &tail);
	      if (format_error) {
		return (FALSE);
	      }
	    }
	  else if (ch == '`')
	    {
	      size_args = TRUE;
	      arg_size = GetArgSizeFromChar(&tail);
	      if (format_error) {
		return(FALSE);
	      }
	      ch = 't';
	    }
	  switch (ch)
	    {
	    case 'a':
	      if (size_args)
		return (TRUE);
	      if (IsVarTerm (args)) {
		Error(INSTANTIATION_ERROR,args,"~a format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (args)) {
		Error(TYPE_ERROR_LIST,args,"~a format/2");
		return(FALSE);
	      }
	      arghd = HeadOfTerm (args);
	      args = TailOfTerm (args);
	      if (IsVarTerm (arghd)) {
		Error(INSTANTIATION_ERROR,arghd,"~a in format/2");
		return(FALSE);		
	      } else if (!IsAtomTerm (arghd)) {
		Error(TYPE_ERROR_ATOM,arghd,"~a in format/2");
		return(FALSE);
	      }
	      plwrite (arghd, PlPutchar, 4);
	      break;
	    case 'c':
	      if (IsVarTerm (args)) {
		Error(INSTANTIATION_ERROR,args,"~c in format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (args)) {
		Error(TYPE_ERROR_LIST,args,"~c in format/2");
		return(FALSE);
	      }
	      arghd = HeadOfTerm (args);
	      args = TailOfTerm (args);
	      if (IsVarTerm (arghd)) {
		Error(INSTANTIATION_ERROR,arghd,"~c in format/2");
		return(FALSE);		
	      } else if (!IsIntTerm (arghd)) {
		Error(TYPE_ERROR_ATOM,arghd,"~a in format/2");
		return(FALSE);
	      } else int2= IntOfTerm(arghd);
	      if (!size_args)
		arg_size = 1;
	      for (i = 0; i < arg_size; i++)
		PlPutc (sno, int2);
	      break;
	    case 'e':
	    case 'E':
	    case 'f':
	    case 'g':
	    case 'G':
	      if (IsVarTerm (args)) {
		Error(INSTANTIATION_ERROR,args,"~%d in format/2", ch);
		return(FALSE);		
	      } else if (!IsPairTerm (args)) {
		Error(TYPE_ERROR_LIST,args,"~%d in format/2", ch);
		return(FALSE);
	      }
	      if (arg_size == 0 || arg_size > 6)
		arg_size = 6;
	      *ptr++ = '.';
	      *ptr++ = '0' + arg_size;
	      *ptr++ = ch;
	      *ptr = 0;
	      arghd = HeadOfTerm (args);
	      args = TailOfTerm (args);
	      if (IsVarTerm(arghd)) {
		Error(INSTANTIATION_ERROR,arghd,"~%c in format/2", ch);
		return(FALSE);		
	      } else if (!IsFloatTerm (arghd)) {
		Error(TYPE_ERROR_FLOAT,arghd,"~%c in format/2", ch);
		return(FALSE);
	      }
	      float_tmp = FloatOfTerm (arghd);
	      sprintf (tmp2, tmp1, float_tmp);
	      ptr = tmp2;
	      while ((ch = *ptr++) != 0)
		PlPutc (sno, ch);
	      break;
	    case 'd':
	      if (IsVarTerm (args)) {
		Error(INSTANTIATION_ERROR,args,"~d format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (args)) {
		Error(TYPE_ERROR_LIST,args,"~d format/2");
		return(FALSE);
	      }
	      arghd = HeadOfTerm (args);
	      args = TailOfTerm (args);
	      if (IsVarTerm (arghd)) {
		Error(INSTANTIATION_ERROR,arghd,"~d in format/2");
		return(FALSE);		
	      } else if (IsIntTerm (arghd)) {
		int2 = IntOfTerm (arghd);
	      } else if (IsLongIntTerm (arghd)) {
		int2 = LongIntOfTerm(arghd);
	      } else {
		Error(TYPE_ERROR_INTEGER,arghd,"~d in format/2");
		return(FALSE);
	      }
	      if (!arg_size)
		plwrite (arghd, PlPutchar, 4);
	      else
		{
		  Int siz;
		  /*
		   * The guys at Quintus have probably
		   * read too much Cobol! 
		   */
		  if (int2 < 0)
		    {
		      int2 = -int2;
		      PlPutc (sno, (int) '-');
		    }
#if SHORT_INTS
		  sprintf (tmp2, "%ld", int2);
#else
		  sprintf (tmp2, "%d", int2);
#endif
		  siz = strlen (tmp2);
		  {
		    char *ptr = tmp2;
		    if (siz <= arg_size)
		      PlPutc (sno, (int) '0');
		    else
		      while (siz > arg_size)
			PlPutc (sno, (int) *ptr++), --siz;
		    PlPutc (sno, (int) '.');
		    while (siz < arg_size)
		      PlPutc (sno, (int) '0'), --arg_size;
		    while (*ptr)
		      PlPutc (sno, (int) (*ptr++));
		  }
		}
	      break;
	    case 'D':
	      if (IsVarTerm (args)) {
		Error(INSTANTIATION_ERROR,args,"~D in format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (args)) {
		Error(TYPE_ERROR_LIST,args,"~D in format/2");
		return(FALSE);
	      }
	      arghd = HeadOfTerm (args);
	      args = TailOfTerm (args);
	      if (IsVarTerm (arghd)) {
		Error(INSTANTIATION_ERROR,arghd,"~D in format/2");
		return(FALSE);		
	      } else if (IsIntTerm (arghd)) {
		int2 = IntOfTerm (arghd);
	      } else if (IsLongIntTerm (arghd)) {
		int2 = LongIntOfTerm(arghd);
	      } else {
		Error(TYPE_ERROR_INTEGER,arghd,"~D in format/2");
		return(FALSE);
	      }
	      {
		Int siz, lsiz;
		char *ptr = tmp2;

		if (int2 < 0)
		  {
		    int2 = -int2;
		    PlPutc (sno, (int) '-');
		  }
		/*
		 * The guys at Quintus have probably
		 * read too much Cobol! 
		 */
#if SHORT_INTS
		sprintf (tmp2, "%ld", int2);
#else
		sprintf (tmp2, "%d", int2);
#endif
		siz = strlen (tmp2);
		if ((lsiz = siz - arg_size) <= 0)
		  PlPutc (sno, (int) '0');
		else
		  {
		    Int imod = lsiz % 3;
		    int output_done = FALSE;

		    for (i = 0; i < lsiz; i++)
		      {
			if (imod-- == 0)
			  {
			    if (output_done)
			      PlPutc (sno, (int) ',');
			    imod = 2;
			  }
			PlPutc (sno, (int) (*ptr++));
			output_done = TRUE;
		      }
		  }
		if (arg_size > 0)
		  PlPutc (sno, (int) '.');
		while (lsiz < 0)
		  PlPutc (sno, (int) '0'), ++lsiz;
		while (*ptr)
		  PlPutc (sno, (int) (*ptr++));
	      }
	      break;
	    case 'r':
	      {
		Int radix = 8;
		ptr = tmp2;
		if (size_args)
		  radix = arg_size;
		if (IsVarTerm (args)) {
		  Error(INSTANTIATION_ERROR,args,"~r in format/2");
		  return(FALSE);		
		} else if (!IsPairTerm (args)) {
		  Error(TYPE_ERROR_LIST,args,"~r in format/2");
		  return(FALSE);
		}
		if (radix > 36 || radix < 2) {
		  Error(DOMAIN_ERROR_RADIX,MkIntTerm(radix),"~r in format/2");
		  return(FALSE);
		}
		arghd = HeadOfTerm (args);
		args = TailOfTerm (args);
		if (IsVarTerm (arghd)) {
		  Error(INSTANTIATION_ERROR,arghd,"~r in format/2");
		  return(FALSE);		
		} else if (IsIntTerm (arghd)) {
		  int2 = IntOfTerm (arghd);
		} else if (IsLongIntTerm (arghd)) {
		int2 = LongIntOfTerm(arghd);
		} else {
		  Error(TYPE_ERROR_INTEGER,arghd,"~r in format/2");
		  return(FALSE);
		}
		if (int2 < 0)
		  {
		    int2 *= -1;
		    PlPutc (sno, (int) '-');
		  }
		else if (int2 == 0)
		  PlPutc (sno, (int) '0');
		while (int2 != 0)
		  {
		    Int numb = int2 % radix;
		    if (numb >= 10)
		      numb += 'a' - 10;
		    else
		      numb += '0';
		    *ptr++ = numb;
		    int2 = int2 / radix;
		  }
		while (--ptr >= tmp2)
		  PlPutc (sno, (int) *ptr);
	      }
	      break;
	    case 'R':
	      {
		Int radix = 8;
		ptr = tmp2;
		if (size_args)
		  radix = arg_size;
		if (IsVarTerm (args)) {
		  Error(INSTANTIATION_ERROR,args,"~R in format/2");
		  return(FALSE);		
		} else if (!IsPairTerm (args)) {
		  Error(TYPE_ERROR_LIST,args,"~R in format/2");
		  return(FALSE);
		}
		if (radix > 36 || radix < 2) {
		  Error(DOMAIN_ERROR_RADIX,MkIntTerm(radix),"~R in format/2");
		  return(FALSE);
		}
		arghd = HeadOfTerm (args);
		args = TailOfTerm (args);
		if (IsVarTerm (arghd)) {
		  Error(INSTANTIATION_ERROR,arghd,"~R in format/2");
		  return(FALSE);		
		} else if (IsIntTerm (arghd)) {
		  int2 = IntOfTerm (arghd);
		} else if (IsLongIntTerm (arghd)) {
		  int2 = LongIntOfTerm(arghd);
		} else {
		  Error(TYPE_ERROR_INTEGER,arghd,"~R in format/2");
		  return(FALSE);
		}
		if (int2 < 0)
		  {
		    int2 *= -1;
		    PlPutc (sno, (int) '-');
		  }
		else if (int2 == 0)
		  PlPutc (sno, (int) '0');
		while (int2 != 0)
		  {
		    Int numb = int2 % radix;
		    if (numb >= 10)
		      numb += 'A' - 10;
		    else
		      numb += '0';
		    *ptr++ = numb;
		    int2 = int2 / radix;
		  }
		while (--ptr >= tmp2)
		  PlPutc (sno, (int) *ptr);
	      }
	      break;
	    case 's':
	      if (IsVarTerm (args)) {
		Error(INSTANTIATION_ERROR,args,"~s in format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (args)) {
		Error(TYPE_ERROR_LIST,args,"~s in format/2");
		return(FALSE);
	      }
	      arghd = HeadOfTerm (args);
	      args = TailOfTerm (args);
	      if (IsVarTerm (arghd)) {
		Error(INSTANTIATION_ERROR,arghd,"~s in format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (arghd)) {
		Error(TYPE_ERROR_LIST,arghd,"~s in format/2");
		return(FALSE);
	      }
	      format_print_str (sno, arg_size, arghd);
	      if (format_error) {
		return(FALSE);
	      }
	      break;
	    case 'i':
	      if (size_args) {
		Error(DOMAIN_ERROR_NOT_ZERO,MkIntTerm(size_args),
		      "~i in format/2");
		return(FALSE);
	      }
	      if (IsVarTerm (args)) {
		Error(INSTANTIATION_ERROR,args,"~i in format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (args)) {
		Error(TYPE_ERROR_LIST,args,"~i in format/2");
		return(FALSE);
	      }
	      args = TailOfTerm (args);
	      break;
	    case 'k':
	      if (size_args) {
		Error(DOMAIN_ERROR_NOT_ZERO,MkIntTerm(size_args),
		      "~k in format/2");
		return(FALSE);
	      }
	      if (IsVarTerm (args)) {
		Error(INSTANTIATION_ERROR,args,"~k in format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (args)) {
		Error(TYPE_ERROR_LIST,args,"~k in format/2");
		return(FALSE);
	      }
	      arghd = HeadOfTerm (args);
	      args = TailOfTerm (args);
	      plwrite (arghd, PlPutchar, (int) 3);
	      break;
	    case 'p':
	      if (size_args) {
		Error(DOMAIN_ERROR_NOT_ZERO,MkIntTerm(size_args),
		      "~p in format/2");
		return(FALSE);
	      }
	      if (IsVarTerm (args)) {
		Error(INSTANTIATION_ERROR,args,"~p in format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (args)) {
		Error(TYPE_ERROR_LIST,args,"~p in format/2");
		return(FALSE);
	      }
	      arghd = HeadOfTerm (args);
	      args = TailOfTerm (args);
	      plwrite (arghd, PlPutchar, (int) 12);
	      break;
	    case 'q':
	      if (size_args) {
		Error(DOMAIN_ERROR_NOT_ZERO,MkIntTerm(size_args),
		      "~q in format/2");
		return(FALSE);
	      }
	      if (IsVarTerm (args)) {
		Error(INSTANTIATION_ERROR,args,"~q in format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (args)) {
		Error(TYPE_ERROR_LIST,args,"~q in format/2");
		return(FALSE);
	      }
	      arghd = HeadOfTerm (args);
	      args = TailOfTerm (args);
	      plwrite (arghd, PlPutchar, (int) 5);
	      break;
	    case 'w':
	      if (size_args) {
		Error(DOMAIN_ERROR_NOT_ZERO,MkIntTerm(size_args),
		      "bad arguments for ~w in format/2");
		return(FALSE);
	      }
	      if (IsVarTerm (args)) {
		Error(INSTANTIATION_ERROR,args,"~w in format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (args)) {
		Error(TYPE_ERROR_LIST,args,"~w in format/2");
		return(FALSE);
	      }
	      arghd = HeadOfTerm (args);
	      args = TailOfTerm (args);
	      plwrite (arghd, PlPutchar, (int) 4);
	      break;
	    case '~':
	      if (size_args) {
		Error(DOMAIN_ERROR_NOT_ZERO,MkIntTerm(size_args),
		      "~~ in format/2");
		return(FALSE);
	      }
	      PlPutc (sno, (int) '~');
	      break;
	    case 'n':
	      if (!size_args)
		arg_size = 1;
	      for (i = 0; i < arg_size; i++)
		PlPutc (sno, (int) '\n');
	      column_boundary = 0;
	      column_fill = ' ';
	      break;
	    case 'N':
	      if (size_args)
		return (TRUE);
	      if (Stream[sno].linepos != 0)
		{
		  PlPutc (sno, (int) '\n');
		  column_fill = ' ';
		}
	      break;
	    case '|':
	      if (size_args)
		{
		  if (Stream[sno].linepos > arg_size)
		    PlPutc (sno, (int) '\n');
		  while (Stream[sno].linepos < arg_size)
		    PlPutc (sno, column_fill);
		}
	      column_boundary = Stream[sno].linepos;
	      break;
	    case '+':
	      if (!size_args)
		arg_size = 8;
	      if (Stream[sno].linepos > column_boundary + arg_size)
		PlPutc (sno, (int) '\n');
	      while (Stream[sno].linepos < column_boundary + arg_size)
		PlPutc (sno, column_fill);
	      column_boundary = Stream[sno].linepos;
	      break;
	    case 't':
	      if (!size_args)
		arg_size = ' ';
	      column_fill = arg_size;
	      break;
	    default:
	      return (FALSE);
	    }
	}
      else if (ch == '\\')
	{
	  if (HeadOfTerm (tail) == MkIntTerm ('c'))
	    {
	      tail = TailOfTerm (tail);
	      if (IsVarTerm (tail)) {
		Error(INSTANTIATION_ERROR,tail,"format/2");
		return(FALSE);		
	      } else if (!IsPairTerm (tail)) {
		Error(TYPE_ERROR_LIST,tail,"format/2");
		return(FALSE);
	      }
	      head = HeadOfTerm (tail);
	      tail = TailOfTerm (tail);
	      if (head != MkIntTerm (10)) {
		Error(DOMAIN_ERROR_NOT_NL,head,
		      "format/2");
		return(FALSE);
	      }
	    }
	  else
	    PlPutc (sno, (int) '\\');
	}
      else
	PlPutc (sno, ch);
      if (IsVarTerm (tail) || !IsPairTerm (tail))
	return (TRUE);
      head = HeadOfTerm (tail);
      tail = TailOfTerm (tail);
    }
  return (TRUE);
}

static Int
p_format(void)
{				/* '$format'(Control,Args)               */
  int sno = c_output_stream;

  PlPutchar_output_stream = c_output_stream;
  return(format(Deref(ARG1),Deref(ARG2), sno));
}


static Int
p_format2(void)
{				/* '$format'(Stream,Control,Args)          */
  PlPutchar_output_stream = CheckStream (ARG1, Output_Stream_f, "format/3");
  if (PlPutchar_output_stream == -1) {
    return(FALSE);
  }
  return(format(Deref(ARG2),Deref(ARG3),PlPutchar_output_stream));
}


static Int
p_skip (void)
{				/* '$skip'(Stream,N)                     */
  int sno = CheckStream (ARG1, Input_Stream_f, "skip/2");
  Int n = IntOfTerm (Deref (ARG2)), ch;

  if (sno < 0)
    return (FALSE);
  if (n < 0 || n > 127)
    return (FALSE);
  while ((ch = PlGetc (sno)) != n && ch != -1);
  return (TRUE);
}

static Int
p_flush (void)
{				/* flush_output(Stream)          */
  int sno = CheckStream (ARG1, Output_Stream_f, "flush_output/1");
  if (sno < 0)
    return (FALSE);
  if (!(Stream[sno].status & (Null_Stream_f|Socket_Stream_f)))
    YP_fflush (Stream[sno].u.file.file);
  return (TRUE);
}

static Int
p_write_depth (void)
{				/* write_depth(Old,New)          */
  Term t1 = Deref (ARG1);
  Term t2 = Deref (ARG2);
  if (!IsVarTerm (t1) && !IsIntTerm (t1))
    return (FALSE);
  if (!IsVarTerm (t2) && !IsIntTerm (t2))
    return (FALSE);
  if (IsVarTerm (t1))
    {
      Term t = MkIntTerm (*max_depth);
      if (!unify_constant(ARG1, t))
	return (FALSE);
    }
  else
    *max_depth = IntOfTerm (t1);
  if (IsVarTerm (ARG2))
    {
      Term t = MkIntTerm (*max_list);
      if (!unify_constant (ARG2, t))
	return (FALSE);
    }
  else
    *max_list = IntOfTerm (t2);
  return (TRUE);
}

static Int
p_change_type_of_char (void)
{				/* change_type_of_char(+char,+type)      */
  Term t1 = Deref (ARG1);
  Term t2 = Deref (ARG2);
  if (!IsVarTerm (t1) && !IsIntTerm (t1))
    return (FALSE);
  if (!IsVarTerm(t2) && !IsIntTerm(t2))
    return (FALSE);
  chtype[IntOfTerm(t1)] = IntOfTerm(t2);
  return (TRUE);
}

static Int
p_type_of_char (void)
{				/* type_of_char(+char,-type)      */
  Term t;

  Term t1 = Deref (ARG1);
  if (!IsVarTerm (t1) && !IsIntTerm (t1))
    return (FALSE);
  t = MkIntTerm(chtype[IntOfTerm (t1)]);
  return (unify(t,ARG2));
}


void
InitBackIO (void)
{
  InitCPredBack ("current_stream", 3, 1, init_cur_s, cont_cur_s, SafePredFlag|SyncPredFlag);
}

void
InitIOPreds (void)
{
  Atom StreamAtom, StreamPosAtom;
  EofAtom = LookupAtom ("end_of_file");
  ReadAtom = LookupAtom ("read");
  WriteAtom = LookupAtom ("write");
  AppendAtom = LookupAtom ("append");
  UserAtom = LookupAtom ("user");
  UsrInAtom = LookupAtom ("user_input");
  UsrOutAtom = LookupAtom ("user_output");
  UsrErrAtom = LookupAtom ("user_error");
  StreamAtom = LookupAtom ("$stream");
  StreamPosAtom = LookupAtom ("$stream_position");
  StreamFunctor = MkFunctor (StreamAtom, 1);
  StreamPosFunctor = MkFunctor (StreamPosAtom, 3);
  StreamEOSFunctor = MkFunctor (LookupAtom("end_of_stream"), 1);


  /* here the Input/Output predicates */
  InitCPred ("$check_stream", 2, p_check_stream, SafePredFlag|SyncPredFlag);
  InitCPred ("$check_stream", 1, p_check_if_stream, SafePredFlag|SyncPredFlag);
  InitCPred ("$stream_flags", 2, p_stream_flags, SafePredFlag|SyncPredFlag);
  InitCPred ("$close", 1, p_close, SafePredFlag|SyncPredFlag);
  InitCPred ("flush_output", 1, p_flush, SafePredFlag|SyncPredFlag);
  InitCPred ("$get", 2, p_get, SafePredFlag|SyncPredFlag);
  InitCPred ("$get0", 2, p_get0, SafePredFlag|SyncPredFlag);
  InitCPred ("$get_byte", 2, p_get_byte, SafePredFlag|SyncPredFlag);
  InitCPred ("$open", 4, p_open, SafePredFlag|SyncPredFlag);
  InitCPred ("$open_null_stream", 1, p_open_null_stream, SafePredFlag|SyncPredFlag);
  InitCPred ("$put", 2, p_put, SafePredFlag|SyncPredFlag);
  InitCPred ("$put_byte", 2, p_put_byte, SafePredFlag|SyncPredFlag);
  InitCPred ("$set_read_error_handler", 1, p_set_read_error_handler, SafePredFlag|SyncPredFlag);
  InitCPred ("$get_read_error_handler", 1, p_get_read_error_handler, SafePredFlag|SyncPredFlag);
  InitCPred ("$read", 3, p_read, SafePredFlag|SyncPredFlag);
  InitCPred ("$set_input", 1, p_set_input, SafePredFlag|SyncPredFlag);
  InitCPred ("$set_output", 1, p_set_output, SafePredFlag|SyncPredFlag);
  InitCPred ("$skip", 2, p_skip, SafePredFlag|SyncPredFlag);
  InitCPred ("$write", 2, p_write, SyncPredFlag);
  InitCPred ("$write", 3, p_write2, SyncPredFlag);
  InitCPred ("$format", 2, p_format, SyncPredFlag);
  InitCPred ("$format", 3, p_format2, SyncPredFlag);
  InitCPred ("$current_line_number", 2, p_cur_line_no, SafePredFlag|SyncPredFlag);
  InitCPred ("$line_position", 2, p_line_position, SafePredFlag|SyncPredFlag);
  InitCPred ("$character_count", 2, p_character_count, SafePredFlag|SyncPredFlag);
  InitCPred ("$start_line", 1, p_startline, SafePredFlag|SyncPredFlag);
  InitCPred ("$show_stream_flags", 2, p_show_stream_flags, SafePredFlag|SyncPredFlag);
  InitCPred ("$show_stream_position", 2, p_show_stream_position, SafePredFlag|SyncPredFlag);
  InitCPred ("$set_stream_position", 2, p_set_stream_position, SafePredFlag|SyncPredFlag);
  InitCPred ("$inform_of_clause", 2, p_inform_of_clause, SafePredFlag|SyncPredFlag);
  InitCPred ("$inform_of_clause", 2, p_inform_of_clause, SafePredFlag|SyncPredFlag);
  InitCPred ("$user_file_name", 2, p_user_file_name, SafePredFlag|SyncPredFlag),
  InitCPred ("$past_eof", 1, p_past_eof, SafePredFlag|SyncPredFlag),
  InitCPred ("$ungetc", 2, p_ungetc, SafePredFlag|SyncPredFlag),
  InitCPred ("current_input", 1, p_current_input, SafePredFlag|SyncPredFlag);
  InitCPred ("current_output", 1, p_current_output, SafePredFlag|SyncPredFlag);
  InitCPred ("always_prompt_user", 0, p_always_prompt_user, SafePredFlag|SyncPredFlag);
  InitCPred ("prompt", 1, p_setprompt, SafePredFlag|SyncPredFlag);
  InitCPred ("prompt", 2, p_prompt, SafePredFlag|SyncPredFlag);
  InitCPred ("write_depth", 2, p_write_depth, SafePredFlag|SyncPredFlag);
  InitCPred ("$change_type_of_char", 2, p_change_type_of_char, SafePredFlag|SyncPredFlag);
  InitCPred ("$type_of_char", 2, p_type_of_char, SafePredFlag|SyncPredFlag);
  InitCPred ("$add_alias_to_stream", 2, p_add_alias_to_stream, SafePredFlag|SyncPredFlag);
  InitCPred ("$check_if_valid_new_alias", 1, p_check_if_valid_new_alias, TestPredFlag|SafePredFlag|SyncPredFlag);
  InitCPred ("$fetch_stream_alias", 2, p_fetch_stream_alias, SafePredFlag|SyncPredFlag);

  InitSockets ();
  InitPlIO ();
}
