/*************************************************************************
 *									 *
 *	 YAP Prolog 							 *
 *									 *
 *	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
 *									 *
 * Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
 *									 *
 **************************************************************************
 *									 *
 * File:		load_foreign.c					 *
 * comments:	dynamic loader of external routines			 *
 *									 *
 *************************************************************************/
#ifdef SCCS
static char     SccsId[] = "%W% %G%.2";
#endif


#include "Yap.h"
#include "Yatom.h"
#include "io.h"
#include <stdlib.h>

#include "Foreign.h"

char LoadMsg[512];
char YapExecutable[512];

STD_PROTO(Int p_load_foreign, (void));

Int
p_load_foreign(void)
{
  StringList ofiles = NIL;
  StringList libs = NIL;
  char *InitProcName;
  YapInitProc InitProc = NULL;
  Term t, t1;
  StringList new;
  Int returncode = FALSE;

  strcpy(LoadMsg,"Invalid arguments");

  /* collect the list of object files */
  t = Deref(ARG1);
  while(1) {
    if(IsVarTerm(t)) goto Error;
    if(IsAtomTerm(t) && AtomOfTerm(t)==AtomNil) break;
    if(!IsPairTerm(t)) goto Error;
    t1 = HeadOfTerm(t);
    t = TailOfTerm(t);
    if(IsVarTerm(t1) || !IsAtomTerm(t1)) goto Error;
    new = (StringList) malloc(sizeof(StringListItem));
    new->next = ofiles;
    new->s = RepAtom(AtomOfTerm(t1))->StrOfAE;
    ofiles = new;
  }

  /* collect the list of library files */
  t = Deref(ARG2);
  while(1) {
    if(IsVarTerm(t)) goto Error;
    if(IsAtomTerm(t) && AtomOfTerm(t)==AtomNil) break;
    if(!IsPairTerm(t)) goto Error;
    t1 = HeadOfTerm(t);
    t = TailOfTerm(t);
    if(IsVarTerm(t1) || !IsAtomTerm(t1)) goto Error;
    new = (StringList) malloc(sizeof(StringListItem));
    new->next = libs;
    new->s = RepAtom(AtomOfTerm(t1))->StrOfAE;
    libs = new;
  }

  /* get the initialization function name */
  t1 = Deref(ARG3);
  if(IsVarTerm(t1) || !IsAtomTerm(t1)) goto Error;
  InitProcName = RepAtom(AtomOfTerm(t1))->StrOfAE;
  
  /* call the OS specific function for dynamic loading */
  if(LoadForeign(ofiles,libs,InitProcName,&InitProc)==LOAD_SUCCEEDED) {
    (*InitProc)();
    returncode = TRUE;
  }

 Error:
  /* TODO: handle error message in a more controlled way */
  if(!returncode) {
    YP_fprintf(YP_stderr,"[load_foreign_files: %s ]\n",LoadMsg);
  }
  while(ofiles) {
    new = ofiles;
    ofiles = ofiles->next;
    free(new);
  }
  while(libs) {
    new = libs;
    libs = libs->next;
    free(new);
  }

  return returncode;
}



void
InitLoadForeign(void)
{
  YAPFindExecutable(yap_args[0]);
  InitCPred("load_foreign_files", 3, p_load_foreign, SafePredFlag|SyncPredFlag);
}

void 
ShutdownLoadForeign(void)
{
}







