/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		boot.yap						 *
* Last rev:	8/2/88							 *
* mods:									 *
* comments:	boot file for Prolog					 *
*									 *
*************************************************************************/
'$chhhhhh'(_,_,J,I) :-
         J >= I,
        erase(R), fail.

'$live' :-
	display_yap_version,
	'$set_value'(spy_creep,0),
	prompt('  ?- '),
	% after an abort, make sure all spy points are gone.
	'$set_read_error_handler'(fail),
	'$clean_debugging_info',
	'$get_value'('$user_module',V),
	( V=[] -> '$current_module'(_,primitives);
		  '$current_module'(_,V), '$compile_mode'(_,0)
	),
	repeat,
		set_input(user),set_output(user),
		'$current_module'(Module),
		( Module=user ->
		    '$compile_mode'(_,0)
		;
		    '$format'(user_error,"[~w]~n", [Module])
		),
		'$catch'('$enter_top_level',Error,'$Error'(Error)).

% Start file for yap

/*		I/O predicates						*/

/* stream predicates							*/

open(Source,M,T) :- var(Source), !,
	throw(error(instantiation_error,open(Source,M,T))).
open(Source,M,T) :- var(M), !,
	throw(error(instantiation_error,open(Source,M,T))).
open(Source,M,T) :- nonvar(T), !,
	throw(error(type_error(variable,T),open(Source,M,T))).
open(user,read,'$stream'(0)) :- !.
open(user,'$csult','$stream'(0)) :- !.
open(user,_,'$stream'(1)) :- !.
open(user_input,read,'$stream'(0)) :- !.
open(user_input,'$csult','$stream'(0)) :- !.
open(user_input,append,'$stream'(0)) :- !.
open(user_input,write,V) :- !,
	'$get_value'(fileerrors,1),
	throw(error(permission_error(open,source_sink,user_input),open(user_input,write,V))).
open(user_output,write,'$stream'(1)) :- !.
open(user_output,append,'$stream'(1)) :- !.
open(user_output,read,'$stream'(1)) :-
	'$get_value'(fileerrors,1),
	throw(error(permission_error(open,source_sink,user_input),open(user_input,write,V))).
open(File,Mode,Stream) :-
	'$open'(File,Mode,Stream,0).

close(V) :- var(V), !,
	throw(error(instantiation_error,close(V))).
close(user) :- !.
close(user_input) :- !.
close(user_output) :- !.
close(user_error) :- !.
close('$stream'(0)) :- !.
close('$stream'(1)) :- !.
close('$stream'(2)) :- !.
close(File) :-
	atom(File), 
	current_stream(File,_,Stream),
	!, close(Stream).
close(Stream) :-
	'$close'(Stream).

set_input(Stream) :-
	'$set_input'(Stream).
	
set_output(Stream) :-
	'$set_output'(Stream).

/* meaning of flags for '$write' is
	 1	quote illegal atoms
	 2	ignore operator declarations
	 4	output '$VAR'(N) terms as A, B, C, ...
	 8	use portray(_)
*/

write(T) :- current_output(S), '$write'(S,4,T).

write(Stream,T) :- 
	'$write'(Stream,4,T),
	fail.
write(_,_).

put(Stream,N) :-  N1 is N, '$put'(Stream,N1).

nl(Stream) :- '$put'(Stream,10).

nl :- current_output(Stream), '$put'(Stream,10).

/* main execution loop							*/
'$read_vars'(Stream,T,V) :-
	current_input(Old),
	set_input(Stream),
	'$read'(true,T,V),
	set_input(Old).

'$enter_top_level' :-
		prompt(_,'   ?- '),
		prompt('   | '),
		'$read_vars'(user_input,Command,Varnames),
		'$set_value'(spy_sl,0),
		'$set_value'(spy_fs,0),
		'$set_value'(spy_sp,0),
		'$set_value'(spy_gn,1),
		'$set_value'(spy_creep,0),
		'$set_value'(spy_cl,1),
		'$set_value'(spy_leap,0),
		'$setflop'(0),
		prompt(_,'   |: '),
		'$command'(Command,Varnames,top),
		'$sync_mmapped_arrays',
		'$set_value'('$live',false).

%
% remove any debugging info after an abort.
%
'$clean_debugging_info' :-
	'$recorded'('$spy',_,R),
	erase(R),
	fail.
'$clean_debugging_info'.

'$erase_sets' :- 
		eraseall('$'),
		eraseall('$$set'),
		eraseall('$$one'), 
		eraseall('$reconsulted'), fail.
'$erase_sets' :- \+ '$recorded'('$path',_,_), '$recorda'('$path',"",_).
'$erase_sets'.

display_yap_version :- '$version'.

'$version' :- 
	'$get_value'('$version_name',VersionName),
	'$format'(user_error, "[ YAP version ~w ]", [VersionName]),
	fail.
'$version' :- '$recorded'('$version',T,_),
	nl,
	write(T), nl, fail.
'$version' :- write(user_error,'
').

yap_version(T) :- '$recordz'('$version',T,_), fail.
yap_version(_).

repeat.
repeat.
repeat.
repeat.
repeat.
repeat.
repeat.
repeat.
repeat.
repeat.
repeat.
repeat.
repeat :- repeat.


'$start_corouts' :- '$recorded'('$corout','$corout'(Name,_,_),R), Name \= main, finish_corout(R),
		fail.
'$start_corouts' :- 
		eraseall('$corout'),
		eraseall('$result'),
		eraseall('$actual'),
		fail.
'$start_corouts' :- '$recorda'('$actual',main,_),
	'$recordz'('$corout','$corout'(main,main,'$corout'([],[],[],[],[],[],[],[],[],[],[],[],[],[],[],[],[],[],[])),Ref),
	'$recorda'('$result',going,_).



'$command'(C,_,top) :- var(C), !,
	throw(error(instantiation_error,meta_call(C))).
'$command'(C,_,top) :- number(C), !,
	throw(error(type_error(callable,C),meta_call(C))).
'$command'(R,_,top) :- db_reference(R), !,
	throw(error(type_error(callable,R),meta_call(R))).
'$command'(end_of_file,_,_) :- !.
'$command'([X|L],_,top) :- !, '$csult'([X|L]), !, fail.
'$command'((:- module(N,P)),_,Context) :- !, '$module'(Context,N,P), !, fail.
'$command'((:-G),V,_) :- !, '$yes_no'(G,(:-)), fail.
'$command'((?-G),V,_) :- !, '$query'(G,V), fail.
'$command'(G,V,Option) :- '$continue_with_command'(Option,V,G).

'$continue_with_command'(reconsult,V,G) :-
	'$prepare_term'(G,V,G0,G1),
	'$$compile'(G1,G0,5), fail.
'$continue_with_command'(consult,V,G) :-
	'$prepare_term'(G,V,G0,G1),
	'$$compile'(G1,G0,4), fail.
'$continue_with_command'(top,V,G) :-
	'$query'(G,V), fail.

'$prepare_term'(G,V,G0,G1) :-
	( '$get_value'('$syntaxcheckflag',on) ->
		'$check_term'(G,V) ; true ),
	'$expand_term'(G, G0, G1).

% proccess an input clause
'$$compile'(true,_,_) :- !.
'$$compile'(G,G0,L) :-
    '$head_and_body'(G,H,_), 
    '$inform_of_clause'(H,L),
    ( '$is_dynamic'(H) -> '$assertz_dynamic'(L,G,G0) ;
/*       ( '$current_module'(M),
	 '$get_value'(verbose,on) ->
         	write(user_error,asserting(M,G)), nl
       	        ; true),*/
      '$$compile_stat'(G,G0,L,H) ).

% process a clause for a static predicate 
'$$compile_stat'(G,G0,L,H) :-
      '$compile'(G,L),
      ( '$some_recordedp'(H)
       ->   % if we already have clauses stored
	   ( '$get_value'('$source_mode',on) -> % and we are in source mode
	        % just test if we don't have to erase the previous clauses
	        ( '$get_value'('$abol',true) -> '$erase_source'(H); true ),
	        % and store our clause
		'$head_and_body'(G0,H0,B0),
	        '$record_stat_source'(H,(H0:-B0),L)
              ;
	        % if we are not in source mode
		'$erase_source'(H) )
       ;
             % no clauses recorded but in source mode
	   ( '$get_value'('$source_mode',on) ->
	        % just store our clause
		'$head_and_body'(G0,H0,B0),
	        '$record_stat_source'(H,(H0:-B0),L)
           ;
             % no clauses stored and not in source mode
		true
	   )
	).


'$erase_source'(G) :- '$recordedp'(G,_,R), erase(R), fail.
'$erase_source'(_).

'$check_if_reconsulted'(N,A) :-
	'$recorded'('$reconsulted',X,_),
	( X = N/A , !;
	  X = '$', !, fail;
	  fail
	).

'$inform_as_reconsulted'(N,A) :-
	'$recorda'('$reconsulted',N/A,_).

'$clear_reconsulting' :-
	'$recorded'('$reconsulted',X,Ref),
	erase(Ref),
	X == '$', !,
	( '$recorded'('$reconsulting',_,R) -> erase(R) ).

/* Executing a query */


% ***************************
% * -------- YAPOR -------- *
% ***************************

'$query'(G,V) :- '$start_yapor', '$parallel_query'(G,V).

'$parallel_query'(G,[]) :- !, '$parallel_yes_no'(G).
'$parallel_query'(G,V) :- call(G), $parallel_new_answer(V).

% MUITO IMPORTANTE: em modo de tabulacao o cut que se segue pode invalidar a 
% correcta avaliacao e completude das tabelas de certos predicados em avaliacao

'$parallel_yes_no'(G) :- call(G), !, $parallel_answer_yes.
'$parallel_yes_no'(_) :- $parallel_answer_no.


% *****************************
% * -------- TABLING -------- *
% *****************************

'$query'(G,V) :- '$tabling_present', !, '$tabling_query'(G,V).

'$tabling_query'(G,[]) :- !, '$tabling_yes_no'(G,(?-)).
'$tabling_query'(G,V) :- 
	                 ( '$user_call'(G),
		           '$write_answer'(V),
                 	   '$show_frozen',
		           '$another',
		           !, 
		           $abolish_incomplete_tables,
                           fail 
                         ;
		           $check_incomplete_tables,
                           '$present_answer'(_, no),
		           fail
	                 ).

'$tabling_yes_no'([X|L],_) :- !, '$csult'([X|L]).
'$tabling_yes_no'(G,C) :- '$user_call'(G), !,
	                  '$show_frozen',
	                  $abolish_incomplete_tables,
                          '$present_answer'(C, yes),
	                  fail.
'$tabling_yes_no'(G,_) :- $check_incomplete_tables,
                          '$present_answer'(_, no).

% *****************************

'$yes_no'([X|L],_) :- !, '$csult'([X|L]).
'$yes_no'(G,C) :- '$user_call'(G), !,
	'$show_frozen',
	'$present_answer'(C, yes),
	fail.
'$yes_no'(G,_) :-
	'$present_answer'(_, no).


'$query'(G,[]) :- !,
	'$yes_no'(G,(?-)).
'$query'(G,V) :-
	(	'$user_call'(G),
		'$write_answer'(V),
		'$show_frozen',
		'$another',
		!, fail ;
		'$present_answer'(_, no),
		fail
	).

'$show_frozen' :-
	'$all_frozen_goals'(LGs0),
	LGs0 = [_|_],
	'$show_list_of_frozen_goals'(LGs0).
'$show_frozen'.

%
% present_answer has three components. First it flushes the streams,
% then it presents the goals, and last it shows any goals frozen on
% the arguments.
%
'$present_answer'(_,_):-
	flush_output(user_output),
	flush_output(user_error),
	fail.
'$present_answer'((?-), Answ) :-
	nl(user_error),
	'$get_value'('$break',BL),
	( BL \= 0 -> 	'$format'(user_error, "[~w] ",[BL]) ;
			true ),
	write(user_error,Answ),	nl(user_error).

'$another' :- write(user_error,' ? '), '$get0'(user_input,C),
			(       C==59 -> '$skip'(user_input,10),fail;
				C==10 -> nl(user_error) ;
				'$skip'(user_input,10), '$ask_again_for_another'
			).

'$ask_again_for_another' :-
	write(user_error,'Action (";" for more choices, <return> for exit)'),
	'$another'.

'$write_answer'(_) :-
	flush_output(user_output),
	flush_output(user_error),
	nl(user_error),
	fail.
'$write_answer'(Vs) :-
	'$write_answer_var_by_var'(Vs).

'$write_answer_var_by_var'([]).
'$write_answer_var_by_var'([[Name|Value]|L]) :- 
	'$writes'(Name,user_error),
	write(user_error,' = '),
	'$write_answer'(Value,L,L1),
	(	L1==[], !;
		nl(user_error),
		'$write_answer_var_by_var'(L1)
	).

'$write_answer'(Value, [], []) :- !,
	write(user_error,Value).
'$write_answer'(Value, [[Name|Value2]|L],M) :- Value == Value2, !,
	'$writes'(Name,user_error),
	write(user_error,' = '),
	'$write_answer'(Value,L,M).
'$write_answer'(Value,[P|L],[P|M]) :- '$write_answer'(Value,L,M).

'$writes'([],_) :- !.
'$writes'([X|L],Stream) :- put(Stream,X), '$writes'(L,Stream).

call(G) :- '$execute'(G).

'$user_call'(G) :- '$execute'(G).

% '$user_call'(V) :- var(V), !,
% 	throw(error(instantiation_error,call(V))).
% '$user_call'(A) :- number(A), !,
% 	throw(error(type_error(callable,A),call(A))).
% '$user_call'(R) :- db_reference(R), !,
% 	throw(error(type_error(callable,R),call(R))).
% '$user_call'(A) :- '$get_value'(spy_creep,V), V \= 0, !,
% 	'$save_current_choice_point'(CP), '$spied_call'(A,CP).
% '$user_call'(A) :- '$save_current_choice_point'(CP), '$call'(A,CP).

'$meta_call'(G) :-
	'$save_current_choice_point'(CP),
	'$call'(G, CP).

'$spied_meta_call'(G) :-
	'$save_current_choice_point'(CP),
	'$spied_call'(G, CP).

'$call'(M:G,CP) :- !,
	( '$current_module'(M) ->
	    '$check_callable'(G,G),
	    '$call'(G,CP)
        ;
	     '$current_module'(Old,M),
	     '$check_callable'(G,G),
	     ( '$call'(G,CP); '$current_module'(_,Old), fail ),
	     ( '$current_module'(_,Old); '$current_module'(_,M), fail)
        ).
'$call'(repeat,_) :- !,
	'$repeat'.
'$call'((A,B),CP) :- !,
	'$check_callables'(A,B,(A,B)),
	'$call'(A,CP), '$call'(B,CP).
'$call'((X->Y),CP) :- !,
	'$check_callables'(X,Y,(X->Y)),
	 CP1 is local_sp,
	'$call'(X,CP), '$cut_by'(CP1), '$call'(Y,CP).
'$call'((X->Y; Z),CP) :- !,
	'$check_callables'(X,Y,Z,(X->Y;Z)),
	( '$call'(X,CP), !, '$call'(Y,CP); '$call'(Z,CP)).
'$call'((A;B),CP) :- !,
	'$check_callables'(A,B,(A;B)),
	('$call'(A,CP) ; '$call'(B,CP)).
'$call'((A|B),CP) :- !,
	'$check_callables'(A,B,(A|B)),
	('$call'(A,CP) ; '$call'(B,CP)).
'$call'(\+ X,_) :- !,
	\+ '$user_call'(X).
'$call'(not X,_) :- !,
	\+ '$user_call'(X).
'$call'(!,CP) :- !,
	'$cut_by'(CP).
'$call'([A|B],_) :- !,
	'$csult'([A|B]).
'$call'(A, CP) :-
	( '$undefined'(A) ->
		functor(A,F,N), '$current_module'(M),
		( '$recorded'('$import','$import'(S,M,F,N),_) ->
		  '$call'(S:A,CP) ;
		  '$undefp'([M|A])
	        )
	    ;
	       '$execute0'(A)
	 ).


'$check_callables'(T1,T2,G) :-
	'$check_callable'(T1,G),
	'$check_callable'(T2,G).

'$check_callables'(T1,T2,T3,G) :-
	'$check_callable'(T1,G),
	'$check_callable'(T2,G),
	'$check_callable'(T3,G).

'$check_callable'(V,G) :- var(V), !,
	'$current_module'(Mod),
	throw(error(instantiation_error,Mod:G)).
'$check_callable'(A,G) :- number(A), !,
	'$current_module'(Mod),
	throw(error(type_error(callable,G),Mod:A)).
'$check_callable'(R,G) :- db_reference(R), !,
	'$current_module'(Mod),
	throw(error(type_error(callable,G),Mod:R)).
'$check_callable'(_,_).

'$repeat'.
'$repeat' :- '$repeat'.

'$spied_call'(M:G,CP) :- !,
	( '$current_module'(M) ->
	    '$check_callable'(G,G),
	    '$spied_call'(G,CP)
        ;
	     '$current_module'(Old,M),
	     '$check_callable'(G,G),
	     ( '$spied_call'(G,CP); '$current_module'(_,Old), fail ),
	     ( '$current_module'(_,Old); '$current_module'(_,M), fail)
        ).
'$spied_call'(fail,_) :- !, fail.
'$spied_call'(false,_) :- !, false.
'$spied_call'(true,_) :- !.
'$spied_call'(otherwise,_) :- !.
'$spied_call'((A,B),CP) :- !,
	'$check_callables'(A,B,(A,B)),
	'$spied_call'(A,CP), '$spied_call'(B,CP).
'$spied_call'((X->Y),CP) :- !,
	'$check_callables'(X,Y,(X->Y)),
	CP1 is local_sp,
	'$spied_call'(X,CP), '$cut_by'(CP1), '$spied_call'(Y,CP).
'$spied_call'((X->Y; Z),CP) :- !,
	'$check_callables'(X,Y,Z,(X->Y;Z)),
	(
	    '$spied_call'(X,CP), !, '$spied_call'(Y,CP) ;
	    '$spied_call'(Z,CP)
	).
'$spied_call'((A;B),CP) :- !,
	'$check_callables'(A,B,(A;B)),
	('$spied_call'(A,CP) ; '$spied_call'(B,CP)).
'$spied_call'((A|B),CP) :- !,
	'$check_callables'(A,B,(A|B)),
	('$spied_call'(A,CP) ; '$spied_call'(B,CP)).
'$spied_call'(\+ X,_) :- !, \+ '$spied_call'(X,CP).
'$spied_call'(not X,_) :- !, not '$spied_call'(X,CP).
'$spied_call'(!,CP) :- $cut_by(CP).
'$spied_call'(repeat,_) :- !, '$repeat'.
'$spied_call'([A|B],_) :- !, '$csult'([A|B]).
'$spied_call'(A,CP) :-
	( '$undefined'(A) ->
		functor(A,F,N), '$current_module'(M),
		( '$recorded'('$import','$import'(S,M,F,N),_) ->
		  '$spied_call'(S:A,CP) ;
		  '$spy'(A)
	        )
	    ;
	       '$spy'(A)
	 ).

% Called by the abstract machine, if no clauses exist for a predicate
'$undefp'([M|G]) :-
	functor(G,F,N),
	'$recorded'('$import','$import'(S,M,F,N),_), !,
	'$user_call'(S:G,CP).
'$undefp'([_|G]) :- '$is_dynamic'(G), !, fail.
'$undefp'([M|G]) :-
	'$recorded'('$unknown','$unknown'(M:G,US),_), !,
	'$user_call'(US).


/* This is the break predicate,
	it saves the importante data about current streams and
	debugger state */

break :- '$get_value'('$break',BL), NBL is BL+1,
	'$get_value'(spy_sl,SPY_SL),
	'$get_value'(spy_fs,SPY_FS),
	'$get_value'(spy_sp,SPY_SP),
	'$get_value'(spy_gn,SPY_GN),
	'$get_value'(spy_creep,SPY_CREEP),
	'$get_value'(spy_cl,SPY_CL),
	'$get_value'(spy_leap,_Leap),
	'$set_value'('$break',NBL),
	current_output(OutStream), current_input(InpStream),
	'$format'(user_error, "[ Break (level ~w) ]~n", [NBL]),
	'$live',
	!,
	'$set_value'('$live',true),
	'$set_value'(spy_sl,SPY_SL),
	'$get_value'(spy_fs,SPY_FS),
	'$set_value'(spy_sp,SPY_SP),
	'$set_value'(spy_gn,SPY_GN),
	'$set_value'(spy_creep,SPY_CREEP),
	'$set_value'(spy_cl,SPY_CL),
	'$set_value'(spy_leap,_Leap),
	set_input(InpStream), set_output(OutStream),
	'$set_value'('$break',BL).


'$csult'(V) :- var(V), !,
	throw(error(instantiation_error,consult(V))).
'$csult'([]) :- !.
'$csult'([-F|L]) :- !, reconsult(F), '$csult'(L).
'$csult'([F|L]) :- consult(F), '$csult'(L).

consult(V) :- var(V), !,
	throw(error(instantiation_error,consult(V))).
consult([]) :- !.
consult([F|Fs]) :- !,
	consult(F),
	consult(Fs).
consult(X) :- atom(X), !,
	'$find_in_path'(X,Y),
	( open(Y,'$csult',Stream), !,
		'$record_loaded'(Stream),
		'$consult'(X,Stream), close(Stream)
	;
		throw(error(permission_error(input,stream,Y),consult(X)))
	).
consult(library(X)) :- !,
	'$find_in_path'(library(X),Y),
	( open(Y,'$csult',Stream), !,
		'$record_loaded'(Stream),
		'$consult'(library(X),Stream), close(Stream)
	;
		throw(error(permission_error(input,stream,Y),consult(library(X))))
	).
consult(V) :- 
	throw(error(type_error(atom,V),consult(V))).

'$consult'(F,Stream) :-
	'$getcwd'(OldD),
	'$set_consulting_file'(Stream),
	H0 is heapused, T0 is cputime,
	current_stream(File,_,Stream),
	'$start_consult'(consult,File,LC),
	'$get_value'('$consulting',Old),
	'$set_value'('$consulting',true),
	'$recorda'('$initialization','$',_),
	( '$get_value'($verbose,on) ->
		tab(user_error,LC),
		'$format'(user_error, "[ consulting ~w... ]~n", [F])
	    ; true ),
	'$loop'(Stream,consult),
	'$end_consult',
	'$exec_initialization_goals',
	( LC == 0 -> prompt(_,'   |: ') ; true),
	( '$get_value'($verbose,on) ->
		tab(user_error,LC) ;
	true ),
	H is heapused-H0, T is cputime-T0,
	'$format'(user_error, "[ ~w consulted ~w bytes in ~w seconds ]~n", [F,H,T]),
	'$set_value'('$consulting',Old),
	'$cd'(OldD),
	!.


'$loaded'(Stream) :-
	'$user_file_name'(Stream,F),
	'$recorded'('$loaded',F,_), !.

'$record_loaded'(user).
'$record_loaded'(user_input).
'$record_loaded'(Stream) :-
	'$user_file_name'(Stream,F),
	'$loaded'(Stream), !.
'$record_loaded'(Stream) :-
	'$user_file_name'(Stream,F),
	'$recorda'('$loaded',F,_).

'$set_consulting_file'(user) :- !,
	'$set_value'('$consulting_file',user_input).
'$set_consulting_file'(user_input) :- !,
	'$set_value'('$consulting_file',user_input).
'$set_consulting_file'(Stream) :-
	'$user_file_name'(Stream,F),
	'$set_value'('$consulting_file',F),
	'$set_consulting_dir'(F).

%
% Use directory where file exists
%
'$set_consulting_dir'(F) :-
	atom_chars(F,S),
	'$strip_file_for_scd'(S,Dir,Unsure,Unsure),
	'$cd'(Dir).

%
% The algorithm: I have two states, one for what I am sure will be an answer,
% the other for what I have found so far.
%
'$strip_file_for_scd'([], [], _, _).
'$strip_file_for_scd'([D|L], Out, Out, Cont) :-
	'$dir_separator'(D), !,
	'$strip_file_for_scd'(L, Cont, [D|C2], C2).
'$strip_file_for_scd'([F|L], Out, Cont, [F|C2]) :-
	'$strip_file_for_scd'(L, Out, Cont, C2).
	

'$loop'(Stream,Status) :-
	'$current_module'(OldModule),
	repeat,
		( current_stream(_,_,Stream) -> true
		 ; '$current_module'(_,OldModule), '$abort_loop'(Stream)
		),
		prompt('|     '), prompt(_,'| '),
		'$catch'('$enter_command'(Stream,Status), Error,
			 '$LoopError'(Error)),
	!,
	'$current_module'(_,OldModule).

'$enter_command'(Stream,Status) :-
	'$read_vars'(Stream,Command,Vars),
	'$command'(Command,Vars,Status).

'$abort_loop'(Stream) :-
	throw(permission_error(input,closed_stream,Stream), loop).

/* General purpose predicates				*/

'$append'([], L, L) .
'$append'([H|T], L, [H|R]) :-
	'$append'(T, L, R).

'$head_and_body'((H:-B),H,B) :- !.
'$head_and_body'(H,H,true).

%
% split head and body, generate an error if body is unbound.
%
'$check_head_and_body'((H:-B),H,B,P) :- !,
	'$check_head'(H,P).
'$check_head_and_body'(H,H,true,P) :-
	'$check_head'(H,P).

'$check_head'(H,P) :- var(H), !,
	throw(error(instantiation_error,P)).
'$check_head'(H,P) :- number(H), !,
	throw(error(type_error(callable,H),P)).
'$check_head'(H,P) :- db_reference(H), !,
	throw(error(type_error(callable,H),P)).
'$check_head'(_,_).

% Path predicates

'$exists'(F,Mode) :- '$get_value'(fileerrors,V), '$set_value'(fileerrors,0),
	( open(F,Mode,S), !, close(S), '$set_value'(fileerrors,V);
	  '$set_value'(fileerrors,V), fail).


'$find_in_path'(user,user) :- !.
'$find_in_path'(user_input,user) :- !.
'$find_in_path'(library(File),NewFile) :- !,
	library_directory(Dir),
	atom_chars(File,FileS),
	atom_chars(Dir,DirS),
	'$dir_separator'(A),
	'$append'(DirS,[A|FileS],NewS),
	atom_chars(NewFile,NewS),
	'$exists'(NewFile,'$csult'), !.
'$find_in_path'(File,File) :- '$exists'(File,'$csult'), !.
'$find_in_path'(File,NewFile) :- name(File,FileStr),
	'$search_in_path'(FileStr,NewFile),!.
'$find_in_path'(File,File).

'$search_in_path'(File,New) :-
	'$recorded'('$path',Path,_), '$append'(Path,File,NewStr),
	name(New,NewStr),'$exists'(New,'$csult').

path(Path) :- findall(X,'$in_path'(X),Path).

'$in_path'(X) :- '$recorded'('$path',S,_),
		( S == ""  -> X = '.' ;
		  name(X,S) ).

add_to_path(New) :- add_to_path(New,last).

add_to_path(New,Pos) :- '$check_path'(New,Str), '$add_to_path'(Str,Pos).

'$add_to_path'(New,_) :- '$recorded'('$path',New,R), erase(R), fail.
'$add_to_path'(New,last) :- '$recordz'('$path',New,_).
'$add_to_path'(New,first) :- '$recorda'('$path',New,_).

remove_from_path(New) :- '$check_path'(New,Path),
			'$recorded'('$path',Path,R), erase(R).

'$check_path'(At,SAt) :- atom(At), !, name(At,S), '$check_path'(S,SAt).
'$check_path'([],[]).
'$check_path'([Ch],[Ch]) :- '$dir_separator'(Ch), !.
'$check_path'([Ch],[Ch,A]) :- !, integer(Ch), '$dir_separator'(A).
'$check_path'([N|S],[N|SN]) :- integer(N), '$check_path'(S,SN).


% term expansion
%
% return two arguments: Expanded0 is the term after "USER" expansion.
%                       Expanded is the final expanded term.
%
'$expand_term'(Term, Expanded0, Expanded) :-
	expand_term(Term, Expandednomod),
	'$expand_term_modules'(Expandednomod,Expanded0),
	'$expand_array_accesses_in_term'(Expanded0,Expanded1),
	'$expand_term_arith'(Expanded1,Expanded).

expand_term(Term,Expanded) :-
	( \+ '$undefined'(term_expansion(_,_)) ,
	  call(term_expansion(Term,UserExpanded)) ;
	  Term = UserExpanded ), !,
	'$expand_term_grammar'(UserExpanded,Expanded).

%
% Grammar Rules expansion
%
'$expand_term_grammar'((A-->B), C) :-
	'$translate_rule'((A-->B),C), !.
'$expand_term_grammar'(A, A).

%
% Arithmetic expansion
%
'$expand_term_arith'(G1, G2) :-
	'$get_value'('$c_arith',true),
	'$c_arith'(G1, G2), !.
'$expand_term_arith'(G,G).


%
% Arithmetic expansion
%
'$expand_array_accesses_in_term'(Expanded0,ExpandedF) :-
	'$array_refs_compiled',
	'$c_arrays'(Expanded0,ExpandedF), !.
'$expand_array_accesses_in_term'(Expanded,Expanded).

%
% Module system expansion
%
'$expand_term_modules'(A,B) :- '$module_expansion'(A,B), !.
'$expand_term_modules'(A,A).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   catch/throw implementation

catch(G,C,A) :- var(G), !,
	throw(error(instantiation_error,catch(G,C,A))).
catch(G,C,A) :- number(G), !,
	throw(error(type_error(callable,G),catch(G,C,A))).
catch(R,C,A) :- db_reference(R), !,
	throw(error(type_error(callable,R),catch(R,C,A))).
catch(G,C,A) :-
	'$catch'(G,C,A).

'$catch'(G,C,A) :-
	'$get_value'('$catch_counter', I),
	I1 is I+1,
	'$set_value'('$catch_counter', I1),
	'$catch'(G,C,A,I).

'$catch'(G,C,A,I) :-
        % on entry we push the catch choice point
        X is '$last_choice_pt',
	'$catch_call'(X,G,I).
% someone sent us a throw.
'$catch'(_,C,A,_) :-
        ('$recorded'('$throw',X,R)->true),
        erase(R),
        ( C=X -> call(A) ; throw(X)).
% normal exit: make sure we only erase what we should erase!
'$catch'(_,_,_,I) :-
        '$recorded'('$catch','$catch'(_,J),R), J >= I,
        erase(R), fail.

'$catch_call'(X,G,I) :-
        '$recorda'('$catch','$catch'(X,I),_),
        call(G),
        ( % on exit remove the catch
          ('$recorded'('$catch','$catch'(X,I),R)->true),
          erase(R)
        ;
          % on backtracking reinstate the catch before backtracking to G
          '$recorda'('$catch','$catch'(X,I),_),
          fail
        ).

throw(A) :-
%(findall(Z, '$recorded'('$catch',Z,_), L2) -> L2 = L1 ; L1 = []),
%write('doing throw for '(A,L1)), nl,
	% fetch the point to jump to
        '$recorded'('$catch','$catch'(X,_),R), !,
        erase(R),
	% now explain why we are jumping.
        '$recordz'('$throw',A,_),
        '$cut_by'(X),
        fail.
throw(G) :-
        write(user_error,system_error_at(G)),
        abort.

'$check_list'(V, _) :- var(V), !.
'$check_list'([], _) :- !.
'$check_list'([A|B], T) :- !,
	'$check_list'(B,T).
'$check_list'(S, T) :-
	throw(error(type_error(list,S),T)).

'$exec_initialization_goals' :-
	recorded('$initialization',G,R),
	erase(R),
	(G = '$' -> ! ; '$catch'(once(G), Error,
			 '$LoopError'(Error)), fail).

