/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		modules.pl						 *
* Last rev:								 *
* mods:									 *
* comments:	module support						 *
*									 *
*************************************************************************/

% module handling

use_module(File) :- 
	(  '$recorded'('$module','$module'(_,File,Publics),_) ->
		use_module(File,Publics)
	;
		ensure_loaded(File)
	).

use_module(File,Imports) :-
	'$current_module'(M),
	'$find_in_path'(File,X),
	( open(X,$csult,Stream), !,
	'$consulting_file_name'(Stream,TrueFileName),
	( '$loaded'(Stream) -> true
	     ;
	     '$record_loaded'(Stream),
	       % the following avoids import of all public predicates
	     '$recorda'($importing,$importing(TrueFileName),R),
	     '$reconsult'(File,Stream)
	 ),
	 close(Stream),
	 ( var(R) -> true; erased(R) -> true; erase(R)),
	 ( '$recorded'($module,$module(TrueFileName,Mod,Publics),_) ->
	     $use_preds(Imports,Publics,Mod,M)
	 ;
	 format('[ use_module/2 can not find a module in file ~w]~n',File),
	 fail
     )
    ;  
	format('[ Error: can not consult file ~w]~n',File), fail
    ).
	
'$consulting_file_name'(Stream,F)  :-
	'$user_file_name'(Stream, F).

'$module'(reconsult,N,P) :- !,
	'$abolish_module_data'(N),
	'$module_dec'(N,P).
'$module'(consult,N,P) :-
	( '$recorded'($module,$module(F,N,_),_),
	     format('[ Module ~w was allready defined in file ~w]~n',[N,F]),
		'$abolish_module_data'(N),
		fail
	;
	   	true
	),
	'$module_dec'(N,P).

'$module_dec'(N,P) :-
	$current_module(Old,N),
	$get_value($consulting_file,F),
	( recordzifnot($module,$module(N),_) -> true; true),
	recorda($module,$module(F,N,P),_),
	( '$recorded'($importing,$importing(F),_) ->
	         true
	;
	 		'$import'(P,N,Old)
	).

'$import'([],_,_) :- !.
'$import'([N/K|L],M,T) :-
	integer(K), atom(N), !,
	( $check_import(M,T,N,K) ->
%	    format('[Importing ~w to ~w]~n',[M:N/K,T]),
	     recorda($import,$import(M,T,N,K),_)
	 ;
	    true
	),
	'$import'(L,M,T).
'$import'([PS|L],M,T) :-
	format('[Illegal pred specification(~w) in module declaration for module ~w]~n',[PS,M]),
	'$import'(L,M,T).

$check_import(M,T,N,K) :-
    '$recorded'($import,$import(M1,T,N,K),R), M1 \= M, /* ZP */ !,
    format('NAME CLASH: ~w was already imported to module ~w;~n',[M1:N/K,T]),
    format('            Do you want to import it from ~w ? [y or n] ',M),
    repeat,
	get0(C), $skipeol(C),
	( C is "y" -> erase(R), !;
	  C is "n" -> !, fail;
	  write(' Please answer with ''y'' or ''n'' '), fail
	).
$check_import(_,_,_,_).

% $use_preds(Imports,Publics,Mod,M)
$use_preds([],_,_,_) :- !.
$use_preds([P|Ps],Publics,Mod,M) :- !,
	$use_preds(P,Publics,Mod,M),
	$use_preds(Ps,Publics,Mod,M).
$use_preds(N/K,Publics,M,Mod) :-
    (  $member(N/K,Publics) ->
			( $check_import(M,Mod,N,K) ->
%			      format('[Importing ~w to ~w]~n',[M:N/K,Mod]),
%	              $trace_module(importing(M:N/K,Mod)),
	              recorda($import,$import(M,Mod,N,K),_)
	 		;
	    	      true
	    	)
	;
		format('[ Error: there is no public predicate ~w in module ~w]~n',
			   [N/K,M])
	).
    	  


'$abolish_module_data'(M) :-
	$current_module(T),
	( '$recorded'($import,$import(M,T,_,_),R), erase(R), fail; true),
	'$recorded'($module,$module(_,M,_),R),
	erase(R),
	fail.
'$abolish_module_data'(_).


% expand module names in a clause
'$module_expansion'((H:-B),(H:-B1)) :-
	'$module_u_vars'(H,UVars),	 % collect head variables in
					 % expanded positions
	'$module_expansion'(B,B1,HVars). % expand body
%	$trace_module((H:-B),(H:-B1)).


$trace_module(X) :-
	telling(F),
	tell('P0:debug'),
	write(X),nl,
	tell(F), fail.
$trace_module(X).

$trace_module(X,Y) :- X==Y, !.
$trace_module(X,Y) :-
	telling(F),
	tell('~/.dbg.modules'),
	write('***************'), nl,
	portray_clause(X),
	portray_clause(Y),
	tell(F),fail.
$trace_module(X,Y).
		
% expand module names in a body
'$module_expansion'(V,call(M:V),HVars) :- var(V), !, $current_module(M).
'$module_expansion'((A,B),(A1,B1),HVars) :- !,
	'$module_expansion'(A,A1,HVars),
	'$module_expansion'(B,B1,HVars).
'$module_expansion'((A;B),(A1;B1),HVars) :- !,
	'$module_expansion'(A,A1,HVars),
	'$module_expansion'(B,B1,HVars).
'$module_expansion'((A->B),(A1->B1),HVars) :- !,
	'$module_expansion'(A,A1,HVars),
	'$module_expansion'(B,B1,HVars).
'$module_expansion'(true,true,HVars) :- !.
'$module_expansion'(fail,fail,HVars) :- !.
'$module_expansion'(false,false,HVars) :- !.
'$module_expansion'(M:G,call(M:G),HVars) :- var(M), !.
'$module_expansion'(M:(M1:G),G1,HVars) :- !,
	'$module_expansion'(M1:G,G1,HVars).
'$module_expansion'(M:G,G1,HVars) :- !,
	( functor(G,F,N), '$recorded'($import,$import(M1,M,F,N),_) ->
		$module_expansion(M1:G,G1,HVars)
	;
	  $meta_expansion(M,G,G0,HVars),
	  G1 = M:G0
	).
'$module_expansion'(G,G1,HVars) :- 
	$system_predicate(G),
	$meta_expansion(primitives,G,G1,HVars), !.
'$module_expansion'(G,G1,HVars) :- 
	$current_module(Mod),
	functor(G,F,N),
	'$recorded'($import,$import(M,Mod,F,N),_), !,
	$module_expansion(M:G,G1,HVars), !.
'$module_expansion'(G,G,HVars) :-
	$current_module(M),
	$meta_expansion(M,G,G1,HVars).

% meta_predicate declaration
% records $meta_predicate(SourceModule,Functor,Arity,Declaration)

meta_predicate Ps :- $meta_predicate(Ps).

$meta_predicate((P,Ps)) :- !,
	$meta_predicate(P),
	$meta_predicate(Ps).
$meta_predicate(P) :-
	functor(P,F,N),
	source_module(M),
	( '$recorded'($meta_predicate,$meta_predicate(M,F,N,_),R), erase(R), fail;
	  true
	),
	recorda($meta_predicate,$meta_predicate(M,F,N,P),_).
	

% return list of vars in expanded postions on G
'$module_u_vars'(H,UVars) :-
	functor(H,F,N),
	source_module(M),
	'$recorded'($meta_predicate,$meta_predicate(M,F,N,D),_), !,
	$module_u_vars(N,D,H,HVars).
'$module_u_vars'(H,[]).

'$module_u_vars'(0,_,_,[]) :- !.
'$module_u_vars'(I,D,H,[V|L]) :-
	arg(I,D,X), ( X=':' ; integer(X)),
	arg(I,H,Y), var(V), !,
	I1 is I-1,
	'$module_u_vars'(I1,D,H,L).
$module_u_vars(I,D,H,V) :-
	I1 is I-1,
	'$module_u_vars'(I1,D,H,L).

% expand arguments of a meta-predicate
% $meta_expansion(ModuleWhereDefined,Goal,ExpandedGoal,MetaVariables)

$meta_expansion(Mod,G,G1,HVars) :- 
	source_module(M),  % why not ... M\=user, M\=primitives,
	functor(G,F,N),
	'$recorded'($meta_predicate,$meta_predicate(Mod,F,N,D),_), !,
	functor(G1,F,N),
%	format('[expanding ~w',[G]),
	$meta_expansion(N,D,G,G1,HVars).
%	format(' gives ~w]~n]',[G1]).

$meta_expansion(_,G,G,_). 

% expand argument
$meta_expansion(0,_,_,_,_) :- !.
$meta_expansion(I,D,G,G1,HVars) :- 
	arg(I,D,X), (X=':' ; integer(X)),
	arg(I,G,A), $do_expand(A,HVars), !,
	source_module(M),
	( var(A) -> 	A1=M:A;		% if a variable
	  A=[_|_] ->	 A1=M:A;	% or list (for the toolkit)
	  $module_expansion(M:A,A1,HVars)
	),
	arg(I,G1,A1),
	I1 is I-1,
	$meta_expansion(I1,D,G,G1,HVars).
$meta_expansion(I,D,G,G1,HVars) :- 
	arg(I,G,A),
	arg(I,G1,A),
	I1 is I-1,
	$meta_expansion(I1,D,G,G1,HVars).

% check if an argument should be expanded
$do_expand(V,HVars) :- var(V), !, $not_in_vars(V,HVars).
$do_expand(M:F,_) :- !, fail.
$do_expand(X,_).

$not_in_vars(_,[]).
$not_in_vars(V,[X|L]) :- X\==V, $not_in_vars(V,L).

source_module(A) :- $current_module(A).

current_module(Mod) :- 
	'$recorded'($module,$module(Mod),_).

current_module(Mod,TFN) :- '$recorded'($module,$module(TFN,Mod,_Publics),_).

module(Mod) :- 
	atom(Mod),
	( recordzifnot($module,$module(Mod),_) -> true; true),
	$current_module(_,Mod).


:- meta_predicate
	abolish(:),
	abolish(:,+),
	all(?,:,?),
	assert(:),
	assert(:,+),
	asserta(:),
	asserta(:,+),
	assertz(:),
	assertz(:,+),
	bagof(?,:,?),
	call(:),
	clause(:,?),
	clause(:,?,?),
	current_predicate(?,:),
	findall(?,:,?),
	listing(?),
	nospy(:),
        not(:),
	retract(:),
	retract(:,?),
	retractall(:),
	setof(?,:,?),
	spy(:),
	'^'(+,:),
	'\+'(:).

$member(X,[X|_]) :- !.
$member(X,[_|L]) :- $member(X,L).

